/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.
    

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 261                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-07-17 13:57:32 +02#$: Date of last commit
$Id: circbuf.cpp 261 2018-07-17 11:57:32Z gmbertani $
*******************************************************************************/

#include "settings.h"
#include "circbuf.h"

CircBuf::CircBuf(bool keepFull, int maxRecords, QString logFilePath)
{    
    setMax(maxRecords);
    kpFull = keepFull;
    outPath = logFilePath;

    full = false;
    MYINFO << "Circbuf " << this << " keepFull: " << kpFull << " created on file " << outPath;
}


CircBuf::~CircBuf()
{
    if(out.device() != nullptr)
    {
        out.flush();
        outFile.flush();
        outFile.close();
        if(kpFull == false)
        {
            q.clear();
            curr.clear();
            scan.clear();
        }
    }
}

int CircBuf::setFile(QString& path, bool append)
{
    if(outPath != "")
    {
        out.flush();
        outFile.flush();
        outFile.close();
    }

    //The file is closed but the data are kept in queue
    //To flush the queue, unsetFile() must be called first
    outPath = path;
    outFile.setFileName(outPath);


    if(append == false)
    {
        //create a brand new file
        //if another one already exists, it will be overwritten
        if ( outFile.open(QIODevice::WriteOnly|QIODevice::Truncate) == 0 )
        {
            MYCRITICAL << "cannot create a new file " << outPath << endl;
            return -1;
        }
    }
    else
    {
        //reopen an existing file if exists, otherwise creates it
        if ( outFile.open(QIODevice::WriteOnly|QIODevice::Append) == 0 )
        {
            MYCRITICAL << "cannot create a new file " << outPath << endl;
            return -1;
        }
    }

    out.setDevice(&outFile);
    return 0;
}

int CircBuf::unSetFile()
{
    if(outPath != "")
    {
        MYDEBUG << "Circbuf " << this << " unset file " << outPath;
        out.flush();
        outFile.flush();
        outFile.close();


        outPath = "";
        if(kpFull == false)
        {
            q.clear();
            curr.clear();
            scan.clear();
        }

        return 0;
    }
    return -1;
}




void CircBuf::setMax(int maxRecords)
{
    q.clear();
    if (maxRecords < CIRCBUF_MIN)
    {
        maxRecs = CIRCBUF_MIN;
    }
    else
    {
        maxRecs = maxRecords;
    }
    full = false;
    MYINFO << "Circbuf " << this << " resized to max records: " << maxRecs;
}

int CircBuf::append(const char* format, ...)
{
    QString buf;
    va_list args;
    va_start(args, format);

    buf = QString::vasprintf(format, args);
    curr += buf;

    va_end(args);
    return curr.length();
}



int CircBuf::append(QString& buf)
{
    curr = curr + buf;
    return curr.length();
}


int CircBuf::insert(const char* format, ...)
{
    QString buf;
    int currLen = 0;

    va_list args;
    va_start(args, format);


    buf = QString::vasprintf(format, args);
    curr += buf;
    currLen = curr.length();

    MYDEBUG << "Circbuf " << this << " inserting a record " << currLen << " bytes long";

    commit();

    va_end(args);
    return currLen;
}





void CircBuf::commit(bool prepend)
{
    if(prepend == true)
    {
        q.prepend(curr);
    }
    else
    {
        q.enqueue(curr);
    }

    curr.clear();

    MYDEBUG << "Circbuf " << this << " committed scan# " << q.size();

    if(q.size() >= maxRecs)
    {
        full = true;

        int excess = q.size() - maxRecs;
        while(excess > 0)
        {
            if(prepend == true)
            {
                //discards the newest record
                q.takeLast();
                excess = q.size() - maxRecs;
                MYDEBUG << "Circbuf " << this << " full, discarding newest scan (prepend=" << prepend << "), left: " << excess;
            }
            else
            {
                //discards the oldest record
                q.dequeue();
                excess = q.size() - maxRecs;
                MYDEBUG << "Circbuf " << this << " full, discarding oldest scan, left: " << excess;
            }
        }
    }

}


void CircBuf::rollback(uint recs)
{
    for(uint i = 0; i < recs; i++)
    {
        if(q.count() >= 1)
        {
            q.removeLast();
        }
    }

    MYDEBUG << "Circbuf " << this << " removed latest " << recs << " scans, now " << q.size() << " scans are still in queue";

    if(q.size() < maxRecs)
    {
        full = false;
    }

}



int CircBuf::dump(QString shotName, int numRecs)
{
    MYDEBUG << this << ":dump( " << shotName << ", " << numRecs << " )";
    int written = 0;
    if(outFile.isOpen() == false)
    {
        MYDEBUG << "Circbuf " << this << " dump() requested with file not yet assigned, doing nothing";
        return written;
    }

    if(kpFull == true)
    {
        //with kpFull, the records in queue are copied to disk
        //but the queue is not flushed. This allows to keep the
        //dump files size almost equal, once the queue reaches
        //its maximum size.
        //This behavior is good for plot files, not for CSV
        //Please note, a single spectra scan line is composed by many lines


        foreach (scan, q)
        {
            if(shotName.isEmpty() == true)
            {
                out << scan << endl;
                MYDEBUG << "Circbuf " << this << " copied record# " << written << " to file " << outPath;
            }
            else
            {
                out << scan << shotName << endl;
                MYDEBUG << "Circbuf " << this << " copied record# " << written << " to file " << outPath << " shotName: " << shotName;
            }

            written++;
            if(numRecs > 0 &&
               written >= numRecs)
            {
                break;
            }
        }

    }
    else
    {
        //without kpFull, the records in queue are flushed on the disk
        //leaving the queue empty. This behavior is good for CSV files
        while(q.empty() == false)
        {
            scan = q.dequeue();
            if(shotName.isEmpty() == true)
            {
                out << scan << endl;
                MYDEBUG << "Circbuf " << this << " extracted oldest record# " << written << " to file " << outPath;
            }
            else
            {
                //the shot name is appended as last field of the record
                out << scan << shotName << endl;
                MYDEBUG << "Circbuf " << this << " extracted oldest record# " << written << " to file " << outPath << " shotName: " << shotName;
            }
            written++;
            if(numRecs > 0 &&
               written >= numRecs)
            {
                break;
            }

        }
        MYDEBUG << "Circbuf " << this << " queue empty";
        full = false;
    }
    out.flush();
    return written;
}

