/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.


    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 370                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-11-27 00:17:03 +01#$: Date of last commit
$Id: control.cpp 370 2018-11-26 23:17:03Z gmbertani $
*******************************************************************************/



#include "setup.h"
#include "settings.h"
#include "circbuf.h"
#include "radio.h"
#include "waterfall.h"
#include "postproc.h"
#include "mainwindow.h"
#include "control.h"

//#define _MIDNIGHT_SWAP_TEST 1

Control::Control(Settings* appSettings, XQDir wDir, QObject *parent) :
    QObject(parent)
{
    setObjectName("Control object");

    MY_ASSERT( nullptr !=  appSettings);
    as = appSettings;

    statPlot = nullptr;
    console = as->getConsoleMode();

    XQDir d = qApp->applicationDirPath();
    QString notifySound = d.absoluteFilePath(NOTIFY_SOUND);
    ping = new QSound(notifySound);
    MY_ASSERT( nullptr !=  ping)

    //frequency domain scans pool
    ps = new Pool<Scan*>;
    MY_ASSERT( nullptr !=  ps)

    //radio FFT manager
    r = new Radio(as, ps, wDir, this);
    MY_ASSERT( nullptr !=  r)

    //circular buffers
    //scans are the maximum number of records that can be stored
    //in cb without overwriting. Resulting DATs will be longer than
    //expected because the interval is only the GUI refreshing interval
    //but the samples could take more time to be extracted/decoded/written
    //on disk.

    dumpScans = 0;
    cbDAT = new CircBuf(true, dumpScans);
    MY_ASSERT( nullptr !=  cbDAT)

    //statistic CSV is opened in append mode
    cbCSV = new CircBuf(false, dumpScans);
    MY_ASSERT( nullptr !=  cbCSV)

    sShotCoverage = 0;

    //the file names are reinitialized each time "Start" is pressed
    sessionName = "";
    csvFileName = "";
    cmdFileName = "";
    dumpFileName = "";
    wfShotFileName = "";
    timeStamp = "";
    dateTimeStamp = "";
    logSwap = false;

    am = AM_INVALID;

    loop = 0;
    scans = 0;

    upDays = 0;
    autoCount = 0;
    manCount = 0;
    prog = 0;
    tRange = 100;   //100%

    //loops countdown stopped
    waitingShots.clear();
    waitingDumps.clear();

    //acquisition stopped
    running = false;
    init = false;
    overload = true;
    newHdr = false;
    testMode = false;
    autoStopReq = false;
    waterfallChange = false;
    enableCapture = false;

    //other temporary vars
    fromHz      = 0;
    toHz      = DEFAULT_RATE;
    zoomLo      = 0;
    zoomHi      = DEFAULT_RATE;
    rangeLo     = 0;
    rangeHi     = DEFAULT_RATE;
    centerFreq  = 0;
    maxFreq     = 0;
    FFTbins     = int(DEFAULT_RATE / DEFAULT_RESOLUTION);
    loggedBins  = FFTbins;
    wfWidth     = 1024;
    wfZoom      = DEFAULT_ZOOM;
    wfOffset    = 0;
    wfTune      = DEFAULT_TUNE;
    hzShift     = 0;
    restart     = false;

    binSize     = (DEFAULT_RESOLUTION / 10e8);
    avgDbfs     = 0;
    variance    = 0;
    stddev      = 0;
    maxDbfs     = MIN_DBFS;
    minDbfs     = MAX_DBFS;
    maxDiff     = 0.0;

    avgN        = MIN_DBFS;
    avgS        = MIN_DBFS;
    avgDiff     = 0;

    peakS       = 0;
    peakN       = 0;
    peakDiff    = 0;
    peakFreq    = 0;

    autoUpThr = 0.0;
    autoDnThr = 0.0;

    totalTime = 0;
    scanTime = 0;
    avgScanTime = 0;
    scanDelta = 0;
    dBfsDelta = 0;
    csvDelta = 0;
    outDelta = 0;

    eventDetected = 0;
    wf     = nullptr;

    bpp         = 0;
    hpp         = 0;

    firstLoggedBin = 0;
    lastLoggedBin  = loggedBins - 1;

    scanIdxBak    = -1;

    //NaNs and infinites statistic counter
    nans        = 0;
    infinite    = 0;
    noDataBuf   = 0;

    //forces device reopening
    scc = 1;

    //retrieves the working directory
    //QString s;
    //s = QString("%1/%2").arg( XQDir::homePath() ).arg( RESOURCE_PATH );
    //wd.setCurrent( s );

    wd = wDir;
    wd.makeAbsolute();

    pp = new PostProc( as, this, wd );
    MY_ASSERT( nullptr !=  pp);
    pp->loadData();

    //connects the available slots
    connect( &samplingTimer, SIGNAL( timeout() ), this, SLOT( slotControlLoop() ), Qt::DirectConnection );
    connect( r,                 SIGNAL( fault(int) ),    this,  SLOT( slotStopAcquisition(int) ));

    connect( as,                SIGNAL( notifySetInterval() ),  this, SLOT( slotSetInterval() ) );
    connect( as,                SIGNAL( notifySetShots() ),     this, SLOT( slotSetShots() ) );
    connect( as,                SIGNAL( notifySetRecTime() ),   this, SLOT( slotSetRecTime() ) );
    connect( as,                SIGNAL( notifySetAfter() ),     this, SLOT( slotSetAfter() ) );
    connect( as,                SIGNAL( notifySetUpThreshold() ), this, SLOT( slotSetUpThreshold() ) );
    connect( as,                SIGNAL( notifySetDnThreshold() ), this, SLOT( slotSetDnThreshold() ) );
    connect( as,                SIGNAL( notifySetTrange() ),    this, SLOT( slotSetTrange() ) );
    connect( as,                SIGNAL( notifySavedAs() ),      this, SLOT( slotStopTimeChange() ) );
    connect( as,                SIGNAL( notifyOpened() ),       this, SLOT( slotStopTimeChange() ) );


    //automatic report request
    //connect( pp,     SIGNAL( autoReport() ), as->getMainWindow(), SLOT( slotReport() ) );


    //posticipate the initialization in order to wait for main window coming up
    samplingTimer.setSingleShot(false);
    samplingTimer.setInterval( as->getInterval() );
}

Control::~Control()
{
    delete r;
    delete cbCSV;
    delete cbDAT;
    delete ps;
    delete ping;

    if(statPlot != nullptr)
    {
        statPlot->close();
        delete statPlot;
    }
}




QString Control::getDateTime()
{
    QString s = cal.currentDateTimeUtc().toString(Qt::TextDate);
    return s;
}

QString Control::getDateTimeStamp()
{
    QString s = cal.currentDateTimeUtc().toString(Qt::ISODate);
    s.replace(':', '-');
    return s;
}


QString Control::getDateStamp()
{
    QString s = cal.date().toString(Qt::ISODate);
    return s;
}


//---slots----------------

#if 0

void Control::slotMainWindowReady()
{
    //connecting the report button
    MY_ASSERT( nullptr !=   as->getMainWindow() );
    connect( as->getMainWindow(), SIGNAL( createReport(const char*)), pp, SLOT( createFullReport(const char*)) );
}

#endif

void Control::slotAcquisition(bool go)
{

    if(go == true)
    {
        init = true;
        running = true;
        restart = true;
        loop = 0;
        scans = 0;
        noDataBuf = 0;
        infinite = 0;
        nans = 0;
        avgN = MIN_DBFS;
        avgDiff = 0;
        eventDetected = 0;
        autoStopReq = false;
        manStopReq = false;
        waterfallChange = false;
        autoCount = 0;
        manCount = 0;
        enableCapture = false;

        statClock.start();
        scanTimer.start();

        autoUpThr = as->getUpThreshold();
        autoDnThr = as->getDnThreshold();

        //change working directory to selected one
        bool ok = XQDir::setCurrent( wd.currentPath() );
        MY_ASSERT(ok == true)

        //initialize UTC calendar
        cal = QDateTime::currentDateTimeUtc();


        if(console == false)
        {
            if(wf == nullptr)
            {
                wf = as->getWaterfallWindow();
                connect( wf, &Waterfall::waterfallResized, [=](void) { waterfallChange = true; } );
            }

            //clean the WF before acquisition starts
            wf->cleanAll();
        }

        if(scc > 0)
        {
            if(r->isDeviceOpen() == false)
            {
                //restarting after midnight swap
                r->slotSetDevice();
            }

            if( r->slotSetSampleRate() == true )
            {
                if( r->slotSetBandwidth() == true )
                {
                    if(r->slotSetResolution() == true)
                    {
                        if(r->slotSetGain() == true)
                        {
                            if(r->slotSetError() == true)
                            {
                                if(r->slotSetTune() == true)
                                {
                                    scc = 0;
                                }
                            }
                        }
                    }

                }
            }
            if(scc > 0)
            {
                MYCRITICAL << "failed restart, scc=" << scc;
            }
        }

        // events delayers to take shots and dumps
        waitingShots.clear();
        waitingDumps.clear();

        //applying preferences:

        wd.refresh();
        if(as->getEraseLogs() != 0)
        {
            //erase old statistic CSV files
            QStringList files = wd.entryList();
            QString file;
            int goDelete;
            foreach (file, files)
            {
                goDelete = 0;

                goDelete += file.endsWith(TABLE_EXT, Qt::CaseInsensitive);
                if(goDelete == 0)
                {
                    continue;
                }

                file = wd.absoluteFilePath(file);
                if(wd.remove( file ) == false)
                {
                    MYWARNING << "cannot remove file " << file;
                }
            }
        }

        if(as->getEraseShots() != 0)
        {
            //erase old shots, feeds and plots
            QStringList files = wd.entryList();
            QString file;
            int goDelete;
            foreach (file, files)
            {
                goDelete = 0;

                goDelete += file.endsWith(GP_DATA_EXT, Qt::CaseInsensitive);
                goDelete += file.endsWith(GP_CMD_EXT, Qt::CaseInsensitive);
                goDelete += file.endsWith(HTML_EXT, Qt::CaseInsensitive);
                goDelete += file.endsWith(PDF_EXT, Qt::CaseInsensitive);
                goDelete += file.endsWith(SSHOT_EXT, Qt::CaseInsensitive);
                goDelete += (file == DAILYCTR_NAME);
                if(goDelete == 0)
                {
                    continue;
                }
                file = wd.absoluteFilePath(file);
                if( wd.remove(file) == false )
                {
                    MYWARNING << "cannot remove file " << file;
                }
            }

        }

        //theorical number of scans stored in the given time
        dumpScans = static_cast<int>( ceil( static_cast<float>(as->getRecTime() * 1000) / as->getInterval() ) );
        cbDAT->setMax(dumpScans);

        //updates FFT related members
        //slotFFTchanged();
        //loggedBins = static_cast<int>( as->getBandwidth() / r->getBinSize() );

        dataDumps.clear();

        //assigns output file names - the actual datetime must be embedded in filenames
        cmdFileName.clear();
        datFileName.clear();
        dumpFileName.clear();

        //determine the acquisition mode
        emit status( tr("Starting acquisition...") );
        am = static_cast<E_ACQ_MODE>( as->getAcqMode() );
        switch( am )
        {
            case AM_CONTINUOUS:
                if ( as->getShots() == 0 && as->getUpThreshold() == 0.0F )
                {
                    //continuous mode
                    MYINFO << "*** Continuous acquisition mode (shots=0, upThreshold=0, recTime=0) ";

                    cbCSV->setMax(dumpScans);
                    makeSessName( getDateTimeStamp() );

                }
                else
                {
                    am = AM_INVALID;
                    faultStatus = tr("Please set: shots=0 and upThreshold=0");
                }
                break;

            case AM_PERIODIC:
                if ( as->getShots() > 0 && as->getUpThreshold() == 0.0F )
                {
                    //periodic shots mode
                    MYINFO << "*** Periodic acquisition mode ( shots>0, upThreshold=0, recTime>0 )";
                    cbCSV->setMax(dumpScans);
                    makeSessName( getDateTimeStamp() );
                }
                else
                {
                    am = AM_INVALID;
                    faultStatus = tr("Please set: shots>0 and upThreshold=0");
                }
                break;

            case AM_AUTOMATIC:
                if ( as->getShots() > 0 && as->getUpThreshold() != 0.0F )
                {
                    //automatic events detection
                    MYINFO << "*** Automatic peaks detection mode ( shots>0, upThreshold>0, recTime>0 )";
                    cbCSV->setMax(3);   //each event generates 3 rows in statistic CSV
                    makeSessName( getDateStamp() );
                }
                else
                {
                    am = AM_INVALID;
                    faultStatus = tr("Please set: shots>0 and upThreshold>0");
                }
                break;

            default:
                break;
        }

        csvFileName = makeFileName(CSV_PREFIX, TABLE_EXT);

        if( as->getGNUplot() == Qt::Checked )
        {
            //gnuplot output files:
            //2D Color mapped spectrograms (PT_3D_MAPPED / SPECMAP_PREFIX)
            //are similar to waterfall output, with time on Y
            //frequency in X and the dBfs signal represented with
            //a color scale. At the end of each horizontal scan
            //the S, N and S-N values are stored.

            //3D spectrograms (PT_3D_REAL / SPEC3D_PREFIX)
            //like color mapped but the image is not flat because the
            //dBfs power is plotted as colored dot on Z axis and projected
            //in perspective

            //In 2D plot files (PT_2D_POWER / POW2D_PREFIX)
            //the data are reduced to 4 columns containing
            //time [ms], peak value (Signal), average value (Noise) and difference
            //of each scan, resulting in an skyPipe-like diagram

            //If shots or threshold are set, a new output
            //data file will be created each time a new event
            //have been catched. Otherwise the file will be continuous
            //(and huge after some time!)

            //The data file is marked with
            //the same datetime and progressive number of the
            //relative screenshot.

            //selection of data file prefix
            switch(as->getPlotType())
            {
                case PT_2D_MAPPED:
                    gpPrefix = SPECMAP_PREFIX;
                    break;

                case PT_3D_REAL:
                    gpPrefix = SPEC3D_PREFIX;
                    break;

                case PT_2D_POWER:
                    gpPrefix = POW2D_PREFIX;
                    break;
            }

            //besides data, gnuplot needs sone commands to correctly
            //display output data. So, a command file is created also.
            //The prologue is written immediately, while the epilogue
            //will be added when acquisition stops.

            cmdFileName = makeFileName(gpPrefix, GP_CMD_EXT);

            //gnuplot continuous data file
            datFileName =  makeFileName(gpPrefix, GP_DATA_EXT);

            //gnuplot dumps (GNUplot alternative to screenshots, useful while running in Console UI)
            dumpFileName =  makeFileName(gpPrefix, GP_DATA_EXT);
            //the shot number will be appended later, with extension

            gpPrologue();

        }


        //warning: this must be the main thread
        MYINFO << "Starting acquisition." ;
        MYINFO << "One read every " << as->getInterval() << " ms." ;
        samplingTimer.setInterval( as->getInterval() );
        samplingTimer.start();
        r->start();

        emit start();

        if(console == false)
        {
            MainWindow* mw = as->getMainWindow();
            MY_ASSERT( nullptr !=  mw);
            connect(pp, SIGNAL(status(QString)), mw, SLOT(slotShowStatus(QString)), Qt::QueuedConnection);
        }

        //change working directory to selected one
        ok = XQDir::setCurrent( wd.currentPath() );
        MY_ASSERT(ok == true)

        MY_ASSERT( csvFileName.isEmpty() == false )
        QString csvAbsPath = wd.absoluteFilePath( csvFileName );

        //the shots acquisition should continue if
        //the previous session was not completed
        //(shots taken less than expected) and the
        //date didn't change

        int result = 0;
        if( cal.date() == as->getLastEventDate() )
        {
            result++;
        }
        if( as->getLastEventNr() < as->getShots() )
        {
            result++;
        }
        if( QFile::exists(csvAbsPath) == true )
        {
            result++;
        }
        if( as->getEraseLogs() == Qt::Unchecked && as->getEraseShots() == Qt::Unchecked )
        {
            result++;
        }

        if(result == 4)
        {
            prog = static_cast<uint>( as->getLastEventNr() );        //continues from the last event known (for automatic mode only)
            autoCount = static_cast<int>(prog);
            MYINFO << "Events progressive initialized to #" << prog << " statistic CSV opened in append";

            result = cbCSV->setFile( csvAbsPath, true );
        }
        else
        {
            prog = 0;
            autoCount = 0;
            as->setLastEventNr(0);

            newHdr = true;
            MYINFO << "Events progressive zeroed, creating a new statistic CSV";

            //creates a new CSV file
            cbCSV->unSetFile();
            result = cbCSV->setFile( csvAbsPath );
        }
        MY_ASSERT(result == 0)


        //CSV column header
        csvFields.clear();
        csvFields.append( tr("Event nr.") );                        //field #1
        csvFields.append( tr("Date") );                             //field #2
        csvFields.append( tr("UTC Time") );                         //field #3
        csvFields.append( tr("Unix Timestamp s") );                 //field #4
        csvFields.append( tr("Tune") );                             //field #5 *
        csvFields.append( tr("Bandwidth") );                        //field #6 *
        csvFields.append( tr("Central Hz") );                       //field #7 *
        csvFields.append( tr("Lowest Hz") );                        //field #8
        csvFields.append( tr("Highest Hz") );                       //field #9
        csvFields.append( tr("zoomed BW Hz") );                     //field #10
        csvFields.append( tr("step Hz") );                          //field #11
        csvFields.append( tr("Threshold mode") );                   //field #12
        csvFields.append( tr("Up threshold [dBfs]") );              //field #15
        csvFields.append( tr("Dn threshold [dBfs]") );              //field #16
        csvFields.append( tr("Auto up threshold %") );              //field #13
        csvFields.append( tr("Auto dn threshold %") );              //field #14
        csvFields.append( tr("Range low Hz") );                     //field #17
        csvFields.append( tr("Range high Hz") );                    //field #18
        csvFields.append( tr("Top peak (S) [dBfs]") );              //field #19
        csvFields.append( tr("Average peak [dBfs]") );              //field #20
        csvFields.append( tr("Average noise (N) [dBfs]") );         //field #21
        csvFields.append( tr("Difference (S-N) [dBfs]") );          //field #22
        csvFields.append( tr("Average difference (S-N) [dBfs]") );  //field #23
        csvFields.append( tr("Top peak[Hz]") );                     //field #24
        csvFields.append( tr("reserved") );                         //field #25 *  spare column for future use
        csvFields.append( tr("Std.deviation") );                    //field #26 *  last field for continuous and periodic
        csvFields.append( tr("Lasting [mS]") );                     //field #27
        csvFields.append( tr("Freq shift [Hz]") );                  //field #28
        csvFields.append( tr("Echo area [FFT points]") );           //field #29
        csvFields.append( tr("Interval area [FFT points]") );       //field #30
        csvFields.append( tr("Peaks count") );                      //field #31
        csvFields.append( tr("LOS speed [m/s]") );                  //field #32
        csvFields.append( tr("Event status") );                     //field #33
        csvFields.append( tr("Shot name") );                        //field #34

        emit run( static_cast<int>(prog), as->getShots() );
    }
    else
    {
        MYINFO << "manual acquisition stop request" ;
        if(eventDetected)
        {
            emit status( QString(tr("Acquisition will stop in some seconds, be patient...")));
        }
        else
        {
            emit status( tr("Stopping acquisition...") );
        }
        emit GUIbusy();
        manStopReq = true;
    }
}

void Control::slotControlLoop()
{
    bool peak = false;

    scanDelta = 0;
    dBfsDelta = 0;
    csvDelta = 0;
    outDelta = 0;
    deadTime = statClock.restart();
    MYDEBUG << "loop: " << loop;

    //the scanning lifesignal goes direcly to stdout
    switch( loop % 8 )
    {
    case 0:
    case 1:
        printf("\r\\");
        break;

    case 2:
    case 3:
        printf("\r|");
        break;

    case 4:
    case 5:
        printf("\r/");
        break;

    case 6:
    case 7:
        printf("\r-");
        break;

    }
    fflush(stdout);

    if(console == false)
    {
        peak = wfdBfs();
    }
    else
    {
        peak = consoleDBfs();
    }

    EvTuple et;

    if( r->isFinished() )
    {
        MYINFO << "Acquisition stopping request due to radio thread termination" ;
        autoStopReq = true;
        slotStopAcquisition(0);
    }
    else
    {
        switch (am)
        {
            case AM_CONTINUOUS:
                if (init == true)
                {
                    //forgets about the last progressive
                    prog = 0;
                    autoCount = 0;
                    as->setLastEventNr( static_cast<int>(prog) );

                    if( as->getGNUplot() == Qt::Checked )
                    {
                        //unique gnuplot dat file (power2d)
                        cbDAT->unSetFile();

                        QString absPath = wd.absoluteFilePath( datFileName );
                        cbDAT->setFile(  absPath );
                        cbDAT->setKeepMode(false);
                    }

                    //continuous acquisition (no shots) each scan line is immediately
                    //dumped to file
                    dumpScans = 1;

                    //the initialization terminates once a sufficient number of scans
                    //have been collected in the queues
                    init = false;
                }

                //immediate - not delayed
                cbDAT->dump();

                //continuous mode can be stopped only manually or from server
                if(manStopReq == true)
                {
                    slotStopAcquisition(0);
                    //return;
                }
                break;

            case AM_PERIODIC:
                //produces screenshots and/or dumps at regular intervals, that is the shorter
                //between the screenshot coverage and dump coverage
                if (init == true)
                {
                    //Starts the countdown to next shot
                    if( waitingDumps.count() == 0 )
                    {
                        prog = 0;   //forgets about last event number, these are scans now
                        autoCount = 0;
                        as->setLastEventNr(static_cast<int>(prog));
                        et.time = time(nullptr);
                        et.prog = static_cast<int>(prog);
                        waitingDumps.enqueue( et );
                        MYINFO << "init ==> Enqueued et.prog=" << et.prog << " et.time=" << et.time;
                        cbDAT->setKeepMode(true);
                    }

                    //The initialization terminates once a sufficient number of scans
                    //have been collected in the queues. After that, the program
                    //can start recording
                    init = !( scans >=  static_cast<uint>( as->getAvgdScans() ));
                }

                if( waitingDumps.count() > 0  )
                {
                    //please take in count that dumps coverage can be longer
                    //than screenshot coverage, so if both are enabled,
                    //in the screenshots could result time holes because
                    //the time between shots is the same of dumps
                    //For this reason, only the dumps queue is used here:
                    et = waitingDumps.head();
                    time_t now = time(nullptr);
                    MYDEBUG << "== Peeking et.prog=" << et.prog << " et.time=" << et.time;
                    int secToDump = static_cast<int>( as->getRecTime() - (now - et.time) );
                    MYDEBUG << "now: " << now << " seconds, next dump in: " << secToDump  << " seconds";

                    if( secToDump <= 0 )
                    {
                        if( autoCount < as->getShots() )
                        {
                            if(init == false)   //skips the first empty shot
                            {
                                autoCount++;
                                if( as->getScreenshots() == Qt::Checked )
                                {
                                    slotTakeWfAutoShot();
                                }
                            }

                            if( as->getGNUplot() == Qt::Checked )
                            {
                                //for gnuplot output only: DAT file is broken down in many dumps paired with relative screenshots
                                slotTakeDump();
                            }

                            //the prog counter in this case does not count events but scans
                            if(console == false)
                            {
                                cbCSV->dump(wfShotFileName);
                            }
                            else
                            {
                                cbCSV->dump(dumpFileName);
                            }
                            MYINFO << "Triggering time reached, taking shot#" << autoCount << " of "  << as->getShots();

                            et = waitingDumps.dequeue();
                            MYINFO << "<== Dequeued et.prog=" << et.prog << " et.time=" << et.time;

                            et.time = time(nullptr);
                            et.prog = autoCount+1;
                            waitingDumps.enqueue( et );
                            MYINFO << "==> Enqueued et.prog=" << et.prog << " et.time=" << et.time;
                        }
                        else
                        {
                            waitingDumps.dequeue();
                            MYINFO << "Stopping acquisition, max shots/dumps exceeded: (" << as->getShots() << ")." ;
                            autoStopReq = true;
                        }
                    }
                }



                //periodic mode can be stopped in any moment
                if(manStopReq == true || autoStopReq == true)
                {
                    slotStopAcquisition(0);
                    //return;
                }
                break;

            case AM_AUTOMATIC:
                if (init == true)
                {
                    if( as->getGNUplot() == Qt::Checked )
                    {
                        cbDAT->setKeepMode(true);
                    }

                    //the initialization terminates once a sufficient number of scans
                    //have been collected in the queues
                    init = !( scans >= static_cast<uint>( as->getAvgdScans() ));
                }

                if( autoStopReq == false &&
                    prog > static_cast<uint>( as->getShots() ) )
                {
                    MYINFO << "Acquisition stopping request due to  max shots exceeded: (" << as->getShots() << ")." ;
                    autoStopReq = true;
                }

                if( peak == true &&
                    autoStopReq == false &&
                    manStopReq == false )
                {
                    //new event detected (raising edge)

                    //Starts the countdown
                    //with afterPeak == 0 the trigger will be immediate
                    et.time = time(nullptr);
                    et.prog = static_cast<int>(prog);
                    waitingShots.enqueue( et );
                    waitingDumps.enqueue( et );
                    MYINFO << "Event #" << prog << " detected at scan #" << scans << " pushing countdown" ;
                    MYINFO << "Now there are still " << waitingShots.count() << " events waiting a shot" ;
                    MYINFO << "and " << waitingDumps.count() << " events waiting a dump" ;
                }


                //remember: "shots" means screenshots png while "dump" means a gnuplot dat file.

                if( waitingShots.count() > 0  )
                {
                    MYDEBUG << "events waiting for shot: " << waitingShots.count();

                    et = waitingShots.head();
                    int delaySecs = (as->getAfter() * sShotCoverage) / 100000; //from ms to secs
                    int secToShot = static_cast<int>(delaySecs - (time(nullptr) - et.time));
                    MYDEBUG << "next shot in: " << secToShot  << " seconds";

                    //waits time expiring AND event closed before triggering the shot
                    if( secToShot < 0 && eventDetected == 0 )
                    {
                        //shot time!
                        if( et.prog <= as->getShots() )
                        {
                            waitingShots.dequeue();

                            autoCount = et.prog;
                            MYINFO << "Countdown expired, triggering shot#" << autoCount << " of "  << as->getShots() ;
                            if( as->getScreenshots() == Qt::Checked )
                            {
                                slotTakeWfAutoShot();
                            }

                            //the CSV is updated simultaneously
                            if(console == false)
                            {
                                cbCSV->dump(wfShotFileName, 3);
                            }
                            else
                            {
                                cbCSV->dump(dumpFileName, 3);
                            }
                        }
                    }
                }

                if( waitingDumps.count() > 0  )
                {
                    MYDEBUG << "events waiting for dump: " << waitingDumps.count();

                    et = waitingDumps.head();
                    int delaySecs = (as->getAfter() * as->getRecTime()) / 100;
                    int secToDump = static_cast<int>(delaySecs - (time(nullptr) - et.time));
                    MYDEBUG << "next dump in: " << secToDump  << " seconds";

                    //waits time expiring AND event closed before triggering the dump
                    if( secToDump < 0 && eventDetected == 0 )
                    {
                        //dump time!
                        if( et.prog <= as->getShots() )
                        {
                            waitingDumps.dequeue();

                            autoCount = et.prog;
                            MYINFO << "Countdown expired, triggering dump#" << autoCount << " of "  << as->getShots() ;

                            if( as->getGNUplot() == Qt::Checked )
                            {
                                slotTakeDump();
                            }

                        }
                    }
                }

                //in automatic mode, the acquisition
                //waits until the pending events are closed and all the text
                //rows stored in circbufs
                //have been written on disk before stopping.

                if(autoStopReq == true || manStopReq == true)
                {

                    //wait for pending events closing

                    if( waitingShots.count() == 0 && waitingDumps.count() == 0 && eventDetected == 0 )
                    {
                        emit status( tr("Acquisition stopping...") );
                        MYINFO << "stopping radio thread to skip further events detections";
                        r->stop();
                        r->wait(10000);

                        slotStopAcquisition(0);
                        //return;
                    }


                }
                break;

            default:                
                MYCRITICAL << "Wrong parameters combination for required configuration, please fix and restart acquisition";
                autoStopReq = true;
                slotStopAcquisition(0);
                emit status( faultStatus );

                //retriggers MainWindow::slotSetAcqMode()
                am = static_cast<E_ACQ_MODE>( as->getAcqMode() );
                as->setAcqMode(am);
                break;

        }

        //checks for midnight UTC: when it comes,
        //closes the csv and continuous gnuplot files
        //then opens two brand new ones

        QTime utc = cal.currentDateTimeUtc().time();
        if(
#ifndef _MIDNIGHT_SWAP_TEST

            utc.hour()  == 0  &&
            manStopReq == false &&
#else
            manStopReq == true &&
#warning MIDNIGHT SWAP TEST ACTIVE
#endif
            logSwap == false &&
            eventDetected == 0 &&
            autoStopReq == false &&
            pp->isRunning() == false)
        {

            emit GUIbusy();

            MYINFO << "*********** MIDNIGHT UTC ******************";

            //stops acquisition
            //because all circbufs files must be closed
            //before daily archiving
            slotStopAcquisition(MIDNIGHT_SWAP); //stop because midnight swap - it must NOT repeat the archiving afterwards


            //Reinitialize UTC calendar
            //needed only for debug, when tampering
            //the system datetime. Uninfluent in
            //other cases.
            cal = QDateTime::currentDateTimeUtc();

            //there must be no files opened while archiving

            pp->setWdir(wd);
            pp->setTask( PP_MIDNIGHT );
            pp->start( QThread::IdlePriority );
            while (pp->isRunning() == false)
            {
                QThread::msleep(100);
            }
            MYINFO << "+++++ PostProc thread started";
            pp->wait();

            MYINFO << "+++++ PostProc thread terminated";

            //once postprocessing is finished, saves the counters to disk
            pp->saveData();

            //reset progressive counter
            MYINFO << "automatic progressive number zeroed (was " << prog << ")";
            prog = 0;
            as->setLastEventNr(0);
            autoCount = 0;

            //midnight swapping done
            logSwap = true;
            upDays++;

            //restarts acquisition
            slotAcquisition(true);

            emit GUIfree();

            MYINFO << "************* midnight utc swapping done ****************";
        }

#ifndef _MIDNIGHT_SWAP_TEST
        if( utc.hour() != 0  && logSwap == true )
#else
        if( manStopReq == true  && logSwap == true )
#endif
        {
            logSwap = false;
            MYINFO << "files ready for next swapping";
        }

    }

    if(autoStopReq == false)
    {
        if( totalTime > as->getInterval() && overload == false)
        {
            MYDEBUG << "last scan took " << totalTime
                << " ms with interval set to " << as->getInterval() << " ms";

            overload = true;
        }
        else if( init == true && manStopReq == false)
        {
            overload = false;
            switch(am)
            {
            case AM_CONTINUOUS:
            default:
                emit status( tr("Acquisition running continuous mode") );
                break;

            case AM_PERIODIC:
                emit status( tr("Acquisition running periodic mode") );
                break;

            case AM_AUTOMATIC:
                emit status( tr("Acquisition running automatic mode") );
                break;
            }

        }

    }

    loop++;
    qApp->processEvents();
    printStats();
    //statPlot->flush(); superfluous

    if((autoStopReq||manStopReq) &&
            samplingTimer.isActive() == false)
    {
        autoStopReq = false;
        manStopReq = false;
    }


    totalTime = deadTime + scanDelta + dBfsDelta  + csvDelta + outDelta;
    if(scanTime > 0)
    {
        avgScanTime = (scanTime+avgScanTime) / 2;
        scanTime = 0;

        if(loop % MEASURE_DELAY_LOOP == 0)
        {
            measureScanTime();
        }
    }
}


void Control::slotStopAcquisition(int what)
{   
    bool acqFault = false;

    if(what == 0)
    {
        MYINFO << "Acquisition stopped regularly" ;
    }
    else if(what == MIDNIGHT_SWAP)
    {
        MYINFO << "Acquisition stopped automatically due to midnight swap";
    }
    else
    {
        MYCRITICAL << "Acquisition stopped due to a fault detected in radio thread, what=" << what ;
        autoStopReq = true;
        acqFault = true;
    }

#ifdef _DEBUG
    printStats();
#endif

    emit status( tr("Stopping acquisition...") );

    samplingTimer.stop();
    waitingShots.clear();
    waitingDumps.clear();

    QThread::sleep(1); //waits for sampling timer stopped

    //stops radio acquisition (or test pattern generator)
    //if not yet stopped
    if(r->isRunning() == true)
    {
        r->stop();
    }

    cbCSV->unSetFile();
    cbDAT->unSetFile();

    /*
    if( as->getGNUplot() == Qt::Checked )
    {
        gpEpilogue();
        MYINFO << "gnuplot epilogue written." ;
    }
    */

    pp->stop();

#if 0
    //in debug, updates the archive
    //each time the acquisition stops

    //REMEMBER TO DELETE the DailyCounters.ds file when
    //rebuilding the archive after a flattening
    if(what != MIDNIGHT_SWAP)
    {   
        emit status( tr("Updating daily report...") );
        pp->setTask( PP_MIDNIGHT );
        pp->start( QThread::IdlePriority );
        pp->wait();
        pp->saveData();

    }
#endif

    if(autoStopReq == true)
    {
        //resets the Start pushbutton
        //as->getMainWindow()->stop();
        emit stop();
    }

    if(!acqFault)
    {
        //resets the Start pushbutton
        //as->getMainWindow()->stop();
        emit stop();
    }
    else
    {
        emit status( tr("Acquisition FAULT") );
    }
    emit GUIfree();
    qApp->processEvents();

    //forces device reopening
    scc = 1;
}



QString Control::makeFileName(const QString& prefix, const QString& ext, uint progId)
{
    QString fileName;

    if(progId == 0)
    {
        /* fileName = fileName.sprintf("%s%s%s",
            qPrintable(prefix),
            qPrintable(sessionName),
            qPrintable(ext));
         */
        fileName = QString("%1%2%3")
                .arg(prefix)
                .arg(sessionName)
                .arg(ext);

    }
    else
    {
        /* fileName = fileName.sprintf("%s%s_%05u%s",
            qPrintable(prefix),
            qPrintable(sessionName),
            progId,
            qPrintable(ext));
        */
        fileName = QString("%1%2_%3%4")
                .arg(prefix)
                .arg(sessionName)
                .arg(progId, 5, 10, QChar('0'))
                .arg(ext);



    }

    return fileName;
}


void Control::makeSessName(QString dateTime)
{
    QString ams;
    switch(am)
    {
    case AM_CONTINUOUS:
    default:
        ams = tr("continuous") ;
        break;

    case AM_PERIODIC:
        ams = tr("periodic") ;
        break;

    case AM_AUTOMATIC:
        ams = tr("automatic") ;
        break;
    }

    sessionName = QString("%1_%2_%3")
            .arg( as->getConfigName() )
            .arg( ams )
            .arg( dateTime );
}

void Control::slotSetInterval()
{
    MYDEBUG << "forwarding to slotSetRecTime()...";
    slotSetRecTime();
    slotStopTimeChange();
}


void Control::slotSetShots()
{
    slotStopTimeChange();
}


void Control::slotSetRecTime()
{
    MYDEBUG << "scans = " << dumpScans << " calling slotFFTchanged()...";

    //slotFFTchanged();
    slotStopTimeChange();
}


void Control::slotSetAfter()
{
    slotStopTimeChange();
}


void Control::slotSetUpThreshold()
{
    slotStopTimeChange();
}

void Control::slotSetDnThreshold()
{
    slotStopTimeChange();
}

void Control::slotSetTrange()
{
    tRange = as->getTrange();
    slotStopTimeChange();
}





void Control::slotTakeWfAutoShot()
{
    if(console == false && wf != nullptr)
    {
        wfShotFileName.clear();
        wfShotFileName = makeFileName(AUTOSHOT_PREFIX, SSHOT_EXT, static_cast<uint>(autoCount));
        wf->takeShot( autoCount, as->getShots(),  peakS, peakN, avgDiff, wd.absoluteFilePath(wfShotFileName) );
        emit status( tr("Automatic shot taken") );
        notify();
    }
}

void Control::slotTakeWfManShot()
{
    if(console == false && wf != nullptr)
    {
        notify();

        QString wfManShotFileName;

        manCount++;
        wfManShotFileName = makeFileName(MANSHOT_PREFIX, SSHOT_EXT, static_cast<uint>(manCount));
        wf->takeShot( manCount, 0,  0, 0, 0, wd.absoluteFilePath(wfManShotFileName) );
        emit status( tr("Manual shot taken") );
    }
}


void Control::notify()
{
    if (as->getPing() == Qt::Checked)
    {

        if(  QAudioDeviceInfo::availableDevices(QAudio::AudioOutput).isEmpty() == true )
        {
            MYWARNING << "Default sound system not available, trying alternatives";

            //alternative commands to play a wav file if Qt reference sound system is not present
            //this is a linux problem, since it presents a wide choice of audio drivers

            //It should not happen, except in case the Qt libraries have been built
            //with sound system disabled

            QString wav = ping->fileName();

    #if WANT_PULSEAUDIO

            wav = QString("/usr/bin/paplay %1/%2").arg(SOUNDS_PATH).arg(wav);
            QProcess::startDetached( wav );

    #elif WANT_SOX
            wav = QString("/usr/bin/play %1/%2").arg(SOUNDS_PATH).arg(wav);
            QProcess::startDetached( wav );
    #else
            MYDEBUG << "silently playing " << wav;
    #endif

        }
        else
        {
            ping->play();
        }
    }
}

void Control::slotTakeDump()
{
    //completes the gnuplot dump filename
    //with the autoshot number then dumps the data
    //on disk
    dumpFileName.clear();
    dumpFileName = makeFileName(gpPrefix, GP_DATA_EXT, static_cast<uint>(autoCount));
    QString absPath = wd.absoluteFilePath( dumpFileName );

    cbDAT->unSetFile();
    cbDAT->setFile( absPath );
    cbDAT->dump();

    dataDumps.append(dumpFileName);
    MYINFO << "Created dump file: " << dumpFileName;


}

void Control::slotStopTimeChange()
{
    //configuration change occurred while acquisition was stopped
    running = false;
    autoCount = 0;          //shots counter are reset
    manCount = 0;           //shots counter are reset
    scc++;
    if(console == false && wf != nullptr)
    {
        wf->slotUpdateGrid();
    }
    MYDEBUG << "scc = " << scc;
}


void Control::measureScanTime()
{
    int checkRequired = 0;
    int checkCalculated = 0;

    if(!console && as->getScreenshots() == Qt::Checked)
    {
        /*
         * The screenshot number of scans is fixed because it matches the waterfall widget's height.
         * is the time coverage that needs to be calculated at runtime (unless we're in console mode)
         */

        //required number of scans stored in the waterfall
        //(waterfall's widget height)
        int wfScans = wf->scansNumber();
        sShotCoverage = (wfScans * as->getInterval()) ;
        MYINFO << "Required screenshots coverage in ms=" << sShotCoverage << " for " << wfScans << " scans";

        if(avgScanTime > 0)
        {
             //effective number of scans stored in a shot
            sShotCoverage = static_cast<int>( wfScans * avgScanTime );
            MYINFO << "Average scan time measured: " << avgScanTime << "ms, effective screenshot coverage:" << sShotCoverage << " ms";
            checkCalculated++;
        }

        checkRequired++;
    }

    if(as->getGNUplot() == Qt::Checked)
    {
        /*
         * The GNUplot dump duration is set in mainwindow as parameter, but the number of scans
         * that a dump can store depends of the acquisition speed, that is dynamic and is
         * calculated here
         */
        //required number of scans stored in a shot
        //(GNUplot dumps)
        dumpScans = static_cast<int>( ceil( static_cast<float>(as->getRecTime() * 1000) / as->getInterval() ) );
        MYINFO << "Required dump coverage (shot duration) in seconds=" << as->getRecTime() << " for " << dumpScans << " scans";

        if(avgScanTime > 0)
        {
             //effective number of scans stored in a shot
            dumpScans = static_cast<int>( ceil( static_cast<float>(as->getRecTime() * 1000) / avgScanTime ) );
            MYINFO << "Effective scans per dump:" << dumpScans;

            cbDAT->setMax(dumpScans);
            if(am != AM_AUTOMATIC)
            {
                //in automatic mode the statistic CSV stores events data, not scans data.
                cbCSV->setMax(dumpScans);
            }

            checkCalculated++;
        }

        checkRequired++;
    }

    if(checkRequired == checkCalculated)
    {
        init = false;
    }

}

//---- end slots ---------------------


bool Control::wfdBfs()
{

    int i, scanIdx;
    double binFreq;
    float power, pixPower;
    float dBfs = 0;
    Scan* scan;
    QString str;

    MY_ASSERT(console == false);  //should never execute this method in console mode

    cal = QDateTime::currentDateTimeUtc();
    double msTimeT = (cal.toMSecsSinceEpoch() / 1000.0);
    QString ts;
    ts = ts.asprintf("%.3f", msTimeT);
    //ts = QString("%1").arg(msTimeT, 4, 'g', 3, QChar('0'));
    MYDEBUG << "wfDbfs() ts=" << ts;

    scanIdx = ps->getData();
    if(scanIdx == -1)
    {
        noDataBuf++;
        if(scanIdxBak != -1)
        {
#ifdef SCAN_REPEATING
            //data not ready: reuses the last
            //data known
            scanIdx = scanIdxBak;
            scanIdxBak = -1;
            MYDEBUG << "No data - reusing last data";
#else
            return 0;
#endif
        }
        else
        {
            //no data to show available yet
            MYDEBUG << "No data at all";
            return 0;
        }
    }
    else
    {
        MYDEBUG << "Data ok";
        //noDataBuf = 0;
    }

    //at least one data scan available
    scan = ps->getElem( scanIdx );
    MY_ASSERT( nullptr != scan )
    if(scan->getBufSize() == 0)
    {
        //invalid data, discard it
        MYINFO << "invalid data, releasing scanIdx=" << scanIdx;
        ps->release(scanIdx);
        scanIdx = -1;
        scanIdxBak = -1;
        return 0;
    }

    for(i = 1;;i++)
    {
        int toAddIdx = ps->getData();
        if(toAddIdx == -1)
        {
            //no more scans in queue
            if(i > 1)
            {
                MYDEBUG << "integration by " << i;
            }
            break;
        }

        Scan* toAdd = ps->getElem( toAddIdx );
        MY_ASSERT( nullptr != toAdd )
        if(toAdd->getBufSize() == 0 || scan->getBufSize() != toAdd->getBufSize())
        {
            //invalid data, discard it
            MYINFO << "invalid data, releasing toAddIdx=" << toAddIdx;
            ps->release(toAddIdx);
            toAddIdx = -1;
            return 0;
        }

        //integrates all the pending scans before displaying them
#ifdef INTEGRATION_BY_PEAKS
        scan->peaks( toAdd );
#else
        scan->average( toAdd );
#endif
        ps->release(toAddIdx);
    } 

    const float* scanData = nullptr;
    scanData = scan->constData();
    MY_ASSERT( nullptr !=  scanData)

    scanDelta = statClock.restart();

    //date+time used in CSV files
    dateTimeStamp   = scan->getTimeStamp();

    //separe date from time - only time is shown on waterfall
    QStringList sl = dateTimeStamp.split(CSV_SEP);
    timeStamp = sl.at(1);


    avgDbfs = MIN_DBFS;
    maxDbfs = MIN_DBFS;
    variance = 0;
    stddev = 0;

    maxFreq = 0;
    maxDiff = 0.0;

    if(
      /*
      FFTbins != s->getBufSize() ||
      as->fpCmp(binSize, s->getBinSize()) != 0 ||
      */
      as->fpCmpf(binSize, as->getResolution(), 5) ||
      wfZoom != as->getZoom() ||
      wfBw != as->getBandwidth() ||
      wfOffset != as->getOffset() ||
      wfTune != as->getTune() ||
      waterfallChange == true ||
      restart == true)
    {
        if ( as->fpCmpf( scan->getBinSize(), as->getResolution(), 5 ) != 0 )
        {
            //the buffer size must match the resolution chosen
            //otherwise it means it's an old buffer that cannot
            //be displayed anymore
            MYDEBUG << __func__ << " discarding old buffer, each bin is "
                   << scan->getBinSize()
                   << " Hz wide, while actual resolution is "
                   <<  as->getResolution() << " Hz";

            if(scanIdxBak != -1)
            {
                //releases the backupped data and
                //backs up the current data
                MYDEBUG << "releasing avgIdxBak=" << scanIdxBak;
                ps->release(scanIdxBak);
            }

            scanIdxBak = scanIdx;

            return false;
        }

        MYINFO << __func__ << " runtime change ";
        //runtime change
        FFTbins = scan->getBufSize();
        binSize = scan->getBinSize();
        wfZoom = as->getZoom();
        wfBw = as->getBandwidth();
        wfOffset = as->getOffset();
        wfTune = as->getTune();

        loggedBins =  static_cast<int>(wfBw / binSize);

        //retrieves the waterfall window pointer
        //and updates the waterfall pane width in pixel
        //that is mandatory to calculate correctly the bpp
        if (wf == nullptr)
        {
            wf = as->getWaterfallWindow();
            wf->slotUpdateGrid();
        }

        bpp = getBpp( loggedBins );
        hpp = bpp * binSize;
        MYINFO << "wfBw=" << wfBw << " bpp=" << bpp << " binSize=" << binSize
               << " hpp=" << hpp << " FFTbins=" << FFTbins ;



        //when scan starts or any upper slider changes on waterfall
        //the loggedbins and firstloggedbins are recalculated
        float err = as->getError();
        float fracErr = 0.0;
        float fracPpmToHz = 0.0;
        float intErr = 0.0; //not used

        //the fractional part of ppm error is used to apply a transparent offset
        //to the waterfall, in order to correct the frequency more accurately
        //than dongle does
        fracErr = modf(err, &intErr);

        //converts the fractional ppm error to Hz
        fracPpmToHz = (as->getTune() / 1000000) * fracErr;

        //recalculate the band's start and the end frequencies
        int rawFromHz = r->getLowBound();
        fromHz = static_cast<int>(rawFromHz+fracPpmToHz);

        int rawToHz = r->getHighBound();
        toHz = static_cast<int>(rawToHz+fracPpmToHz);

        //peak detection range (tRange) from % to Hz:
        as->getZoomedRange( zoomLo, zoomHi );
        int interval = 0;

        int zoomRange = (zoomHi - zoomLo);
        interval = (zoomRange * as->getTrange()) / 100;

        MYINFO << "raw: fromHz=" << rawFromHz << "  toHz=" << rawToHz;
        MYINFO << "fracErr=" << fracErr << " fracPpmToHz=" << fracPpmToHz;
        MYINFO << "corrected: fromHz=" << fromHz << " toHz=" << toHz;
        MYINFO << "zoomLo=" << zoomLo << " zoomHi=" << zoomHi << " zoomRange=" << zoomRange;

        rangeLo = zoomLo + (zoomRange / 2) - (interval / 2);
        rangeHi = rangeLo + interval;
        MYINFO << "range " << as->getTrange() << " % = " << rangeLo << "..." << rangeHi << " = " << interval << " Hz ";

        //zoom is in _tenths_ of hz
        int binsInZoom = static_cast<int>( ((10*wfBw) / wfZoom) / binSize );
        firstLoggedBin = static_cast<int>( (zoomLo - fromHz) / binSize );
        if(firstLoggedBin < 0)
        {
            firstLoggedBin = 0;
        }

        lastLoggedBin = firstLoggedBin+binsInZoom-1;
        if( as->getGNUplot() == Qt::Checked )
        {
            cbDAT->setMax(dumpScans);
            cbDAT->commit();
            cbDAT->dump();
        }

        MYINFO << "binsInZoom=" << binsInZoom << " firstLoggedBin=" << firstLoggedBin
               << " lastLoggedBin=" << lastLoggedBin
               << " loggedBins=" << loggedBins;

        //testMode =( r->getDevices().count() > 0 ) ?  false : true;
        eventDetected = 0;

        if(wf != nullptr)
        {
            //If widget's width is smaller than loggedBins
            //the dBfs value must be the average of N adjacent bins.
            // In the opposite case, when width
            //is larger than loggedBins, the scaling is purely graphical

            wf->setDetectionRange(rangeLo, rangeHi);
            wf->slotUpdateGrid();
        }

        centerFreq = (zoomLo + (zoomHi-zoomLo) / 2);
        waterfallChange = false;
        restart = false;
    }

    if(newHdr == true)
    {
        MYINFO << "Appending new header in CSV";
        QString header = csvFields.join(CSV_SEP);
        cbCSV->insert( qPrintable(header) );
        cbCSV->dump();
        newHdr = false;
    }

    //the log files record all the logged bins, while the
    //waterfall widget shows only the portion (bandwidth and offset)
    //required


    int valuesFixed = 0;
    int notchIndex = -1;
    int l = firstLoggedBin;
    int avgBins = 0;
    bool notched = false;
    bool inPeak = false;
    float denom = 0.0;
    float preRem = 0.0;
    float postRem = 0.0;
    float pixelCount = 0.0;
    float repRem = 0.0;
    float lastDiff = 0.0;

    HzTuple ht = {0,0,0,0};

    Notch n;
    QList<Notch> notches = as->getNotches();

    binFreq = zoomLo;
    power = 0;
    pixPower = 0;

    if(notches.count() > 0)
    {
        //one or more notches are defined
        notchIndex = 0;
        n = notches.at(notchIndex);
    }

    //manages output to waterfall widget. Since X resolution [pixels] is fixed,
    //I must reduce the output bins. I.E. if bins resolution is 32768 and waterfall
    //width is 2048 pixels, each waterfall pixel must show the _average_ of 16 bins:
    //(smart scaling with bpp = 16)
    //note that bpp is often a real number, with integer+fractional parts

    for ( i = firstLoggedBin ; i < lastLoggedBin; )
    {
        postRem = 0.0;
        denom = 0.0;
        power = 0.0;
        pixPower = 0.0;

        if(notchIndex != -1)
        {
            if(binFreq >= n.begin &&
               notched == false)
            {
                //entering in a notched band
                notched = true;
            }
            else if(binFreq > n.end &&
               notched == true)
            {
                //exiting a notched band
                notched = false;
                notchIndex++;
                if(notchIndex < notches.count())
                {
                    n = notches.at(notchIndex);
                }
                else
                {
                    //end of notches
                    notchIndex = -1;
                }

            }
        }


        if(preRem > 0.0F)
        {
            //remainder of previous pixel
            power = scanData[l] * preRem;

            denom = preRem;
            i++;
            l++;
            if(i > loggedBins)
            {
                //end of scan line, cannot compute the right value for this
                //pixel: set it to average value
                if(wf != nullptr)
                {
                    if(notched == true)
                    {
                        avgDbfs = avgN;
                    }
                    wf->slotAppendPixel(avgDbfs);
                }
                goto deepBreak;
            }

        }

        if( as->fpCmpf(denom,0.0) == 0 || (denom + 1.0F ) < bpp )
        {
            //power of integer bins is summed
            do
            {
                power = power + scanData[l];
                denom = denom + 1.0F;
                i++;
                l++;
                if(i > loggedBins)
                {
                    //end of scan line, cannot compute the right value for this
                    //pixel: set it to average value

                    if(notched == true)
                    {
                        avgDbfs = avgN;
                    }

                    if(wf != nullptr)
                    {
                        wf->slotAppendPixel(avgDbfs);
                    }

                    goto deepBreak;
                }
            }
            while ((denom + 1.0F) < bpp);

        }


        if(bpp < 1.0F)
        {
            //less than one bin per pixel: the same pixel value is redrawn until denom == bpp
            //the following pixel have the same linear power


            if (powerToDbfs( power, dBfs ) == false)
            {
                //fixes the data with default value
                dBfs = avgN;
                valuesFixed++;
                //fixDone = true;
                MYDEBUG << "\aERROR: dBfs is NaN or infinite, forcing dBfs to " << dBfs ;
            }

            do
            {
                pixelCount = pixelCount + 1.0F;      //progressive scan pixels count

                if(notched == true)
                {
                    dBfs = avgN;
                }
                if(wf != nullptr)
                {
                    wf->slotAppendPixel(dBfs);
                }
                binFreq = binFreq + static_cast<double>(hpp);

                repRem += bpp;
            }
            while (repRem < 1.0F);

            repRem -= 1.0F;
        }
        else
        {

            //one or more bins per pixel: the resulting pixel is the average of bins

            postRem = bpp - denom;
            power = power + scanData[l] * postRem;
            denom = denom + postRem;
            //MY_ASSERT(  as->fpCmpf(denom, bpp) == 0 )

            preRem = 1.0F - postRem;
            if(preRem > 1.0F)
            {
                //bppFrac is 0.0
                preRem = 0.0F;
                i++;
                l++;
            }

            //now denom must be == bpp, saying how many bins are stored in a pixel
            pixPower = (power / denom);         //average linear power

            pixelCount = pixelCount + 1.0F;      //progressive scan pixels count

            if ( as->fpCmpf(pixPower, 0.0F, 7) == 0 )
            {
                //avoids infinites
                 dBfs = avgN;
                 valuesFixed++;
                 //fixDone = true;
                 MYDEBUG << "\aERROR: dBfs is infinite, forcing dBfs to " << dBfs ;
            }
            else
            {
                //pixPower = power / static_cast<float>( FFTbins );
                if (powerToDbfs( power, dBfs ) == false)
                {
                    //fixes the data with default value
                    dBfs = avgN;
                    valuesFixed++;
                    //fixDone = true;
                    MYDEBUG << "\aERROR: dBfs is NaN or infinite, forcing dBfs to " << dBfs ;
                }
            }


            //composing a waterfall scan pixel by pixel
            if(notched == true)
            {
                dBfs = avgN;
            }
            if(wf != nullptr)
            {
                wf->slotAppendPixel(dBfs);
            }
            binFreq = binFreq + static_cast<double>(hpp);

        }

deepBreak:

        if((autoStopReq == true || manStopReq == true) && lasting.elapsed() > (as->getOverdenseSec() * 1000))
        {
            //in automatic mode, if a carrier has been detected as an echo, the capturing
            //mode can be interrupted by pressing stop and wait until the maximum time allowed for an
            //overdense event expires. After that limit, each bin is set to avgN causing a
            //forced event close
            dBfs = avgN;
        }


        if(newHdr == false && scanIdxBak != -1) //avoids to append duplicate records
        {
            if( as->getGNUplot() == Qt::Checked &&
                    (gpPrefix == SPECMAP_PREFIX || gpPrefix == SPEC3D_PREFIX))
            {
                if(pixelCount < wfWidth)
                {
                    //skips the last record because it will be written after calculating
                    //the min max and average dBfs to mark the end of scan line
                    cbDAT->append("%s    %i   %.2f\n", qPrintable(ts), static_cast<int>(binFreq), static_cast<double>(dBfs));
                }
            }

        }

        int bf = static_cast<int>(binFreq);
        if( bf < zoomLo || bf > zoomHi )
        {
            MYDEBUG << "BUG: lo: " << zoomLo << " hi:" << zoomHi << " binFreq:" << bf;
            break;
        }

        //statistics

        //the average noise is calculated on each entire scan
        //as a running average
        //after summing all the bins displayed
        avgDbfs = avgDbfs + dBfs;

        avgDiff = avgS - avgN;

        avgBins++;

        //the peaks and the S value are calculated only inside the detection range
        //the same for stddev
        if(bf >= rangeLo && bf <= rangeHi)
        {
            variance = variance + powf((dBfs - avgN), 2);

            if( dBfs > maxDbfs )
            {
                maxFreq = static_cast<int>(binFreq);
                maxDbfs = dBfs;
            }     

            if( dBfs < minDbfs )
            {
                minDbfs = dBfs;
            }

            //detects the first and the last frequency in the scan that crosses
            //the lower threshold. Considering these pixels for the contiguous scans
            //interested by the event, it will be possible to calculate the number of
            //bins (FFT points) involved (Area)

            float diff = 0;
            if(as->getThresholdsBehavior() == TB_ABSOLUTE)
            {
                diff = dBfs;    //absolute thresholds
            }
            else if(as->getThresholdsBehavior() == TB_DIFFERENTIAL)
            {
                diff = dBfs - avgN;     //difference against fixed thresholds
            }
            else if (as->getThresholdsBehavior() == TB_AUTOMATIC)
            {
                diff = dBfs - avgN;     //difference against automatic thresholds
            }

            //takes the first lower threshold positive crossing in this scan
            if (diff > autoDnThr &&
                lastDiff <= autoDnThr &&
                ht.hzFrom == 0)
            {
                ht.hzFrom = static_cast<int>(binFreq);
                ht.hzPeak = 0;
                ht.hzTo = 0;
                ht.count = 0;
                inPeak = false;
            }

            //counts each time the signal reaches a peak above the upper
            //threshold in this scan
            if (diff > autoUpThr &&
                ht.hzFrom > 0 &&
                inPeak == false)
            {
                //upper threshold positive crossing, increments the counter only
                ht.count++;
                inPeak = true;
            }

            if(inPeak == true)
            {
                if(diff < autoUpThr)
                {
                    //upper threshold negative crossing
                    inPeak = false;
                }

                //Temporary maxDiff, calculated on the average noise of the previous scan.
                maxDiff = (maxDbfs - avgN);

                if(diff >= maxDiff)
                {
                    //stores the frequency of the peak
                    ht.hzPeak = static_cast<int>(binFreq);
                }
            }

            //takes the last lower threshold negative crossing in this scan
            if (diff <= autoDnThr &&
                lastDiff > autoDnThr &&
                ht.hzFrom > 0)
            {
                ht.hzTo = static_cast<int>(binFreq);
            }
            lastDiff = diff;
        }

        if(ht.hzFrom != 0 && ht.hzTo <= ht.hzFrom)
        {
            //the signal is still high at the end of range, force closing
             ht.hzTo = rangeHi;
        }
    }

    scans++;
    scanTime = scanTimer.restart();
    dBfsDelta = statClock.restart();

    //only at the end of the scan it's possible to determinate
    //the average N then the maximum S-N difference
    //and be sure the threshold has been crossed

    //the N value calculated on this scan is inserted in a FIFO
    avgDbfs /= avgBins;
    Nfifo.enqueue(avgDbfs);
    if( Nfifo.size() > as->getAvgdScans() )
    {
        //discards the oldest element to keep the
        //FIFO length constant.
        Nfifo.dequeue();
        if(enableCapture == false &&
                static_cast<int>(avgN) < as->getNoiseLimit())
        {
            MYINFO << "Average N value available =" << avgN << " dBfs, value alloweed for capturing";
            enableCapture = true;
        }
        else if(
                am == AM_AUTOMATIC &&
                as->getNoiseLimit() < 0 &&
                static_cast<int>(avgN) > as->getNoiseLimit() &&
                enableCapture == true
                )
        {
            MYINFO << "Noise level " << avgN << "dBfs exceeds the maximum level alloweed for capturing" << as->getNoiseLimit() << " dBfs";
            MYWARNING << "Capture inhibited until the noise level will fall below the limit";
            emit status( tr("Band too noisy, capture inhibited") );
            enableCapture = false;
            eventDetected = 0;
        }
    }

    variance /= (loggedBins-1);
    stddev = sqrt(variance);
    MYDEBUG << "avgN=" << avgN << " maxDbfs=" << maxDbfs << " minDbfs=" << minDbfs << " variance=" << variance << " stddev=" << stddev;

    maxDiff = maxDbfs - avgN;           //canonical S-N calculated inside the peak detection range


    //the S values are averaged too, these values
    //are used in case of automatic thresholds
    Sfifo.enqueue(maxDbfs);
    if( Sfifo.size() > as->getAvgdScans() )
    {
        //discards the oldest element to keep the
        //FIFO length constant.
        Sfifo.dequeue();
    }

    bool raisingEdge = false;

    if( init == false )
    {
        //doesn't record anything during initialization

        //updating CSV scan file
        if(am == AM_CONTINUOUS)
        {
            //continuous mode: the <prog> counter counts the scans.
            //one CSV row is produced at each scan
            if(newHdr == false)
            {
                //in continuous mode, the prog counts the scans, not the events
                prog++;

                //prints the statistic values at each scan
                //note that floats are localized, in order to use the local decimal separator
                str = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                        .arg(CSV_SEP)
                        .arg(prog)
                        .arg(dateTimeStamp)
                        .arg(qPrintable(ts))
                        .arg(as->getTune())
                        .arg(as->getBandwidth())
                        .arg(centerFreq)
                        .arg(zoomLo)
                        .arg(zoomHi)
                        .arg(zoomHi-zoomLo)
                        .arg(static_cast<double>(binSize));


                str += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                        .arg(CSV_SEP)
                        .arg("-")
                        .arg(0)
                        .arg(0)
                        .arg(static_cast<double>(as->getUpThreshold()))
                        .arg(static_cast<double>(as->getDnThreshold()))
                        .arg(rangeLo)
                        .arg(rangeHi);

                str += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                        .arg(CSV_SEP)
                        .arg(static_cast<double>(maxDbfs))
                        .arg(static_cast<double>(avgS))
                        .arg(static_cast<double>(avgN))
                        .arg(static_cast<double>(maxDiff))
                        .arg(static_cast<double>(avgDiff))
                        .arg(maxFreq)
                        .arg(0)
                        .arg(stddev);

                //fields for automatic mode only
                str += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                        .arg(CSV_SEP)
                        .arg(0)
                        .arg(0)
                        .arg(0)
                        .arg(0)
                        .arg(0)
                        .arg(0)
                        .arg("-");

                cbCSV->insert( qPrintable(str) );

                if(prog % 30 == 0)
                {
                    //dumps the data on file every 30 scans
                    cbCSV->dump();
                }
            }
        }
        else if ( am == AM_PERIODIC )
        {
static int contaggio = 0;
            //in periodic mode, the prog counts the scans, not the events.
            //the CSV record when the shot is made must carry its filename
            prog++;

            peakN = avgN;
            peakS = maxDbfs;
            peakDiff = (peakS - peakN);
            peakFreq = maxFreq;

            //prints the statistic values at each scan
            //note that floats are localized, in order to use the local decimal separator
            str = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                    .arg(CSV_SEP)
                    .arg(prog)
                    .arg(dateTimeStamp)
                    .arg(qPrintable(ts))
                    .arg(as->getTune())
                    .arg(as->getBandwidth())
                    .arg(centerFreq)
                    .arg(zoomLo)
                    .arg(zoomHi)
                    .arg(zoomHi-zoomLo)
                    .arg(static_cast<double>(binSize));

            str += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                    .arg(CSV_SEP)
                    .arg("-")
                    .arg(0)
                    .arg(0)
                    .arg(static_cast<double>(as->getUpThreshold()))
                    .arg(static_cast<double>(as->getDnThreshold()))
                    .arg(rangeLo)
                    .arg(rangeHi);

            str += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                    .arg(CSV_SEP)
                    .arg(static_cast<double>(maxDbfs))
                    .arg(static_cast<double>(avgS))
                    .arg(static_cast<double>(avgN))
                    .arg(static_cast<double>(maxDiff))
                    .arg(static_cast<double>(avgDiff))
                    .arg(maxFreq)
                    .arg(0)
                    .arg(stddev);

            //fields for automatic mode only
            str += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                    .arg(CSV_SEP)
                    .arg(0)
                    .arg(0)
                    .arg(0)
                    .arg(0)
                    .arg(0)
                    .arg(0)
                    .arg("-");


            cbCSV->insert( qPrintable(str) );
contaggio++;
        }

        //event detection raising edge (automatic mode)
        //peaks that cross the upper threshold
        //are considered only within the detection range
        if( am == AM_AUTOMATIC &&
            enableCapture == true  &&
            scanIdxBak != -1 &&
            init == false ) //avoids first false event at acquisition start
        {

            if( eventDetected == 0 &&
                (
                    ( as->getThresholdsBehavior() == TB_DIFFERENTIAL && maxDiff > autoUpThr ) ||
                    ( as->getThresholdsBehavior() == TB_AUTOMATIC && maxDiff > autoUpThr) ||
                    ( as->getThresholdsBehavior() == TB_ABSOLUTE && maxDbfs > autoUpThr )
                )
              )
            {
                prog++;
                MYWARNING << "Event " << prog << " started at " << maxFreq << " Hz";
                MYINFO << "maxDbfs (S)=" << maxDbfs << " maxDiff (S-N)=" << maxDiff;
                eventBorder.clear();
                eventBorder.append(ht);
                peakStr.clear();
                eventDetected = maxFreq;
                lasting.start();

                hzShift = centerFreq - maxFreq;

                if(newHdr == false)
                {
                    //prints the statistic values at raising edge
                    str = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                            .arg(CSV_SEP)
                            .arg(prog)
                            .arg(dateTimeStamp)
                            .arg(qPrintable(ts))
                            .arg(as->getTune())
                            .arg(as->getBandwidth())
                            .arg(centerFreq)
                            .arg(zoomLo)
                            .arg(zoomHi)
                            .arg(zoomHi-zoomLo)
                            .arg(static_cast<double>(binSize));


                    str += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(as->thresholdBehaviorString())
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoUpThr) : static_cast<double>(as->getUpThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoDnThr) : static_cast<double>(as->getDnThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getUpThreshold()) : 0.0)
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getDnThreshold()) : 0.0)
                            .arg(rangeLo)
                            .arg(rangeHi);

                    str += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                            .arg(CSV_SEP)
                            .arg(static_cast<double>(maxDbfs))
                            .arg(static_cast<double>(avgS))
                            .arg(static_cast<double>(avgN))
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) )  //when TB_ABSOLUTE is set, the difference is not reported
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(avgDiff) )  //when TB_ABSOLUTE is set, the average difference is not reported
                             .arg(maxFreq)
                            .arg(0)
                            .arg(stddev);


                    //fields for automatic mode only
                    str += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(0)
                            .arg(hzShift)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg("Raise");

                    cbCSV->insert( qPrintable(str) );

                    //by default, the next record (the peak) is initialized as the raising edge.
                    //if a higher peak will be detected later, the peak record will be updated
                    //before being written on disk

                    //initialize event stats
                    peakN = avgN;
                    peakS = maxDbfs;
                    peakDiff = (peakS - peakN);
                    peakFreq = maxFreq;

                    peakStr = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                            .arg(CSV_SEP)
                            .arg(prog)
                            .arg(dateTimeStamp)
                            .arg(qPrintable(ts))
                            .arg(as->getTune())
                            .arg(as->getBandwidth())
                            .arg(centerFreq)
                            .arg(zoomLo)
                            .arg(zoomHi)
                            .arg(zoomHi-zoomLo)
                            .arg(static_cast<double>(binSize));


                    peakStr += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(as->thresholdBehaviorString())
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoUpThr) : static_cast<double>(as->getUpThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoDnThr) : static_cast<double>(as->getDnThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getUpThreshold()) : 0.0)
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getDnThreshold()) : 0.0)
                            .arg(rangeLo)
                            .arg(rangeHi);

                    peakStr += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                            .arg(CSV_SEP)
                            .arg(static_cast<double>(peakS))
                            .arg(static_cast<double>(avgS))
                            .arg(static_cast<double>(peakN))
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) )  //when TB_ABSOLUTE is set, the difference is not reported
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(avgDiff) )  //when TB_ABSOLUTE is set, the average difference is not reported
                             .arg(peakFreq)
                            .arg(0)
                            .arg(stddev);


                    //fields for automatic mode only
                    peakStr += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(0)
                            .arg(hzShift)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg("Peak");


                    hzShift = 0;
                    raisingEdge = true;

                }
            }
            else if( eventDetected != 0 &&
                     manStopReq == false &&
                     autoStopReq == false &&
                     (
                        ( as->getThresholdsBehavior() == TB_DIFFERENTIAL && maxDiff > autoDnThr ) ||                       
                        ( as->getThresholdsBehavior() == TB_AUTOMATIC && maxDiff > autoDnThr ) ||
                        ( as->getThresholdsBehavior() == TB_ABSOLUTE && maxDbfs > autoDnThr )
                     )
                   )

            {
                //event running, continue detection of contour
                MYDEBUG << "Event " << prog << " continues...";

                if( eventDetected < 0 )
                {
                    //this event is joined to a previous peak
                    //so I must discard the latest 2 records in circbuf
                    //(the peak and fall)
                    //and continue the detection
                    cbCSV->rollback(2);
                    eventDetected = maxFreq;
                    MYINFO << "Joining to event " << prog;
                }

                eventBorder.append(ht);
                hzShift = centerFreq - maxFreq;

                //update event stats
                if(maxDiff >= peakDiff)
                {
                    peakN = avgN;
                    peakS = maxDbfs;
                    peakDiff = (peakS - peakN);
                    peakFreq = maxFreq;


                    int msLasting = lasting.elapsed();

                    peakStr = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                            .arg(CSV_SEP)
                            .arg(prog)
                            .arg(dateTimeStamp)
                            .arg(qPrintable(ts))
                            .arg(as->getTune())
                            .arg(as->getBandwidth())
                            .arg(centerFreq)
                            .arg(zoomLo)
                            .arg(zoomHi)
                            .arg(zoomHi-zoomLo)
                            .arg(static_cast<double>(binSize));

                    peakStr += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(as->thresholdBehaviorString())
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoUpThr) : static_cast<double>(as->getUpThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoDnThr) : static_cast<double>(as->getDnThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getUpThreshold()) : 0.0)
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getDnThreshold()) : 0.0)
                            .arg(rangeLo)
                            .arg(rangeHi);

                    peakStr += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                            .arg(CSV_SEP)
                            .arg(static_cast<double>(peakS))
                            .arg(static_cast<double>(avgS))
                            .arg(static_cast<double>(peakN))
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) )  //when TB_ABSOLUTE is set, the difference is not reported
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(avgDiff) )  //when TB_ABSOLUTE is set, the average difference is not reported
                             .arg(peakFreq)
                            .arg(0)
                            .arg(stddev);

                    //fields for automatic mode only
                    peakStr += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(msLasting)
                            .arg(hzShift)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg("Peak");

                }

            }
            else if( eventDetected > 0 &&
                     (
                        ( as->getThresholdsBehavior() == TB_DIFFERENTIAL && maxDiff <= autoDnThr ) ||
                        ( as->getThresholdsBehavior() == TB_AUTOMATIC && maxDiff <= autoDnThr ) ||
                        ( as->getThresholdsBehavior() == TB_ABSOLUTE && maxDbfs <= autoDnThr )
                     )
                   )
            {
                //event detection falling edge:
                //when maxDiff falls below the lower threshold
                //after a peak has been detected
                MYWARNING << "Event " << prog << " stopped at " << maxFreq << " Hz";
                eventDetected = -maxFreq;
                int msLasting = lasting.elapsed();

                //The doppler calculation assumes that the central frequency in waterfall
                //is the same of transmitter.
                //TODO: the calculation must keep in count the position of the
                //trasmitter vs. receiver
                hzShift = centerFreq - maxFreq;
                const int C = SPEED_OF_LIGHT;
                float doppler = (hzShift * C) / centerFreq;


                eventBorder.append(ht);

                //calculates area as nr. of bins interested in the echo
                int echoArea = 0;

                //and the entire interval area, a rectangle covering
                //the interval range x event lasting
                int intervalArea = 0;
                int peaksCount = 0;
                foreach(ht, eventBorder)
                {
                    if(ht.hzTo == 0)
                    {
                        ht.hzTo = ht.hzPeak;
                    }
                    float hzCoverage = ht.hzTo - ht.hzFrom;
                    if(hzCoverage < 0)
                    {
                        MYDEBUG << "--------BAD echoArea -------------";
                        MYDEBUG << "ht.hzFrom: " << ht.hzFrom;
                        MYDEBUG << "ht.hzPeak: " << ht.hzPeak;
                        MYDEBUG << "ht.hzTo: " << ht.hzTo;
                        MYDEBUG << "ht.count: " << ht.count;
                    }
                    else if(hzCoverage == 0)
                    {
                        //this happens when hzTo == hzFrom
                        //a correction is needed:
                        hzCoverage = 1;
                    }

                    echoArea += (hzCoverage / binSize)  ;
                    MYDEBUG << "hzCoverage = " << hzCoverage << " Hz, echoArea = " << echoArea << " FFT points (bins)";
                    int hzFull = (rangeHi - rangeLo);
                    if(hzFull < 0)
                    {
                        MYDEBUG << "--------BAD intervalArea -------------";
                        MYDEBUG << "ht.hzFrom: " << ht.hzFrom;
                        MYDEBUG << "ht.hzPeak: " << ht.hzPeak;
                        MYDEBUG << "ht.hzTo: " << ht.hzTo;
                        MYDEBUG << "ht.count: " << ht.count;
                    }
                    intervalArea += (hzFull / binSize);
                    MYDEBUG << "hzFull = " << hzFull << "Hz, intervalArea = " << intervalArea << " FFT points (bins)";
                    peaksCount += ht.count;
                }
                MYINFO << "Resulting echo area: " << echoArea << " of " << intervalArea << " [points]";

                if(newHdr == false &&
                   scanIdxBak != -1 )
                {
                    //prints the statistic values at event peak
                    cbCSV->insert( qPrintable(peakStr) );

                    //prints the statistic values at falling edge
                    str = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                            .arg(CSV_SEP)
                            .arg(prog)
                            .arg(dateTimeStamp)
                            .arg(qPrintable(ts))
                            .arg(as->getTune())
                            .arg(as->getBandwidth())
                            .arg(centerFreq)
                            .arg(zoomLo)
                            .arg(zoomHi)
                            .arg(zoomHi-zoomLo)
                            .arg(static_cast<double>(binSize));

                    str += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(as->thresholdBehaviorString())
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoUpThr) : static_cast<double>(as->getUpThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoDnThr) : static_cast<double>(as->getDnThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getUpThreshold()) : 0.0)
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getDnThreshold()) : 0.0)
                            .arg(rangeLo)
                            .arg(rangeHi);

                    str += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                            .arg(CSV_SEP)
                            .arg(static_cast<double>(maxDbfs))
                            .arg(static_cast<double>(avgS))
                            .arg(static_cast<double>(avgN))
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) )  //when TB_ABSOLUTE is set, the difference is not reported
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(avgDiff) )  //when TB_ABSOLUTE is set, the average difference is not reported
                            .arg(maxFreq)
                            .arg(0)
                            .arg(stddev);

                    //fields for automatic mode only
                    str += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(msLasting)
                            .arg(hzShift)
                            .arg(echoArea)
                            .arg(intervalArea)
                            .arg(peaksCount)
                            .arg(static_cast<double>(doppler))
                            .arg("Fall");


                    cbCSV->insert( qPrintable(str) );
                    //note: the CSV on disk is updated once a day
                    //and flushed at acq stop

                }

                //must wait some time before declaring the event as completely closed
                joinTime.restart();
            }
            else if(eventDetected < 0 && joinTime.elapsed() > as->getJoinTime())
            {
                //last peak event completely closed
                MYWARNING << "Event# " << prog << " closed";
                eventDetected = 0;
                eventBorder.clear();

                //the event progressive is saved as local setting
                as->setLastEventNr( static_cast<int>(prog) );

            }
            else
            {
                if( as->getThresholdsBehavior() == TB_AUTOMATIC )
                {
                    MYDEBUG << "no event pending, recalculate automatic thresholds";
                    autoUpThr = avgDiff + ((avgDiff * as->getUpThreshold()) / 100.0F);
                    autoDnThr = avgDiff + ((avgDiff * as->getDnThreshold()) / 100.0F);
                }
            }
        }

    }

    if(wf != nullptr)
    {
        wf->slotPlotPower(maxDbfs, avgN, avgDiff, raisingEdge);          //update power/time graph
        wf->slotRefresh( timeStamp, raisingEdge );          //update freq/power graph
        wf->setCapturing(!enableCapture, eventDetected);    //update the capturing / too noisy labels
        wf->setScreenshotCoverage(sShotCoverage / 1000);    //update the screenshot coverage label
    }

    csvDelta = statClock.restart();

    if(newHdr == false && scanIdxBak != -1) //avoids to append duplicate records
    {
        if( as->getGNUplot() == Qt::Checked )
        {
            //end of scan line
            if( gpPrefix == POW2D_PREFIX )
            {
                cbDAT->insert( "%s    %.2f   %.2f   %.2f", qPrintable(ts), static_cast<double>(avgN), static_cast<double>(maxDbfs),
                               (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) );
                                //when TB_ABSOLUTE is set, the difference is not reported

                //cbDAT->dump();
                 //cbDAT->commit();
            }
            else if( gpPrefix == SPECMAP_PREFIX || gpPrefix == SPEC3D_PREFIX )
            {
                cbDAT->append( "%s    %i   %.2f    %s   %.2f   %.2f   %.2f\n", qPrintable(ts),  static_cast<int>(binFreq), static_cast<double>(dBfs),
                               qPrintable(ts), static_cast<double>(avgN), static_cast<double>(maxDbfs),
                               (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) );
                                //when TB_ABSOLUTE is set, the difference is not reported
                cbDAT->append("");   //end of scan line (for surface plots) is marked with an empty line
                cbDAT->commit();
            }
            else qFatal("no gnuplot format defined");
        }
    }

    scan = nullptr;
    scanData = nullptr;

    //the averaged N value to be used in next scan
    //is calculated by averaging all the
    //values stored in FIFO
    avgN = avgDbfs;
    foreach (avgDbfs, Nfifo)
    {
        avgN += avgDbfs;
    }
    avgN /= (Nfifo.count()+1);

    //the averaged S value to be used in next scan
    //is calculated by averaging all the
    //values stored in FIFO
    avgS = maxDbfs;
    foreach (maxDbfs, Sfifo)
    {
        avgS += maxDbfs;
    }
    avgS /= (Sfifo.count()+1);


    if(scanIdxBak != -1)
    {
        //releases the backupped data and
        //backs up the current data
        MYDEBUG << "releasing avgIdxBak=" << scanIdxBak;
        ps->release(scanIdxBak);
    }

    scanIdxBak = scanIdx;

    outDelta = statClock.restart();

    return raisingEdge;
}


bool Control::consoleDBfs()
{

    int i, scanIdx;
    double binFreq;
    float power;
    float dBfs = 0;
    Scan* scan;
    QString str;

    cal = QDateTime::currentDateTimeUtc();
    double msTimeT = (cal.toMSecsSinceEpoch() / 1000.0);
    QString ts;
    ts = ts.asprintf("%.3f", msTimeT);
    //ts = QString("%1").arg(msTimeT, 4, 'g', 3, QChar('0'));
    MYDEBUG << "wfDbfs() ts=" << ts;

    scanIdx = ps->getData();
    if(scanIdx == -1)
    {
        noDataBuf++;
        if(scanIdxBak != -1)
        {
#ifdef SCAN_REPEATING
            //data not ready: reuses the last
            //data known
            scanIdx = scanIdxBak;
            scanIdxBak = -1;
            MYDEBUG << "No data - reusing last data";
#else
            return 0;
#endif
        }
        else
        {
            //no data to show available yet
            MYDEBUG << "No data at all";
            return 0;
        }
    }
    else
    {
        MYDEBUG << "Data ok";
        //noDataBuf = 0;
    }

    //at least one data scan available
    scan = ps->getElem( scanIdx );
    MY_ASSERT( nullptr != scan )
    if(scan->getBufSize() == 0)
    {
        //invalid data, discard it
        MYINFO << "invalid data, releasing scanIdx=" << scanIdx;
        ps->release(scanIdx);
        scanIdx = -1;
        scanIdxBak = -1;
        return 0;
    }

    for(i = 1;;i++)
    {
        int toAddIdx = ps->getData();
        if(toAddIdx == -1)
        {
            //no more scans in queue
            if(i > 1)
            {
                MYDEBUG << "integration by " << i;
            }
            break;
        }

        Scan* toAdd = ps->getElem( toAddIdx );
        MY_ASSERT( nullptr != toAdd )
        if(toAdd->getBufSize() == 0 || scan->getBufSize() != toAdd->getBufSize())
        {
            //invalid data, discard it
            MYINFO << "invalid data, releasing toAddIdx=" << toAddIdx;
            ps->release(toAddIdx);
            toAddIdx = -1;
            return 0;
        }

        //integrates all the pending scans before displaying them
#ifdef INTEGRATION_BY_PEAKS
        scan->peaks( toAdd );
#else
        scan->average( toAdd );
#endif
        ps->release(toAddIdx);
    }

    const float* scanData = nullptr;
    scanData = scan->constData();
    MY_ASSERT( nullptr !=  scanData)

    scanDelta = statClock.restart();

    //date+time used in CSV files
    dateTimeStamp   = scan->getTimeStamp();

    //separe date from time - only time is shown on waterfall
    QStringList sl = dateTimeStamp.split(CSV_SEP);
    timeStamp = sl.at(1);


    avgDbfs = MIN_DBFS;
    maxDbfs = MIN_DBFS;
    variance = 0;
    stddev = 0;


    maxFreq = 0;
    maxDiff = 0.0;
    if( init == true && restart == true )
    {
        if ( as->fpCmpf( scan->getBinSize(), as->getResolution(), 5 ) != 0 )
        {
            //the buffer size must match the resolution chosen
            //otherwise it means it's an old buffer that cannot
            //be processed anymore
            MYDEBUG << __func__ << " discarding old buffer, each bin is "
                   << scan->getBinSize()
                   << " Hz wide, while actual resolution is "
                   <<  as->getResolution() << " Hz";

            if(scanIdxBak != -1)
            {
                //releases the backupped data and
                //backs up the current data
                MYDEBUG << "releasing avgIdxBak=" << scanIdxBak;
                ps->release(scanIdxBak);
            }

            scanIdxBak = scanIdx;

            return false;
        }

        MYINFO << __func__ << " initializing variables";
        //runtime change
        FFTbins = scan->getBufSize();
        binSize = scan->getBinSize();
        wfZoom = as->getZoom();
        wfBw = as->getBandwidth();
        wfOffset = as->getOffset();
        wfTune = as->getTune();

        loggedBins =  static_cast<int>(wfBw / binSize);

        MYINFO << "wfBw=" << wfBw << " binSize=" << binSize
                << " FFTbins=" << FFTbins ;

        float err = as->getError();
        float fracErr = 0.0;
        float fracPpmToHz = 0.0;
        float intErr = 0.0;

        //the fractional part of ppm error is used to apply a transparent offset
        //to the waterfall, in order to correct the frequency more accurately
        //than dongle does
        fracErr = modf(err, &intErr);

        //converts the fractional ppm error to Hz, intErr is not used
        fracPpmToHz = (as->getTune() / 1000000) * fracErr;

        //recalculate the band's start and the end frequencies
        int rawFromHz = r->getLowBound();
        fromHz = static_cast<int>(rawFromHz+fracPpmToHz);

        int rawToHz = r->getHighBound();
        toHz = static_cast<int>(rawToHz+fracPpmToHz);

        //peak detection range (tRange) from % to Hz:
        as->getZoomedRange( zoomLo, zoomHi );
        int interval = 0;

        int zoomRange = (zoomHi - zoomLo);
        interval = (zoomRange * as->getTrange()) / 100;

        MYINFO << "raw: fromHz=" << rawFromHz << "  toHz=" << rawToHz;
        MYINFO << "fracErr=" << fracErr << " fracPpmToHz=" << fracPpmToHz;
        MYINFO << "corrected: fromHz=" << fromHz << " toHz=" << toHz;
        MYINFO << "zoomLo=" << zoomLo << " zoomHi=" << zoomHi << " zoomRange=" << zoomRange;

        rangeLo = zoomLo + (zoomRange / 2) - (interval / 2);
        rangeHi = rangeLo + interval;
        MYINFO << "range " << as->getTrange() << " % = " << rangeLo << "..." << rangeHi << " = " << interval << " Hz ";


        //zoom is in _tenths_ of hz
        int binsInZoom = static_cast<int>( ((10*wfBw) / wfZoom) / binSize );
        firstLoggedBin = static_cast<int>( (zoomLo - fromHz) / binSize );
        if(firstLoggedBin < 0)
        {
            firstLoggedBin = 0;
        }

        lastLoggedBin = firstLoggedBin+binsInZoom-1;
        if( as->getGNUplot() == Qt::Checked )
        {
            cbDAT->setMax(dumpScans);
            cbDAT->commit();
            cbDAT->dump();
        }

        MYINFO << "binsInZoom=" << binsInZoom << " firstLoggedBin=" << firstLoggedBin
               << " lastLoggedBin=" << lastLoggedBin
               << " loggedBins=" << loggedBins;

        centerFreq = (zoomLo + (zoomHi-zoomLo) / 2);
        MYINFO << "Central freq=" << centerFreq << " Hz";
        eventDetected = 0;
        restart = false;
    }

    if(newHdr == true)
    {
        MYINFO << "Appending new header in CSV";
        QString header = csvFields.join(CSV_SEP);
        cbCSV->insert( qPrintable(header) );
        cbCSV->dump();
        newHdr = false;
    }


    //the log files don't care of pixels

    int valuesFixed = 0;
    int notchIndex = -1;
    int avgBins = 0;
    bool notched = false;
    bool inPeak = false;
    float lastDiff = 0.0;
    HzTuple ht = {0,0,0,0};

    Notch n;
    QList<Notch> notches = as->getNotches();

    binFreq = zoomLo;
    power = 0;

    if(notches.count() > 0)
    {
        //one or more notches are defined
        notchIndex = 0;
        n = notches.at(notchIndex);
    }


    for ( i = firstLoggedBin ; i < lastLoggedBin; i++ )
    {
        power = 0.0;

        if(notchIndex != -1)
        {
            if(binFreq >= n.begin &&
               notched == false)
            {
                //entering in a notched band
                notched = true;
            }
            else if(binFreq > n.end &&
               notched == true)
            {
                //exiting a notched band
                notched = false;
                notchIndex++;
                if(notchIndex < notches.count())
                {
                    n = notches.at(notchIndex);
                }
                else
                {
                    //end of notches
                    notchIndex = -1;
                }

            }
        }
        power = scanData[i];

        if(i > loggedBins)
        {
            //end of scan line
            break;
        }

        if(notched == true)
        {
            avgDbfs = avgN;
        }


        if (powerToDbfs( power, dBfs ) == false)
        {
            //fixes the data with default value
            dBfs = avgN;
            valuesFixed++;
            MYDEBUG << "\aERROR: dBfs is NaN or infinite, forcing dBfs to " << dBfs ;
        }

        if(notched == true)
        {
            dBfs = avgN;
        }
        binFreq = binFreq + static_cast<double>( as->getResolution() );


        if((autoStopReq == true || manStopReq == true) && lasting.elapsed() > (as->getOverdenseSec() * 1000))
        {
            //in automatic mode, if a carrier has been detected as an echo, the capturing
            //mode can be interrupted by pressing stop and wait until the maximum time allowed for an
            //overdense event expires. After that limit, each bin is set to avgN causing a
            //forced event close
            dBfs = avgN;
        }

        if(newHdr == false && scanIdxBak != -1) //avoids to append duplicate records
        {
            if( as->getGNUplot() == Qt::Checked &&
                    (gpPrefix == SPECMAP_PREFIX || gpPrefix == SPEC3D_PREFIX))
            {
                if(i < lastLoggedBin)
                {
                    //skips the last record because it will be written later, after calculating
                    //the min max and average dBfs to mark the end of scan line
                    cbDAT->append("%s    %i   %.2f\n", qPrintable(ts), static_cast<int>(binFreq), static_cast<double>(dBfs));
                }
            }

        }

        int bf = static_cast<int>(binFreq);
        if( bf < zoomLo || bf > zoomHi )
        {
            MYDEBUG << "BUG: lo: " << zoomLo << " hi:" << zoomHi << " binFreq:" << bf;
            break;
        }

        //statistics

        //the average noise is calculated on each entire scan
        //as a running average
        //after summing all the bins displayed
        avgDbfs = avgDbfs + dBfs;

        avgDiff = avgS - avgN;

        avgBins++;

        //the peaks and the S value are calculated only inside the detection range
        //the same for stddev
        if(bf >= rangeLo && bf <= rangeHi)
        {
            MYDEBUG << "rangeLo=" << rangeLo << " rangeHi=" << rangeHi << " bf=" << bf << "Hz";
            variance = variance + powf((dBfs - avgN), 2);

            if( dBfs > maxDbfs )
            {
                maxFreq = static_cast<int>(binFreq);
                maxDbfs = dBfs;
            }

            if( dBfs < minDbfs )
            {
                minDbfs = dBfs;
            }

            //detects the first and the last frequency in the scan that crosses
            //the lower threshold. Considering these pixels for the contiguous scans
            //interested by the event, it will be possible to calculate the number of
            //bins involved (Area)

            float diff = 0;
            if(as->getThresholdsBehavior() == TB_ABSOLUTE)
            {
                diff = dBfs;    //absolute thresholds
            }
            else if(as->getThresholdsBehavior() == TB_DIFFERENTIAL)
            {
                diff = dBfs - avgN;     //difference against fixed thresholds
            }
            else if (as->getThresholdsBehavior() == TB_AUTOMATIC)
            {
                diff = dBfs - avgN;     //difference against automatic thresholds

            }

            //takes the first lower threshold positive crossing in this scan
            if (diff > autoDnThr &&
                lastDiff <= autoDnThr &&
                ht.hzFrom == 0)
            {
                ht.hzFrom = static_cast<int>(binFreq);
                ht.hzPeak = 0;
                ht.hzTo = 0;
                ht.count = 0;
                inPeak = false;
            }

            //counts each time the signal reaches a peak above the upper
            //threshold in this scan
            if (diff > autoUpThr &&
                ht.hzFrom > 0 &&
                inPeak == false)
            {
                //upper threshold positive crossing, increments the counter only
                ht.count++;
                inPeak = true;
            }

            if(inPeak == true)
            {
                if(diff < autoUpThr)
                {
                    //upper threshold negative crossing
                    inPeak = false;
                }

                //Temporary maxDiff, calculated on the average noise of the previous scan.
                maxDiff = (maxDbfs - avgN);

                if(diff >= maxDiff)
                {
                    //stores the frequency of the peak
                    ht.hzPeak = static_cast<int>(binFreq);
                }
            }

            //takes the last lower threshold negative crossing in this scan
            if (diff <= autoDnThr &&
                lastDiff > autoDnThr &&
                ht.hzFrom > 0)
            {
                ht.hzTo = static_cast<int>(binFreq);
            }
            lastDiff = diff;
        }

        if(ht.hzFrom != 0 && ht.hzTo <= ht.hzFrom)
        {
            //the signal is still high at the end of range, force closing
             ht.hzTo = rangeHi;
        }

    }  //end of for() loop

    scans++;
    scanTime = scanTimer.restart();
    dBfsDelta = statClock.restart();
    //only at the end of the scan it's possible to determinate
    //the average N then the maximum S-N difference
    //and be sure the threshold has been crossed

    //the N value calculated on this scan is inserted in a FIFO
    avgDbfs /= avgBins;
    Nfifo.enqueue(avgDbfs);
    if( Nfifo.size() > as->getAvgdScans() )
    {
        //discards the oldest element to keep the
        //FIFO length constant.
        Nfifo.dequeue();
        if(enableCapture == false &&
                static_cast<int>(avgN) < as->getNoiseLimit())
        {
            MYINFO << "Average N value available =" << avgN << " dBfs, value alloweed for capturing";
            enableCapture = true;
        }
        else if(
                am == AM_AUTOMATIC &&
                as->getNoiseLimit() < 0 &&
                static_cast<int>(avgN) > as->getNoiseLimit() &&
                enableCapture == true
                )
        {
            MYINFO << "Noise level " << avgN << "dBfs exceeds the maximum level alloweed for capturing" << as->getNoiseLimit() << " dBfs";
            MYWARNING << "Capture inhibited until the noise level will fall below the limit";
            emit status( tr("Band too noisy, capture inhibited") );
            enableCapture = false;
            eventDetected = 0;
        }
    }

    variance /= (loggedBins-1);
    stddev = sqrt(variance);

    maxDiff = maxDbfs - avgN;           //canonical S-N calculated inside the peak detection range

    MYINFO << "avgN=" << avgN << " maxDbfs=" << maxDbfs << " maxDiff=" << maxDiff << " maxFreq=" << maxFreq ;

              //" minDbfs=" << minDbfs << " variance=" << variance << " stddev=" << stddev ;

    //the S values are averaged too, these values
    //are used in case of automatic thresholds
    Sfifo.enqueue(maxDbfs);
    if( Sfifo.size() > as->getAvgdScans() )
    {
        //discards the oldest element to keep the
        //FIFO length constant.
        Sfifo.dequeue();
    }

    bool raisingEdge = false;

    if( init == false )
    {
        //doesn't record anything during initialization

        //updating CSV scan file
        if(am == AM_CONTINUOUS)
        {
            //continuous mode: the <prog> counter counts the scans.
            //one CSV row is produced at each scan
            if(newHdr == false)
            {
                //in continuous mode, the prog counts the scans, not the events
                //despite the "eventNr" in the header
                prog++;

                //prints the statistic values at each scan
                //note that floats are localized, in order to use the local decimal separator
                str = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                        .arg(CSV_SEP)
                        .arg(prog)
                        .arg(dateTimeStamp)
                        .arg(qPrintable(ts))
                        .arg(as->getTune())
                        .arg(as->getBandwidth())
                        .arg(centerFreq)
                        .arg(zoomLo)
                        .arg(zoomHi)
                        .arg(zoomHi-zoomLo)
                        .arg(static_cast<double>(binSize));


                str += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                        .arg(CSV_SEP)
                        .arg("-")
                        .arg(0)
                        .arg(0)
                        .arg(static_cast<double>(as->getUpThreshold()))
                        .arg(static_cast<double>(as->getDnThreshold()))
                        .arg(rangeLo)
                        .arg(rangeHi);

                str += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                        .arg(CSV_SEP)
                        .arg(static_cast<double>(maxDbfs))
                        .arg(static_cast<double>(avgS))
                        .arg(static_cast<double>(avgN))
                        .arg(static_cast<double>(maxDiff))
                        .arg(static_cast<double>(avgDiff))
                        .arg(maxFreq)
                        .arg(0)
                        .arg(stddev);

                //fields for automatic mode only
                str += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                        .arg(CSV_SEP)
                        .arg(0)
                        .arg(0)
                        .arg(0)
                        .arg(0)
                        .arg(0)
                        .arg(0)
                        .arg("-");

                cbCSV->insert( qPrintable(str) );

                if(prog % 30 == 0)
                {
                    //dumps the data on file every 30 scans
                    cbCSV->dump();
                }
            }
        }
        else if ( am == AM_PERIODIC )
        {
            //in periodic mode, the prog counts the scans, not the events.
            //the CSV record when the shot is made must carry its filename
            prog++;

            peakN = avgN;
            peakS = maxDbfs;
            peakDiff = (peakS - peakN);
            peakFreq = maxFreq;

            //prints the statistic values at each scan
            //note that floats are localized, in order to use the local decimal separator
            str = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                    .arg(CSV_SEP)
                    .arg(prog)
                    .arg(dateTimeStamp)
                    .arg(qPrintable(ts))
                    .arg(as->getTune())
                    .arg(as->getBandwidth())
                    .arg(centerFreq)
                    .arg(zoomLo)
                    .arg(zoomHi)
                    .arg(zoomHi-zoomLo)
                    .arg(static_cast<double>(binSize));

            str += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                    .arg(CSV_SEP)
                    .arg("-")
                    .arg(0)
                    .arg(0)
                    .arg(static_cast<double>(as->getUpThreshold()))
                    .arg(static_cast<double>(as->getDnThreshold()))
                    .arg(rangeLo)
                    .arg(rangeHi);

            str += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                    .arg(CSV_SEP)
                    .arg(static_cast<double>(maxDbfs))
                    .arg(static_cast<double>(avgS))
                    .arg(static_cast<double>(avgN))
                    .arg(static_cast<double>(maxDiff))
                    .arg(static_cast<double>(avgDiff))
                    .arg(maxFreq)
                    .arg(0)
                    .arg(stddev);

            //fields for automatic mode only
            str += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                    .arg(CSV_SEP)
                    .arg(0)
                    .arg(0)
                    .arg(0)
                    .arg(0)
                    .arg(0)
                    .arg(0)
                    .arg("-");


            cbCSV->insert( qPrintable(str) );
        }

        //event detection raising edge (automatic mode)
        //peaks that cross the upper threshold
        //are considered only within the detection range
        if( am == AM_AUTOMATIC &&
            enableCapture == true  &&
            scanIdxBak != -1 &&
            init == false ) //avoids first false event at acquisition start
        {

            if( eventDetected == 0 &&
                (
                    ( as->getThresholdsBehavior() == TB_DIFFERENTIAL && maxDiff > autoUpThr ) ||
                    ( as->getThresholdsBehavior() == TB_AUTOMATIC && maxDiff > autoUpThr) ||
                    ( as->getThresholdsBehavior() == TB_ABSOLUTE && maxDbfs > autoUpThr )
                )
              )
            {
                prog++;
                MYWARNING << "Event " << prog << " started at " << maxFreq << " Hz";
                MYINFO << "maxDbfs (S)=" << maxDbfs << " maxDiff (S-N)=" << maxDiff << " autoUpThr=" << autoUpThr;
                eventBorder.clear();
                eventBorder.append(ht);
                peakStr.clear();
                eventDetected = maxFreq;
                lasting.start();

                hzShift = centerFreq - maxFreq;

                if(newHdr == false)
                {
                    //prints the statistic values at raising edge
                    str = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                            .arg(CSV_SEP)
                            .arg(prog)
                            .arg(dateTimeStamp)
                            .arg(qPrintable(ts))
                            .arg(as->getTune())
                            .arg(as->getBandwidth())
                            .arg(centerFreq)
                            .arg(zoomLo)
                            .arg(zoomHi)
                            .arg(zoomHi-zoomLo)
                            .arg(static_cast<double>(binSize));


                    str += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(as->thresholdBehaviorString())
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoUpThr) : static_cast<double>(as->getUpThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoDnThr) : static_cast<double>(as->getDnThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getUpThreshold()) : 0.0)
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getDnThreshold()) : 0.0)
                            .arg(rangeLo)
                            .arg(rangeHi);


                    str += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                            .arg(CSV_SEP)
                            .arg(static_cast<double>(maxDbfs))
                            .arg(static_cast<double>(avgS))
                            .arg(static_cast<double>(avgN))
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) )  //when TB_ABSOLUTE is set, the difference is not reported
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(avgDiff) )  //when TB_ABSOLUTE is set, the average difference is not reported
                             .arg(maxFreq)
                            .arg(0)
                            .arg(stddev);

                    //fields for automatic mode only
                    str += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(0)
                            .arg(hzShift)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg("Raise");

                    cbCSV->insert( qPrintable(str) );

                    //by default, the next record (the peak) is initialized as the raising edge.
                    //if a higher peak will be detected later, the peak record will be updated
                    //before being written on disk

                    //initialize event stats
                    peakN = avgN;
                    peakS = maxDbfs;
                    peakDiff = (peakS - peakN);
                    peakFreq = maxFreq;

                    peakStr = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                            .arg(CSV_SEP)
                            .arg(prog)
                            .arg(dateTimeStamp)
                            .arg(qPrintable(ts))
                            .arg(as->getTune())
                            .arg(as->getBandwidth())
                            .arg(centerFreq)
                            .arg(zoomLo)
                            .arg(zoomHi)
                            .arg(zoomHi-zoomLo)
                            .arg(static_cast<double>(binSize));


                    peakStr += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(as->thresholdBehaviorString())
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoUpThr) : static_cast<double>(as->getUpThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoDnThr) : static_cast<double>(as->getDnThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getUpThreshold()) : 0.0)
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getDnThreshold()) : 0.0)
                            .arg(rangeLo)
                            .arg(rangeHi);


                    peakStr += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                            .arg(CSV_SEP)
                            .arg(static_cast<double>(peakS))
                            .arg(static_cast<double>(avgS))
                            .arg(static_cast<double>(peakN))
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) )  //when TB_ABSOLUTE is set, the difference is not reported
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(avgDiff) )  //when TB_ABSOLUTE is set, the average difference is not reported
                             .arg(peakFreq)
                            .arg(0)
                            .arg(stddev);


                    //fields for automatic mode only
                    peakStr += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(0)
                            .arg(hzShift)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg("Peak");

                    raisingEdge = true;

                }
            }
            else if( eventDetected != 0 &&
                     manStopReq == false &&
                     autoStopReq == false &&
                     (
                        ( as->getThresholdsBehavior() == TB_DIFFERENTIAL && maxDiff > autoDnThr ) ||
                        ( as->getThresholdsBehavior() == TB_AUTOMATIC && maxDiff > autoDnThr ) ||
                        ( as->getThresholdsBehavior() == TB_ABSOLUTE && maxDbfs > autoDnThr )
                     )
                   )

            {
                //event running, continue detection of contour
                MYDEBUG << "Event " << prog << " continues...";
                MYINFO << "maxDbfs (S)=" << maxDbfs << " maxDiff (S-N)=" << maxDiff << " autoUpThr=" << autoUpThr;
                if( eventDetected < 0 )
                {
                    //this event is joined to a previous peak
                    //so I must discard the latest 2 records in circbuf
                    //(the peak and fall)
                    //and continue the detection
                    cbCSV->rollback(2);
                    eventDetected = maxFreq;
                    MYINFO << "Joining to event " << prog;
                }

                eventBorder.append(ht);
                hzShift = centerFreq - maxFreq;

                //update event stats
                if(maxDiff >= peakDiff)
                {
                    peakN = avgN;
                    peakS = maxDbfs;
                    peakDiff = (peakS - peakN);
                    peakFreq = maxFreq;


                    int msLasting = lasting.elapsed();

                    peakStr =QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                            .arg(CSV_SEP)
                            .arg(prog)
                            .arg(dateTimeStamp)
                            .arg(qPrintable(ts))
                            .arg(as->getTune())
                            .arg(as->getBandwidth())
                            .arg(centerFreq)
                            .arg(zoomLo)
                            .arg(zoomHi)
                            .arg(zoomHi-zoomLo)
                            .arg(static_cast<double>(binSize));

                    peakStr += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(as->thresholdBehaviorString())
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoUpThr) : static_cast<double>(as->getUpThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoDnThr) : static_cast<double>(as->getDnThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getUpThreshold()) : 0.0)
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getDnThreshold()) : 0.0)
                            .arg(rangeLo)
                            .arg(rangeHi);

                    peakStr += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                            .arg(CSV_SEP)
                            .arg(static_cast<double>(peakS))
                            .arg(static_cast<double>(avgS))
                            .arg(static_cast<double>(peakN))
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) )  //when TB_ABSOLUTE is set, the difference is not reported
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(avgDiff) )  //when TB_ABSOLUTE is set, the average difference is not reported
                             .arg(peakFreq)
                            .arg(0)
                            .arg(stddev);

                    //fields for automatic mode only
                    peakStr += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(msLasting)
                            .arg(hzShift)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg(0)
                            .arg("Peak");

                }

            }
            else if( eventDetected > 0 &&
                     (
                        ( as->getThresholdsBehavior() == TB_DIFFERENTIAL && maxDiff <= autoDnThr ) ||
                        ( as->getThresholdsBehavior() == TB_AUTOMATIC && maxDiff <= autoDnThr ) ||
                        ( as->getThresholdsBehavior() == TB_ABSOLUTE && maxDbfs <= autoDnThr )
                     )
                   )
            {
                //event detection falling edge:
                //when maxDiff falls below the lower threshold
                //after a peak has been detected
                MYWARNING << "Event " << prog << " stopped at " << maxFreq << " Hz";
                eventDetected = -maxFreq;
                int msLasting = lasting.elapsed();

                //The doppler calculation assumes that the central frequency in waterfall
                //is the same of transmitter.
                //TODO: the calculation must keep in count the position of the
                //trasmitter vs. receiver

                hzShift = centerFreq - maxFreq;
                const int C = SPEED_OF_LIGHT;
                float doppler = (hzShift * C) / centerFreq;

                eventBorder.append(ht);

                //calculates area as nr. of bins interested in the echo
                int echoArea = 0;

                //and the entire interval area, a rectangle covering
                //the interval range x event lasting
                int intervalArea = 0;
                int peaksCount = 0;
                foreach(ht, eventBorder)
                {
                    if(ht.hzTo == 0)
                    {
                        ht.hzTo = ht.hzPeak;
                    }
                    float hzCoverage = ht.hzTo - ht.hzFrom;
                    if(hzCoverage < 0)
                    {
                        MYDEBUG << "--------BAD echoArea -------------";
                        MYDEBUG << "ht.hzFrom: " << ht.hzFrom;
                        MYDEBUG << "ht.hzPeak: " << ht.hzPeak;
                        MYDEBUG << "ht.hzTo: " << ht.hzTo;
                        MYDEBUG << "ht.count: " << ht.count;
                    }
                    else if(hzCoverage == 0)
                    {
                        //this happens when hzTo == hzFrom
                        //a correction is needed:
                        hzCoverage = 1;
                    }

                    echoArea += (hzCoverage / binSize)  ;
                    MYDEBUG << "hzCoverage = " << hzCoverage << " Hz, echoArea = " << echoArea << " FFT points (bins)";
                    int hzFull = (rangeHi - rangeLo);
                    if(hzFull < 0)
                    {
                        MYDEBUG << "--------BAD intervalArea -------------";
                        MYDEBUG << "ht.hzFrom: " << ht.hzFrom;
                        MYDEBUG << "ht.hzPeak: " << ht.hzPeak;
                        MYDEBUG << "ht.hzTo: " << ht.hzTo;
                        MYDEBUG << "ht.count: " << ht.count;
                    }
                    intervalArea += (hzFull / binSize);
                    MYDEBUG << "hzFull = " << hzFull << "Hz, intervalArea = " << intervalArea << " FFT points (bins)";
                    peaksCount += ht.count;
                }
                MYINFO << "Resulting echo area: " << echoArea << " of " << intervalArea << " [points]";

                if(newHdr == false &&
                   scanIdxBak != -1 )
                {
                    //prints the statistic values at event peak
                    cbCSV->insert( qPrintable(peakStr) );

                    //prints the statistic values at falling edge
                    str = QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1%9%1%10%1%L11%1")
                            .arg(CSV_SEP)
                            .arg(prog)
                            .arg(dateTimeStamp)
                            .arg(qPrintable(ts))
                            .arg(as->getTune())
                            .arg(as->getBandwidth())
                            .arg(centerFreq)
                            .arg(zoomLo)
                            .arg(zoomHi)
                            .arg(zoomHi-zoomLo)
                            .arg(static_cast<double>(binSize));

                    str += QString("%2%1%L3%1%L4%1%L5%1%L6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(as->thresholdBehaviorString())
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoUpThr) : static_cast<double>(as->getUpThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(autoDnThr) : static_cast<double>(as->getDnThreshold()))
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getUpThreshold()) : 0.0)
                            .arg(as->getThresholdsBehavior() == TB_AUTOMATIC ? static_cast<double>(as->getDnThreshold()) : 0.0)
                            .arg(rangeLo)
                            .arg(rangeHi);

                    str += QString("%L2%1%L3%1%L4%1%L5%1%L6%1%7%1%L8%1%L9%1")
                            .arg(CSV_SEP)
                            .arg(static_cast<double>(maxDbfs))
                            .arg(static_cast<double>(avgS))
                            .arg(static_cast<double>(avgN))
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) )  //when TB_ABSOLUTE is set, the difference is not reported
                            .arg( (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(avgDiff) )  //when TB_ABSOLUTE is set, the average difference is not reported
                            .arg(maxFreq)
                            .arg(0)
                            .arg(stddev);

                    //fields for automatic mode only
                    str += QString("%2%1%3%1%4%1%5%1%6%1%7%1%8%1")
                            .arg(CSV_SEP)
                            .arg(msLasting)
                            .arg(hzShift)
                            .arg(echoArea)
                            .arg(intervalArea)
                            .arg(peaksCount)
                            .arg(static_cast<double>(doppler))
                            .arg("Fall");


                    cbCSV->insert( qPrintable(str) );
                    //note: the CSV on disk is updated once a day
                    //and flushed at acq stop

                }

                //must wait some time before declaring the event as completely closed
                joinTime.restart();
            }
            else if(eventDetected < 0 && joinTime.elapsed() > as->getJoinTime())
            {
                //last peak event completely closed
                MYWARNING << "Event# " << prog << " closed";
                eventDetected = 0;
                eventBorder.clear();

                //the event progressive is saved as local setting
                as->setLastEventNr( static_cast<int>(prog) );

            }
            else
            {

                if( as->getThresholdsBehavior() == TB_AUTOMATIC )
                {
                    MYDEBUG << "no event pending, recalculate automatic thresholds";
                    autoUpThr = avgDiff + ((avgDiff * as->getUpThreshold()) / 100.0F);
                    autoDnThr = avgDiff + ((avgDiff * as->getDnThreshold()) / 100.0F);
                }
            }

        }

    }

    csvDelta = statClock.restart();

    if(newHdr == false && scanIdxBak != -1) //avoids to append duplicate records
    {
        if( as->getGNUplot() == Qt::Checked )
        {
            //end of scan line
            if( gpPrefix == POW2D_PREFIX )
            {
                cbDAT->insert( "%s    %.2f   %.2f   %.2f", qPrintable(ts), static_cast<double>(avgN), static_cast<double>(maxDbfs),
                               (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) );
                                //when TB_ABSOLUTE is set, the difference is not reported

                //cbDAT->dump();
                 //cbDAT->commit();
            }
            else if( gpPrefix == SPECMAP_PREFIX || gpPrefix == SPEC3D_PREFIX )
            {
                cbDAT->append( "%s    %i   %.2f    %s   %.2f   %.2f   %.2f\n", qPrintable(ts),  static_cast<int>(binFreq), static_cast<double>(dBfs),
                               qPrintable(ts), static_cast<double>(avgN), static_cast<double>(maxDbfs),
                               (as->getThresholdsBehavior() == TB_ABSOLUTE)? 0.0 : static_cast<double>(maxDiff) );
                                //when TB_ABSOLUTE is set, the difference is not reported
                cbDAT->append("");   //end of scan line (for surface plots) is marked with an empty line
                cbDAT->commit();
            }
            else qFatal("no gnuplot format defined");
        }
    }

    scan = nullptr;
    scanData = nullptr;

    //the averaged N value to be used in next scan
    //is calculated by averaging all the
    //values stored in FIFO
    avgN = avgDbfs;
    foreach (avgDbfs, Nfifo)
    {

        avgN += avgDbfs;

    }
    avgN /= (Nfifo.count()+1);

    //the averaged S value to be used in next scan
    //is calculated by averaging all the
    //values stored in FIFO
    avgS = maxDbfs;
    foreach (maxDbfs, Sfifo)
    {
        avgS += maxDbfs;
    }
    avgS /= (Sfifo.count()+1);


    if(scanIdxBak != -1)
    {
        //releases the backupped data and
        //backs up the current data
        MYDEBUG << "releasing avgIdxBak=" << scanIdxBak;
        ps->release(scanIdxBak);
    }

    scanIdxBak = scanIdx;
    outDelta = statClock.restart();

    return raisingEdge;
}




bool Control::powerToDbfs(float pixPower, float& dBfs)
{
    bool result = true;

    dBfs  = 10 * log10(pixPower / r->getDClinearPower());
    //dBfs  =  pixPower;

    if(std::isnormal(dBfs) == 0)
    {
        if(std::isnan(dBfs) != 0)
        {
            nans++;
            result = false;
        }
        if(std::isfinite(dBfs) == 0)
        {
            infinite++;
            result = false;
        }
    }
    return result;
}

int Control::roundToInt(float n)
{
    float m = 0;
    float frac = modf(n, &m);

    if(fabs(frac) < 0.50F)
    {
        return static_cast<int>( floor( n ) );
    }
    else
    {
        return static_cast<int>( ceil( n ) );
    }

}


bool Control::gpPrologue()
{

    QFile f( wd.absoluteFilePath(cmdFileName) );
    if( f.open(QIODevice::WriteOnly) == false )
    {
        MYCRITICAL << "error opening " << cmdFileName << " for writing" ;
        return false;
    }

    QTextStream gp(&f);

    cal = QDateTime::currentDateTimeUtc();

    //common part
    gp << "#GNUPLOT command file automatically generated by Echoes " << APP_VERSION << " on " << cal.toString() << endl;
    gp << "#set terminal " << GNUPLOT_TERM << " size 1024,768 font \"fixed,10\" " << endl;
    gp << "set termoption noenhanced" << endl;
    gp << "set border 4095" << endl;
    gp << "set ytics out" << endl;
    gp << "set mxtics 10" << endl;
    gp << "set mytics 10" << endl;
    gp << "set mztics 10" << endl;
    gp << "set ztics out" << endl;

    if(gpPrefix == SPECMAP_PREFIX)
    {
        //2D color-mapped spectra: each scan is composed by multiple rows, terminated with
        //the average (N) maximum (S) and difference (S-N)
        gp << "set title \"Spectrogram  UTC: " << cal.toString(Qt::SystemLocaleShortDate) << " \" offset 0,3" << endl;
        gp << "set xlabel \"Hz\" font \"fixed,16\" offset 0,-2" << endl;
        gp << "set ylabel \"UTC Hour\"  font \"fixed,16\"" << endl;
        gp << "set key left outside" << endl;
        gp << "set grid xtics ytics ztics" << endl;
        gp << "set cblabel \"dBfs\" font \"fixed,16\" offset 1,0" << endl;
        gp << "set xtics out rotate border offset 0,-1 format \"%.0f\" " << endl;
        gp << "unset surface" << endl;
        gp << "set view 0,0" << endl;
        gp << "set pm3d at s map" << endl;
        gp << "set format y \"%H:%M:%S\" timedate" << endl;
        gp << "set cbrange [" << MIN_DBFS << ":" << MAX_DBFS << "]" << endl;
    }
    else if(gpPrefix == SPEC3D_PREFIX)
    {
        //3D-spectra: file format is identical to SPECMAP but power is plotted
        //on Z-axis with colored dots
        gp << "set title \"3D Spectrogram  UTC: " << cal.toString(Qt::SystemLocaleShortDate) << " \" offset 0,3" << endl;
        gp << "set xlabel \"Hz\" font \"fixed,16\" offset 0,-2" << endl;
        gp << "set ylabel \"UTC Hour\"  font \"fixed,16\"" << endl;
        gp << "set zlabel \"dBfs\" font \"fixed,16\" " << endl;
        gp << "set key left outside" << endl;
        gp << "set grid xtics ytics ztics" << endl;
        gp << "set cblabel \"dBfs\" font \"fixed,16\" offset 1,0" << endl;
        gp << "set xtics out rotate border offset 0,-1 format \"%.0f\" " << endl;
        gp << "unset surface" << endl << endl;
        gp << "set view 45,45, 1,1" << endl;
        gp << "set pm3d implicit at s" << endl << endl;
        gp << "set format y \"%H:%M:%S\" timedate" << endl;
        gp << "set cbrange [" << MIN_DBFS << ":" << MAX_DBFS << "]" << endl;
    }
    else if(gpPrefix == POW2D_PREFIX)
    {
        //2D-power: the average (N) maximum (S) and difference (S-N) one row per scan
        gp << "set title \"Power plot, UTC: " << cal.toString(Qt::SystemLocaleShortDate) << " \" offset 0,3" << endl;
        gp << "set xlabel \"UTC Hour\" font \"fixed,16\" offset 0,-2" << endl;
        gp << "set ylabel \"dBfs\" font \"fixed,16\" offset 0,-2" << endl;
        gp << "set grid xtics ytics" << endl;
        gp << "set xtics out rotate border offset 0,-1" << endl;
        gp << "set format x \"%H:%M:%S\" timedate" << endl;
        gp << "set format y \"%.2f\"" << endl;
    }

    gp << "set palette model RGB" << endl;
    gp << "set cbrange [" << MIN_DBFS << ":" << MAX_DBFS << "]" << endl;

    gp << "set palette defined (";

    if(console == false && wf != nullptr)
    {
        //builds the palette taking the actual setting used in waterfall
        gp << wf->GNUplotPalette();
    }
    else
    {
        //gets the gnuplot palette from configuration
        gp << as->getGNUplotPalette();

    }
    gp << ")" << endl;


    QString fileNameTemplate = QString("%1%2_%05u%3")
            .arg(gpPrefix)
            .arg(sessionName)
            .arg(GP_DATA_EXT);

    gp << "do for [i=1:" << as->getShots() << "] {" << endl;
    gp << "\tplotfile=sprintf(\"" << fileNameTemplate << "\", i)" << endl;

    if(gpPrefix == SPECMAP_PREFIX || gpPrefix == SPEC3D_PREFIX)
    {
        //3d-spectra: each scan is composed by multiple rows, terminated with
        //the average (N) maximum (S) and difference (S-N), like POW2D does
        gp << "\tset title sprintf(\"Spectrogram " << cal.toString(Qt::SystemLocaleShortDate) << " file: %s\", plotfile)" << endl ;
        gp << "\tsplot plotfile using 2:1:3" << endl;

    }


    if(gpPrefix == POW2D_PREFIX)
    {
        //2D-power: the average (N) maximum (S) and difference (S-N) one row per scan
        gp << "\tset multiplot layout 3,1" << endl;
        gp << "\tset lmargin at screen 0.15" << endl;
        gp << "\tset xrange [] writeback" << endl;
        gp << "\tset title sprintf(\"Power plot " << cal.toString(Qt::SystemLocaleShortDate) << " file: %s\", plotfile)" << endl ;
        gp << "\tplot plotfile using 1:2 with lines title \"maximum power S\" " << endl;
        gp << "\tplot plotfile using 1:2 with lines title \"average power N\" " << endl;
        gp << "\tplot plotfile using 1:4 with lines title \"max difference S-N\" " << endl;
        gp << "\tunset multiplot" << endl;
    }

    gp << "\tpause -1 \"hit return to continue\" " << endl;
    gp << "}" << endl;

    gp.flush();
    f.close();
    return true;
}


bool Control::gpEpilogue()
{
    //MY_ASSERT( QFile::exists(cmdFileName) == true );

    QFile f( wd.absoluteFilePath(cmdFileName) );
    if( f.open(QIODevice::WriteOnly | QIODevice::Append) == false )
    {
        return false;
        //MYCRITICAL << "error opening " << cmdFileName << " for append" ;
    }

    QTextStream gp(&f);

    int i = 0;
    int shotsMade = dataDumps.count();

    gp << "set palette model RGB" << endl;
    gp << "set cbrange [" << MIN_DBFS << ":" << MAX_DBFS << "]" << endl;

    gp << "set palette defined (";

    if(console == false && wf != nullptr)
    {
        //builds the palette taking the actual setting used in waterfall
        gp << wf->GNUplotPalette();
    }
    else
    {
        //gets the gnuplot palette from configuration
        gp << as->getGNUplotPalette();

    }
    gp << ")" << endl;

    if(gpPrefix == SPECMAP_PREFIX || gpPrefix == SPEC3D_PREFIX)
    {
        //3d-spectra: each scan is composed by multiple rows, terminated with
        //the average (N) maximum (S) and difference (S-N), like POW2D does


        if(shotsMade > 0)
        {
            for( i = 0; i < shotsMade; i++ )
            {
                QFileInfo fi( dataDumps.takeFirst() );
                dumpFileName = fi.fileName();
                gp << "print \"plotting shot " << i+1 << " of " << shotsMade << endl;
                gp << "set title \"Spectrogram  UTC: " << cal.toString(Qt::SystemLocaleShortDate) << " file: " << dumpFileName << " \" offset 0,3" << endl;
                gp << "splot \"" << dumpFileName << "\" using 2:1:3 notitle" << endl;
                gp << "pause -1 \"hit return to continue\" " << endl;
            }
        }
        else
        {
            gp << "splot \"" << datFileName << "\" using 2:1:3 notitle" << endl;
            gp << "pause -1 \"hit return to continue\" " << endl;
        }
    }

    if(gpPrefix == POW2D_PREFIX)
    {
        //2D-power: the average (N) maximum (S) and difference (S-N) one row per scan

        if(shotsMade > 0)
        {
            for( i = 0; i < shotsMade; i++ )
            {
                QFileInfo fi( dataDumps.takeFirst() );
                dumpFileName = fi.fileName();
                gp << "print \"plotting shot " << i+1 << " of " << shotsMade << endl;
                gp << "set title \"Power plot UTC: " << cal.toString(Qt::SystemLocaleShortDate) << " file: " << dumpFileName << " \" offset 0,3" << endl;
                gp << "plot \"" << dumpFileName << "\" using 1:3 with lines title \"maximum power S\", \"";
                gp << dumpFileName << "\" using 1:2 with lines title \"average power N\", \"";
                gp << dumpFileName << "\" using 1:4 with lines title \"max difference S-N\"" << endl;
                gp << "pause -1 \"hit return to continue\" " << endl;
            }
        }
        else
        {
            gp << "plot \"" << datFileName << "\" using 1:3 with lines title \"maximum power S\", \"";
            gp << datFileName << "\" using 1:2 with lines title \"average power N\", \"";
            gp << datFileName << "\" using 1:4 with lines title \"max difference S-N\"" << endl;
            gp << "pause -1 \"hit return to continue\"" << endl;
        }
    }
    gp.flush();
    return true;

}


float Control::getBpp(int loggedBins)
{
    MYDEBUG << "getBpp( loggedBins = " << loggedBins << " )";

    QRect wSize = as->getWwPaneGeometry();
    MYDEBUG << "getBpp wSize = " << wSize;

    wfWidth = wSize.width();
    MYDEBUG << "getBpp wfWidth = " << wfWidth << " pixels";

    float bpp = static_cast<float>(loggedBins) / wfWidth;
    bpp /= as->getZoom();
    bpp *= 10;  //zoom is in _tenths_ of Hz

    //bppFrac = modf(bpp, &bppInt);
    MYINFO << "bpp = " << bpp;
    return bpp;
}


bool Control::getStats( StatInfos* stats )
{
    stats->sc.deadTime = deadTime;
    stats->sc.scanDelta = scanDelta;
    stats->sc.dBfsDelta = dBfsDelta;
    stats->sc.csvDelta = csvDelta;
    stats->sc.outDelta = outDelta;

    stats->sc.totalTime = totalTime;
    stats->sc.infinite = infinite;
    stats->sc.nans = nans;
    stats->sc.noDataBufs = noDataBuf;
    stats->sc.scans = scans;
    stats->sc.loop = loop;

    return r->getStats(stats);

}

void Control::printStats()
{
    StatInfos s;
    getStats(&s);

    int effPerc = 0;
    if(s.sc.loop > 0)
    {
        effPerc = 100 - ((s.sc.noDataBufs * 100) / s.sc.loop);
    }


    MYDEBUG << endl
            << "Control - loop:" << s.sc.loop << endl
            << "Control - scans:" << s.sc.scans << endl
            << " deadTime:" << s.sc.deadTime << endl
            << " scanDelta:" << s.sc.scanDelta << endl
            << " dBfsDelta:" << s.sc.dBfsDelta << endl
            << " csvDelta:" << s.sc.csvDelta << endl
            << " outDelta:" << s.sc.outDelta << endl
            << " totalTime:" << s.sc.totalTime << endl
            << " infinite:" << s.sc.infinite << endl
            << " NaNs:" << s.sc.nans  << endl
            << " noDataBufs:" << s.sc.noDataBufs << endl
            << " efficiency " << effPerc << "%%";

    MYDEBUG << endl
            << "Radio - waitIQms:" << s.sr.waitIQms << endl
            << " inputSamples:" << s.sr.inputSamples << endl
            << " outputSamples:" << s.sr.outputSamples << endl
            << " FFTbins:" << s.sr.FFTbins << endl
            << " binSize:" << s.sr.binSize << endl
            << " overflows:" << s.sr.overflows << endl
            << " NaNs:" << s.sr.nans << endl
            << " waitScanMs:" << s.sr.waitScanMs << endl
            << " elapsedLoops:" << s.sr.elapsedLoops << endl
            << " acqTime:" << s.sr.acqTime << endl
            << " resampMs:" << s.sr.resampMs << endl
            << " FFTms:" << s.sr.FFTms << endl;

    if (testMode == false)
    {
        MYDEBUG << "SyncRx - timeFreeBuf:" << s.ss.timeFreeBuf
                << " droppedSamples:" << s.ss.droppedSamples << " " << endl;
    }

    QString ps;

#ifdef _DEBUG
    if(statPlot == nullptr)
    {
        QString pltPath = wd.absoluteFilePath(INNER_STATISTIC_PLOT);
        statPlot = new QFile(pltPath);
        MY_ASSERT( nullptr !=  statPlot)
        bool ok = statPlot->open(QIODevice::WriteOnly);
        if(ok == true)
        {
            MYINFO << "echoes inner stat file opened; " << pltPath;
            ps = "Control:loop;Control:scans;totalTime;deadTime;scanDelta;dBfsDelta;csvDelta;outDelta;infinite;nans;noDataBufs;;Radio:inputSamples;outputSamples;FFTbins;binSize;acqTime;overflows;nans;waitScanMs;waitIQms;elapsedLoops;resampMs;FFTms;;Syncrx:timeFreeBuf;droppedSamples\n";
            statPlot->write(qPrintable(ps), ps.length());
            ps = ";;time spent in acq.loop;time spent waiting loop;time spent waiting scan; time spent displaying scan;time spent updating csv;time spent updating dumps;;;;;;;;;;;time spent in radio loop;;;\n";
            statPlot->write(qPrintable(ps), ps.length());
            statPlot->flush();

        }
        else
        {
            MYCRITICAL << "unable to open inner stat file " << pltPath
                       << "error=" << statPlot->errorString();
            delete statPlot;
            statPlot = nullptr;
        }
    }
    else
    {
        QString ps = QString("%1;%2;%3;%4;%5;%6;%7;%8;%9;%10;%11;;%12;%13;%14;%15;%16;%17;%18;%19;%20;%21;%22;%23;;")
            .arg(s.sc.loop)
            .arg(s.sc.scans)
            .arg(s.sc.totalTime)
            .arg(s.sc.deadTime)
            .arg(s.sc.scanDelta)
            .arg(s.sc.dBfsDelta)
            .arg(s.sc.csvDelta)
            .arg(s.sc.outDelta)
            .arg(s.sc.infinite)
            .arg(s.sc.nans)
            .arg(s.sc.noDataBufs)
            .arg(s.sr.inputSamples)
            .arg(s.sr.outputSamples)
            .arg(s.sr.FFTbins)
            .arg(static_cast<double>(s.sr.binSize))
            .arg(s.sr.acqTime)
            .arg(s.sr.overflows)
            .arg(s.sr.nans)
            .arg(s.sr.waitScanMs)
            .arg(s.sr.waitIQms)
            .arg(s.sr.elapsedLoops)
            .arg(s.sr.resampMs)
            .arg(s.sr.FFTms);

        if (testMode == false)
        {
            QString ps2 = QString("%1;%2;")
                 .arg(s.ss.timeFreeBuf)
                 .arg(s.ss.droppedSamples);
            ps.append(ps2);
        }
        MYDEBUG << "writing echoes inner stats: " << ps << " in statPlot: " << statPlot;
        statPlot->write(qPrintable(ps), ps.length());
        ps = "\n";
        statPlot->write(qPrintable(ps), ps.length());
        statPlot->flush();
        MYDEBUG << "..done";

    }
#endif //_DEBUG
}


