/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.


    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 268                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-07-19 07:45:48 +02#$: Date of last commit
$Id: dclient.cpp 268 2018-07-19 05:45:48Z gmbertani $
*******************************************************************************/

#include "setup.h"
#include "rtl-sdr.h"
#include "dclient.h"



int DongleClient::clearUDPinputSocketStatic(QUdpSocket* socket)
{
    int recv = MAX_DGRAM_SIZE;
    int wastedBytes = 0;
    QHostAddress sender;
    quint16 senderPort = 0;
    MY_ASSERT( nullptr !=  socket)
    QByteArray baDgram(MAX_DGRAM_SIZE, 0);
    while(recv == MAX_DGRAM_SIZE)
    {
        recv = socket->readDatagram(baDgram.data(), baDgram.size(),
                        &sender, &senderPort);
        if (recv > 0)
        {
            wastedBytes += recv;
        }
    }
    MYINFO << "Cleared UDP input buffer, wasted: " << wastedBytes << " bytes";
    return wastedBytes;
}


bool DongleClient::lookForServerStatic(Settings* appSettings, QList<DeviceMap>* devices)
{
    MYINFO << __func__;

    MY_ASSERT( nullptr !=  devices)
    MY_ASSERT( nullptr !=  appSettings)

    bool found = false;
    QString netDev;
    QHostAddress serverAddr = appSettings->getServerAddress();
    quint16 serverPort = static_cast<quint16>( appSettings->getServerPort() );
    if( serverAddr.toIPv4Address() != 0 && serverPort != 0 )
    {
        int idx = 0;
        if(devices->count() > 0)
        {
            idx = devices->last().index + 1;
        }

        DeviceMap dm;

        foreach(dm, *devices)
        {
            if(dm.name.contains("UDP"))
            {
                //dongle server already scanned
                found = true;
            }
        }

        if(found == false)
        {

            //checks if a dongle server has been defined
            netDev = QString("%1:UDP server %2:%3").arg(idx).arg(serverAddr.toString()).arg(serverPort);
            MYINFO << "### Looking for dongle server at " << netDev;

            QByteArray baMyConfigName, baCommand;
            qint64 sent = 0;

            baMyConfigName.append( appSettings->getConfigName() );
            baMyConfigName.append( "." );
            baMyConfigName.append( appSettings->getConfigName() );
            QUdpSocket* socket = new QUdpSocket();
            MY_ASSERT( nullptr !=  socket)

            //stops eventual incoming data flux from server previously activated
            //to avoid flooding
            baCommand.append( appSettings->getConfigName() );
            baCommand.append( "." );
            baCommand.append( "STOP" );
            sent = socket->writeDatagram(baCommand, serverAddr, serverPort);
            clearUDPinputSocketStatic(socket);

            sent = socket->writeDatagram(baMyConfigName, serverAddr, serverPort);
            if(sent != baMyConfigName.size())
            {
                MYCRITICAL << "### Sending not complete, sent (=" << sent << ") bytes < baMyConfigName.size()(=" << baMyConfigName.size() << ") bytes : " << socket->errorString();
            }

            //waits answer from server in 10s
            if( socket->waitForReadyRead(10000) )
            {
                while (socket->hasPendingDatagrams())
                {
                    qint64 size = socket->pendingDatagramSize();
                    if(size <= 0)
                    {
                        continue;
                    }

                    QByteArray baSrvConf(static_cast<int>(size), '\0');

                    socket->readDatagram(baSrvConf.data(), baSrvConf.size(),
                                         &serverAddr, &serverPort);
                    QDataStream hdr(baSrvConf);
                    qint16 syncWord = -1;
                    hdr >> syncWord;
                    if(syncWord == SYNC_SRVCONF)
                    {
                        //the configuration got is stored but not applied
                        //because this is a static function, it must be used
                        //to check if there is a server that can be connected
                        //but real connection (intended as handshake)
                        //is done by constructor.

                        found = true;

                        if(devices->count() > 0)
                        {
                            dm = devices->last();
                            dm.index++;
                        }
                        else
                        {
                            dm.index = 0;
                        }

                        dm.name = netDev;
                        devices->append(dm);
                        MYINFO << "### Server radio configuration received";
                    }
                    else
                    {
                        MYWARNING << "### Discarded data packet arrived is not a server configuration, its sync is " << syncWord
                                  << " instead of " << SYNC_SRVCONF;
                    }
                }

            }
        }
    }

    if(found == false)
    {
        MYWARNING << "### no dongle servers configured in " << appSettings->getConfigName();
    }

    return found;
}


DongleClient::DongleClient(Settings* appSettings, Pool<IQbuf*>* bPool, QThread *parentThread) :
     IQdevice(appSettings, bPool, parentThread)
{
    setObjectName("DongleClient thread");
    owner = parentThread;

    MYINFO << "parentThread=" << parentThread << " owner= " << owner << " parent()=" << parent();

    droppedSamples = 0;
    timeFreeBuf = 0;
    recvBufferCount = 0;
    clientSocket = nullptr;
    serverConnected = false;
    baIQbuf.clear();
    srvIQbufSize = 0;
    iqBufSize = 0;
    rrIn = nullptr;
    rrOut = nullptr;
    rr = nullptr;
    decim = 0;

    if(as->getServerAddress().toIPv4Address() != 0)
    {
        //this instance is a client, so it tries to connect the given
        //dongle server
        //waiting for server's configuration
        serverConnected = waitServerConfig();

        //enlarges the sending buffer to 1MB
        clientSocket->setSocketOption( QAbstractSocket::ReceiveBufferSizeSocketOption, 0x100000 );
        MYINFO << " client socket opened, ReceiveBufferSizeSocketOption = " << clientSocket->socketOption( QAbstractSocket::ReceiveBufferSizeSocketOption );

        QString s;
        rates.append( s.setNum( as->getSampleRate() )  );
        gains.append( s.setNum( as->getGain() / 10 ) );
    }


}

DongleClient::~DongleClient()
{
    sendCommand("QUIT");

    //destroying client
    if(clientSocket != nullptr)
    {
        MYINFO << "Client socket closed";
        serverConnected = false;
        delete clientSocket;
        clientSocket = nullptr;
    }
}


void DongleClient::run()
{
    MYINFO << "Dongle client thread started";
    int bufIdx = -1;
    IQbuf* buf = nullptr;

    //reconnects at each acq restart
    clearUDPinputSocketStatic(clientSocket);

    sendCommand("START");
    timeAcq.start();

    for(;;)
    {
        if( isFinished() == true || isInterruptionRequested() == true )
        {
            MYINFO << "DongleClient thread stopped";
            sendCommand("STOP");
            clearUDPinputSocketStatic(clientSocket);
            break;
        }

        if(serverConnected)
        {

            bufIdx = pb->take();

            if(bufIdx != -1)
            {
                timeFreeBuf = static_cast<int>(timeAcq.restart());

                buf = pb->getElem(bufIdx);
                if(buf != nullptr)
                {
                    int retVal = readSamples(buf);
                    if(retVal > 0)
                    {
                        pb->forward(bufIdx);
                    }
                    else
                    {
                        //no data to forward, frees the buffer
                        pb->release(bufIdx);
                    }
                }
                else
                {
                    //cannot forward data to radio
                    //due to missing free buffers
                    droppedSamples += buf->doublets();
                }
                timeAcq.restart();
            }
        }
        else
        {
            MYWARNING << "Unrequested server disconnection, stopping DongleClient thread";
            clearUDPinputSocketStatic(clientSocket);
            break;
        }
    }

}


bool DongleClient::getStats( StatInfos* stats )
{
    stats->ss.droppedSamples = droppedSamples;
    stats->ss.timeFreeBuf = timeFreeBuf;
    return true;
}



int DongleClient::readSamples(IQbuf* buf)
{
    QMutexLocker ml(&mutex);

    QHostAddress sender;
    quint16 senderPort = 0;
    int retVal = 0;
    bool skipBuffer = false;
    int sliceSize = MAX_DGRAM_SIZE;
    int totSlices = 0;
    int sliceIndex = 0;
    QByteArray baDgram(sliceSize, '\0');
    qint64 recv = 0;

    MYDEBUG << __func__;
    try
    {
        if( clientSocket->pendingDatagramSize() == 0 )
        {
            return retVal;
        }

        while(retVal == 0)
        {
            baDgram.resize( static_cast<int>(MAX_DGRAM_SIZE) );
            recv = clientSocket->readDatagram(baDgram.data(), baDgram.size(),
                        &sender, &senderPort);

            if(recv > 0)
            {
                //datagram received
                //the datagram size matches the size expected
                if( recv == strlen("STOP") )
                {
                    //the datagram contains a stop acquisition command
                    if(baDgram.contains("STOP"))
                    {
                        MYINFO << "### The server has stopped the acquisition";
                        sliceIndex = 0;
                        baIQbuf.clear();
                        retVal = -1;
                    }

                }
                else if(recv < MAX_DGRAM_SIZE)
                {
                     //must read the datagram content to understand its structure
                    QDataStream ds(baDgram);

                    qint16 syncWord = -1;
                    ds >> syncWord;

                    if(syncWord == SYNC_IQBUF)
                    {
                        //it's an iqbuf header
                        hdr.timeStamp.clear();

                        //extracts the header data
                        ds >> hdr.uid;
                        ds >> hdr.timeStamp;

                        buf->setTimeStamp(hdr.timeStamp);
                        buf->setBufSize( iqBufSize );
                        totSlices = static_cast<int>(ceil( (srvIQbufSize * 2) / sliceSize ));
                        sliceIndex = totSlices;
                    }
                    else if(syncWord == SYNC_SRVCONF)
                    {
                        //it's a configuration change
                        QString     deviceName;
                        int         sampleRate;
                        int         tune;
                        int         gain;
                        float       ppmError;
                        QList<uint32_t> uRates;
                        QList<float> fGains;

                        ds >> deviceName;   //simply stored
                        ds >> sampleRate        ; as->setSampleRate(sampleRate);
                        ds >> gain              ; as->setGain(gain);
                        ds >> ppmError          ; as->setError(ppmError);
                        ds >> tune              ; as->setTune(tune);
                        ds >> srvIQbufSize      ;
                        ds >> fGains;
                        ds >> uRates;

                        myCfgChanged = true;
                        MYINFO << "Server radio configuration changed at runtime, received " << recv << " bytes";
                        MYINFO << "gains: " << fGains;
                        MYINFO << "rates: " << uRates;

                        setPackedGains(fGains);
                        setPackedRates(uRates);
                        slotParamsChange();
                    }
                    baIQbuf.clear();
                }
                else if(sliceIndex > 0)
                {
                    //it's an iqbuf slice

                    if(recv == sliceSize )
                    {
                        baIQbuf.append(baDgram);
                        sliceIndex--;

                        if( sliceIndex == 0 &&
                            baIQbuf.size() != (srvIQbufSize * 2) )
                        {
                            MYWARNING << "expected buffer size and received data don't match "
                                      << "expected = " << (srvIQbufSize * 2)
                                      << "received bytes =" << baIQbuf.size();
                            skipBuffer = true;
                            baIQbuf.clear();
                            retVal = -2;
                        }
                    }
                    else
                    {
                        MYWARNING << "unexpected shorter slice#" << sliceIndex
                                  << " = " << recv << " bytes, discarding buffer";
                        baIQbuf.clear();
                        QDataStream ds(baDgram);
                        qint16 syncWord = -1;
                        ds >> syncWord;
                        if(syncWord == SYNC_IQBUF)
                        {
                            MYWARNING << "unexpected header at slice#" << sliceIndex
                                  << " = " << recv << " bytes, discarding buffer";
                            retVal = -3;
                        }
                        else
                        {
                            MYWARNING << "unexpected shorter slice#" << sliceIndex
                                  << " = " << recv << " bytes, discarding buffer";
                            retVal = -4;
                        }
                    }

                    if(sliceIndex == 0 && skipBuffer == false)
                    {
                        for ( int f = 0, j = 0 ; f < srvIQbufSize ; f++ , j += 2 )
                        {
                            rrIn[f].real = static_cast<float>( static_cast<int>(baIQbuf.at(j))   - 128);
                            rrIn[f].imag = static_cast<float>( static_cast<int>(baIQbuf.at(j+1)) - 128);
                        }

                        rresamp_crcf_execute(rr, rrIn, rrOut);   //note order rr, inputs, outputs

                        for ( int f = 0, j = 0 ; f < iqBufSize ; f++ , j += 2 )
                        {
                            uchar i = static_cast<uchar>((static_cast<int8_t>(rrOut[f].real) + 128));
                            uchar q = static_cast<uchar>((static_cast<int8_t>(rrOut[f].imag) + 128));

                            buf->data()[j] = i;
                            buf->data()[j+1] = q;

                        }

                        recvBufferCount++;
                        MYDEBUG << "Received buffer #" << recvBufferCount;

                        retVal = baIQbuf.size();
                        baIQbuf.clear();
                    }

                }

            }
        }
    }
    catch ( QException& e )
    {
        MYCRITICAL << "Exception " << e.what();
    }

    return retVal;
}




bool DongleClient::sendMyConfigName()
{
    MYINFO << __func__;
    QByteArray baMyConfigName;
    qint64 sent = 0;

    if(serverConnected == false)
    {
        if(clientSocket == nullptr)
        {
            clientSocket = new QUdpSocket(/*no parent to avoid "Cannot create children..." */);
            MY_ASSERT( nullptr !=  clientSocket)
        }

        baMyConfigName.append( as->getConfigName() );
        baMyConfigName.append( "." );
        baMyConfigName.append( as->getConfigName() );

        sent = clientSocket->writeDatagram(baMyConfigName, as->getServerAddress(), static_cast<quint16>(as->getServerPort()));
        if(sent != baMyConfigName.size())
        {
            MYCRITICAL << "Sending not complete, sent (=" << sent << ") bytes < baMyConfigName.size()(=" << baMyConfigName.size() << ") bytes";
            return false;
        }
        MYINFO << "configuration sent to " << as->getServerAddress() << ":" << as->getServerPort();
    }
    return true;
}

void DongleClient::sendCommand(QString cmd)
{
    MYINFO << __func__ << "(" << cmd << ")";
    QByteArray baCommand;
    qint64 sent = 0;

    if(serverConnected)
    {
        baCommand.append( as->getConfigName() );
        baCommand.append( "." );
        baCommand.append( cmd );
        sent = clientSocket->writeDatagram(baCommand, as->getServerAddress(), static_cast<quint16>(as->getServerPort()));
        if(sent != baCommand.size())
        {
            MYCRITICAL << "Sending not complete, sent (=" << sent << ") bytes < baCommand.size()(=" << baCommand.size() << ") bytes";
        }
    }
}

bool DongleClient::waitServerConfig()
{
    bool result = false;
    QHostAddress serverAddr;
    quint16 serverPort;

    sendMyConfigName();

    if(clientSocket->state() == QAbstractSocket::BoundState)
    {
        MYINFO << "Waiting server config...";
        if( clientSocket->waitForReadyRead() )
        {
            while (clientSocket->hasPendingDatagrams())
            {
                qint64 size = clientSocket->pendingDatagramSize();
                QByteArray baData(static_cast<int>(size), '\0');

                clientSocket->readDatagram(baData.data(), baData.size(),
                                     &serverAddr, &serverPort);
                QDataStream cfg(baData);
                qint16 syncWord = -1;
                cfg >> syncWord;
                if(syncWord == SYNC_SRVCONF)
                {
                    QString     deviceName;
                    int         sampleRate;
                    int         tune;
                    int         gain;
                    float       ppmError;
                    QList<uint32_t> uRates;
                    QList<float> fGains;

                    cfg >> deviceName;   //simply stored
                    cfg >> sampleRate   ; as->setSampleRate(sampleRate);
                    cfg >> gain         ; as->setGain(gain);
                    cfg >> ppmError     ; as->setError(ppmError);
                    cfg >> tune         ; as->setTune(tune);
                    cfg >> srvIQbufSize ;
                    cfg >> fGains;
                    cfg >> uRates;

                    MYINFO << "gains: " << fGains;
                    MYINFO << "rates: " << uRates;

                    setPackedGains(fGains);
                    setPackedRates(uRates);

                    slotParamsChange();
                    baIQbuf.clear();
                    result = true;
                    MYINFO << "Server radio configuration received, client connected to server";
                }
                else
                {
                    MYWARNING << "Discarded unknown data packet a, its size is " << baData.size();
                }

            }

        }
    }
    return result;
}


void DongleClient::slotParamsChange()
{
    bw = as->getBandwidth();
    lowBound   =  as->getTune() - (bw / 2);
    highBound  =  as->getTune() + (bw / 2);

    MYINFO << "Subsampled range (band)" << lowBound << " to " << highBound << " Hz";

    // recalc client iqBufSize
    float x = sr;
    x *= as->getFFTbins();
    x /= bw;
    iqBufSize = as->iround( x );
    MYINFO << "iqBufSize is " << iqBufSize;
    if(iqBufSize % MAX_DGRAM_SIZE != 0)
    {
        //iqBufSize must always be a multiple of 512
        int red = as->iround( static_cast<float>(iqBufSize) / MAX_DGRAM_SIZE );
        iqBufSize = red * MAX_DGRAM_SIZE;
        MYINFO << "iqBufSize not multiple of " << MAX_DGRAM_SIZE << ", patched to " << iqBufSize;
    }

    //the next datagrams will be iqbuf slices
    //let's prepare to receive them

    if(rr != nullptr)
    {
        rresamp_crcf_destroy(rr);
        rr = nullptr;
    }

    double ratio =  iqBufSize;
    ratio = ratio / srvIQbufSize;
    MYINFO << "Rational resampling, " << iqBufSize << " / " << srvIQbufSize
        << " = " << ratio;

    rr = rresamp_crcf_create_default(static_cast<uint>(iqBufSize), static_cast<uint>(srvIQbufSize));  //note order output,input
    MY_ASSERT(rr != nullptr)
    decim = rresamp_crcf_get_decim(rr);
    MYINFO << "Decimation factor Q: " << decim;
    uint interp = rresamp_crcf_get_interp(rr);
    MYINFO << "Interpolation factor P: " << interp;
    uint blockLen = rresamp_crcf_get_block_len(rr);
    MYINFO << "Block len Q: " << blockLen;

    if(rrOut != nullptr)
    {
        delete[] rrOut;
        rrOut = nullptr;
    }
    rrOut = new liquid_float_complex[iqBufSize];
    MY_ASSERT(rrOut != nullptr)
    memset(rrOut, 0, static_cast<size_t>(iqBufSize) * sizeof(liquid_float_complex));

    if(rrIn != nullptr)
    {
        delete[] rrIn;
        rrIn = nullptr;
    }
    rrIn = new liquid_float_complex[srvIQbufSize];
    MY_ASSERT(rrIn != nullptr)
    memset(rrIn, 0, static_cast<size_t>(srvIQbufSize) * sizeof(liquid_float_complex));

    emit status( tr("Ready.") );

    //ending of parameter change sequence
    ready = true;
}


bool DongleClient::slotSetBandwidth()
{
    QMutexLocker ml(&mutex);
    ready = false;
    int newBw = as->getBandwidth();

    if( bw != newBw && bw <= sr )
    {
        bw = newBw;
        slotParamsChange();
    }
    return true;
}

bool DongleClient::slotSetResolution()
{
    QMutexLocker ml(&mutex);
    slotParamsChange();
    return true;
}

