/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.


    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html



*******************************************************************************
$Rev:: 351                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-10-22 00:13:47 +02#$: Date of last commit
$Id: funcgen.cpp 351 2018-10-21 22:13:47Z gmbertani $
*******************************************************************************/

#include <time.h>
#include "setup.h"
#include "funcgen.h"

FuncGen::FuncGen(Settings* appSettings, Pool<IQbuf*>* bPool, QThread *parentThread) :
    IQdevice(appSettings, bPool, parentThread)
{
    setObjectName("FuncGen thread");

    MYINFO << "parentThread=" << parentThread << " parent()=" << parent();

    ready = false;

    iqBufSize = 0;
    stored = nullptr;
    acqTime = 0;
    amplitude = 0.0;
    actualFreq = 0.0;

    sweepFreq = 0;
    loopsToNextSweep = 0;
    highBound = 0;
    lowBound = 0;
    bw = 0;

    testMode = TM_NONE;
    meteor = FG_NONE;
    countDown = 0;
    burning = 0;
    ablation = 0;
    ionization = 0;

    buildingNewWave = false;

    clearQ = 0x5;
    clearI = 0x5;

    MYINFO << "I/Q buffer size set to " << MAX_SAMPLE_BUF_SIZE << " I/Q doublets ("
           << MAX_SAMPLE_BUF_SIZE * 2 << " bytes)";

    stored = nullptr;

    deviceName = "Test patterns";

    //fills the gains list with sinusoid amplitudes
    //(positive signed char values)
    gains.append("1.00");
    gains.append("2.00");
    gains.append("5.00");
    gains.append("10.00");
    gains.append("20.00");
    gains.append("30.00");
    gains.append("50.00");
    gains.append("75.00");
    gains.append("100.00");
    gains.append("127.00");

    //sample rates list
    rates.append("250000");
    rates.append("300000");
    rates.append("1000000");
    rates.append("1200000");
    rates.append("1500000");
    rates.append("1800000");
    rates.append("2000000");
    rates.append("2200000");
    rates.append("2500000");
    rates.append("2800000");
    rates.append("3000000");
    rates.append("3200000");


}

FuncGen::~FuncGen()
{
    //waits for the end of owned thread ( run() termination )
    if( wait(10000) == false )
        MYWARNING << className() << " destructor - can't stop the thread";
}


void FuncGen::run()
{
    bool init = true;

    MYINFO << "Function generator thread started";
    int bufIdx = -1;

    iqBufSize = DEFAULT_BW;

    //static IQ buffer for periodic waveforms
    stored = new IQbuf( iqBufSize );
    MY_ASSERT( nullptr !=  stored);

    //enough small to be fast and enough big to allow
    //down resampling:
    iqBufSize = static_cast<int>(as->getSampleRate() / as->getResolution());

    //acquisition time in ms:
    acqTime = (1000 * iqBufSize) / as->getSampleRate();

    countDown = 0;
    burning = 0;
    ablation = 0;
    ionization = 0;
    amplitude = 0.0;
    actualFreq = 0.0;

    lowBound   =  as->getTune() - (bw / 2);
    highBound  =  as->getTune() + (bw / 2);

    if(as->getShots() > 0 && as->getUpThreshold() > 0)
    {
        MYINFO << "shots required and up threshold set, entering meteor simulation to test the automatic capture";
        testMode = TM_METEOR;
    }
    else if(as->getShots() > 0)
    {
        MYINFO << "Shots required without thresholds, entering sweeping sinewave generation to test periodic mode";
        testMode = TM_SWEEP;
    }
    else
    {
        MYINFO << "no shots required, entering sinewave generation to test continuous mode";
        testMode = TM_SINE;
    }


    for(;;)
    {
        if( isFinished() == true || isInterruptionRequested() == true )
        {
            MYDEBUG << "FuncGen: exiting and terminating thread.";
            break;
        }



        IQbuf* buf = nullptr;

        bufIdx = pb->take();
        if(bufIdx != -1)
        {

            buf = pb->getElem(bufIdx);
            if(buf != nullptr)
            {

                iqBufSize = static_cast<int>(as->getSampleRate() / as->getResolution());

                if( buf->doublets() != iqBufSize || init == true )
                {
                    buf->setBufSize( iqBufSize );
                    stored->setBufSize( iqBufSize );
                    acqTime = (1000 * iqBufSize) / as->getSampleRate();
                    init = false;
                }

                tsGen = QDateTime::currentDateTimeUtc();
                QString s = tsGen.toString( "dd/MM/yyyy;hh:mm:ss.zzz");
                buf->setTimeStamp( s );

                switch(testMode)
                {
                case TM_NONE:
                case TM_DC:
                    break;

                case TM_SINE:
                    getWave(
                        static_cast<float>(as->getSampleRate() / 3 ),
                        static_cast<float>(as->getGain()),
                        buf );
                    break;

                case TM_SWEEP:
                    if(loopsToNextSweep == 0)
                    {
                        loopsToNextSweep = 50;
                        if(sweepFreq >= as->getTune())
                        {
                            sweepFreq = 0;
                        }
                        sweepFreq += static_cast<int>(as->getResolution() * 10);

                    }
                    else
                    {
                        loopsToNextSweep--;
                    }

                    getWave(
                        static_cast<float>(sweepFreq),
                        static_cast<float>(as->getGain()),
                        buf );

                    break;

                case TM_METEOR:

                    getEvent(
                        static_cast<float>(as->getTune()),
                        static_cast<float>(as->getGain()),
                        buf );

                    switch(meteor)
                    {
                    case FG_NONE:
                        //no events - the test produces a fixed pattern
                        //controlled by the caller
                        //goSine();
                        break;

                    case FG_SET:
                        //previous event concluded
                        //triggers a new one
                        //reloads casual countdown
                        srand( static_cast<uint>( time(nullptr) ) );
                        countDown = ((rand() % 3) + 1) * //countdown is in 50ms units
                                (as->getRecTime() * 20); //- 1 to 3 times the rec time
                        meteor = FG_COUNTING;
                        break;

                    case FG_COUNTING:
                        //waits countdown. Counters are decremented in getEvent() to get
                        //synced with radio thread
                        if(countDown <= 0)
                        {
                            //trigger
                            //calculate burning scans
                            burning = (rand() % 10) + 2;
                            actualFreq = as->getSampleRate() / ((rand() % 100) + 1);

                            //produces the short vertical strip with a fixed sinewave
                            goSine();
                            meteor = FG_BURNING;
                        }
                        else
                        {
                            makeNoise();
                        }
                        break;

                    case FG_BURNING:
                        //burning phase
                        if(burning <= 0)
                        {
                            //prepares ablation sweep width
                            //looks better with lowest SR
                            ablation = static_cast<int>(
                                        actualFreq + (rand() % static_cast<int>(actualFreq / 100))
                                        +1 );

                            meteor = FG_ABLATION;
                        }
                        break;

                    case FG_ABLATION:
                        //ablation is very fast - it lasts one scan only and at full scale amplitude
                        {
                            amplitude = 127;
                            goSweep();
                            amplitude = (rand() % 30);
                            ionization = (rand() % 10) + 2;
                            meteor = FG_IONIZATION;
                        }
                        break;

                    case FG_IONIZATION:
                        //ionization phase
                        if(ionization > 0)
                        {
                            //approximate decreasing repetition of ablation scan
                            //actualFreq += (rand() % 10000);
                            //ablation -= (rand() % 5000);
                            amplitude = (amplitude > 0)? amplitude - (rand() % 2) : 0;
                            goSweep();

                        }
                        else
                        {
                            //event's end
                            //clears output wave
                            makeNoise();
                            meteor = FG_SET;

                        }
                        break;
                    }

                }

                memcpy( buf->data(), stored->data(), static_cast<size_t>(buf->bSize()) );
                pb->forward(bufIdx);

            }

        }

        QThread::yieldCurrentThread();

    }

    if(stored != nullptr)
    {
        delete stored;
        stored = nullptr;
    }

}




bool FuncGen::isDeviceOpen()
{
    return true;
}

void FuncGen::getWave(float freq, float amp, IQbuf *buf)
{
    QMutexLocker ml(&mt);

    //comparing floats in this way is not safe but
    //they are in fact integers boxed in floats
    if( as->fpCmpf(freq, actualFreq) != 0 || as->fpCmpf(amp, amplitude) != 0 )
    {
        //changed parameters: zeroes the buffer

        actualFreq = freq;
        amplitude = amp;
        //makeNoise();
        buildingNewWave = true;
    }

    goSine();

    //outputs the stored wave
     memcpy(buf->data(), stored->data(), static_cast<size_t>( buf->bSize() ) );
}


void FuncGen::getEvent(float freq, float amp, IQbuf *buf)
{
    QMutexLocker ml(&mt);

    if(meteor == FG_NONE)
    {
        makeNoise();
        memcpy(buf->data(), stored->data(), static_cast<size_t>( buf->bSize() ) );
        meteor = FG_SET;
    }

    if(meteor == FG_SET)
    {     
        //changed parameters:
        actualFreq = freq;
        amplitude = amp;

        //aborts pending events and zeroes the buffer
        makeNoise();
    }

    //ms counting
    msleep(50);
    if(countDown > 0)
    {
        countDown--;
    }
    else if(burning > 0)
    {
        burning--;
    }
    else if(ionization > 0)
    {
        ionization--;
    }

    memcpy(buf->data(), stored->data(), static_cast<size_t>( buf->bSize() ));

}



void FuncGen::goSweep()
{
    QElapsedTimer clock;

    clock.start();

    //fills the input buffer with a sine wave sweeping from actualFreq
    //to actualFreq+ablation
    //Please note that one sample is TWO bytes (I+Q)
    //only I (real) samples are generated
    //the Q (imaginary) samples are all zeroed (+128 unsigned)

    const float twopi = 6.28318530F;
    float t = 0, z = 0, x = 0, y = 0, oldY = 0, dt = 0, df = 0, currFreq = 0;

    dt = 1.0F / as->getSampleRate();
    df = (2 * ablation) / static_cast<float>( iqBufSize );
    currFreq = actualFreq;

    MYDEBUG << "Calculating sine wave table, sweeping from " << actualFreq << " to " << ablation << " Hz";
    MYDEBUG << "Amplitude=" << amplitude;
    MYDEBUG << "dt = " << dt;

    int periods = 0;
    int phase = 0;
    x = 0;

    //QString wave = "sweeptab.dat";
    //QFile f(wd.absoluteFilePath( wave ));
    //f.open(QIODevice::WriteOnly);
    //QTextStream s(&f);

    for (int n = 0; n < stored->bSize(); n += 2)
    {
        t = x * dt;                 // 1/sample_rate = time between samples
        z = twopi*currFreq*t + phase;
        oldY = y;
        y = amplitude * sinf( z );

        stored->data()[n] = static_cast<uchar>( y + 128 );     // real part - use either sine or cosine
        stored->data()[n+1] = 128;                             // imaginary part = 80h fixed

        //s << x << " " << stored[n] << endl;

        if( oldY <= 0 && y > 0 ) //zero-cross
        {
            periods++;

        }
        currFreq += df;
        x++;
    }

    MYDEBUG << "Sweeping wave generation took " << clock.elapsed() << " ms";
}

void FuncGen::goSine()
{
    QElapsedTimer clock;

    clock.start();

    //fills the input buffer with a sine wave that covers
    //with precise periods the entire buffer.
    //The number of periods to store in buffer can be specified,
    //while the resulting frequency is calculated accordingly.
    //Please note that one sample is TWO bytes (I+Q)
    //only I (real) samples are generated
    //the Q (imaginary) samples are all zeroed (+128 unsigned)

    const float twopi = 6.28318530F;
    float t = 0, z = 0, x = 0, y = 0, oldY = 0, lasting = 0, ppp = 0, dt = 0, tp = 0;

    dt = 1.0F / as->getSampleRate();
    lasting = 1.0F / actualFreq;
    ppp = static_cast<float>(as->getSampleRate() / actualFreq);
    tp = acqTime / lasting;

    MYDEBUG << "Calculating sine wave table, actualFreq=" << actualFreq;
    MYDEBUG << "Amplitude=" << amplitude;
    MYDEBUG << "Points per period = " << ppp;
    MYDEBUG << "Period lasting [s] =" << lasting;
    MYDEBUG << "dt = " << dt;
    MYDEBUG << "Expected total periods = " << tp;

    int periods = 0;

    int phase = 0;
    x = 0;

    //QString wave;
    //wave = "sinetab.dat";
    //QFile f(wd.absoluteFilePath( wave ));
    //f.open(QIODevice::WriteOnly);
    //QTextStream s(&f);
    for (int n = 0; n < stored->bSize(); n += 2)
    {
        t = x * dt;                 // 1/sample_rate = time between samples
        z = twopi*actualFreq*t + phase;
        oldY = y;
        y = amplitude * sinf( z );

        stored->data()[n] = static_cast<uchar>( y + 128 );     // real part - use either sine or cosine
        stored->data()[n+1] = 128;                             // imaginary part = 80h fixed

        //s << x << " " << stored[n] << endl;

        if( oldY <= 0 && y > 0 ) //zero-cross
        {
            periods++;
        }

        x++;
    }

    //f.close();

    MYDEBUG << "Sine wave generation took " << clock.elapsed() << " ms";
    MYDEBUG << "actualFreq=" << actualFreq << " Hz, " << periods << " full periods in buffer.";

}




void FuncGen::goDC()
{
    //fills the input buffer with a positive constant value.
    //Please note that one sample is TWO bytes (I+Q)
    //only I (real) samples are filled
    //the Q (imaginary) samples are all zeros

    MYDEBUG << "Inserting a constant DC offset in buffer, a=" << amplitude;

    int a = as->iround(amplitude);
    for (int n = 0; n < stored->bSize(); n += 2)
    {
        stored->data()[n] = static_cast<uchar>( a + 128 );
        stored->data()[n+1] = 128;
    }

}

void FuncGen::makeNoise()
{
    //fills the input buffer with randomized low numbers
    //to simulate background noise, both I and Q
    MYDEBUG << "Background noise";

    for (int n = 0; n < stored->bSize(); n += 2)
    {
        stored->data()[n] = rand() % clearI;
        stored->data()[n+1] = rand() % clearQ;
    }

}

void FuncGen::retune()
{
    //sweeping restarts from the lowest frequency
    //displayable (always 1hz)
    MYDEBUG << "retune()" ;
    sweepFreq = lowBound;
    loopsToNextSweep = 0;
}

bool FuncGen::getStats( StatInfos* stats )
{
    Q_UNUSED(stats)
    return false;
}

void FuncGen::slotParamsChange()
{
    bw = as->getBandwidth();
    lowBound   =  as->getTune() - (bw / 2);
    highBound  =  as->getTune() + (bw / 2);

    MYINFO << "Subsampled range (band)" << lowBound << " to " << highBound << " Hz";

    retune();

    emit status( tr("Ready.") );
    ready = true;
}

bool FuncGen::slotSetTune()
{
    ready = false;
    //int newTune = as->getTune();

    if(isRunning() == true)
    {
        if(pb != nullptr)
        {
            //flushes the unprocessed iq buffers
            pb->flush();
        }
    }

    if(true /*newTune > MAX_SAMPLE_RATE_HZ */)
    {
        MYWARNING << "Test device tune can't be changed";
        //sr = MAX_SAMPLE_RATE_HZ;
        //bw = sr;
        tune = bw/2;
        //as->setBandwidth(bw);
        //as->setSampleRate(sr);
        as->setTune(tune);
        //as->setResBins(512);
    }
    /*
    else
    {
        tune = newTune;
        as->setTune(tune);
    }
    */
    MYINFO << "Actual tuning: " << tune << " Hz";

    //Tuning OK
    slotParamsChange();
    return true;
}

bool FuncGen::slotSetSampleRate()
{
    ready = false;
    QString newSr;
    newSr.setNum( as->getSampleRate() );
    if( rates.contains( newSr ))
    {
        sr = as->getSampleRate();
        //bw = sr;
        tune = bw/2;
        //as->setBandwidth(bw);
        as->setTune(tune);

        slotParamsChange();
        return true;
    }
    return false;
}

bool FuncGen::slotSetGain()
{
    ready = false;
    QString newGain;
    newGain.setNum( as->getGain() / 10 ) ; //tenths of dB --> dB
    if( gains.contains( newGain ))         //for gain strings comparation
    {
        gain = as->getGain();
        as->setGain( gain );
        slotParamsChange();
    }
    else
    {
        newGain = gains.last();
        gain =  static_cast<int>(newGain.toFloat() * 10);
        as->setGain( gain ); //dB --> tenths of dB
    }
    return true;
}

bool FuncGen::slotSetBandwidth()
{
    ready = false;
    int newBw = as->getBandwidth();

    if(newBw > sr)
    {
        bw = sr;
    }
    else if(newBw > 0)
    {
        bw = newBw;
    }
    else
    {
        return false;
    }

    as->setBandwidth(bw);
    slotParamsChange();
    return true;
}
