/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.


    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html




*******************************************************************************
$Rev:: 382                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-12-20 08:43:50 +01#$: Date of last commit
$Id: mainwindow.cpp 382 2018-12-20 07:43:50Z gmbertani $
*******************************************************************************/
#include <QFileDialog>
#include "ui_mainwindow.h"
#include "siteinfos.h"
#include "waterfall.h"
#include "radio.h"
#include "control.h"
#include "postproc.h"
#include "mainwindow.h"


MainWindow::MainWindow(Settings* appSettings, Control* appControl, QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    setObjectName("MainWindow widget");
    ui->setupUi(this);

    init = true;
    outputEnabled = false;
    deviceChanged = false;
    busyInst = 0;


    MY_ASSERT( nullptr !=  appControl)
    ac = appControl;

    MY_ASSERT( nullptr !=  appSettings)
    as = appSettings;

    tune = as->getTune();

    wf = new Waterfall(as, ac, this);
    MY_ASSERT( nullptr !=  wf)
    wf->hide();

    //make wf pointer available for Control
    as->setWaterfallWindow( wf );

    si = new SiteInfos(as);
    MY_ASSERT( nullptr !=  si)
    si->hide();

    r = ac->getRadio();
    MY_ASSERT( nullptr !=  r)

    sysCheck = new QTimer;
    MY_ASSERT( nullptr !=  sysCheck)

    PostProc* pp = ac->postProcessor();
    MY_ASSERT( nullptr !=  pp)

    //GUI widgets arrangement by tabs:
    guiWidgetsList.resize(GT_TOTAL);

    guiWidgetsList[GT_GLOBAL]={ui->pbOpen, ui->pbSaveAs, ui->pbAbout, ui->pbQuit, ui->pbStart};

    guiWidgetsList[GT_DEVICE]={ ui->cbDevice, ui->cbSR,
                                ui->pbm1kT, ui->pbm10kT,  ui->pbm100kT,  ui->pbm1MT,  ui->pbm10MT,
                                ui->pbp1kT, ui->pbp10kT,  ui->pbp100kT,  ui->pbp1MT,  ui->pbp10MT,                           
                                ui->sbPpm, ui->cbGain };

    //FFT tab
    guiWidgetsList[GT_FFT]={ui->cbFtw,  ui->pbIncRes,  ui->pbDecRes, ui->chkBypass,
                                ui->pbm1k, ui->pbm10k, ui->pbm100k, ui->pbm1M,
                                ui->pbp1k, ui->pbp10k, ui->pbp100k, ui->pbp1M };

    //Output
    guiWidgetsList[GT_OUTPUT]={  ui->sbInterval,  ui->sbAfter,  ui->sbMaxShots,  ui->sbRecTime,
                                ui->cbThresholds, ui->sbUpThreshold, ui->sbDnThreshold, ui->sbAvg, ui->sbTrange,
                                ui->sbJoinTime, ui->sbMinFree, ui->cbPlotType, ui->chkGnuplot, ui->chkSShots };

    //Report
    guiWidgetsList[GT_REPORT]={ ui->dtFrom, ui->dtTo, ui->pbSiteInfos, ui->chkSite,
                            ui->chkSummary, ui->chkColorgraph, ui->chkOverdensePlots, ui->chkUnderdensePlots,
                            ui->chkDetails, ui->sbUnderThr, ui->sbOverThr, ui->chkCarrierFilter,
                            ui->chkLightingsFilter, ui->pbReport, /* ui->chkRestart, */ui->pbFlatten,
                            ui->chkAutoReport };

    //Preferences
    guiWidgetsList[GT_PREFERENCES]={ ui->chkFticks,  ui->chkPticks, ui->chkTticks, ui->chkTooltips,
                        ui->chkEraseOldLogs, ui->chkEraseOldShots, ui->chkEnablePing,
                        ui->sbLasting, ui->cbPalette, ui->sbNoiseLimit, ui->leServerIP, ui->sbServerPort,
                        ui->pbGnuplot };



    //plugging signals:

    //pushbuttons
    connect( ui->pbAbout,       SIGNAL( pressed() ), this,      SLOT( slotAboutDialog() ) );
    connect( ui->pbOpen,        SIGNAL( pressed() ), this,      SLOT( slotOpenConfig() ) );
    connect( ui->pbSaveAs,      SIGNAL( pressed() ), this,      SLOT( slotSaveAs() ) );
    connect( ui->pbStart,       SIGNAL( pressed() ), this,      SLOT( slotStart() ) );
    connect( ui->pbStop,        SIGNAL( pressed() ), this,      SLOT( slotStop() ) );
    connect( ui->pbQuit,        SIGNAL( pressed() ), this,      SLOT( slotQuit() ) );
    connect( ui->pbGnuplot,     SIGNAL( pressed() ), this,      SLOT( slotFindGNUplot() ) );


    //device
    connect( ui->pbp10MT,       SIGNAL( pressed() ), this,       SLOT( slotTuneAdd10M() ) );
    connect( ui->pbp1MT,       SIGNAL( pressed() ), this,       SLOT( slotTuneAdd1M() ) );
    connect( ui->pbp100kT,       SIGNAL( pressed() ), this,       SLOT( slotTuneAdd100k() ) );
    connect( ui->pbp10kT,        SIGNAL( pressed() ), this,       SLOT( slotTuneAdd10k() ) );
    connect( ui->pbp1kT,        SIGNAL( pressed() ), this,       SLOT( slotTuneAdd1k() ) );
    connect( ui->pbm10MT,       SIGNAL( pressed() ), this,       SLOT( slotTuneSub10M() ) );
    connect( ui->pbm1MT,       SIGNAL( pressed() ), this,       SLOT( slotTuneSub1M() ) );
    connect( ui->pbm100kT,       SIGNAL( pressed() ), this,       SLOT( slotTuneSub100k() ) );
    connect( ui->pbm10kT,        SIGNAL( pressed() ), this,       SLOT( slotTuneSub10k() ) );
    connect( ui->pbm1kT,        SIGNAL( pressed() ), this,       SLOT( slotTuneSub1k() ) );

    connect( ui->pbp1M,        SIGNAL( pressed() ), this,       SLOT( slotBWadd1M() ) );
    connect( ui->pbp100k,      SIGNAL( pressed() ), this,       SLOT( slotBWadd100k() ) );
    connect( ui->pbp10k,       SIGNAL( pressed() ), this,       SLOT( slotBWadd10k() ) );
    connect( ui->pbp1k,        SIGNAL( pressed() ), this,       SLOT( slotBWadd1k() ) );
    connect( ui->pbm1M,        SIGNAL( pressed() ), this,       SLOT( slotBWsub1M() ) );
    connect( ui->pbm100k,      SIGNAL( pressed() ), this,       SLOT( slotBWsub100k() ) );
    connect( ui->pbm10k,       SIGNAL( pressed() ), this,       SLOT( slotBWsub10k() ) );
    connect( ui->pbm1k,        SIGNAL( pressed() ), this,       SLOT( slotBWsub1k() ) );

    connect( ui->lcdBW,        SIGNAL( customContextMenuRequested(const QPoint&) ), this, SLOT( slotEnterBW  (const QPoint&) ) );
    connect( ui->lcdTune,      SIGNAL( customContextMenuRequested(const QPoint&) ), this, SLOT( slotEnterTune(const QPoint&) ) );
    connect( ui->chkBypass,    SIGNAL( stateChanged(int) ), this, SLOT( slotSetSsBypass(int) ) );

    //variable elements comboboxes: strings must be converted to indexes before forwarding events to as
    connect( ui->cbDevice,      SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( slotSetDevice     (const QString &) ) );
    connect( ui->cbSR,          SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( slotSetSampleRate (const QString &) ) );
    connect( ui->cbGain,        SIGNAL( currentIndexChanged(const QString &) ), this, SLOT( slotSetGain       (const QString &) ) );
    connect( ui->sbPpm,         SIGNAL( valueChanged(double) ), as, SLOT( setError  (double) ) );

    //FFT settings
    connect( ui->cbFtw,         SIGNAL( currentIndexChanged(int) ), as, SLOT( setWindow(int) ) );
    connect( ui->pbIncRes,      SIGNAL( pressed() ), as,        SLOT( incResolution() ) );
    connect( ui->pbDecRes,      SIGNAL( pressed() ), as,        SLOT( decResolution() ) );

    //output settings
    connect( ui->rbContinuous,  &QRadioButton::pressed, [=] () { this->slotSetAcqMode(AM_CONTINUOUS); });
    connect( ui->rbPeriodic,    &QRadioButton::pressed, [=] () { this->slotSetAcqMode(AM_PERIODIC); });
    connect( ui->rbAutomatic,   &QRadioButton::pressed, [=] () { this->slotSetAcqMode(AM_AUTOMATIC); });

    connect( ui->sbInterval,    SIGNAL( valueChanged(int) ), as, SLOT( setInterval(int) ) );
    connect( ui->sbMaxShots,    SIGNAL( valueChanged(int) ), as, SLOT( setShots     (int) ) );
    connect( ui->sbRecTime,     SIGNAL( valueChanged(int) ), as, SLOT( setRecTime   (int) ) );
    connect( ui->sbAfter,       SIGNAL( valueChanged(int) ), as, SLOT( setAfter     (int) ) );
    connect( ui->cbThresholds,  SIGNAL( currentIndexChanged(int) ), as, SLOT( setThresholdsBehavior(int)));
    connect( ui->sbUpThreshold, SIGNAL( valueChanged(double) ), as, SLOT( setUpThreshold (double) ) );
    connect( ui->sbDnThreshold, SIGNAL( valueChanged(double) ), as, SLOT( setDnThreshold (double) ) );
    connect( ui->sbTrange,      SIGNAL( valueChanged(int) ), as, SLOT( setTrange    (int) ) );
    connect( ui->chkGnuplot,    SIGNAL( stateChanged(int) ), as, SLOT( setGNUplot(int)) );
    connect( ui->chkSShots,     SIGNAL( stateChanged(int) ), as, SLOT( setScreenshots(int)) );
    connect( ui->sbMinFree,     SIGNAL( valueChanged(int) ), as, SLOT( setMinFree(int)) );
    connect( ui->chkEraseOldLogs,     SIGNAL( stateChanged(int) ), as, SLOT( setEraseLogs   (int) ) );
    connect( ui->chkEraseOldShots,    SIGNAL( stateChanged(int) ), as, SLOT( setEraseShots   (int) ) );
    connect( ui->cbPlotType,    SIGNAL(currentIndexChanged(int)), as, SLOT( setPlotType(int) ) );
    connect( ui->sbLasting,     SIGNAL( valueChanged(int) ), as, SLOT( setDataLasting(int) ) );
    connect( ui->sbAvg,         SIGNAL( valueChanged(int) ), as, SLOT( setAvgdScans(int) ) );
    connect( ui->sbJoinTime,    SIGNAL( valueChanged(int) ), as, SLOT( setJoinTime(int) ) );

    //report settings
    connect( ui->dtFrom,        SIGNAL(dateChanged(QDate)), as, SLOT( setCoverFrom(const QDate&) ) );
    connect( ui->dtTo,          SIGNAL(dateChanged(QDate)), as, SLOT( setCoverTo(const QDate&) ) );
    connect( ui->sbUnderThr,    SIGNAL( valueChanged(int) ), as, SLOT( setUnderdenseMs   (int) ) );
    connect( ui->sbOverThr,     SIGNAL( valueChanged(int) ), as, SLOT( setOverdenseSec   (int) ) );
    connect( ui->chkColorgraph, SIGNAL( stateChanged(int) ), as, SLOT( setColorGraph     (int) ) );
    connect( ui->chkSite,       SIGNAL( stateChanged(int) ), as, SLOT( setSiteInfos      (int) ) );
    connect( ui->chkSummary,    SIGNAL( stateChanged(int) ), as, SLOT( setSummary        (int) ) );
    connect( ui->chkDetails,    SIGNAL( stateChanged(int) ), as, SLOT( setShotsDetails   (int) ) );
    connect( ui->pbReport ,     SIGNAL( pressed() ), this, SLOT( slotReport() ) );
    connect( ui->pbSiteInfos ,  SIGNAL( pressed() ), this, SLOT( slotSiteInfos() ) );
    connect( ui->pbFlatten,     SIGNAL( pressed() ), this, SLOT( slotFlatten() ) );

    connect( ui->chkOverdenseShots,   SIGNAL( stateChanged(int) ), as, SLOT( setOverdenseShots (int) ) );
    connect( ui->chkUnderdenseShots,  SIGNAL( stateChanged(int) ), as, SLOT( setUnderdenseShots(int) ) );
    connect( ui->chkOverdensePlots ,  SIGNAL( stateChanged(int) ), as, SLOT( setOverdensePlots (int) ) );
    connect( ui->chkUnderdensePlots,  SIGNAL( stateChanged(int) ), as, SLOT( setUnderdensePlots(int) ) );
    connect( ui->chkRestart,    SIGNAL( stateChanged(int) ), as, SLOT( setRestart(int) ) );
    connect( ui->chkAutoReport, SIGNAL( stateChanged(int) ), as, SLOT( setAutoReport(int) ) );
    connect( ui->chkCarrierFilter, SIGNAL( stateChanged(int) ), as, SLOT( setCarrierFilter(int) ) );
    connect( ui->chkLightingsFilter, SIGNAL( stateChanged(int) ), as, SLOT( setLightingsFilter(int) ) );
    connect( ui->chkSatFilter, SIGNAL( stateChanged(int) ), as, SLOT( setSaturationFilter(int) ) );


    //preferences
    connect( ui->chkFticks,    SIGNAL( stateChanged(int) ), as, SLOT( setHz             (int) ) );
    connect( ui->chkPticks,    SIGNAL( stateChanged(int) ), as, SLOT( setDbfs           (int) ) );
    connect( ui->chkTticks,    SIGNAL( stateChanged(int) ), as, SLOT( setSec            (int) ) );
    connect( ui->chkTooltips,    SIGNAL( stateChanged(int) ), as, SLOT( setTooltips     (int) ) );
    connect( ui->chkEnablePing,  SIGNAL( stateChanged(int) ), as, SLOT( setPing         (int) ) );
    connect( ui->cbPalette,     SIGNAL( currentIndexChanged(int) ), as, SLOT( setPalette       (int) ) );
    connect( ui->sbNoiseLimit,  SIGNAL( valueChanged(int) ), as, SLOT( setNoiseLimit(int)));
    connect( ui->sbServerPort,  SIGNAL( valueChanged(int) ), as, SLOT( setServerPort(int)));
    connect( qApp,  SIGNAL( focusChanged(QWidget*, QWidget*) ), this, SLOT( slotFocusChanged(QWidget*, QWidget*)) );

    //statusbar
    connect( r,  SIGNAL( status(const QString&, int) ), this, SLOT( slotShowStatus(const QString&, int) ), Qt::QueuedConnection );
    connect( appControl,  SIGNAL( status(const QString&) ), this, SLOT( slotShowStatus(const QString&) ), Qt::QueuedConnection );
    connect( appControl,  SIGNAL( stop() ), this, SLOT( slotAcqStopped() ), Qt::QueuedConnection );
    connect( appControl,  SIGNAL( start() ), this, SLOT( slotAcqStarted() ), Qt::QueuedConnection );

    //feedbacks
    connect( as,  SIGNAL( notifySetResolution() ),  this, SLOT( slotAdjustResolution() ) );

    connect( as,  SIGNAL( notifyOpened() ),         this, SLOT( slotReopen() ) );
    connect( as,  SIGNAL( notifySetUpThreshold()),  this, SLOT( slotValidateUpThreshold() ) );
    connect( as,  SIGNAL( notifySetDnThreshold()),  this, SLOT( slotValidateDnThreshold() ) );
    connect( as,  SIGNAL( notifySetThresholdsBehavior()), this, SLOT( slotSetSbThresholdsParam() ) );

    connect( pp,  SIGNAL( autoReport() ), this, SLOT( slotAutoReport() ) );
    connect( si,  SIGNAL( logoChanged()), wf, SLOT( slotUpdateLogo() ) );
    connect( sysCheck, SIGNAL(timeout()), this, SLOT( slotUpdateSysInfo()) );

    connect( ac, SIGNAL( GUIbusy() ), this, SLOT(showBusy()));
    connect( ac, SIGNAL( GUIfree() ), this, SLOT(showFree()));

    connect( r,  SIGNAL( GUIupdateNeeded() ), this, SLOT( slotGUIinit() ), Qt::QueuedConnection );


    ast = AST_INIT;

    qApp->installEventFilter(this);

    QScreen *screen = QGuiApplication::primaryScreen();
    QRect  screenGeometry = screen->geometry();
    int screenHeight = screenGeometry.height();

    if( height() > screenHeight )
    {
        //for small screens: shrinks vertically the main window
        //to avoid the start button go out the screen

        resize( width(), screenHeight );
    }


    slotReopen();

    MYDEBUG << "done." << endl;
}

MainWindow::~MainWindow()
{
    delete ui;
}

//to disable tooltips globally
bool MainWindow::eventFilter(QObject *obj, QEvent *event)
{
    if(as->getTooltips() != 0)
    {
        if (event->type() == QEvent::ToolTip)
        {
            return true;
        }
        else
        {
            return QMainWindow::eventFilter(obj, event);
        }
    }
    else
    {
        return QMainWindow::eventFilter(obj, event);
    }
}



void MainWindow::slotGUIinit()
{
    int i = -1;
    float d = 0.0;
    QString s = "";

    MYDEBUG << "initializing GUI" << endl;

    blockAllSignals(true);

    outputEnabled = false;

    ui->tabWidget->setCurrentIndex(0);

    //device settings

    if(as->getBandwidth() > MAX_SAMPLE_RATE_HZ)
    {
        bw = MAX_SAMPLE_RATE_HZ;
        as->setBandwidth(bw);
    }

    //then inizializes the value of every control, starting from comboboxes:

    QStringList sl = r->deviceOpened()->getRates();
    ui->cbSR->clear();
    ui->cbSR->addItems( sl );
    MY_ASSERT( ui->cbSR->count() == sl.count() )
    s.setNum( as->getSampleRate() );
    i = ui->cbSR->findText( s, Qt::MatchFixedString );
    if(i == -1)
    {
        MYDEBUG << "string " << s << " not found in ui->cbSR" << endl;
    }
    else
    {
        ui->cbSR->setCurrentIndex( i );
    }


    ui->cbGain->clear();
    sl = r->deviceOpened()->getGains();
    ui->cbGain->addItems( sl );
    MY_ASSERT( ui->cbGain->count() == sl.count() )
    float fGain = static_cast<float>(as->getGain()) / 10;
    if(fGain >= 0)
    {
        s.setNum(fGain, 'f', 2);
    }
    else
    {
        s = "AUTO";
    }

    i = ui->cbGain->findText( s );
    if(i == -1)
    {
        MYDEBUG << "string " << s << " not found in ui->cbGain" << endl;
        s = sl.last();
        i = ui->cbGain->findText( s );
    }
    ui->cbGain->setCurrentIndex( i );

    s = as->getDevice();
    i = ui->cbDevice->findText( s, Qt::MatchFixedString );
    if(i == -1)
    {
        MYDEBUG << "string " << s << " not found in ui->cbDevice" << endl;
    }
    else
    {

        ui->cbDevice->setCurrentIndex( i );
    }

    Qt::CheckState k = static_cast<Qt::CheckState>( as->getSsBypass() );
    ui->chkBypass->setChecked(k);

    bw = as->getBandwidth();
    ui->lcdBW->display(bw);

    tune = as->getTune();
    ui->lcdTune->display(tune);

    d = as->getError();
    ui->sbPpm->setValue( static_cast<double>(d) );


    //cbFtw and cbOutput have fixed values hardcoded in ui
    //that match exactly FFT_WINDOWS and OUT_TYPES

    //FFT settings:
    i =  static_cast<int>( as->getWindow() );
    ui->cbFtw->setCurrentIndex( i );

    as->updateResolution();
    d = as->getResolution();   
    s.setNum(d, 'f', 2);
    ui->lbResolution->setText(s);

    s.setNum( as->getFFTbins() );
    ui->lbPoints->setText(s);


    //output settings

    i = as->getAcqMode();
    switch(i)
    {
    default:
    case AM_CONTINUOUS:
         ui->rbContinuous->setChecked(true);
        break;

    case AM_PERIODIC:
         ui->rbPeriodic->setChecked(true);
        break;

    case AM_AUTOMATIC:
         ui->rbAutomatic->setChecked(true);
        break;
    }


    i = as->getInterval();
    ui->sbInterval->setMinimum(TIME_BASE);
    ui->sbInterval->setValue(i);

    i = as->getShots();
    ui->sbMaxShots->setValue(i);

    i = as->getRecTime();
    ui->sbRecTime->setValue(i);

    i = as->getAfter();
    ui->sbAfter->setValue(i);

    slotSetSbThresholdsParam();

    d = as->getUpThreshold();
    ui->sbUpThreshold->setValue(static_cast<double>(d));

    d = as->getDnThreshold();
    ui->sbDnThreshold->setValue(static_cast<double>(d));

    i = as->getTrange();
    ui->sbTrange->setValue(i);

    k = static_cast<Qt::CheckState>( as->getGNUplot() );
    ui->chkGnuplot->setChecked(k);   

    k = static_cast<Qt::CheckState>( as->getScreenshots() );
    ui->chkSShots->setChecked(k);

    s = as->getGNUplotPath();

    if( s.isEmpty() == true || QFile::exists(s) == false )
    {
        //gnuplot not found
        ui->pbGnuplot->show();
    }
    else
    {
        //gnuplot found, non need to localize it
        ui->pbGnuplot->hide();
    }


    i = as->getPlotType();
    ui->cbPlotType->setCurrentIndex(i);

    i = as->getDataLasting();
    ui->sbLasting->setValue(i);

    i = as->getNoiseLimit();
    ui->sbNoiseLimit->setValue(i);

    i = as->getAvgdScans();
    ui->sbAvg->setValue(i);

    i = as->getJoinTime();
    ui->sbJoinTime->setValue(i);

    i = as->getMinFree();
    ui->sbMinFree->setValue(i);

    slotUpdateSysInfo();

    //report settings

    //at startup the coverage takes the same
    //amount of days stored in rts but shifted
    //forward to match today as ending date
    QDate from = as->getCoverFrom();
    QDate to = as->getCoverTo();
    qint64 daysCovered = from.daysTo(to);
    QDate todayUTC = QDateTime::currentDateTimeUtc().date();
    if(to < todayUTC)
	{
        to =  todayUTC;
	}
    ui->dtTo->setDate(to);
    from = to.addDays(-daysCovered);
    ui->dtFrom->setDate(from);

    k = static_cast<Qt::CheckState>( as->getSummary() );
    ui->chkSummary->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getColorGraph() );
    ui->chkColorgraph->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getSiteInfos() );
    ui->chkSite->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getOverdenseShots() );
    ui->chkOverdenseShots->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getUnderdenseShots() );
    ui->chkUnderdenseShots->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getOverdensePlots() );
    ui->chkOverdensePlots->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getUnderdensePlots() );
    ui->chkUnderdensePlots->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getShotsDetails() );
    ui->chkDetails->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getRestart() );
    ui->chkRestart->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getAutoReport() );
    ui->chkAutoReport->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getCarrierFilter() );
    ui->chkCarrierFilter->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getLightingsFilter() );
    ui->chkLightingsFilter->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getSaturationFilter() );
    ui->chkSatFilter->setChecked(k);

    i = as->getUnderdenseMs();
    ui->sbUnderThr->setValue(i);

    i = as->getOverdenseSec();
    ui->sbOverThr->setValue(i);

    //preferences
    k = static_cast<Qt::CheckState>( as->getSec() );
    ui->chkTticks->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getHz() );
    ui->chkFticks->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getDbfs() );
    ui->chkPticks->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getTooltips() );
    ui->chkTooltips->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getPing() );
    ui->chkEnablePing->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getEraseLogs() );
    ui->chkEraseOldLogs->setChecked(k);

    k = static_cast<Qt::CheckState>( as->getEraseShots() );
    ui->chkEraseOldShots->setChecked(k);

    i = as->getThresholdsBehavior();
    ui->cbThresholds->setCurrentIndex(i);

    i = as->getPalette();
    ui->cbPalette->setCurrentIndex(i);

    i = as->getServerPort();
    ui->sbServerPort->setValue(i);


    s = as->getServerAddress().toString();
    ui->leServerIP->setText(s);

    wf->show(); 

    //restore geometry
    setGeometry( as->getMainGeometry() );

    //Initialize subwindows
    wf->GUIinit();
    si->GUIinit();

    //Configuration applied. Any further change will
    //enable the autosave feature
    if(init && !deviceChanged)
    {
        as->clearChanged();
    }

    blockAllSignals(false);

    setTitle();

    init = false;
}




void MainWindow::GUIrefresh()
{
    MYDEBUG << "refreshing GUI" << endl;
    Qt::CheckState k;

    blockAllSignals(true);

    slotUpdateSysInfo();

    //report settings
    QDate dt = as->getCoverFrom();
    ui->dtFrom->setDate(dt);

    dt = as->getCoverTo();
    ui->dtTo->setDate(dt);

    k = static_cast<Qt::CheckState>( as->getAutoReport() );
    ui->chkAutoReport->setChecked(k);

    blockAllSignals(false);
}




//buttons handling:

void MainWindow::slotQuit()
{
     MYDEBUG << "Quit";

    //stops the threads smoothly
    //r->stopMe();
    //r->wait(10000);

    ac->slotStopAcquisition(0);

    PostProc* pp = ac->postProcessor();
    MY_ASSERT(pp != nullptr)
    pp->saveData();

    //closes the application
    qApp->closeAllWindows();
}


void MainWindow::slotAboutDialog()
{
    QString desc;
    QTextStream ts(&desc);
    XQDir wd = ac->workingDir();
    QString gpl = "Copyright (C) 2021  Giuseppe Massimo Bertani "
        "gmbertani(a)users.sourceforge.net "
        "This program is free software: you can redistribute it and/or modify "
        "it under the terms of the GNU General Public License as published by "
        "the Free Software Foundation, version 3 of the License."
        "This program is distributed in the hope that it will be useful, "
        "but WITHOUT ANY WARRANTY; without even the implied warranty of "
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the "
        "GNU General Public License for more details. "
        "You should have received a copy of the GNU General Public License "
        "along with this program.  If not, http://www.gnu.org/copyleft/gpl.html ";


    ts << "ECHOES - " << tr(TARGET_DESC) << endl;
    ts << "V." << APP_VERSION << "." << endl;
    ts << "Build date:" << BUILD_DATE << "." << endl;

#ifdef _DEBUG
     ts << "UNOFFICIAL-COMPILED FOR DEBUGGING PURPOSES" << endl;
#endif
    ts << "Build architecture: " << QSysInfo::buildAbi() << endl;
    ts << "Qt version: " << QT_VERSION_STR << endl;
    ts << endl << gpl << endl << endl;
    ts << tr("Working directory: ") << endl ;
    ts << wd.absolutePath() << endl << endl;

    ts << "Currently running on " << QSysInfo::productType() << " v." << QSysInfo::productVersion()
        << " (" <<  QSysInfo::currentCpuArchitecture() << ")" << endl;

    QMessageBox::about
    (
        this,
        tr("Informations about Echoes"),
        desc
    );

}


void MainWindow::slotReopen()
{
    MYDEBUG << "Reopen" << endl;

    bool ok = false;

    as->blockSignals(true);
    showBusy();

    ac->slotStopAcquisition(0);

    wf->slotConfigChanged();

    //by default opens the first RTLSDR device found
    //if none is found, the first device will be the dongle server.
    //If none available, as fallback opens the test patterns device

    //load combobox values first
    DeviceMap dm, curr;
    
    ui->cbDevice->blockSignals(true);
    ui->cbDevice->clear();
    foreach(dm, r->getDevices(true))
    {
        MYINFO << "adding " << dm.name;
        ui->cbDevice->addItem( dm.name );
        if(dm.name == ui->cbDevice->currentText())
        {
            curr = dm;
        }
    }

    MY_ASSERT( ui->cbDevice->count() == r->getDevices().count() ) //paranoia: the max nr.of elements has been set in designer

    int idx = 0;
    MYINFO << "ui->cbDevice->currentText()=" << ui->cbDevice->currentText() << "  as->getDevice()=" <<  as->getDevice();
    if( as->getDevice() != ui->cbDevice->currentText() )
    {
        idx = ui->cbDevice->findText( as->getDevice() );
        if (idx < 0)
        {
            //The device specified in rts file is not listed in the combobox or has a different index number
            //
            //Retry to find it, after removing the index number
            QString deviceInCb;
            for(idx = 0; idx <  ui->cbDevice->count(); idx++)
            {
                deviceInCb = ui->cbDevice->itemText(idx);
                if( deviceInCb.contains( as->getDeviceNoIdx() ))
                {
                    break;
                }
                MYWARNING << "Device " << as->getDeviceNoIdx() << " not in " << deviceInCb;
            }

            if(idx >= ui->cbDevice->count())
            {
                QMessageBox::warning(this, QObject::tr("Warning"),
                    QObject::tr("Mismatched dongle in configuration file.\n"
                                "Please select an available device\n"
                                "before starting acquisition")
                );

                ast = AST_ERROR;
                slotShowStatus( tr("Please select another input device"), ast);
                curr = r->getDevices().at( r->getDevices().length() - 1 );
                ui->cbDevice->setCurrentIndex(curr.index);
                ui->cbDevice->blockSignals(false);
                as->blockSignals(false);
                showFree();
                return;

            }
        }

        if (idx >= 0)
        {
            curr = r->getDevices().at(idx);
            MYINFO << "Device " << as->getDeviceNoIdx() << " found";
        }
    }


    ui->cbDevice->setCurrentIndex(idx);    
    ui->cbDevice->blockSignals(false);

    while(idx < r->getDevices().count())
    {
        APP_STATE ast = r->slotSetDevice( curr.index );
        if(ast == AST_OPEN)
        {
            slotShowStatus( tr("Input device opened"), ast);

            if( r->slotSetSampleRate() == true )
            {
                if( r->slotSetBandwidth() == true )
                {
                    if(r->slotSetResolution() == true)
                    {
                        if(r->slotSetGain() == true)
                        {
                            if(r->slotSetError() == true)
                            {
                                if(r->slotSetTune() == true)
                                {
                                    ok = true;
                                    as->setDevice( r->getDeviceName() );
                                    slotShowStatus( tr("Input device opened"), ast);
                                    MYINFO << "Device " << as->getDevice() << " opened";
                                    break;
                                }
                            }
                        }
                    }
                }
            }
        }
        QString cfgFullPath =  as->getConfigFullPath();
        as->loadConfig( cfgFullPath );
        idx++;
    }

    if(ok == false && ast < AST_OPEN)
    {
        MYCRITICAL << "Failed input device configuration, ast=" << ast;
        slotQuit();
        exit(0);
    }

    slotGUIinit();

    sysCheck->setInterval(SYS_CHECK_INTERVAL);
    sysCheck->setSingleShot(false);
    sysCheck->start();

    showFree();

    as->blockSignals(false);
}

void MainWindow::slotOpenConfig()
{

#ifdef NO_NATIVE_DIALOGS
    //native dialogs buggy under linux
    QString fileName = QFileDialog::getOpenFileName(
        this,
        tr("Open configuration file"),
        ac->workingDir().absolutePath(),
        tr("Echoes configuration file (*.rts *.RTS)"),
        0,
        QFileDialog::DontUseNativeDialog);

#else

    QString fileName = QFileDialog::getOpenFileName(
        this,
        tr("Open configuration file"),
        ac->workingDir().absolutePath(),
        tr("Echoes configuration file (*.rts *.RTS)"));
#endif

    if( fileName.isEmpty() )
    {
        //no file specified
        return;
    }

    slotShowStatus(tr("Loading configuration..."));
    as->loadConfig(fileName);

    PostProc* pp = ac->postProcessor();
    MY_ASSERT(pp != nullptr)
    pp->loadData();

    ui->lbConfig->setToolTip(tr("Configuration file loaded: ") + fileName);
    slotShowStatus(tr("Ready."));
    setTitle();

}


void MainWindow::slotFindGNUplot()
{

#ifdef NO_NATIVE_DIALOGS
    //native dialogs buggy under linux

    QString fileName = QFileDialog::getOpenFileName(
        this,
        tr("Find GNUplot executable"),
        as->getGNUplotPath(), //ac->workingDir().absolutePath(),
        tr("GNUplot executable file (gnuplot gnuplot.exe)"),
        0,
        QFileDialog::DontUseNativeDialog);
#else
    QString fileName = QFileDialog::getOpenFileName(
        this,
        tr("Find GNUplot executable"),
        as->getGNUplotPath(), //ac->workingDir().absolutePath(),
        tr("GNUplot executable file (*)"));
#endif
    if( fileName.isEmpty() )
    {
        //no file specified
        return;
    }
    MYINFO << "Found GNUplot at " << fileName;
    slotShowStatus(tr("Locating GNUplot..."));
    as->setGNUplotPath( fileName );
    slotShowStatus(tr("Ready."));
}




void MainWindow::slotFlatten()
{
    PostProc* pp = ac->postProcessor();
    if(pp == nullptr)
    {
        MYWARNING << "PostProcessor not yet instantiated, can't flatten";
        return;
    }

    QString dirName = QFileDialog::getExistingDirectory(
        this,
        tr("Find a folder where to save the flatten archive"),
        ac->workingDir().absolutePath());

    if( dirName.isEmpty() )
    {
        //no folder specified
        return;
    }
    MYINFO << "Found folder " << dirName;
    slotShowStatus(tr("Flattening data archive..."));

    XQDir dest(dirName);
    uint cnt = dest.count();
    if(cnt > 2) //2 are the current directory and its parent, always present
    {
        int sb = QMessageBox::warning(this,
                    tr("Flattening data archive"),
                    tr("The directory %1 is not empty.\nFlattening will destroy all the existing data.\nContinue?")
                    .arg(dirName),
                    QMessageBox::Yes|QMessageBox::No );

        if( sb == QMessageBox::No )
        {
            return;
        }
        MYINFO << "deleting existing directory " << dirName << " and its contents";
        if( dest.removeRecursively() == false )
        {
            QMessageBox::critical(this,
                tr("Error"),
                tr("Unable to clean directory %1 .\nAborting.")
                .arg(dirName),
                QMessageBox::Ok );
            MYCRITICAL << "Unable to clean directory " << dirName << " aborting";
            return;
        }
    }
    MYINFO << "creating a brand new " << dirName;
    if( dest.mkpath(dirName) == false )
    {
        QMessageBox::critical(this,
            tr("Error"),
            tr("Unable to create flatten archive under %1 .\nAborting.")
            .arg(dirName),
            QMessageBox::Ok );
        MYCRITICAL << "Unable to create flatten archive under " << dirName << " aborting";
        return;
    }

    showBusy();
    pp->setWdir(ac->workingDir());
    pp->setFlatDir(dirName);
    pp->setTask( PP_FLATTEN );
    pp->start( QThread::IdlePriority );
    while (pp->isRunning() == false)
    {
        QThread::msleep(100);
    }
    MYINFO << "+++++ PostProc thread started";
    pp->wait();

    MYINFO << "+++++ PostProc thread terminated";
    showFree();

    slotShowStatus(tr("Ready."));
}



void MainWindow::slotSaveAs()
{
#ifdef NO_NATIVE_DIALOGS
    //native dialogs buggy under linux
    QString filePath = QFileDialog::getSaveFileName(
        this,
        tr("Save configuration as"),
        ac->workingDir().absolutePath(),
        tr("Echoes configuration file (*.rts *.RTS)"),
        0,
        QFileDialog::DontUseNativeDialog);
#else
    QString filePath = QFileDialog::getSaveFileName(
        this,
        tr("Save configuration as"),
        ac->workingDir().absolutePath(),
        tr("Echoes configuration file (*.rts *.RTS)"));
#endif

    if( filePath.isEmpty() )
    {
        //nessun nome di file specificato.
        return;
    }

    slotShowStatus(tr("Saving configuration..."));

    QFileInfo fi(filePath);
    QString baseName = fi.baseName();

    //as->setConfigName( ac->workingDir(), baseName );
    as->setConfigName( fi.dir(), baseName );

    PostProc* pp = ac->postProcessor();
    MY_ASSERT(pp != nullptr)
    pp->saveData();

    setTitle();
    slotShowStatus(tr("Ready."));
}

void MainWindow::slotStart()
{
    MYINFO << "Pressed Start";
    if(ast == AST_INIT)
    {
        slotReopen();
    }

    if(ast == AST_OPEN)
    {
        //acquisition start
        ac->slotAcquisition(true);
        slotShowStatus( tr("Acquisition running"), AST_STARTED );
    }
}


void MainWindow::slotStop()
{
    MYINFO << "Pressed Stop";

    if(ast == AST_STARTED)
    {
        ac->slotAcquisition(false);
    }
}


void MainWindow::slotAcqStopped()
{
    //called by Control to notify about an
    //automatic acquisition stop

    slotShowStatus( tr("Acquisition stopped"), AST_OPEN );
}


void MainWindow::slotAcqStarted()
{
    //called by Control to notify about an
    //automatic acquisition restart (after midnight UTC)

    slotShowStatus( tr("Acquisition restarted"), AST_STARTED );
}

void MainWindow::slotSiteInfos()
{
    si->show();
}


void MainWindow::slotReport()
{
    PostProc* pp = ac->postProcessor();

    //Checks the dates that the report should cover and corrects
    //them in case they are in future.
    //This check must NOT be performed if autoreport is active
    checkReportCoverage();

    QString docName;

    QDate from = as->getCoverFrom();
    QDate to = as->getCoverTo();
    int coverage = static_cast<int>( as->getCoverFrom().daysTo( as->getCoverTo() ) );
    if(coverage < 1 || coverage > as->getDataLasting())
    {

        slotShowStatus(tr("Wrong report coverage "), AST_ERROR );
        MYWARNING << "Wrong coverage for report: " << coverage << " days";
        MYWARNING << "Archive can contain max " << as->getDataLasting() << " days";
        ui->dtTo->setFocus();
        return;
    }

    docName = QString("%1_%2_to_%3.html").arg( as->getConfigName() ).arg(from.toString(Qt::ISODate)).arg(to.toString(Qt::ISODate));


    if(pp != nullptr)
    {
        if( pp->isRunning() == true )
        {
            //manual report is not active while processing daily data at midnight UTC.
            MYINFO << "Manual report not active while processing daily data";
            slotShowStatus( tr("Try later"));
        }
        else if( pp->archiveExists(from,to) == false )
        {
            //same if there is no archive yet for the current configuration

            MYINFO << "No archives yet to process for the current configuration " << as->getConfigName()
                   << " and the requested date interval " << from.toString() << " to " << to.toString();
            slotShowStatus( QString(tr("No <%1> data to report yet")).arg(as->getConfigName()) );
        }
        else
        {
            //note that the report must be created from GUI thread
            //because it creates some temporary widgets
            slotShowStatus( tr("Creating report"));
            showBusy();

            pp->createFullReport( docName );

            //updates the interval to generate the next report
            QDate newFrom = as->getCoverTo().addDays(1);
            MYINFO << "Next report coverage from " << newFrom;
            QDate newTo = newFrom.addDays( coverage );
            MYINFO << " to " << newTo;

            showFree();
            slotShowStatus( tr("Ready"));
        }

        if(as->getRestart() != 0 && ast != AST_STARTED)
        {
            MYINFO << "Acquisition self-restart";
            slotStart();
        }
    }
}


void MainWindow::slotAutoReport()
{
    PostProc* pp = ac->postProcessor();

    QString docName;

    QDate from = as->getCoverFrom();
    QDate to = as->getCoverTo();
    int coverage = static_cast<int>( as->getCoverFrom().daysTo( as->getCoverTo() ) );
    if(coverage < 1)
    {
        slotShowStatus(tr("Wrong parameter"), AST_ERROR );
        MYWARNING << "Wrong coverage for report: " << coverage << " days";
        ui->dtTo->setFocus();
        return;
    }

    docName = QString("%1_%2_to_%3.html").arg( as->getConfigName() ).arg(from.toString(Qt::ISODate)).arg(to.toString(Qt::ISODate));


    if(pp != nullptr)
    {
        if( pp->archiveExists(from,to) == false )
        {
            //if there is no archive yet for the current configuration
            //the report can't be produced.
            MYINFO << "No archives to process for the current configuration " << as->getConfigName()
                   << " and the requested date interval " << from.toString() << " to " << to.toString();
            slotShowStatus( tr("No data to report yet"));
        }
        else
        {
            //note that the report must be created from GUI thread
            //because it creates some temporary widgets
            slotShowStatus( tr("Creating report"));
            showBusy();

            pp->createFullReport( docName );

            //updates the interval to generate the next report
            QDate newFrom = as->getCoverTo().addDays(1);
            MYINFO << "Next report coverage from " << newFrom;
            QDate newTo = newFrom.addDays( coverage );
            MYINFO << " to " << newTo;

            if(as->getAutoReport() == Qt::Checked)
            {
                //automatic report, updating dates for next report.
                //note that the coverTo date must be set BEFORE the coverFrom
                //to avoid overlapping and automatic corrections by Settings object.
                as->setCoverTo( newTo );
                as->setCoverFrom( newFrom );
                GUIrefresh();
            }
            showFree();
            slotShowStatus( tr("Ready"));
        }

        if(as->getRestart() != 0 && ast != AST_STARTED)
        {
            MYINFO << "Acquisition self-restart";
            slotStart();
        }
    }
}



void MainWindow::slotEnterBW(const QPoint& pos)
{
    bool ok;

    MYDEBUG << pos;

    int ssr = QInputDialog::getInt(this, tr("Enter Bandwidth [Hz]"),
        tr("Frequency [Hz]:"), bw, 0, MAX_SAMPLE_RATE_HZ, 1, &ok);

    if (ok == true)
    {
        bw = ssr;
        ui->lcdBW->display(bw);
        as->setBandwidth(bw);
    }
}


void MainWindow::slotEnterTune(const QPoint& pos)
{
    bool ok;


    MYDEBUG << pos;

    int freq = QInputDialog::getInt(this, tr("Enter tune"),
        tr("Frequency [Hz]:"), tune, 0, MAX_TUNE_HZ, 1, &ok);

    if (ok == true)
    {
        tune = freq;
        ui->lcdTune->display(tune);
        as->setTune(tune);
    }
}

void MainWindow::slotSetSsBypass( int value )
{
    if(value == Qt::Checked)
    {
        ui->pbp1M->setEnabled(false);
        ui->pbp100k->setEnabled(false);
        ui->pbp10k->setEnabled(false);
        ui->pbp1k->setEnabled(false);

        ui->pbm1M->setEnabled(false);
        ui->pbm100k->setEnabled(false);
        ui->pbm10k->setEnabled(false);
        ui->pbm1k->setEnabled(false);
        bw = as->getSampleRate();
        as->setBandwidth( bw );
        ui->lcdBW->display(bw);
    }
    else
    {
        ui->pbp1M->setEnabled(true);
        ui->pbp100k->setEnabled(true);
        ui->pbp10k->setEnabled(true);
        ui->pbp1k->setEnabled(true);

        ui->pbm1M->setEnabled(true);
        ui->pbm100k->setEnabled(true);
        ui->pbm10k->setEnabled(true);
        ui->pbm1k->setEnabled(true);
     }

     as->setSsBypass( value == Qt::Checked );
}

//---------------------------
void MainWindow::slotTuneAdd10M()
{

    //+10Mhz steps:
    tune = as->getTune();
    if((tune+10e6) <= MAX_TUNE_HZ)
    {
        tune += 10e6;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }
}

void MainWindow::slotTuneAdd1M()
{
    //+1MHz steps
    tune = as->getTune();
    if((tune+1e6) <= MAX_TUNE_HZ)
    {
        tune += 1e6;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }
}

void MainWindow::slotTuneAdd100k()
{

    //+100kHz steps:
    tune = as->getTune();
    if((tune+100e3) <= MAX_TUNE_HZ)
    {
        tune += 100e3;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }
}

void MainWindow::slotTuneAdd10k()
{

    //+10kHz steps:

    tune = as->getTune();
    if((tune+10000) <= MAX_TUNE_HZ)
    {
        tune += 10000;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }
}

void MainWindow::slotTuneAdd1k()
{

    //+1kHz steps:

    tune = as->getTune();
    if((tune+1000) <= MAX_TUNE_HZ)
    {
        tune += 1000;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }
}

void MainWindow::slotTuneSub10M()
{
    //-10MHz steps:

    tune = as->getTune();

    if(tune >= 10e6 &&
       (tune-10e6 > as->getSampleRate()) )
    {
        tune -= 10e6;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }
}

void MainWindow::slotTuneSub1M()
{
    //-1MHz steps
    tune = as->getTune();
    if(tune >= 1e6 &&
       (tune-1e6) > as->getSampleRate() )
    {
        tune -= 1e6;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }
}

void MainWindow::slotTuneSub100k()
{

    // -100kHz steps:

    tune = as->getTune();
    if(tune >= 100e3 &&
       (tune-100e3) > as->getSampleRate() )
    {
        tune -= 100e3;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }
}

void MainWindow::slotTuneSub10k()
{

    //-10kHz steps:
    tune = as->getTune();

    if(tune >= 10000 &&
       (tune-10000) > as->getSampleRate() )
    {
        tune -= 10000;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }

}

void MainWindow::slotTuneSub1k()
{

    //-1kHz steps:
    tune = as->getTune();

    if(tune >= 1000 &&
       (tune-1000) > as->getSampleRate() )
    {
        tune -= 1000;
        as->setTune(tune);
        tune = as->getTune();
        ui->lcdTune->display(tune);
    }

}
//---------------------------


void MainWindow::slotBWadd1k()
{
    if(as->getSsBypass() == true)
    {
        return;
    }

    //+1khz steps:
    bw = as->getBandwidth();
    if((bw+1000) <= as->getSampleRate())
    {
        bw += 1000;
        as->setBandwidth(bw);
        bw = as->getBandwidth();
        ui->lcdBW->display(bw);
    }
}

void MainWindow::slotBWadd1M()
{
    if(as->getSsBypass() == true)
    {
        return;
    }

    //+1MHz steps
    bw = as->getBandwidth();
    if((bw+1e6) <= as->getSampleRate())
    {
        bw += 1e6;
        as->setBandwidth(bw);
        bw = as->getBandwidth();
        ui->lcdBW->display(bw);
    }
}

void MainWindow::slotBWadd100k()
{
    if(as->getSsBypass() == true)
    {
        return;
    }

    //+100kHz steps:
    bw = as->getBandwidth();
    if((bw+100e3) <= as->getSampleRate())
    {
        bw += 100e3;
        as->setBandwidth(bw);
        bw = as->getBandwidth();
        ui->lcdBW->display(bw);
    }
}

void MainWindow::slotBWadd10k()
{
    if(as->getSsBypass() == true)
    {
        return;
    }

    //+10kHz steps:
    bw = as->getBandwidth();
    if((bw+10000) <= as->getSampleRate())
    {
        bw += 10000;
        as->setBandwidth(bw);
        bw = as->getBandwidth();
        ui->lcdBW->display(bw);
    }
}

void MainWindow::slotBWsub1k()
{
    if(as->getSsBypass() == true)
    {
        return;
    }

    //-1k steps:
    bw = as->getBandwidth();
    if(bw > 1000 /* &&
       ((bw-1000) > as->getSampleRate()) */)
    {
        bw -= 1000;
        as->setBandwidth(bw);
        bw = as->getBandwidth();
        ui->lcdBW->display(bw);
    }
}

void MainWindow::slotBWsub1M()
{
    if(as->getSsBypass() == true)
    {
        return;
    }

    //-1MHz steps
    //-1k steps:
    bw = as->getBandwidth();
    if(bw > 1e6 /* &&
       ((bw-1e6) > as->getSampleRate())*/ )
    {
        bw -= 1e6;
        as->setBandwidth(bw);
        bw = as->getBandwidth();
        ui->lcdBW->display(bw);
    }
}

void MainWindow::slotBWsub100k()
{
    if(as->getSsBypass() == true)
    {
        return;
    }

    // -100kHz steps:
    //-1k steps:
    bw = as->getBandwidth();
    if(bw > 100e3 /* &&
       ((bw-100e3) > as->getSampleRate()) */)
    {
        bw -= 100e3;
        as->setBandwidth(bw);
        bw = as->getBandwidth();
        ui->lcdBW->display(bw);
    }
}

void MainWindow::slotBWsub10k()
{
    if(as->getSsBypass() == true)
    {
        return;
    }

    //-10kHz steps:
    //-1k steps:
    bw = as->getBandwidth();
    if(bw > 10000 /* &&
       ((bw-10000) > as->getSampleRate()) */)
    {
        bw -= 10000;
        as->setBandwidth(bw);
        bw = as->getBandwidth();
        ui->lcdBW->display(bw);
    }
}







//--------------------------
bool MainWindow::slotSetDevice( const QString& text )
{
    int value = ui->cbDevice->findText( text, Qt::MatchFixedString );
    if(value == -1)
    {
        MYDEBUG << "string " << text << " not found in ui->cbDevice" << endl;
        return false;
    }

    as->setDevice(text, false);
    deviceChanged = true;

    slotShowStatus( tr("Device switched"), AST_INIT );
    slotReopen();
    return true;
}


void MainWindow::slotSetSampleRate( const QString& text )
{
    int value = ui->cbSR->findText( text, Qt::MatchFixedString );
    if(value == -1)
    {
        MYDEBUG << "string " << text << " not found in ui->cbSR" << endl;
    }
    else
    {

        QString valueString = ui->cbSR->itemText(value);
        bool ok = false;
        value = valueString.toInt(&ok);
        if(ok == false)
        {
            MYDEBUG << "string " << valueString << " not found in ui->cbSR" << endl;
        }
        else
        {
            //other controls on waterfall must be set accordingly:
            as->setSampleRate( value, true );
            as->setOffset(0);
            as->setBandwidth(value);
            bw = value;
            ui->lcdBW->display(bw);
            slotAdjustResolution();

        }
    }
}


void MainWindow::slotSetGain( const QString& text )
{

    int idx = ui->cbGain->findText( text, Qt::MatchFixedString );
    if(idx == -1)
    {
        MYDEBUG << "string \'" << text << "\' not found in ui->cbGain" << endl;
    }
    else
    {
        if(idx == 0)
        {
            //AUTO gain
            MYDEBUG << "selected AUTO gain" << endl;
            as->setGain(DEFAULT_GAIN);
        }
        else
        {

            QString valueString = ui->cbGain->itemText(idx);
            bool ok = false;
            //valueString = valueString.left( valueString.indexOf('.')+2 ); //convert back to tenths of db before conversion
            //valueString.remove('.');
            float value = valueString.toFloat(&ok) * 10;
            if(ok == false)
            {
                MYDEBUG << "string " << valueString << " cannot convert to float" << endl;
            }
            else
            {
                as->setGain( static_cast<int>(value) );
            }

        }
    }
}


void MainWindow::slotAdjustResolution()
{
    //automatically sets the label value after
    //change by the +/- pushbuttons

    ui->lbResolution->setNum( static_cast<double>( as->pround( as->getResolution(), 2 ) ) );
    ui->lbPoints->setNum(  as->getFFTbins() );
}


void MainWindow::slotValidateUpThreshold()
{

}

void MainWindow::slotValidateDnThreshold()
{
}

void MainWindow::slotShowStatus(const QString& status, int appState )
{
    QWidget* w = nullptr;

    if(appState != AST_NONE && appState != AST_ERROR)
    {
        ast = appState;
        emit stateChanged(ast);
    }

    //ui->lbStatus->clear();

    if(appState == AST_ERROR)
    {
        //temporary error, the device status remains unaffected
        //this status is not kept memorized, neither notified
        //ui->lbStatus->setBackgroundRole(QPalette::HighlightedText);
        ui->lbStatus->setStyleSheet("background-color: red; color: yellow;");
    }
    else if(ast == AST_CANNOT_OPEN)
    {
        /*
         * Device found, but error opening it
         * in this situation the user must be able to change device only
         */
        for(int i = GT_GLOBAL; i < GT_TOTAL; i++)
        {
            foreach(w, guiWidgetsList[i])
            {
                w->setEnabled(false);
            }
        }
        ui->pbQuit->setEnabled(true);
        ui->cbDevice->setEnabled(true);
        //ui->lbStatus->setBackgroundRole(QPalette::HighlightedText);
        ui->lbStatus->setStyleSheet("background-color: red; color: yellow;");
    }
    else if(ast > AST_CANNOT_OPEN)
    {
        //Device detected and opened, all controls are enabled
        //is responsability of each device to consider or not
        //the actions on these controls.

        for(int i = GT_GLOBAL; i < GT_TOTAL; i++)
        {
            bool enable = true;
            if(i == GT_DEVICE && as->getDevice().contains("UDP server"))
            {
                //on running client sessions the controls in device tab are disabled
                enable = false;

            }

            foreach(w, guiWidgetsList[i])
            {
                w->setEnabled(enable);
            }
        }

        //the device tab is enabled when acquisition is stopped
        //to allow change device
        ui->cbDevice->setEnabled(true);

        //acquisition started, locks some controls
        if(ast == AST_STARTED)
        {
            //the reporting controls are never locked because they
            //must apply on PostProc


            QList<GUItabs> toBeDisabledClient = {GT_GLOBAL, GT_DEVICE};
            QList<GUItabs> toBeDisabledServer = {GT_GLOBAL};
            QList<GUItabs>* toBeDisabled = &toBeDisabledServer;
            if(as->getDevice().contains("UDP server"))
            {
               toBeDisabled = &toBeDisabledClient;
            }

            GUItabs i;
            foreach (i, *toBeDisabled)
            {
                foreach(w, guiWidgetsList[i])
                {
                    w->setEnabled(false);
                }
            }
            ui->pbQuit->setEnabled(true);
            ui->pbStart->setEnabled(true);
            //update windows titlebars
            setTitle();
        }
        ui->lbStatus->setStyleSheet("color: #11ff00; background-color: rgb(63, 63, 63);");
        //ui->lbStatus->setBackgroundRole(QPalette::WindowText);

    }

    if( QString::compare(ui->lbStatus->text(), status, Qt::CaseInsensitive) != 0 )
    {
        MYINFO << "Status bar: from<" << ui->lbStatus->text() << ">"
                << " to <" << status << ">" ;
        ui->lbStatus->setText(status);
    }

    //workaround for tooltips bug to avoid text truncation
    //see: http://www.qtcentre.org/threads/43959-QToolTip-doesnt-wrap-text
    QString filler;
    filler.fill(' ', status.size());
    QString tt = status+filler;

    ui->lbStatus->setToolTip(tt);
    qApp->processEvents();

}

void MainWindow::slotUpdateSysInfo()
{
    if(init)
    {
        //measurement not are performed while initializing
        return;
    }

    if(outputEnabled == false)
    {
        slotSetAcqMode( static_cast<E_ACQ_MODE>( as->getAcqMode() ) );
        outputEnabled = true;
    }

    ui->lbMeasuredInterval->setText( QString("%1 mS").arg( ac->measuredInterval() ) );

    XQDir archiveDir(
            QString("%1/%2")
            .arg( ac->workingDir().absolutePath() )
            .arg( as->getConfigName() )
    );

    ui->lbFreeStorage->setText("0 MB");
    QStorageInfo sti(archiveDir);
    if( sti.isValid() )
    {
        double gigaFree = sti.bytesFree();
        gigaFree /= 0x40000000L;
        QString s = QString("%L1 GB").arg( gigaFree, 0, 'g', 3 );
        ui->lbFreeStorage->setText( s );

        if( ast == AST_STARTED &&
            static_cast<double>(as->getMinFree()) > gigaFree )
        {
            MYWARNING << "Available space on drive has fallen below " << as->getMinFree() << " GB, stopping acquisition.";
            slotStart();
            QMessageBox::warning(this, tr("Warning"), tr("Available space on drive has fallen\n below %1 GB limit.\nAcquisition stopped.")
                .arg( static_cast<double>(as->getMinFree()) ), QMessageBox::Ok);
        }
    }
    else
    {
        archiveDir.mkpath( archiveDir.absolutePath() );
    }


}



void MainWindow::slotFocusChanged(QWidget*old, QWidget*now)
{
    Q_UNUSED(now);
    if(old == ui->leServerIP)
    {
        as->setServerAddress( ui->leServerIP->text() );
    }
}


void MainWindow::blockAllSignals(bool block)
{
    QWidget* w = nullptr;
    for(int i = GT_GLOBAL; i < GT_TOTAL; i++)
    {
        foreach(w, guiWidgetsList[i])
        {
            w->blockSignals(block);
        }
    }

    wf->blockAllSignals(block);

}


void MainWindow::setTitle()
{
    QString title;
    QTextStream ts(&title);
    ts << "Echoes v." << APP_VERSION ;
    setWindowTitle(title);
    title.clear();
    ts << "Echoes waterfall window - (" << as->getConfigName() << ")";
    wf->setWindowTitle(title);
    ui->lbConfig->setText( as->getConfigName() );

}

void MainWindow::moveEvent ( QMoveEvent * event )
{
    Q_UNUSED(event)
    if(ast != AST_INIT)
    {
        as->setMainGeometry( geometry() );
    }
}

void MainWindow::resizeEvent ( QResizeEvent* event )
{
    Q_UNUSED(event)  //avoids warning unused-parameter
    if(ast != AST_INIT)
    {
        as->setMainGeometry( geometry() );

    }
}


void MainWindow::showBusy()
{
    QCursor cur = cursor();
    cur.setShape(Qt::WaitCursor);
    setCursor(cur);
    qApp->processEvents();

    MY_ASSERT(busyInst >= 0)
    busyInst++;
}


void MainWindow::showFree()
{
    if(busyInst < 1)
    {
        return;
    }

    busyInst--;

    QCursor cur = cursor();
    cur.setShape(Qt::ArrowCursor);
    setCursor(cur);
    qApp->processEvents();
}

void MainWindow::checkReportCoverage()
{
    //max report coverage:
    //starting from today, it can cover back to 31 days

    QDate from = ui->dtFrom->date();
    QDate to = ui->dtTo->date();

    qint64 dayDiff = from.daysTo(to);

    if(dayDiff > MAXIMUM_REPORT_COVERAGE)
    {

        QString warn = QString( tr("Unable to generate reports covering more than %1 days.") ).arg(MAXIMUM_REPORT_COVERAGE);

        MYWARNING << warn;

        QMessageBox::warning
                (this,
                 TARGET_NAME,
                 warn,
                 QMessageBox::Cancel);
        return;
    }

/*
    if( to >= QDate::currentDate() )
    {
        //excludes today and future
        to = QDate::currentDate();
        to = to.addDays(-1);
        MYWARNING << "End date corrected to " << to.toString();
        ui->dtTo->setDate(to);
    }
*/
    if( to > QDate::currentDate() )
    {
        //excludes future
        to = QDate::currentDate();
        MYWARNING << "End date corrected to " << to.toString();
    }

    if( from >= to )
    {
        //from must be antecedent to
        from = to.addDays(-1);
        MYWARNING << "Begin date corrected to " << from.toString();
    }

    if (dayDiff < 0)
    {

        //if difference is negative, the start date is corrected
        //as one day before the end date
        //Corrects silently

        from = to.addDays(-1);
        MYWARNING << "Begin date corrected to " << from.toString();
        ui->dtFrom->setDate(from);
        return;
    }

    //stores the interval in settings, volatile
    as->setCoverFrom( from );
    as->setCoverTo( to );

    ui->dtFrom->setDate(from);
    ui->dtTo->setDate(to);
}

void MainWindow::slotSetSbThresholdsParam()
{

    //sets the thresholds spinboxes intervals
    //depending of the nature differential (S-N)
    //or relative (S) of these thresholds

    E_THRESHOLDS_BEHAVIOR tb = static_cast<E_THRESHOLDS_BEHAVIOR>( as->getThresholdsBehavior() );
    if(tb == TB_ABSOLUTE)
    {
        //absolute
        ui->sbUpThreshold->setMinimum(MIN_DBFS);
        ui->sbUpThreshold->setMaximum(MAX_DBFS);
        ui->sbUpThreshold->setSuffix(" dBfs");

        ui->sbDnThreshold->setMinimum(MIN_DBFS);
        ui->sbDnThreshold->setMaximum(MAX_DBFS);
        ui->sbDnThreshold->setSuffix(" dBfs");

    }
    else if(tb == TB_DIFFERENTIAL)
    {
        //differential
        ui->sbUpThreshold->setMinimum(0);
        ui->sbUpThreshold->setMaximum(MAX_DBFS);
        ui->sbUpThreshold->setSuffix(" dBfs");


        ui->sbDnThreshold->setMinimum(0);
        ui->sbDnThreshold->setMaximum(MAX_DBFS);
        ui->sbDnThreshold->setSuffix(" dBfs");

    }
    else if(tb == TB_AUTOMATIC)
    {
        //automatic (percentages)
        ui->sbUpThreshold->setMinimum(0);
        ui->sbUpThreshold->setMaximum(100);
        ui->sbUpThreshold->setSuffix(" %");


        ui->sbDnThreshold->setMinimum(0);
        ui->sbDnThreshold->setMaximum(100);
        ui->sbDnThreshold->setSuffix(" %");

    }
}

void MainWindow::slotSetAcqMode( E_ACQ_MODE am )
{
    as->setAcqMode(am);

    switch(am)
    {

        case AM_CONTINUOUS:
            MYINFO << "Selected continuous acquisition mode";
            ui->rbContinuous->setChecked(true);
            break;

        case AM_PERIODIC:
            MYINFO << "Selected periodic acquisition mode";
            ui->rbPeriodic->setChecked(true);
            break;

        case AM_AUTOMATIC:
            MYINFO << "Selected automatic acquisition mode";
            ui->rbAutomatic->setChecked(true);
            break;

        case AM_INVALID:
            MYINFO << "Parameters not compatible with mode chosen";
            ui->rbPeriodic->setChecked(false);
            ui->rbAutomatic->setChecked(false);
            ui->rbContinuous->setChecked(false);
            break;
    }
}
