/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.
    

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 255                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-07-17 13:37:43 +02#$: Date of last commit
$Id: pool.h 255 2018-07-17 11:37:43Z gmbertani $
*******************************************************************************/

#ifndef POOL_H
#define POOL_H

#include <QtCore>
#include "setup.h"
#include "settings.h"

//T must be a pointer!
template <class T> class Pool
{

protected:

    //pool of T
    QVector<T>* pool;

    QQueue<int>* freeBufs;          ///contains the free buffer indexes ready to be taken
    QQueue<int>* writtenBufs;       ///contains the buffer written ready to be processed

    QMutex mutex;


public:

    ///
    /// \brief Pool
    ///
    ///
    ///
    explicit Pool()
    {
        QMutexLocker ml(&mutex);

        pool = new QVector<T>;
        MY_ASSERT( nullptr !=   pool );

        freeBufs = new QQueue<int>;
        MY_ASSERT( nullptr !=   freeBufs );

        writtenBufs = new QQueue<int>;
        MY_ASSERT( nullptr !=   writtenBufs );

    }

    ~Pool()
    {
        if(pool != 0)
        {
            clear();
            delete pool;
        }

        if(writtenBufs != 0)
        {
            delete writtenBufs;
            writtenBufs = 0;
        }

        if(freeBufs != 0)
        {
            delete freeBufs;
            freeBufs = 0;
        }

    }

    //general rule for return values:
    // boolean methods: false means failure
    // integer methods: -1 means failure
    // pointer methods: zero means failure


    ///
    /// \brief clear
    /// \return
    ///destroys all elements in the pool
    bool clear()
    {
        QMutexLocker ml(&mutex);

        MYDEBUG << "pool: " << this << " clear()";
        if(pool != 0)
        {
            T elem;
            for(int i = 0; i < pool->size(); i++)
            {
                elem = pool->at(i);
                MY_ASSERT( nullptr !=   elem );
                MYDEBUG << "destroying element [" << elem->myUID() << "]";

                delete elem;
            }
            pool->clear();
            freeBufs->clear();
            writtenBufs->clear();
            Q_ASSERT(pool->count() == 0);
            return true;
        }
        return false;

    }


    ///
    /// \brief take
    /// \return
    //Gets a clean buffer to write in
    //returns -1 when all buffers have been written
    //and no free ones are available, otherwise returns
    //the buffer index. The relative pointer is returned
    //by getBuffer()
    int take()
    {
        QMutexLocker ml(&mutex);

        MYDEBUG << "take()";
        int count = freeBufs->size();

        if( count <= 0 )
        {
            //No more buffers available. Pool must be resized.
            MYDEBUG << "no more free buffers in pool";
            QThread::yieldCurrentThread();
            return -1;
        }

        MYDEBUG << "dequeuing the first of " << count << " elements";

        int index = freeBufs->dequeue();

        MYDEBUG << "took [" << index << "], " << count << " elements left";

        pool->value(index)->clear();
        return index;
    }


    ///
    /// \brief forward
    /// \param index
    /// \return
    //enqueue the data buffer to be processed
    bool forward( int index )
    {
        QMutexLocker ml(&mutex);

        MYDEBUG << "forward(" << index << ")";

        int count = freeBufs->size();
        if( count < 0 || count > pool->size() )
        {
            MYDEBUG << "in freebuf: " << count << " elements";
            MYDEBUG << "freeBufs=" << freeBufs;
        }

        if( index < 0 || index > pool->size()  )
        {
            //size cannot exceed the given pool->size()
            MYWARNING << "index " << index << " out of range";
            return false;
        }
        writtenBufs->enqueue( index );


        return true;
    }

    ///
    /// \brief peekFree
    /// \return
    ///
    int peekFree()
    {
        QMutexLocker ml(&mutex);
        return freeBufs->count();
    }

    ///
    /// \brief peekBusy
    /// \return
    ///
    int peekBusy()
    {
        QMutexLocker ml(&mutex);
        return pool->size() - freeBufs->count() - writtenBufs->count();
    }

    ///
    /// \brief peekData
    /// \return
    ///
    //gets the number of data buffers ready without extraction
    int peekWritten()
    {
        QMutexLocker ml(&mutex);

        if( writtenBufs->isEmpty() == true )
        {
            //No data buffers available yet.
            MYDEBUG << "no data buffers available";
            QThread::yieldCurrentThread();

            int count = freeBufs->size();
            MYDEBUG << count << " elements left in freeBufs";
            MYDEBUG << "freeBufs=" << freeBufs;
            return -1;
        }

        return writtenBufs->count();
    }

    ///
    /// \brief getData
    /// \return
    ///
    //gets the index of the next data buffer to be processed
    int getData()
    {
        QMutexLocker ml(&mutex);

        MYDEBUG << "getData()";

        if( writtenBufs->isEmpty() == true )
        {
            //No data buffers available yet.
            MYDEBUG << "getData() no data buffers available";
            QThread::yieldCurrentThread();
            return -1;
        }
        return writtenBufs->dequeue();
    }


    ///
    /// \brief release
    /// \param index
    /// \return
    //releases the processed data buffer for recycling
    //return false in case of failure
    bool release( int index )
    {
        QMutexLocker ml(&mutex);

        if( freeBufs->count() > 0 && freeBufs->contains( index ) == true )
        {
            //No data buffers available yet.
            MYINFO << "pool: " << this << " index " << index << " has already been freed (already present in freeBufs[])";
            QThread::yieldCurrentThread();
            return false;
        }

        if( index < 0)
        {
            MYINFO << "pool: " << this << " index " << index << " is invalid";
            QThread::yieldCurrentThread();
            return false;
        }

        if( writtenBufs->count() > 0 && writtenBufs->contains( index ) == true )
        {
            //No data buffers available yet.
            MYINFO << "pool: " << this << " index " << index << " has not been got/processed (still present in writtenBufs[])";
            QThread::yieldCurrentThread();
            return false;
        }
        T s = pool->value(index);
        MY_ASSERT( nullptr !=  s);
        freeBufs->enqueue( index );
        MYDEBUG << "pool: " << this << "release(" << index << ") OK";
        return true;
    }

    ///
    /// \brief flush
    /// \return
    /// clears the writtenBufs from unprocessed
    /// buffers
    int flush()
    {
        QMutexLocker ml(&mutex);

        MYDEBUG << "flushing all unprocessed written buffers";
        int index;
        int released = 0;
        foreach (index, *writtenBufs)
        {
            freeBufs->enqueue( writtenBufs->dequeue() );
            released++;
            QThread::yieldCurrentThread();
        }
        return released;
    }

    ///
    /// \brief flush
    /// \return
    /// clears the oldest unprocessed
    /// buffer and makes it a free one
    void flushOne()
    {
        QMutexLocker ml(&mutex);
        if(writtenBufs->count() > 0)
        {
            MYDEBUG << "flushing oldest unprocessed written buffer";
            freeBufs->enqueue( writtenBufs->dequeue() );
        }
    }

    ///
    /// \brief size
    /// \return
    ///
    int size() const
    {
        return pool->size();
    }

    ///
    /// \brief getElem
    /// \param index
    /// \return
    //returns the given buffer or zero if the given index is
    //out of range or a free one (not taken).
    //Note that the element is not extracted,
    //simply its pointer is returned.
    T getElem( int index )
    {
        QMutexLocker ml(&mutex);

        if( index < 0 ||
            index > pool->size() )
        {
            MYWARNING << "invalid parameter";
            return 0;
        }

        T s = pool->value(index);
        MY_ASSERT( nullptr !=  s);
        return s;
    }

    ///
    /// \brief insert
    /// \param elem
    ///
    void insert( int index, T elem )
    {
        QMutexLocker ml(&mutex);
        pool->insert(index, elem);
        freeBufs->enqueue(index);
    }


};

#endif // POOL_H
