/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.
    

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 380                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-12-20 08:39:47 +01#$: Date of last commit
$Id: postproc.cpp 380 2018-12-20 07:39:47Z gmbertani $
*******************************************************************************/

#include <QLocale>
//#include "fftw3.h"
#include "control.h"
#include "waterfall.h"
#include "mainwindow.h"
#include "postproc.h"

#include <unistd.h>
#include <errno.h>

//TODO: allow user's customization:
#warning values tuned at detection range 10%
#define _SATURATION_FILTER_THRESHOLD (100.0F)
#define _CARRIER_FILTER_THRESHOLD_1  (2e-06F)
#define _CARRIER_FILTER_THRESHOLD_2  (10.0F)
#define _LIGHTINGS_FILTER_THRESHOLD  (5.0F)
#define _LIGHTINGS_FILTER_MIN_SCANS (1)



void DailyCounters::reset()
{
    //24hr + totals

    allEvents.clear();
    allGoods.clear();
    allOvers.clear();
    allUnders.clear();
    allFakes.clear();

    hEvents.clear();
    hGoods.clear();
    hOvers.clear();
    hUnders.clear();
    hFakes.clear();

    for(uint h=0; h < 24; h++)
    {
        hS[h] = INVALID_DBFS;
        hN[h] = INVALID_DBFS;
        hEvents[h].clear();
        hGoods[h] .clear();
        hOvers[h] .clear();
        hUnders[h].clear();
        hFakes[h] .clear();
    }
}



bool DailyCounters::save(QDataStream& ds)
{
    if(ds.status() == QDataStream::Ok)
    {
        ds << day;
        ds << allEvents;
        ds << allGoods;
        ds << allOvers;
        ds << allUnders;
        ds << allFakes;
        ds << hEvents;
        ds << hGoods;
        ds << hOvers;
        ds << hUnders;
        ds << hFakes;
        ds << hLastings;
        ds << hS;
        ds << hN;
        ds << shotsOver;
        ds << shotsUnder;
        ds << shotsFake;
        ds << dataBegin;
        ds << dataPeak;
        ds << dataEnd;
        return (ds.status() == QDataStream::Ok);
    }
    return false;

}

bool DailyCounters::load(QDataStream &ds)
{

    if(ds.status() == QDataStream::Ok)
    {
        ds >> day;
        ds >> allEvents;
        ds >> allGoods;
        ds >> allOvers;
        ds >> allUnders;
        ds >> allFakes;
        ds >> hEvents;
        ds >> hGoods;
        ds >> hOvers;
        ds >> hUnders;
        ds >> hFakes;
        ds >> hLastings;
        ds >> hS;
        ds >> hN;
        ds >> shotsOver;
        ds >> shotsUnder;
        ds >> shotsFake;
        ds >> dataBegin;
        ds >> dataPeak;
        ds >> dataEnd;
        return (ds.status() == QDataStream::Ok);
    }
    return false;
}


PostProc::PostProc(Settings* appSettings, Control* appControl, XQDir workDir,  QObject *parent) :
    QThread(parent)
{
    //csvIn.setFileName(scanCsvPath);
    //csvIn.open( QIODevice::ReadOnly);
    setObjectName("PostProc thread");

    MY_ASSERT( nullptr !=  appSettings);
    as = appSettings;

    MY_ASSERT( nullptr !=  appControl);
    ac = appControl;

    wd = workDir;

    task = PP_NONE;
    stopProc = false;
    flatDir.setPath("/tmp");

    destRootPath = QString("%1/%2")
            .arg(wd.absolutePath())
            .arg(as->getConfigName());

    /*
     * TODO: tricky system folders management
     * can be improved with QStandardPath but
     * there will be impacts on program installation
     * on various platforms.
     */

#ifdef NIX
    extResourcesPath = QString(AUX_PATH);
#else
    extResourcesPath = qApp->applicationDirPath();
#endif

    MYINFO << "extResourcePath=" << extResourcesPath;
}

PostProc::~PostProc()
{
    if(pdfReport != nullptr)
    {
        delete pdfReport;
    }
}

void PostProc::run()
{
    MYINFO << "Post processing thread started";
    emit status( tr("Postprocessing daily data"));

    destRootPath = QString("%1/%2")
            .arg(wd.absolutePath())
            .arg(as->getConfigName());

    switch(task)
    {
        case PP_MIDNIGHT:
            //root subdir under workdir (level 0)
            midnightTasks();
            break;

        case PP_FLATTEN:
            //flatDir must exist and be empty
            flatten();
            break;

        default:
            break;
    }


    //checks for expired data
    //(oldest than <lasting> days)
    if( as->getDataLasting() > 0 )
    {
        if( wd.pushCd(destRootPath) == true )
        {
            checkExpiredData();
            wd.popCd();
        }
    }

    stopProc = false;
}

bool PostProc::midnightTasks()
{
    MYINFO << "midnightTasks()";
    bool ok = false;

    wd.refresh();

    QString configName = as->getConfigName();

    //gets a list of all the statistic CSV present in working directory
    QStringList stats = selectFiles(CSV_PREFIX, TABLE_EXT);
    QString srcCSV, destCSV;

    foreach( srcCSV, stats )
    {
        //gets the date referred by the file
        if( parseEventDateFromName(srcCSV, CSV_PREFIX, TABLE_EXT, configName, processDT) == 0 )
        {
            MYCRITICAL << "midnightTasks(): error parsing date from string: " << srcCSV;
            return false;
        }


        //create the subdirectories tree for data files

        //date subdir (level 1)
        destDatePath = QString("%1/%2")
                .arg(destRootPath)
                .arg(processDT.toString(Qt::ISODate));


        //2nd level subdirs:

        //statistics directory
        destStatsPath = destDatePath + "/" + tr("stats") + "/";
        ok = wd.mkpath(destStatsPath);
        if(ok == false)
        {
            MYCRITICAL << "midnightTasks() failed creating directory " << destStatsPath;
            return false;
        }

        destCSV = destStatsPath + srcCSV;
        srcCSV = wd.absoluteFilePath(srcCSV);

        ok = QFile::exists( destCSV );
        if(ok == false)
        {
            ok = wd.copy( srcCSV, destCSV );
            if(ok == false)
            {

                MYCRITICAL << "midnightTasks() failed copying file " << srcCSV
                           << " to " << destCSV;

                return false;
            }
        }



        //gnuplot files - SPEC3D and SPECMAP have the same content structure
        //so they will be copied in the same subdir destGPspecPath, together with relative command files
        destGPspecPath = destDatePath + "/" + tr("gnuplot") + "/" + tr("specs");
        ok = moveTo( SPEC3D_PREFIX, GP_DATA_EXT, destGPspecPath );
        if(ok == false)
        {
            MYCRITICAL << "midnightTasks() failed moving data files " << SPEC3D_PREFIX << "* to subdir";
            return false;
        }
        ok = moveTo( SPECMAP_PREFIX, GP_DATA_EXT, destGPspecPath );
        if(ok == false)
        {
            MYCRITICAL << "midnightTasks() failed moving data files " << SPEC3D_PREFIX << "* to subdir";
            return false;
        }
        ok = moveTo( SPEC3D_PREFIX, GP_CMD_EXT, destGPspecPath );
        if(ok == false)
        {
            MYCRITICAL << "midnightTasks() failed moving command files " << SPEC3D_PREFIX << "* to subdir";
            return false;
        }
        ok = moveTo( SPECMAP_PREFIX, GP_CMD_EXT, destGPspecPath );
        if(ok == false)
        {
            MYCRITICAL << "midnightTasks() failed moving command files " << SPEC3D_PREFIX << "* to subdir";
            return false;
        }

        //gnuplot files - POW2D have no spectral informations so they are placed in a dedicated subdir
        //together with relative command files
        destGPpowPath = destDatePath + "/" + tr("gnuplot") + "/" + tr("power");
        ok = moveTo( POW2D_PREFIX, GP_DATA_EXT, destGPpowPath );
        if(ok == false)
        {
            MYCRITICAL << "midnightTasks() failed moving data files " << SPEC3D_PREFIX << "* to subdir";
            return false;
        }
        ok = moveTo( POW2D_PREFIX, GP_CMD_EXT, destGPpowPath );
        if(ok == false)
        {
            MYCRITICAL << "midnightTasks() failed moving command files " << SPEC3D_PREFIX << "* to subdir";
            return false;
        }


        //screenshots directory
        destShotsPath = destDatePath + "/" + tr("screenshots");
        ok = moveTo( AUTOSHOT_PREFIX, SSHOT_EXT, destShotsPath );
        if(ok == false)
        {
            MYCRITICAL << "midnightTasks() failed moving files " << SSHOT_EXT << " to subdir";
            return false;
        }

        //The original CSV is removed at last. This allows to retry later if the
        //file moving process fails
        ok = QFile::remove( srcCSV );
        if(ok == false)

        {
            MYCRITICAL << "midnightTasks() failed removing file " << srcCSV << " to " << destCSV;
            return false;
        }

    }


    //generate the daily report with counts
    createDailyReport();

    if(as->getAutoReport() == Qt::Checked)
    {
        //when the coverTo date for report is TODAY (the new day just started)
        //it triggers a full report. Obviously, the latest data reported
        //will be from yesterday (the day just ended) and the coverFrom date must be
        //antecedent

        QDate dFrom = as->getCoverFrom();
        QDate dTo = as->getCoverTo();
        if(
           dFrom < dTo &&
           dTo <= processDT
        )
        {
            emit autoReport();
        }
        else
        {
            if( dTo > processDT )
            {
                MYWARNING << "Report date interval not yet expired: " << dFrom << " to " << dTo << " skipping autoReport";
            }

            MYWARNING << "Latest processed date: " << processDT;
        }
    }

    return true;
}

bool PostProc::archiveExists(QDate &from, QDate &to)
{
    //the number of subfolders found in archive for the actual configuration
    //must match at least partially the counter structures present in memory sd[]
    //in order to produce a report

    wd.pushCd( as->getConfigName() );
    QStringList arc = selectSubDirs(from,to);
    wd.popCd();

    QList<DailyCounters*> data = selectData( from, to );

    int datas = data.count();
    if(datas > 0)
    {
        //there are data to report related to days
        //included in the given interval
        return true;
    }

    //the given days interval is completely wrong or there are no data
    //in archive to report for these days
    return false;
}


QStringList PostProc::selectFiles(const QString& prefix, const QString& suffix)
{
    QString configName = as->getConfigName();
    QString fileNameMask;

    fileNameMask = QString(tr("%1%2_automatic_20*%3")).arg(prefix).arg( configName ).arg(suffix);
    QStringList filter;
    filter << fileNameMask;
    wd.setNameFilters(filter);
    QStringList selectedFiles = wd.entryList( XQDir::Files, XQDir::Name );
    MYINFO << "Found " << selectedFiles.count() << " files " << prefix << "*" << suffix;
    return selectedFiles;
}


bool PostProc::flatten()
{
    QString flatPath = flatDir.absolutePath();
    return wd.xcopy( destRootPath, flatPath );
}


bool PostProc::moveTo(const QString& prefix, const QString& ext,  QString destPath)
{
    QString src, dest;
    QString srcPath = wd.absolutePath();

    QString mask = tr("%1%2_automatic_%3*%4")
            .arg( prefix )
            .arg( as->getConfigName() )
            .arg( processDT.toString(Qt::ISODate) )
            .arg(ext);

    //Creates subdir if not existing
    bool ok = wd.mkpath(destPath);
    if(ok == false)
    {
        MYCRITICAL << "moveTo() failed creating directory " << destPath;
        return false;
    }


    //collects the interested filenames
    QStringList filters;
    filters.append(mask);

    QStringList files = wd.entryList( filters );

    //moving the files (copy to destDir + removal of src)
    //to the new subdir
    foreach (src, files)
    {
        dest = destPath + "/" + src;
        src = srcPath + "/" + src;

        bool ok = QFile::exists(dest);
        if(ok == false)
        {
            ok = wd.copy(src, dest);
            if(ok == false)
            {
                MYCRITICAL << "moveTo() Error copying file " << src << " to " << dest;
                return false;
            }
        }

        ok = QFile::remove(src);
        if(ok == false)
        {
            MYWARNING << "moveTo() Error removing file " << src ;
            return false;
        }

    }
    return true;
}



int PostProc::checkExpiredData()
{
    MYINFO << "Deleting expired data subdirs";
    emit status( tr("Deleting expired data subdirs"));

    //calculate the oldest valid date
    QDate oldestValid = ac->UTCcalendarClock().date();
    oldestValid = oldestValid.addDays( -as->getDataLasting() );

    //conventional zero-date Jan 1, 2018 (before echoes was published)
    QDate zeroDate(2018,1,1);


    QStringList toDelete = selectSubDirs(zeroDate, oldestValid);
    QString deleteMe;
    if(toDelete.isEmpty() == false)
    {
        MYINFO << "Deleting expired subdirectories:";
        foreach(deleteMe, toDelete)
        {
            QDir tbd(deleteMe);
            QDate date = QDate::fromString(tbd.dirName(), Qt::ISODate);
            MY_ASSERT( date.isValid() )
            if(wd.pushCd(deleteMe) == true)
            {
                MYINFO << "deleting " << deleteMe ;
                //if( wd.removeRecursively()  == false )
                if( wd.removeDir(deleteMe) == false )
                {
                    //this happens, for instance, while trying to delete a file opened by another process
                    QString warn = tr("Error found while deleting oldest data\nsubdirectories under %1\ndeletion could be incomplete o neither happened")
                            .arg( deleteMe);

                    MYWARNING << warn;
                }

                if( sd.keys().contains(date) == true )
                {
                    sd[date]->reset();
                    sd.remove(date);
                }
                wd.popCd();

                /* commented out in saveData()
                //removal of expired backup daily_counters.ds
                QStringList filter;
                filter << QString(DAILYCTR_NAME) + "*.ds."+ deleteMe + "*";
                QStringList expiredBaks = wd.entryList(filter, QDir::Files);
                QString bakToDelete;
                foreach (bakToDelete, expiredBaks)
                {
                    MYINFO << "Removing expired daily_counters backup file " << bakToDelete;
                    if( wd.remove(bakToDelete) == false )
                    {
                        MYWARNING << "Failed removal of " << bakToDelete;
                    }
                }
                */

            }

        }



    }
    emit status( tr("Ready"));
    return 0;
}


QStringList PostProc::selectSubDirs( QDate& from, QDate& to, bool absolutePaths )
{
    QString dirNameMask = "\?\?\?\?-\?\?-\?\?";

    QStringList filter;
    filter << dirNameMask;
    QStringList allDirs = wd.entryList( filter, XQDir::Dirs, XQDir::Name );
    MYINFO << "Found " << allDirs.count() << " daily directories";

    QString testDir;
    QDate testDate;
    QStringList selected;
    foreach(testDir, allDirs)
    {
        testDate = QDate::fromString(testDir, Qt::ISODate);
        if(testDate >= from && testDate <= to)
        {
            if(absolutePaths == true)
            {
                selected << wd.absoluteFilePath(testDir);
            }
            else
            {
                selected << testDir;
            }
            MYINFO << testDir << " selected";
        }
    }

    return selected;
}

bool PostProc::createDailyReport()
{
    QLocale loc;    //for decimal commas / points

    MYINFO << "createDailyReport()";

    wd.pushCd(destRootPath);

    QString dailyCSVname;
    QFile dailyCSV;
    QStringList fields;
    QString s;
    QTextStream ts;


    dailyCSVname = QString("%1%2%3")
            .arg(DAILY_PREFIX)
            .arg(as->getConfigName())
            .arg( TABLE_EXT );

    dailyCSV.setFileName( wd.absoluteFilePath(dailyCSVname) );

    //The dailyCSV is recreated from scratch at each midnight UTC. Eventual existing
    //dailyCSV is overwritten - not appended.

    MYINFO << "Creating file " << dailyCSVname;
    if ( dailyCSV.open(QIODevice::WriteOnly|QIODevice::Text|QIODevice::Truncate) == true )
    {
        ts.setDevice(&dailyCSV);

        //header first
        ts << tr("Echoes Daily Report") << CSV_SEP << endl;

         //first heading row
        fields.clear();
        fields.append( tr("Date:") );
        for(int h=0; h < 24; h++)
        {
            //s.sprintf("%02uh", h);
            s = QString("%1h").arg( h, 2, 10, QChar('0') );
            fields.append( s );
            fields.append( "" );
            fields.append( "" );
            fields.append( "" );
            fields.append( "" );
            fields.append( "" );
            fields.append( "|" );
        }

        fields.append( tr("Daily totals") );
        fields.append( "" );
        fields.append( "" );
        fields.append( "" );
        fields.append( "" );
        fields.append( "" );
        fields.append( "|" );
        s = fields.join(CSV_SEP);
        ts << s << endl;

         //second heading row
        fields.clear();
        fields.append( tr("") );
        for(int h=0; h < 24; h++)
        {
            fields.append( tr("Max S [dBm]") );
            fields.append( tr("Avg N [dBm]") );
            fields.append( tr("Total") );
            fields.append( tr("overdenses") );
            fields.append( tr("underdenses") );
            fields.append( tr("fakes") );
            fields.append( "|" );
        }
        fields.append( tr("Max S [dBm]") );
        fields.append( tr("Avg N [dBm]") );
        fields.append( tr("Total") );
        fields.append( tr("overdenses") );
        fields.append( tr("underdenses") );
        fields.append( tr("fakes") );
        fields.append( "|" );
        s = fields.join(CSV_SEP);
        ts << s << endl;

    }
    else
    {
        MYCRITICAL << "Cannot create file " << dailyCSVname;
        wd.popCd();
        return false;
    }

    //keeps the daily report file open

    //now proceeds with reading statistic CSV files in each subdir and count the events
    //subdivided by category and hour


    QDate from(2018,1,1);
    QDate to = QDate::currentDate();
    QStringList subs = selectSubDirs( from, to );
    DailyCounters *thisDay = nullptr;

    //scans all the available dates
    foreach (destDatePath, subs)
    {

        thisDay = new DailyCounters;
        MY_ASSERT( nullptr !=  thisDay);

        wd.pushCd(destDatePath);
        wd.cd( tr("stats") + "/" );
        //there should be only one statistic CSV per directory
        //If not, skips the subdir
        QStringList csvs = selectFiles( CSV_PREFIX, TABLE_EXT );
        if(csvs.count() != 1)
        {
            MYWARNING << "Unexpected nr. of statistic CSV files ("
                      << csvs.count()
                      << ") under "
                      << wd.absolutePath()
                      << " skipping it...";
            wd.popCd();
            continue;
        }

        QString inCSVname = csvs.first();

        MYINFO << "parsing file " << inCSVname;
        QFile csvIn( wd.absoluteFilePath(inCSVname) );
        if ( csvIn.open(QIODevice::ReadOnly|QIODevice::Text) == false )
        {
            MYWARNING << "dailyReport(): The data file " << inCSVname << " doesn't open, skipping it...";
            wd.popCd();
            continue;
        }


        QByteArray rawLine;
        QStringList data;
        QDateTime recDT;


        bool ok = true;

        uint line = 0;
        QDate refDate;

        //for saturation filter
        float referenceAvgS = MAX_DBFS;
        float nextReferenceAvgS = referenceAvgS;
        float avgS = 0;


        //this is the real parsing, where events classification happens
        csvIn.seek(0);
        float peak = 0, avgN = 0, diff = 0, avgDiff = 0, lastingScans = 0;
        int msPerScan = ac->measuredInterval();

        while( csvIn.atEnd() == false )
        {
            QThread::yieldCurrentThread();

            rawLine = csvIn.readLine();
            s = QString(rawLine);
            line++;

            data = s.split(CSV_SEP, QString::SkipEmptyParts);
            if(data.count() != CF_TOTAL)
            {
                //wrong columns number, skipping record
                MYWARNING << "The line #" << line << " has " << data.count()
                          << " columns instead of " << CF_TOTAL
                          << " skipping it.";
                continue;
            }

            s = data.at(CF_DATE_UTC) + " " + data.at(CF_TIME_UTC);

            recDT = QDateTime::fromString( s , "dd/MM/yyyy hh:mm:ss.zzz" );
            if(recDT.isValid() == false)
            {
                //this is not a data row, probably is the header
                continue;
            }

            if(refDate.year() < 2018)
            {
                refDate = recDT.date();
            }

            if(refDate != recDT.date())
            {
                MYWARNING << "Error: the file " << inCSVname << " contains records with date different from " << refDate << " : " << recDT.date() << " at line " << line;
            }

            uint hour = static_cast<uint>(recDT.time().hour());

            s = data.at( CF_PROG );
            uint prog = s.toUInt(&ok);
            MY_ASSERT(ok == true)

            QString shotFileName = data.at(CF_SHOT_NAME);
            shotFileName = shotFileName.trimmed(); //remove spaces

            if( data.at( CF_EVENT_STATUS ) == "Fall" )
            {
                //these informations are available only on falling front

                s = data.at( CF_LASTING_MS );
                int lastingMs = s.toInt(&ok);
                MY_ASSERT(ok == true)

                s = data.at(CF_ECHO_AREA);
				uint echoArea = loc.toUInt( s, &ok );
                MY_ASSERT(ok == true)

				s = data.at(CF_INTERVAL_AREA);
				uint intervalArea = loc.toUInt( s, &ok );
                MY_ASSERT(ok == true)

				s = data.at(CF_PEAKS_COUNT);
				uint totPeaks = loc.toUInt( s, &ok );
                MY_ASSERT(ok == true)

                if(referenceAvgS == MAX_DBFS)
                {
                    referenceAvgS = avgS;
                }
                if(totPeaks > 1)
                {
                    //discards the new reference keeping the previous one
                    nextReferenceAvgS = referenceAvgS;
                }

                lastingScans = as->iround( static_cast<float>(lastingMs) / msPerScan );

                MYINFO << "day " << recDT << ": event#" << prog << " lasting ms=" << lastingMs << " scans= " << lastingScans
                       << " S=" << peak << " avgS=" << avgS << " referenceAvgS=" << referenceAvgS << " N=" << avgN
                       << " S-N=" << diff << " avg(S-N)=" << avgDiff
                       << " echoArea=" << echoArea << "intervalArea=" << intervalArea << " totPeaks=" << totPeaks;
			
                thisDay->hEvents[hour].append(prog);
                thisDay->allEvents.append(prog);

                //discriminate fake events first


                //fakes filter suited for lightings and electrostatic discharges
                if( as->getLightingsFilter() != 0 &&
                        lastingScans > _LIGHTINGS_FILTER_MIN_SCANS)
                {
                    float areaRatio = static_cast<float>(intervalArea) / echoArea;
                    if( areaRatio < _LIGHTINGS_FILTER_THRESHOLD )
                    {
                        MYINFO << "Event #" << prog << " is probably fake, due to electrostatic discharge, areaRatio=" << areaRatio;
                        thisDay->hFakes[hour].append(prog);
                        thisDay->allFakes.append(prog);
                        thisDay->shotsFake.append( shotFileName );
                        thisDay->dataEnd[prog] = data;
                        nextReferenceAvgS = referenceAvgS; //discards the new reference keeping the previous one
                        continue;
                    }
                }

                //fakes filter suited for close and strong signals
                //causing receiver saturations are detected by checking the
                //average S falling down suddendly

                if( as->getSaturationFilter() != 0 )
                {
                    int fuzzyRatio = as->fuzzyCompare( referenceAvgS, avgS, _SATURATION_FILTER_THRESHOLD );

                    if( fuzzyRatio > 0 )
                    {
                        MYINFO << " Event #" << prog << " is probably a fake due to saturated rx, avgS=" << avgS
                               << " referenceAvgS=" << referenceAvgS << " fuzzyRatio=" << fuzzyRatio;
                        thisDay->hFakes[hour].append(prog);
                        thisDay->allFakes.append(prog);
                        thisDay->shotsFake.append( shotFileName );
                        thisDay->dataEnd[prog] = data;
                        nextReferenceAvgS = referenceAvgS; //discards the new reference keeping the previous one
                        continue;
                    }
                }

                //fakes filter suited for radio carriers tuned at waterfall's center frequency
                if( as->getCarrierFilter() != 0 )
                {
                    float coef = echoArea;
                    coef /= (lastingMs * static_cast<int>(intervalArea));
                    if( coef > 0 && coef < _CARRIER_FILTER_THRESHOLD_1 )
                    {
                        MYINFO << " Event #" << prog << " is probably a fake due to radio carrier, coef=" << coef;
                        thisDay->hFakes[hour].append(prog);
                        thisDay->allFakes.append(prog);
                        thisDay->shotsFake.append( shotFileName );
                        thisDay->dataEnd[prog] = data;
                        nextReferenceAvgS = referenceAvgS; //discards the new reference keeping the previous one
                        continue;
                    }

                    int fuzzyRatio = as->fuzzyCompare( diff, avgDiff, _CARRIER_FILTER_THRESHOLD_2);
                    if( fuzzyRatio == 0 && lastingMs > as->getUnderdenseMs() )
                    {
                        MYINFO  << " Event #" << prog << " is probably a fake due to radio carrier, fuzzyRatio=" << fuzzyRatio;
                        thisDay->hFakes[hour].append(prog);
                        thisDay->allFakes.append(prog);
                        thisDay->shotsFake.append( shotFileName );
                        thisDay->dataEnd[prog] = data;
                        nextReferenceAvgS = referenceAvgS; //discards the new reference keeping the previous one
                        continue;
                    }
                }


                //simple fake filter, lasting based
                if ( (lastingMs / 1000) > as->getOverdenseSec() )
                {
                    thisDay->hFakes[hour].append(prog);
                    thisDay->allFakes.append(prog);
                    thisDay->shotsFake.append( shotFileName );
                    MYINFO << "Event #" << prog << " lasts too long to be a meteor echo, so it's fake, lastingMs=" << lastingMs;
                    nextReferenceAvgS = referenceAvgS; //discards the new reference keeping the previous one
                }

                else if ( lastingMs <= as->getUnderdenseMs() )
                {
                    thisDay->hUnders[hour].append(prog);
                    thisDay->hGoods[hour].append(prog);
                    thisDay->allUnders.append(prog);
                    thisDay->allGoods.append(prog);
                    thisDay->shotsUnder.append( shotFileName );
                    MYINFO << "Event #" << prog << " is underdense";

                }
                else if ( (lastingMs / 1000) <= as->getOverdenseSec() )
                {
                    thisDay->hOvers[hour].append(prog);
                    thisDay->hGoods[hour].append(prog);
                    thisDay->allOvers.append(prog);
                    thisDay->allGoods.append(prog);
                    thisDay->shotsOver.append( shotFileName );
                    MYINFO << "Event #" << prog << " is overdense";
                }

                thisDay->dataEnd[prog] = data;
            }
            else if( data.at( CF_EVENT_STATUS ) == "Raise" )
            {
                //raising front
                thisDay->dataBegin[prog] = data;
            }
            else if( data.at( CF_EVENT_STATUS ) == "Peak" )
            {
                //signal peak
                thisDay->dataPeak[prog] = data;

                if(peak > thisDay->hS[hour])
                {
                    thisDay->hS[hour] = peak;
                }
                if(avgN > thisDay->hN[hour])
                {
                    thisDay->hN[hour] = avgN;
                }

                //these values are meaningful if catched on peaks

                s = data.at( CF_AVG_PEAK_DBFS );
                avgS = loc.toFloat( s, &ok );
                MY_ASSERT(ok == true);
                referenceAvgS = nextReferenceAvgS;
                nextReferenceAvgS = avgS;

                s = data.at( CF_PEAK_DBFS );
                peak = loc.toFloat( s, &ok );
                MY_ASSERT(ok == true)

                s = data.at( CF_AVG_DBFS );
                avgN = loc.toFloat( s, &ok );
                MY_ASSERT(ok == true)

                s = data.at( CF_DIFF_DBFS );
                diff = loc.toFloat( s, &ok );
                MY_ASSERT(ok == true)

                s = data.at( CF_AVG_DIFF );
                avgDiff = loc.toFloat( s, &ok );
                MY_ASSERT(ok == true)

            }
        }

        csvIn.close();

        //going back to destRootPath
        wd.popCd();

        sd[ refDate ] = thisDay;

        //entering destDatePath
        wd.pushCd(destDatePath);

        MYINFO << "archiving events of date " << destDatePath;

        //moves the relative data files to the appropriate subdirectories
        //here, we are still into the date subdir

        MYINFO << "archiving screenshots";
        destShotsPath =  tr("screenshots") +"/";
        archiveData(refDate, destShotsPath);

        MYINFO << "archiving spectra plots";
        destGPspecPath =  tr("gnuplot") +"/" + tr("specs") + "/";
        archiveData(refDate, destGPspecPath);

        MYINFO << "archiving power plots";
        destGPpowPath = tr("gnuplot") + "/" + tr("power");
        archiveData(refDate, destGPpowPath);

        //again back to destRootPath
        wd.popCd();


        //when each input CSV has been read, the collected data are written
        //in the daily CSV
        fields.clear();
        if(recDT.isValid() == false)
        {
            continue;
        }

        fields.append( recDT.date().toString(Qt::TextDate) );

        float dailyMaxS = INVALID_DBFS, dailyAvgN = 0;
        uint avgCount = 0;
        for(uint h=0; h < 24; h++)
        {
            //note that floats are localized, in order to use the local decimal separator

            if( thisDay->hS[h] > dailyMaxS )
            {
                dailyMaxS = thisDay->hS[h];
            }

            if( thisDay->hS[h] > INVALID_DBFS )
            {
                s = loc.toString( thisDay->hS[h], 'f', 2 );
            }
            else
            {
                //invalid values are not represented
                s = "-";
            }
            fields.append(s);

            if( thisDay->hN[h] > INVALID_DBFS )
            {
                dailyAvgN += thisDay->hN[h];
                avgCount++;
                s = loc.toString( thisDay->hN[h], 'f', 2 );
            }
            else
            {
                //invalid values are not represented neither considered for averaging
                s = "-";
            }
            fields.append(s);

            s.setNum( thisDay->hEvents[h].count() );
            fields.append(s);

            s.setNum( thisDay->hOvers[h].count() );
            fields.append(s);

            s.setNum( thisDay->hUnders[h].count() );
            fields.append(s);

            s.setNum( thisDay->hFakes[h].count() );
            fields.append(s);

            fields.append( "|" );
        }

        //daily totals
        dailyAvgN /= avgCount;

        if( dailyMaxS > INVALID_DBFS )
        {
            s = loc.toString( dailyMaxS, 'f', 2 );
        }
        else
        {
            s = "-";
        }
        fields.append(s);

        if( dailyAvgN > INVALID_DBFS )
        {
            s = loc.toString( dailyAvgN, 'f', 2 );
        }
        else
        {
            s = "-";
        }
        fields.append(s);

        s.setNum( thisDay->allEvents.count() );
        fields.append(s);

        s.setNum( thisDay->allOvers.count() );
        fields.append(s);

        s.setNum( thisDay->allUnders.count() );
        fields.append(s);

        s.setNum( thisDay->allFakes.count() );
        fields.append(s);

        fields.append( "|" );

        //one row per day
        s = fields.join(CSV_SEP);
        ts << s << endl;      

    }

    dailyCSV.close();
    wd.popCd();
    //QDir testDir = ac->workingDir();
    //Q_ASSERT( wd == testDir );
    MYINFO << "Parsing terminated, daily report and archiving done.";
    emit status( tr("Ready"));
    return 0;
}

bool PostProc::archiveData(QDate& refDate, QString& destPath)
{
    QString overDir;
    QString underDir;
    QString fakeDir;

    DailyCounters* thisDay = sd[ refDate ];
    MY_ASSERT( nullptr !=  thisDay);

    //Creates the third level of subdirectories under destPath

    overDir  = tr("overdense") +"/";
    underDir = tr("underdense") + "/";
    fakeDir  = tr("fakes") + "/";

    wd.pushCd(destPath);

    bool ok = wd.mkpath(overDir);
    if(ok == false)
    {
        MYCRITICAL << "archiveData() failed creating directory " << wd.absoluteFilePath(overDir);
        return false;
    }

    ok = wd.mkpath(underDir);
    if(ok == false)
    {
        MYCRITICAL << "archiveData() failed creating directory " << wd.absoluteFilePath(underDir);
        return false;
    }

    ok = wd.mkpath(fakeDir);
    if(ok == false)
    {
        MYCRITICAL << "archiveData() failed creating directory " << wd.absoluteFilePath(fakeDir);
        return false;
    }


    //gets a list of data/shot files in this directory
    QString fileNameMask = QString("*_%1_*").arg( as->getConfigName() );
    QStringList filters;
    filters << fileNameMask;
    wd.setNameFilters(filters);
    QStringList selectedFiles = wd.entryList( XQDir::Files, XQDir::Name );
    MYINFO << "Found " << selectedFiles.count() << " data files to categorize";

    //backup copy of current dir to restore the original position
    //in case of errors
    XQDir wdBak = wd;

    //move the screenshots files in the appropriate subdirectories
    if( archiveDailyData(selectedFiles, thisDay->hOvers, tr("overdense")) == true )
    {
        if( archiveDailyData(selectedFiles, thisDay->hUnders, tr("underdense")) == true )
        {
            if( archiveDailyData(selectedFiles, thisDay->hFakes, tr("fakes")) == true )
            {
                ok = false;
            }
        }
    }
    wd.popCd();

    if(ok == false)
    {
        //restoring original position after errors
        wd = wdBak;
    }
    return ok;
}


uint PostProc::parseEventDateFromName(QString fileName, const QString& prefix, const QString& ext,
    const QString& cfg, QDate& fileDate )
{
    int index = 0;
    uint eventNr = 0;
    QString sect;

    index = fileName.contains(prefix);
    if (index != 1)
    {
        return eventNr;
    }

    //remove the prefix:
    fileName.remove(prefix);

    //the CSV name must contain the actual config name
    index = fileName.contains(cfg);
    if (index < 1)
    {
        return eventNr;
    }

    //then removes the config name
    fileName.remove(cfg);

    //the _automatic_ flag
    fileName.remove(tr("_automatic_"));

    //and the extension
    fileName.remove(ext);

    //now splits by underscores:
    QStringList sects = fileName.split('_', QString::SkipEmptyParts);
    if(sects.count() == 1)
    {
        //the filename carries no event nr. so let's assume 1
        eventNr = 1;
    }
    else if(sects.count() == 2)
    {
        //2 fields should remain: the acquisition start datetime and the event nr.

        //parsing event nr first
        bool ok = false;
        sect = sects.at(1);
        eventNr = sect.toUInt(&ok);
        MY_ASSERT(ok == true)
    }

    sect = sects.at(0);
    QStringList dateSects = sect.split('T', QString::SkipEmptyParts);
    sect = dateSects.at(0);
    fileDate = QDate::fromString(sect, Qt::ISODate);
    MY_ASSERT(fileDate.isValid() == true)
    return eventNr;
}



bool PostProc::createFullReport(QString reportName)
{
    wd = ac->workingDir();

    destRootPath = wd.absoluteFilePath( as->getConfigName() );

    //available dates in archive

    QDate dayDir;
    QDate closestDayDir = as->getCoverFrom();
    QList<QDate> dates = sd.keys();
    foreach( dayDir, dates )
    {
        if(dayDir > closestDayDir && dayDir <= as->getCoverTo())
        {
            closestDayDir = dayDir;
        }
    }

    QString reportDir = wd.absoluteFilePath( destRootPath + "/" + closestDayDir.toString(Qt::ISODate) );
    MYINFO << "Creating full report under " << reportDir;

    wd.pushCd( reportDir ); 
    QString fileName;
    foreach(fileName, wd.entryList(QStringList("*.html"), QDir::Files))
    {
        wd.remove(fileName);
    }

    wd.pushCd( extResourcesPath );
    QString pathSrc = wd.absoluteFilePath(REPORT_CSS);
    wd.popCd();
    wd.refresh();
    QString pathDest = wd.absoluteFilePath(REPORT_CSS);

    MYINFO << "Copying " << pathSrc << " to " << pathDest;
    bool ok = wd.copy(pathSrc, pathDest);
    if(ok == false)
    {
        MYWARNING << "Error copying file "
                   << pathSrc
                   << " to "
                   << pathDest;

        wd.popCd();
        return false;
    }

    MYINFO << "Creating report: " << wd.absoluteFilePath(reportName); 

    //loading paragraph models
    QMap<QString,QString> models;
    models[ tr("config_device_settings") ]            = "";
    models[ tr("config_fft_settings") ]               = "";
    models[ tr("config_output_settings") ]            = "";
    models[ tr("config_preferences_settings") ]       = "";
    models[ tr("config_report_settings") ]            = "";
    models[ tr("config_section_title") ]              = "";
    models[ tr("config_waterfall_settings") ]         = "";
    models[ tr("config_notches_settings") ]           = "";
    models[ tr("config_notch_line") ]                 = "";
    models[ tr("hourly_data_overdense_table") ]       = "";
    models[ tr("hourly_data_section_title") ]         = "";
    models[ tr("table_end") ]                         = "";
    models[ tr("hourly_data_total_table") ]           = "";
    models[ tr("hourly_data_underdense_table") ]      = "";
    models[ tr("preamble") ]                          = "";
    models[ tr("begin") ]                             = "";
    models[ tr("end") ]                               = "";
    models[ tr("shot") ]                              = "";
    models[ tr("shots_overdense_title") ]             = "";
    models[ tr("shots_underdense_title") ]            = "";
    models[ tr("shots_date") ]                        = "";
    models[ tr("shot_data") ]                         = "";
    models[ tr("station") ]                           = "";
    models[ tr("summary") ]                           = "";
    models[ tr("summary_table_line") ]                = "";

    QTextStream s;
    QFile f;
    DailyCounters *dayCount;
    QString modelPath;
    QString modelName;
    QString toFormat;

    QString now = QDateTime::currentDateTime().toString();


    bool GNUplotFound = false;
    if(as->getOverdensePlots() != 0 || as->getUnderdensePlots() != 0)
    {
        int ret = QProcess::execute( as->getGNUplotPath(), QStringList("-V") );
        if (ret < 0)
        {
            QMessageBox::warning( as->getMainWindow(), tr("Warning"), tr("The GNUplot executable is not installed or not present in PATH\nThe resulting report won't contain any plot.") );
        }
        else
        {
            GNUplotFound = true;
        }
    }



    QString lang = tr("EN");
    foreach ( modelName, models.keys() )
    {

        modelPath = QString(":/models/html/%1/%2.html").arg(lang).arg(modelName);
        f.setFileName( modelPath );
        bool ok = f.open( QIODevice::ReadOnly );
        if(ok == false)
        {
            MYCRITICAL << "Can't open file " << modelPath;
            continue;
        }
        s.setDevice(&f);
        toFormat = s.readAll();
        models[ modelName ] = toFormat;
        f.close();
    }

    QList<DailyCounters*> data = selectData( as->getCoverFrom(), as->getCoverTo() );

    QString par, reportHtml;
    //uint pageNum = 1;

    //preamble: page styles
    par = models[ tr("preamble") ]
            .arg( APP_VERSION )
            .arg( now )
            .arg( REPORT_CSS );
    reportHtml.append(par);


    // station data and summary
    /*
    par = models[ tr("heading") ]
            .arg(VERSION_NR)
            .arg(RELEASE_NR)
            .arg( QDate::currentDate().toString(Qt::SystemLocaleShortDate) );
    sect.append(par);
    */

    //begin
    par = models[ tr("begin") ];
    reportHtml.append(par);

    if(as->getSiteInfos() != 0)
    {
        MYINFO << "Collecting site infos";

        QString logoPath = as->getLogo();


        //if the logo is set to the default value, pointing to a resource
        //it must be written down to a png file in order to supply html
        //with a valid path on file system
        if( as->getLogo() == DEFAULT_LOGO )
        {
            QPixmap logo( as->getLogo() );
            QString logoFile = as->getLogo().remove(0,2); //remove the :/
            logoFile.append( SSHOT_EXT );
            logoPath = wd.absoluteFilePath(logoFile);
            logo.save(logoPath);
            logoPath = wd.relativeFilePath(logoPath);
        }
        else
        {
            //otherwise the logo icon must be copied in the archive
            QFileInfo fi( logoPath );
            QString logoDest( wd.absoluteFilePath(fi.fileName() ) );
            MYINFO << "Copying " <<  logoPath << " to ./" << logoDest;
            bool ok = wd.copy( logoPath, logoDest );
            if(ok == false)
            {
                MYWARNING << "Error copying file "
                           << logoPath
                           << " to ./"
                           << logoDest;

                //doesn't stop if the logo can't be found
                // wd.popCd();
                //return false;
            }

            logoPath = fi.fileName();
        }

        //station data
        par = models[ tr("station") ]
                .arg( as->getStationName() )
                .arg( logoPath )
                .arg( as->getLat() )
                .arg( as->getLong() )
                .arg( as->getAlt() )
                .arg( as->getContact() )
                .arg( as->getRxSetup() )
                .arg( as->getNotes() );


        reportHtml.append(par);
    }

    if(as->getSummary() != 0)
    {
        MYINFO << "Generating summary";

        par = models[ tr("summary") ]
                .arg( as->getCoverFrom().toString(Qt::SystemLocaleShortDate) )
                .arg( as->getCoverTo().toString(Qt::SystemLocaleShortDate) )
                .arg( as->getConfigName()+CONFIG_EXT )
                .arg( as->getUnderdenseMs() )
                .arg( as->getOverdenseSec() )
                .arg( APP_VERSION )
                .arg( QDate::currentDate().toString(Qt::SystemLocaleShortDate) )
                .arg( QSysInfo::buildAbi() )
                .arg( /* QSysInfo::machineHostName() */ QHostInfo::localHostName() );

        reportHtml.append(par);

        QString underStr, overStr, fakeStr;
        uint eventNr;
        const uint lineWrapAt = 10;
        uint wrapAt;
        foreach(dayCount, data)
        {
            underStr.clear();
            wrapAt = lineWrapAt;
            foreach (eventNr, dayCount->allUnders)
            {
                underStr.append( QString("%1, ").arg(eventNr) );
                wrapAt--;
                if(wrapAt == 0)
                {
                    underStr.append("<BR>");
                    wrapAt = lineWrapAt;
                }
            }
            underStr.chop(2);   //cuts last comma

            overStr.clear();
            wrapAt = lineWrapAt;
            foreach (eventNr, dayCount->allOvers)
            {
                overStr.append( QString("%1, ").arg(eventNr) );
                wrapAt--;
                if(wrapAt == 0)
                {
                    overStr.append("<BR>");
                    wrapAt = lineWrapAt;
                }
            }
            overStr.chop(2);

            fakeStr.clear();
            wrapAt = lineWrapAt;
            foreach (eventNr, dayCount->allFakes)
            {
                fakeStr.append( QString("%1, ").arg(eventNr) );
                wrapAt--;
                if(wrapAt == 0)
                {
                    fakeStr.append("<BR>");
                    wrapAt = lineWrapAt;
                }
            }
            fakeStr.chop(2);


            par = models[ tr("summary_table_line") ]
                .arg( dayCount->day.toString(Qt::SystemLocaleShortDate) )
                .arg( underStr )
                .arg( overStr )
                .arg( fakeStr );
            reportHtml.append(par);
        }

        par = models[ tr("table_end") ];
        reportHtml.append(par);

        if( as->getShotsDetails() != 0 )
        {
            par = models[ tr("config_section_title") ]
                    .arg( as->getConfigName()+CONFIG_EXT );
            reportHtml.append(par);
            par = models[ tr("config_device_settings") ]
                    .arg( as->getDevice() )
                    .arg( as->getSampleRate() )
                    .arg( as->getTune() )
                    .arg( static_cast<double>( as->getGain() ) / 10.0 )
                    .arg( static_cast<double>(as->getError()) );

            reportHtml.append(par);



            QVector<QString>wins;
            QString flags;
            wins.append("RECTANGLE");
            wins.append("HAMMING");
            wins.append("BLACKMAN");
            wins.append("BLACKMAN_HARRIS");
            wins.append("HANN_POISSON");
            wins.append("YOUSSEF");
            wins.append("BARTLETT");
            wins.append("HANN");
            wins.append("BLACKMAN_HARRIS 7");
            wins.append("FLAT TOP");
            flags = tr("NATURAL");

            par = models[ tr("config_fft_settings") ]
                    .arg( wins.at( as->getWindow() ) )
                    .arg( static_cast<double>(as->getResolution()) )
                    .arg( as->getBandwidth() );

            reportHtml.append(par);

            QString pt;
            switch( as->getPlotType() )
            {
                case PT_2D_MAPPED:
                    pt = tr("2D Color-mapped");
                    break;

                case PT_2D_POWER:
                    pt = tr("2D Total power");
                    break;

                case PT_3D_REAL:
                    pt = tr("3D Perspective");
                    break;

            }

            QString measureUnit = (as->thresholdBehaviorString() == tr("automatic"))? "%" : "dBfs";

            par = models[ tr("config_output_settings") ]
                    .arg( chkBoxStateSimple( as->getScreenshots() ) )
                    .arg( chkBoxStateSimple( as->getGNUplot() ) )
                    .arg( pt )
                    .arg( static_cast<double>(as->getInterval()) )
                    .arg( as->getAfter() )
                    .arg( as->getShots() )
                    .arg( as->getRecTime() )
                    .arg( as->thresholdBehaviorString() )
                    .arg( static_cast<double>(as->getUpThreshold()) )                   
                    .arg( measureUnit )
                    .arg( static_cast<double>(as->getDnThreshold()) )
                    .arg( as->getTrange() )
                    .arg( as->getAvgdScans() )
                    .arg( as->getJoinTime() )
                    .arg( static_cast<double>(as->getMinFree()) );


            reportHtml.append(par);

            par = models[ tr("config_report_settings") ]
                    .arg( as->getCoverFrom().toString( Qt::SystemLocaleShortDate) )
                    .arg( as->getCoverTo().toString( Qt::SystemLocaleShortDate) )
                    .arg( chkBoxStateSimple( as->getSummary() ) )
                    .arg( chkBoxStateSimple( as->getColorGraph() ) )
                    .arg( chkBoxStateSimple( as->getSiteInfos() ) )
                    .arg( chkBoxStateSimple( as->getOverdenseShots() ) )
                    .arg( chkBoxStateSimple( as->getOverdensePlots() ) )
                    .arg( chkBoxStateSimple( as->getUnderdenseShots() ) )
                    .arg( chkBoxStateSimple( as->getUnderdensePlots() ) )
                    .arg( chkBoxStateSimple( as->getShotsDetails() ) )
                    .arg( chkBoxStateSimple( as->getRestart() ) )
                    .arg( chkBoxStateSimple( as->getAutoReport() ) )
                    .arg( chkBoxStateSimple( as->getCarrierFilter() ) )
                    .arg( chkBoxStateSimple( as->getLightingsFilter() ) )
                    .arg( chkBoxStateSimple( as->getSaturationFilter() ) );

            reportHtml.append(par);

            par = models[ tr("config_preferences_settings") ]
                    .arg( chkBoxStateSimple( as->getHz() ) )
                    .arg( chkBoxStateSimple( as->getDbfs() ) )
                    .arg( chkBoxStateSimple( as->getSec() ) )
                    .arg( chkBoxStateSimple( as->getTooltips() ) )
                    .arg( chkBoxStateSimple( as->getEraseLogs() ) )
                    .arg( chkBoxStateSimple( as->getEraseShots() ) )
                    .arg( chkBoxStateSimple( as->getPing() ) )
                    .arg( as->getDataLasting() )
                    .arg( as->getPalette() == 0 ? tr("B/W") : tr("Color") )
                    .arg( as->getNoiseLimit() )
                    .arg( as->getServerAddress().toString() )
                    .arg( as->getServerPort() );

            reportHtml.append(par);

            double fZoom = static_cast<double>(as->getZoom())/10.0F;
            int zoomCoverage = 2 * static_cast<int>((as->getBandwidth()/( fZoom )) / 2.0F);

            par = models[ tr("config_waterfall_settings") ]
                    .arg( fZoom )
                    .arg( zoomCoverage )
                    .arg( as->getOffset() )
                    .arg( as->getPowerZoom() )
                    .arg( as->getPowerOffset() )
                    .arg( as->getBrightness() )
                    .arg( as->getContrast() );

            reportHtml.append(par);

            if(as->getNotches().count() > 0)
            {
                //notch filters table
                // header
                par = models[ tr("config_notches_settings") ];
                reportHtml.append(par);
                Notch nt;
                int num = 1;
                foreach(nt, as->getNotches())
                {
                    par = models[ tr("config_notch_line") ]
                            .arg( num++ )
                            .arg( nt.begin )
                            .arg( nt.end );

                    reportHtml.append(par);
                }
                par = models[ tr("table_end") ];
                reportHtml.append(par);

            }

        }
    }

    //par = models[ tr("footer") ].arg(pageNum++);
    //sect.append(par);

    uint figCount = 1;
    if(as->getColorGraph() != 0)
    {
        MYINFO << "Generating hourly count tables";
        /*
        par = models[ tr("heading") ]
                .arg(VERSION_NR)
                .arg(RELEASE_NR)
                .arg( QDate::currentDate().toString(Qt::SystemLocaleShortDate) );
        sect.append(par);
        */

        //creating colorgrams
        QList< QList<uint> > goods, overs, unders, fakes;
        QList< QDate > days;
        DailyCounters* dc;
        QList<uint> dayCounts;

        foreach(dc, data)
        {
            days.append( dc->day );         //days covered

            //goods: all events except fakes
            dayCounts.clear();
            for( uint h = 0; h < 24; h++ )
            {
                dayCounts.append( static_cast<uint>(dc->hGoods[h].count()) );
            }
            goods.append( dayCounts );

            //overs: overdense events
            dayCounts.clear();
            for( uint h = 0; h < 24; h++ )
            {
                dayCounts.append( static_cast<uint>(dc->hOvers[h].count()) );
            }
            overs.append( dayCounts );

            //unders: underdense events
            dayCounts.clear();
            for( uint h = 0; h < 24; h++ )
            {
                dayCounts.append(  static_cast<uint>(dc->hUnders[h].count()) );
            }
            unders.append( dayCounts );

            //fakes: false positives only
            dayCounts.clear();
            for( uint h = 0; h < 24; h++ )
            {
                dayCounts.append(  static_cast<uint>(dc->hFakes[h].count()) );
            }
            fakes.append( dayCounts );

        }

        QString cgram = wd.absoluteFilePath("cgram_goods.png");
        drawCgrams( cgram, days, goods );
        cgram = wd.relativeFilePath(cgram);

        par = models[ tr("hourly_data_section_title") ];
        reportHtml.append(par);

        par = models[ tr("hourly_data_total_table") ].arg(cgram).arg(figCount++);
        reportHtml.append(par);


        //par = models[ tr("footer") ].arg(pageNum++);
        //sect.append(par);

        if( as->getOverdenseShots() != 0 ||  as->getOverdensePlots() != 0 )
        {
            //hourly data table: overdenses
            /*
            par = models[ tr("heading") ]
                    .arg(VERSION_NR)
                    .arg(RELEASE_NR)
                    .arg( QDate::currentDate().toString(Qt::SystemLocaleShortDate) );
            sect.append(par);
            */

            cgram = wd.absoluteFilePath("cgram_overs.png");
            drawCgrams( cgram, days, overs );
            cgram = wd.relativeFilePath(cgram);
            par = models[ tr("hourly_data_overdense_table") ].arg(cgram).arg(figCount++);
            reportHtml.append(par);


            //par = models[ tr("footer") ].arg(pageNum++);
            //sect.append(par);
        }

        if( as->getUnderdenseShots() != 0 || as->getUnderdensePlots() != 0 )
        {

            //hourly data table: underdenses
            /*
            par = models[ tr("heading") ]
                    .arg(VERSION_NR)
                    .arg(RELEASE_NR)
                    .arg( QDate::currentDate().toString(Qt::SystemLocaleShortDate) );
            sect.append(par);

            */

            cgram = wd.absoluteFilePath("cgram_unders.png");
            drawCgrams( cgram, days, unders );
            cgram = wd.relativeFilePath(cgram);
            par = models[ tr("hourly_data_underdense_table") ].arg(cgram).arg(figCount++);

            reportHtml.append(par);


            //par = models[ tr("footer") ].arg(pageNum++);
            //sect.append(par);
        }
    }


    if( as->getOverdenseShots() != 0 || as->getOverdensePlots() != 0 )
    {
        /*
        par = models[ tr("heading") ]
                .arg(VERSION_NR)
                .arg(RELEASE_NR)
                .arg( QDate::currentDate().toString(Qt::SystemLocaleShortDate) );
        sect.append(par);
        */

        par = models[ tr("shots_overdense_title") ];
        reportHtml.append(par);

        //overdense shots and data, day by day
        QStringList csvBeg, csvPeak, csvEnd;
        QString shotName;
        uint eventNr;

        as->getMainWindow()->setCursor( Qt::BusyCursor );
        MYINFO << "Processing overdense shots";
        foreach (dayCount, data)
        {
            destDatePath = QString("%1/%2")
                    .arg(destRootPath)
                    .arg( dayCount->day.toString(Qt::ISODate) );

            XQDir dd( destDatePath );
            if( dd.exists( destDatePath ) == false )
            {
                continue;
            }

            destStatsPath  = destDatePath + "/" + tr("stats");
            destShotsPath  = destDatePath + "/" + tr("screenshots") + "/" + tr("overdense");
            destGPspecPath = destDatePath + "/" + tr("gnuplot")     + "/" + tr("specs") + "/" + tr("overdense");
            destGPpowPath  = destDatePath + "/" + tr("gnuplot")     + "/" + tr("power") + "/" + tr("overdense");

            foreach (eventNr, dayCount->allOvers)
            {
                csvBeg = dayCount->dataBegin.value(eventNr);
                csvPeak = dayCount->dataPeak.value(eventNr);
                csvEnd = dayCount->dataEnd.value(eventNr);

                shotName = csvEnd.at(CF_SHOT_NAME);
                shotName = shotName.trimmed();  //cuts initial space char

                QString eventStr = shotName.split('_').last().left(5);
                bool ok = false;
                uint chkEventNr = eventStr.toUInt(&ok);
                if( ok == false || chkEventNr != eventNr )
                {
                    QString msg = tr("The statistic informations on event#%1\nhas mismatched screenshot associated:\n%2\n\nThis event will be ignored in report")
                        .arg(eventNr)
                        .arg(shotName);

                    MYWARNING << msg;

                    if(as->getAutoReport() == Qt::Checked)
                    {
                        //no dialog in autoreport
                        continue;
                    }

                    QMessageBox::warning( as->getMainWindow(), tr("Warning"), msg );
                    continue;
                }

                par = models[ tr("shots_date") ].arg( shotName ).arg( dayCount->day.toString(Qt::SystemLocaleShortDate) );
                reportHtml.append(par);


                if( as->getOverdenseShots() != 0 )
                {
                    MYINFO << shotName;
                    QString filePath = wd.absoluteFilePath( destShotsPath + "/" + shotName );

                    if( QFile::exists( filePath ) == true )
                    {
                        //note: in data generated under console mode, screenshots are not present
                        filePath = wd.relativeFilePath( filePath );
                        par = models[ tr("shot") ]
                                .arg( filePath )
                                .arg( figCount++ )
                                .arg( shotName );
                        reportHtml.append(par);
                    }
                }

                if( GNUplotFound != 0 && as->getOverdensePlots() != 0 )
                {
                    //QString absFilePath = wd.absoluteFilePath( destGPspecPath + "/" + shotName );
                    QString rs = csvEnd.at(CF_RANGE_LOW);
                    int rangeLo = rs.toInt();
                    rs = csvEnd.at(CF_RANGE_HI);
                    int rangeHi = rs.toInt();
                    rs = csvEnd.at(CF_LASTING_MS);
                    int msLasting = rs.toInt();
                    QString datPngPath = plotToPNG( shotName, rangeLo, rangeHi, msLasting );
                    if ( datPngPath == "" )
                    {
                        //something went wrong, avoids to recall
                        //gnuplot in loop
                        //GNUplotFound = false;
                    }
                    else
                    {
                        QString datPngName = wd.splitPath(datPngPath).last();
                        MYINFO << datPngName;

                        par = models[ tr("shot") ]
                                .arg( datPngPath )
                                .arg( figCount++ )
                                .arg( datPngName );
                        reportHtml.append(par);
                    }
                }

                if(as->getShotsDetails() != 0)
                {
                    par = models[ tr("shot_data") ]
                            .arg( eventNr )                     // arg %1
                            .arg( csvEnd.at(CF_DATE_UTC) )
                            .arg( csvEnd.at(CF_TIME_UTC) )
                            .arg( csvEnd.at(CF_UNIX_TIMESTAMP) )
                            .arg( csvEnd.at(CF_TUNE) )
                            .arg( csvEnd.at(CF_BW_HZ) )
                            .arg( csvEnd.at(CF_CENTRAL_HZ) )
                            .arg( csvEnd.at(CF_LOW_HZ) )
                            .arg( csvEnd.at(CF_HIGH_HZ) )
                            .arg( csvEnd.at(CF_ZOOMED_BW) )     //arg 10
                            .arg( csvEnd.at(CF_STEP_HZ) )
                            .arg( csvEnd.at(CF_THR_BEHAVIOR) )
                            .arg( csvEnd.at(CF_UP_THR_PERC) )
                            .arg( csvEnd.at(CF_DN_THR_PERC) )
                            .arg( csvEnd.at(CF_UP_THR_DBFS) )
                            .arg( csvEnd.at(CF_DN_THR_DBFS) )
                            .arg( csvEnd.at(CF_RANGE_LOW) )
                            .arg( csvEnd.at(CF_RANGE_HI) )      //arg 18

                            .arg( csvBeg.at(CF_PEAK_DBFS) )     //arg %19
                            .arg( csvPeak.at(CF_PEAK_DBFS) )
                            .arg( csvEnd.at(CF_PEAK_DBFS) )

                            .arg( csvBeg.at(CF_AVG_PEAK_DBFS) )     //arg %22
                            .arg( csvPeak.at(CF_AVG_PEAK_DBFS) )
                            .arg( csvEnd.at(CF_AVG_PEAK_DBFS) )

                            .arg( csvBeg.at(CF_AVG_DBFS) )      //arg %25
                            .arg( csvPeak.at(CF_AVG_DBFS) )
                            .arg( csvEnd.at(CF_AVG_DBFS) )

                            .arg( csvBeg.at(CF_DIFF_DBFS) )     //arg %28
                            .arg( csvPeak.at(CF_DIFF_DBFS) )
                            .arg( csvEnd.at(CF_DIFF_DBFS) )

                            .arg( csvBeg.at(CF_AVG_DIFF) )     //arg %31
                            .arg( csvPeak.at(CF_AVG_DIFF) )
                            .arg( csvEnd.at(CF_AVG_DIFF) )

                            .arg( csvBeg.at(CF_PEAK_HZ) )       //arg %34
                            .arg( csvPeak.at(CF_PEAK_HZ) )
                            .arg( csvEnd.at(CF_PEAK_HZ) )

                            .arg( csvBeg.at(CF_STDDEV) )    //arg %37
                            .arg( csvPeak.at(CF_STDDEV) )
                            .arg( csvEnd.at(CF_STDDEV) )

                            .arg( csvBeg.at(CF_LASTING_MS) )    //arg %40
                            .arg( csvPeak.at(CF_LASTING_MS) )
                            .arg( csvEnd.at(CF_LASTING_MS) )

                            .arg( csvEnd.at(CF_SHIFT_HZ) )      //arg %43
                            .arg( csvEnd.at(CF_ECHO_AREA) )
                            .arg( csvEnd.at(CF_INTERVAL_AREA) )
                            .arg( csvEnd.at(CF_PEAKS_COUNT) )
                            .arg( csvEnd.at(CF_LOS_SPEED) );     //arg %47

                    reportHtml.append(par);
                    MYDEBUG << "shot: " << eventNr;
                }

                /*
                par = models[ tr("footer") ].arg(pageNum++);
                sect.append(par);

                par = models[ tr("heading") ]
                        .arg(VERSION_NR)
                        .arg(RELEASE_NR)
                        .arg( QDate::currentDate().toString(Qt::SystemLocaleShortDate) );
                sect.append(par);
                */
            }

        }
        as->getMainWindow()->unsetCursor();
    }


    if( as->getUnderdenseShots() != 0 ||  as->getUnderdensePlots() != 0 )
    {
        /*
        par = models[ tr("heading") ]
                .arg(VERSION_NR)
                .arg(RELEASE_NR)
                .arg( QDate::currentDate().toString(Qt::SystemLocaleShortDate) );
        sect.append(par);
        */

        par = models[ tr("shots_underdense_title") ];
        reportHtml.append(par);

        //underdense shots and data, day by day
        QStringList csvBeg, csvPeak, csvEnd;
        QString shotName;
        uint eventNr;

        as->getMainWindow()->setCursor( Qt::BusyCursor );
        MYINFO << "Processing underdense shots";
        foreach (dayCount, data)
        {
            destDatePath = QString("%1/%2")
                    .arg(destRootPath)
                    .arg( dayCount->day.toString(Qt::ISODate) );

            XQDir dd( destDatePath );
            if( dd.exists( destDatePath ) == false )
            {
                continue;
            }

            destStatsPath  = destDatePath + "/" + tr("stats");
            destShotsPath  = destDatePath + "/" + tr("screenshots") + "/" + tr("underdense");
            destGPspecPath = destDatePath + "/" + tr("plots")     + "/" + tr("specs") + "/" + tr("underdense");
            destGPpowPath  = destDatePath + "/" + tr("plots")     + "/" + tr("power") + "/" + tr("underdense");

            foreach (eventNr, dayCount->allUnders)
            {

                csvBeg = dayCount->dataBegin.value(eventNr);
                csvPeak = dayCount->dataPeak.value(eventNr);
                csvEnd = dayCount->dataEnd.value(eventNr);

                shotName = csvEnd.at(CF_SHOT_NAME);
                shotName = shotName.trimmed();

                QString eventStr = shotName.split('_').last().left(5);
                bool ok = false;
                uint chkEventNr = eventStr.toUInt(&ok);
                if( ok == false || chkEventNr != eventNr )
                {
                    QString msg = tr("The statistic informations on event#%1\nhas mismatched screenshot associated:\n%2\n\nThis event will be ignored in report")
                        .arg(eventNr)
                        .arg(shotName);

                    MYWARNING << msg;

                    if(as->getAutoReport() == Qt::Checked)
                    {
                        //no dialog in autoreport
                        continue;
                    }

                    QMessageBox::warning( as->getMainWindow(), tr("Warning"), msg );
                    continue;
                }


                par = models[ tr("shots_date") ].arg( shotName ).arg( dayCount->day.toString(Qt::SystemLocaleShortDate) );
                reportHtml.append(par);


                if( as->getUnderdenseShots() != 0 )
                {
                    MYINFO << shotName;
                    QString filePath = wd.absoluteFilePath( destShotsPath + "/" + shotName );
                    if( QFile::exists( filePath ) == true )
                    {
                        //note: in data generated under console mode, screenshots are not present
                        filePath = wd.relativeFilePath( filePath );
                        par = models[ tr("shot") ]
                                .arg( filePath )
                                .arg( figCount++ )
                                .arg( shotName );
                        reportHtml.append(par);
                    }
                }

                if( GNUplotFound != 0 && as->getUnderdensePlots() != 0 )
                {
                    //QString absFilePath = wd.absoluteFilePath( destGPspecPath + "/" + shotName );
                    QString rs = csvEnd.at(CF_RANGE_LOW);
                    int rangeLo = rs.toInt();
                    rs = csvEnd.at(CF_RANGE_HI);
                    int rangeHi = rs.toInt();
                    rs = csvEnd.at(CF_LASTING_MS);
                    int msLasting = rs.toInt();
                    QString datPngPath = plotToPNG( shotName, rangeLo, rangeHi, msLasting );
                    if ( datPngPath == "" )
                    {
                        //something went wrong, avoids to recall
                        //gnuplot in loop
                        //GNUplotFound = false;
                    }
                    else
                    {
                        QString datPngName = wd.splitPath(datPngPath).last();
                        MYINFO << datPngName;

                        par = models[ tr("shot") ]
                                .arg( datPngPath )
                                .arg( figCount++ )
                                .arg( datPngName );
                        reportHtml.append(par);

                    }
                }


                if(as->getShotsDetails() != 0)
                {
                    par = models[ tr("shot_data") ]
                            .arg( eventNr )                     // arg %1
                            .arg( csvEnd.at(CF_DATE_UTC) )
                            .arg( csvEnd.at(CF_TIME_UTC) )
                            .arg( csvEnd.at(CF_STEP_HZ) )
                            .arg( csvEnd.at(CF_UP_THR_DBFS) )
                            .arg( csvEnd.at(CF_DN_THR_DBFS) )
                            .arg( csvEnd.at(CF_RANGE_LOW) )
                            .arg( csvEnd.at(CF_RANGE_HI) )

                            .arg( csvBeg.at(CF_PEAK_DBFS) )     //arg %9
                            .arg( csvPeak.at(CF_PEAK_DBFS) )
                            .arg( csvEnd.at(CF_PEAK_DBFS) )

                            .arg( csvBeg.at(CF_AVG_DBFS) )      //arg %12
                            .arg( csvPeak.at(CF_AVG_DBFS) )
                            .arg( csvEnd.at(CF_AVG_DBFS) )

                            .arg( csvBeg.at(CF_DIFF_DBFS) )     //arg %15
                            .arg( csvPeak.at(CF_DIFF_DBFS) )
                            .arg( csvEnd.at(CF_DIFF_DBFS) )

                            .arg( csvBeg.at(CF_PEAK_HZ) )       //arg %18
                            .arg( csvPeak.at(CF_PEAK_HZ) )
                            .arg( csvEnd.at(CF_PEAK_HZ) )

                            .arg( csvBeg.at(CF_LASTING_MS) )    //arg %21
                            .arg( csvPeak.at(CF_LASTING_MS) )
                            .arg( csvEnd.at(CF_LASTING_MS) )

                            .arg( csvEnd.at(CF_SHIFT_HZ) )      //arg %24
                            .arg( csvEnd.at(CF_ECHO_AREA) )     //arg %25
                            .arg( csvEnd.at(CF_INTERVAL_AREA) ) //arg %26
                            .arg( csvEnd.at(CF_PEAKS_COUNT) )   //arg %27
                            .arg( csvEnd.at(CF_LOS_SPEED) );     //arg %28

                    reportHtml.append(par);
                    MYDEBUG << "shot: " << eventNr;
                }
                /*
                par = models[ tr("footer") ].arg(pageNum++);
                sect.append(par);

                par = models[ tr("heading") ]
                        .arg(VERSION_NR)
                        .arg(RELEASE_NR)
                        .arg( QDate::currentDate().toString(Qt::SystemLocaleShortDate) );
                sect.append(par);
                */
            }

        }
        as->getMainWindow()->unsetCursor();

    }

    par = models[ tr("end") ];
    reportHtml.append(par);

    /*BUG: direct pdf looks bad
    QTextDocument doc;
    QString styleSheetName = ad.absoluteFilePath(REPORT_CSS);
    doc.setDefaultStyleSheet( styleSheetName );
    doc.setDocumentMargin( 0.5 );
    doc.setHtml( reportHtml );
    QPdfWriter pdfDev(  wd.absoluteFilePath(REPORT_PDF) );
    doc.print( &pdfDev );
*/
    QTextDocument document;
    document.setHtml(reportHtml);

    //PDF output through QPrinter

    /*
     * still looks ugly,
     * screenshots and shot data overlaps in the same
     * pages. Bad html or smth missing in pdf setup?

    QString pdfName = reportName;
    pdfName.replace(HTML_EXT, PDF_EXT);
    MYINFO << "Saving " << pdfName;
    QPrinter printer(QPrinter::PrinterResolution);
    printer.setOutputFormat(QPrinter::PdfFormat);
    printer.setPageOrientation(QPageLayout::Landscape);
    printer.setPaperSize(QPrinter::A4);
    printer.setOutputFileName( wd.absoluteFilePath(pdfName) );
    printer.setPageMargins( QMarginsF(15, 15, 15, 15) );
    document.print(&printer);
*/

    //direct HTML output
    MYINFO << "Saving " << reportName;
    f.setFileName( wd.absoluteFilePath(reportName) );
    f.open( QIODevice::WriteOnly );
    s.setDevice(&f);
    s << reportHtml << endl;
    f.flush();
    f.close();
    MYINFO << "Done.";

    wd.popCd();
    return true;
}



QList<DailyCounters*> PostProc::selectData(QDate from, QDate to)
{
    MYINFO << "selectData(" << from.toString(Qt::SystemLocaleShortDate)
           << ", " << to.toString(Qt::SystemLocaleShortDate) << ") ";

    QList<DailyCounters*> ldc;
    QList<QDate> dates = sd.keys();

    QDate day;

    foreach(day, dates)
    {
        if(day >= from && day <= to)
        {
            sd[day]->day = day;
            ldc.append( sd[day] );
        }

    }

    return ldc;
}


QString PostProc::mapBackColor(uint maxCount, uint count)
{
    QString rgb;
    int index;

    //zero value is always black
    if(count == 0)
    {
        rgb = "#000000";
    }
    else if (count == maxCount)
    {
        //full scale
        rgb = "#ff1600";
    }
    else
    {
        //Colorgramme scale, 25 levels
        QStringList colorList ={
            "#000000", // [0]
            "#0064fe",
            "#006dfd",
            "#0079f0",
            "#0087fe",
            "#009eff",
            "#00b0ff",
            "#00c6f3",
            "#00eaff",
            "#38fcf5",
            "#4efdd4",
            "#6cfcb8",
            "#83fca1",
            "#a4f892",   // [13]
            "#b8fc7a",
            "#d1fc68",
            "#f4f96d",
            "#ffe600",
            "#ffc500",
            "#ffae00",
            "#ff9400",
            "#ff8600",
            "#ff7300",
            "#ff6300",
            "#ff1600"   //[24] = maxCount
        };

        float ratio = static_cast<float>(maxCount) / 24;
        index = static_cast<int>(count / ratio);
        rgb = colorList.at( index );
    }

    return rgb;
}


QString PostProc::mapForeColor(uint maxCount, uint count)
{
    QString rgb;
    int index;

    //zero value is always white
    if(count == 0)
    {
        rgb = "#ffffff";
    }
    else if (count == maxCount)
    {
        //full scale
        rgb = "#000000";
    }
    else
    {
        //Text colors for colorgramme scale, 25 levels
        QStringList colorList ={
            "#ffffff",          //"#000000", // [0]
            "#ffffff",         //"#0064fe",
            "#ffffff",         //"#006dfd",
            "#ffffff",         //"#0079f0",
            "#ffffff",         //"#0087fe",
            "#ffffff",         //"#009eff",
            "#ffffff",         //"#00b0ff",
            "#ffffff",         //"#00c6f3",
            "#000000",         //"#00eaff",
            "#000000",         //"#38fcf5",
            "#000000",         //"#4efdd4",
            "#000000",         //"#6cfcb8",
            "#000000",         //"#83fca1",
            "#000000",         //"#a4f892",   // [13]
            "#000000",         //"#b8fc7a",
            "#000000",         //"#d1fc68",
            "#000000",         //"#f4f96d",
            "#000000",         //"#ffe600",
            "#000000",         //"#ffc500",
            "#000000",         //"#ffae00",
            "#000000",         //"#ff9400",
            "#000000",         //"#ff8600",
            "#000000",         //"#ff7300",
            "#000000",         //"#ff6300",
            "#000000",         //"#ff1600"   //[24] = maxCount
        };

        float ratio = static_cast<float>(maxCount) / 24;
        index = static_cast<int>(count / ratio);
        rgb = colorList.at( index );
    }

    return rgb;
}




QString PostProc::plotToPNG( QString shotName, int rangeLo, int rangeHi, int msLasting )
{
    Q_UNUSED(msLasting)

    QString baseDatPath;
    QString datPath;
    //bool wantPower = false;
    //bool wantSpectra = false;
    PLOT_TYPES pt;

    //shots could be DATs instead of PNGs if the program has been
    //started in console mode, in that case the shotName is adjusted
    shotName.replace( SSHOT_EXT, GP_DATA_EXT );

    if( shotName.contains(AUTOSHOT_PREFIX) == false )
    {
        //no shot found
        return "";
    }

    //from the shot name, it builds the path of the relative plot file
    //baseDatPath = wd.absoluteFilePath( destGPspecPath + "/" + shotName );
    baseDatPath = wd.relativeFilePath( destGPspecPath + "/" + shotName );

    //assumes it's a 2D mapped plot first
    datPath = baseDatPath;
    datPath.replace( AUTOSHOT_PREFIX, SPECMAP_PREFIX );

    if ( QFile::exists( datPath ) == true )
    {
        //file exists, plot type found.
        pt = PT_2D_MAPPED;
    }
    else
    {
        //if the file does not exist, tries to find a 3D real plot

        datPath = baseDatPath;
        datPath.replace( AUTOSHOT_PREFIX, SPEC3D_PREFIX );

        if ( QFile::exists( datPath ) == true )
        {
            //file exists, plot type found.
            pt = PT_3D_REAL;
        }
        else
        {
            //if the file does not exist, tries to find a 2D power plot
            //baseDatPath = wd.absoluteFilePath( destGPpowPath + "/" + shotName );
            baseDatPath = wd.relativeFilePath( destGPpowPath + "/" + shotName );
            datPath = baseDatPath;
            datPath.replace( AUTOSHOT_PREFIX, POW2D_PREFIX );
            if ( QFile::exists( datPath ) == true )
            {
                //file exists, plot type found.
                pt = PT_2D_POWER;
            }
            else
            {
                return ""; //no plot found
            }
        }
    }

    QString absDatPngPath = datPath;
    absDatPngPath.replace( GP_DATA_EXT, QString(GP_DATA_EXT)+QString(SSHOT_EXT) );

    QString datPngName = wd.splitPath( absDatPngPath ).last();

    QString cmdFileName("temp.plt");
    QFile f( wd.absoluteFilePath(cmdFileName) );
    if( f.open(QIODevice::WriteOnly) == false )
    {
        MYCRITICAL << "error opening " << cmdFileName << " for writing" ;
        return "";
    }

    QTextStream gp(&f);

    //common part
    gp << "#Temporary GNUPLOT command file automatically generated by Echoes " << APP_VERSION  << endl;
    gp << "set terminal png size 1024,768 font \"fixed,10\" " << endl;
    gp << "set termoption noenhanced" << endl;
    gp << "set output \"" << wd.absoluteFilePath(absDatPngPath) << "\"" << endl;
    gp << "set border 4095" << endl;
    gp << "set ytics out" << endl;
    gp << "set ztics out" << endl;
    gp << "set mxtics 10" << endl;
    gp << "set mytics 10" << endl;
    gp << "set mztics 10" << endl;

    if(pt == PT_2D_MAPPED)
    {
        //mapped 2d-spectra: each scan is composed by multiple rows, terminated with
        //the average (N) maximum (S) and difference (S-N), like POW2D does
        gp << "set title \"Spectrogram  file: " << datPngName << " \" offset 0,3" << endl;
        gp << "set xlabel \"Hz\" font \"fixed,16\" offset 0,-2" << endl;
        gp << "set ylabel \"UTC Hour\"  font \"fixed,16\"" << endl;
        gp << "set key left outside" << endl;
        gp << "set grid xtics ytics ztics" << endl;
        gp << "set cblabel \"dBfs\" font \"fixed,16\" offset 1,0" << endl;
        gp << "set xtics out rotate border offset 0,-1 format \"%.0f\" " << endl;
        gp << "unset surface" << endl;
        gp << "set view 0,0" << endl;
        gp << "set pm3d at s map" << endl;
        gp << "set format y \"%H:%M:%S\" timedate" << endl;

        gp << "set palette model RGB" << endl;
        gp << "set cbrange [" << MIN_DBFS << ":" << MAX_DBFS << "]" << endl;
        gp << "set palette defined (";

        //gets the gnuplot palette from configuration
        gp << as->getGNUplotPalette();
        gp << ")" << endl;

        gp << "set xrange [" << rangeLo << ":" << rangeHi << "]" << endl;
        gp << "splot \"" << datPath << "\" using 2:1:3 notitle" << endl;

    }
    else if(pt == PT_3D_REAL)
    {
        //3D-spectra: file format is identical to SPECMAP but power is plotted
        //on Z-axis with colored dots
        gp << "set title \"3D Spectrogram  file: " << datPngName << " \" offset 0,3" << endl;
        gp << "set xlabel \"Hz\" font \"fixed,16\" offset 0,-2" << endl;
        gp << "set ylabel \"UTC Hour\"  font \"fixed,16\"" << endl;
        gp << "set zlabel \"dBfs\" font \"fixed,16\" " << endl;
        gp << "set key left outside" << endl;
        gp << "set grid xtics ytics ztics" << endl;
        gp << "set cblabel \"dBfs\" font \"fixed,16\" offset 1,0" << endl;
        gp << "set xtics out rotate border offset 0,-1 format \"%.0f\" " << endl;
        gp << "unset surface" << endl << endl;
        gp << "set view 45,45, 1,1" << endl;
        gp << "set pm3d implicit at s" << endl << endl;
        gp << "set format y \"%H:%M:%S\" timedate" << endl;

        gp << "set palette model RGB" << endl;
        gp << "set cbrange [" << MIN_DBFS << ":" << MAX_DBFS << "]" << endl;
        gp << "set palette defined (";

        //gets the gnuplot palette from configuration
        gp << as->getGNUplotPalette();
        gp << ")" << endl;

        gp << "set xrange [" << rangeLo << ":" << rangeHi << "]" << endl;
        gp << "splot \"" << datPath << "\" using 2:1:3 notitle" << endl;

    }
    else if(pt == PT_2D_POWER)
    {
        //2D-power: the average (N) maximum (S) and difference (S-N) one row per scan
        gp << "set title \"Power plot\" offset 0,3" << endl;
        gp << "set xlabel \"UTC Hour\" font \"fixed,16\" offset 0,-2" << endl;
        gp << "set ylabel \"dBfs\" font \"fixed,16\" offset 0,-2" << endl;
        gp << "set grid xtics ytics" << endl;
        gp << "set xtics out rotate border offset 0,-1" << endl;
        gp << "set format x \"%H:%M:%S\" timedate" << endl;

        gp << "set palette model RGB" << endl;
        gp << "set cbrange [" << MIN_DBFS << ":" << MAX_DBFS << "]" << endl;
        gp << "set palette defined (";

        //gets the gnuplot palette from configuration
        gp << as->getGNUplotPalette();
        gp << ")" << endl;

        gp << "set title \"Power plot,  file: " << datPngName << " \" offset 0,3" << endl;
        gp << "plot \"" << datPath  << "\" using 1:3 with lines title \"maximum power S\", \"";
        gp << datPath  << "\" using 1:2 with lines title \"average power N\", \"";
        gp << datPath << "\" using 1:4 with lines title \"max difference S-N\"" << endl;
    }
    gp.flush();


    //once the command file has been created
    //invokes GNUplot to generate the PNG
    int ret = QProcess::execute( as->getGNUplotPath(), QStringList( wd.absoluteFilePath( TEMP_GP_CMD )) );
    if (ret < 0)
    {
        QString msg = tr("The GNUplot executable failed to start.\nPlots won't be present in the report");
        MYWARNING << msg;

        if(as->getAutoReport() == Qt::Checked)
        {
            //no dialog in autoreport
            return "";
        }

        QMessageBox::warning( as->getMainWindow(), tr("Error"), msg );
        return "";
    }

    QFile chkGPout(absDatPngPath);

    if ( chkGPout.open(QIODevice::ReadOnly) == false || chkGPout.size() == 0 )
    {
        QString bakCmd = datPngName + GP_CMD_EXT;
        QFile::copy( TEMP_GP_CMD, bakCmd );
        QString msg = tr("GNUplot found an error in\n%1,\n\nThe actual settings could have been changed since that file has been created.\n\nSaving the command file as %2\nfor debug purposes.")
                .arg(datPath)
                .arg(bakCmd);
        MYINFO << msg;
        QMessageBox::warning( as->getMainWindow(), tr("Error"), msg );
        return "";

    }
    return absDatPngPath;

}

/*
  Old colorgram with absolute color value, hours on X axis and days on Y axis

QString PostProc::drawCgrams(QString pngFile, QList<QDate>& days, QList< QList<uint> >& set )
{
    //26 columns, day + one column for each hour count from 00h to 23h + daily totals
    QTableWidget tw(days.count() + 1, 26);
    tw.setStyleSheet( "background-color: white; color: black;" );


    QTableWidgetItem* twi = new QTableWidgetItem( tr("Day / Hour") );
    MY_ASSERT( nullptr !=  twi);
    tw.setItem(0, 0, twi);
    QString hs;
    int hour = 0;
    for(hour = 0; hour < 24; hour++)
    {
        //hs = hs.sprintf("%02uh", hour);
        hs = QString("%1h").arg( hour, 2, 10, QChar('0') );
        twi = new QTableWidgetItem( hs );
        MY_ASSERT( nullptr !=  twi);
        tw.setItem(0, hour+1, twi);
    }
    twi = new QTableWidgetItem( tr("Daily totals") );
    MY_ASSERT( nullptr !=  twi);
    tw.setItem(0, hour+1, twi);

    //completes the image by one row for each day
    int day = 1;
    QList<uint> daySet;
    QDate date;


    foreach (date, days)
    {
        daySet = set.at(day-1);
        twi = new QTableWidgetItem( date.toString(Qt::TextDate) );
        MY_ASSERT( nullptr !=  twi);
        tw.setItem(day, 0, twi);

        uint total = 0;
        for(int hour = 0; hour < 24; hour++)
        {
            total += daySet.at(hour);

            //hs = hs.sprintf("%u", daySet.at(hour));
            hs = QString("%1").arg( daySet.at(hour) );
            twi = new QTableWidgetItem( hs );
            MY_ASSERT( nullptr !=  twi);
            twi->setBackground( QColor( mapColor( daySet.at(hour)))  );
            tw.setItem(day, hour+1, twi);
        }
        //hs = hs.sprintf("%u", total);
        hs = QString("%1").arg(total);
        twi = new QTableWidgetItem( hs );  //total
        MY_ASSERT( nullptr !=  twi);
        tw.setItem(day, hour+1, twi);
        day++;

    }
    tw.verticalHeader()->hide();
    tw.horizontalHeader()->hide();
    tw.setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    tw.setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    tw.resizeColumnsToContents();
    tw.resizeRowsToContents();
    tw.setSizeAdjustPolicy( QAbstractScrollArea::AdjustToContents);
    tw.adjustSize();
    QPixmap pix = tw.grab();
    pix.save(pngFile);
    return pngFile;

}

*/



QString PostProc::drawCgrams(QString pngFile, QList<QDate>& days, QList< QList<uint> >& set )
{
    QDate date;
    int hour = 0;
    int day = 1;
    uint maxValue = 0;
    QList<uint> daySet;

    //26 rows, day + one row for each hour count from 00h to 23h + daily totals
    QTableWidget tw(26, days.count() + 1);
    tw.setStyleSheet( "background-color: white; color: black;" );

    //calculate full-scale value (white color)
    foreach (date, days)
    {
        daySet = set.at(day-1);
        for(int hour = 0; hour < 24; hour++)
        {
            if( daySet.at(hour) > maxValue )
            {
                maxValue = daySet.at(hour);
            }
        }
        day++;
    }
    MYINFO << "maxValue = " << maxValue;

    day = 1;
    QTableWidgetItem* twi = new QTableWidgetItem( tr("Hour / Day") );
    MY_ASSERT( nullptr !=  twi);
    tw.setItem(0, 0, twi);
    QString hs;
    for(hour = 0; hour < 24; hour++)
    {
        //hs = hs.sprintf("%02uh", hour);
        hs = QString("%1h").arg( hour, 2, 10, QChar('0') );
        twi = new QTableWidgetItem( hs );
        MY_ASSERT( nullptr !=  twi);
        tw.setItem(hour+1, 0, twi);
    }
    twi = new QTableWidgetItem( tr("Daily totals") );
    MY_ASSERT( nullptr !=  twi);
    tw.setItem(hour+1, 0, twi);

    //completes the image by one column for each day
    foreach (date, days)
    {
        daySet = set.at(day-1);
        QString dayString;
        dayString.setNum(date.day());
        twi = new QTableWidgetItem( dayString, 0 );
        MY_ASSERT( nullptr !=  twi);
        tw.setItem(0, day, twi);

        uint total = 0;
        for(int hour = 0; hour < 24; hour++)
        {
            total += daySet.at(hour);

            //hs = hs.sprintf("%u", daySet.at(hour));
            hs = QString("%1").arg( daySet.at(hour) );
            twi = new QTableWidgetItem( hs );
            MY_ASSERT( nullptr !=  twi);
            twi->setBackground( QColor( mapBackColor( maxValue, daySet.at(hour)))  );
            twi->setForeground( QColor( mapForeColor( maxValue, daySet.at(hour)))  );
            tw.setItem(hour+1, day, twi);
        }
        //hs = hs.sprintf("%u", total);
        hs = QString("%1").arg(total);
        twi = new QTableWidgetItem( hs );  //total
        MY_ASSERT( nullptr !=  twi);
        tw.setItem(hour+1, day, twi);
        day++;

    }
    tw.verticalHeader()->hide();
    tw.horizontalHeader()->hide();
    tw.setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    tw.setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    tw.resizeColumnsToContents();
    tw.resizeRowsToContents();
    tw.setSizeAdjustPolicy( QAbstractScrollArea::AdjustToContents);
    tw.adjustSize();
    QPixmap pix = tw.grab();
    pix.save(pngFile);
    return pngFile;

}




QString PostProc::chkBoxStateSimple( int chkState )
{
    switch(chkState)
    {
        case Qt::Checked:
            return tr("yes");

        case Qt::Unchecked:
            return tr("no");

        case Qt::PartiallyChecked:
            return tr("partial");

    }
    return tr("???");
}


QString PostProc::plotCgrams(QString pngFile, QList<QDate>& days, QList< QList<uint> >& set )
{
    //first, the data must be written down to a dat file
    //26 columns, day + one column for each hour count from 00h to 23h + daily totals

    QString datFileName = pngFile + GP_DATA_EXT ;
    QFile f( wd.absoluteFilePath(datFileName) );
    if( f.open(QIODevice::WriteOnly) == false )
    {
        MYCRITICAL << "error opening " << datFileName << " for writing" ;
        return "";
    }

    QTextStream gp(&f);


    gp << "#Temporary GNUPLOT data file automatically generated by Echoes " << APP_VERSION  << endl;

    int day = 1;
    QList<uint> daySet;
    QDate date;

    gp << "-" << '\t';
    for(int hour = 0; hour < 24; hour++)
    {
        gp << hour << '\t';
    }
    gp << "total" << endl;

    foreach (date, days)
    {
        daySet = set.at(day-1);
        gp << date.toString(Qt::SystemLocaleShortDate) << '\t';

        uint total = 0;
        for(int hour = 0; hour < 24; hour++)
        {
            total += daySet.at(hour);
            gp << daySet.at(hour) << '\t';
        }
        gp << total << endl;
        day++;
    }
    gp.flush();
    f.close();


    QString cmdFileName =  wd.absoluteFilePath(TEMP_GP_CMD);
    f.setFileName( cmdFileName );
    if( f.open(QIODevice::WriteOnly) == false )
    {
        MYCRITICAL << "error opening " << cmdFileName << " for writing" ;
        return "";
    }

    gp.setDevice(&f);

    //common part
    gp << "#Temporary GNUPLOT command file automatically generated by Echoes " << APP_VERSION << endl;
    gp << "set terminal png size 1024,768 font \"fixed,10\" " << endl;
    gp << "set termoption noenhanced" << endl;
    gp << "set output \"" << wd.absoluteFilePath(pngFile) << "\"" << endl;
    gp << "set style data histogram" << endl;
    gp << "set style histogram cluster gap 1" << endl;
    gp << "set style fill solid border -1" << endl;
    gp << "set boxwidth 0.9" << endl;
    gp << "set border 4095" << endl;
    gp << "set xlabel \"Days\" offset 0,-4 " << endl;
    gp << "set ylabel \"Count\" offset -2,0 " << endl;
    gp << "set grid xtics ytics" << endl;
    gp << "set xtic rotate by -45 scale 0" << endl;
    gp << "plot \"" << datFileName << "\" using 1:xticlabels(1), for [i=2:*] '' using i title column" << endl;
    gp.flush();
    f.close();

    //once the command file has been created
    //invokes GNUplot to generate the PNG
    int ret = QProcess::execute( as->getGNUplotPath(), QStringList( cmdFileName ) );
    if (ret < 0)
    {
        QMessageBox::warning( as->getMainWindow(), tr("Error"), tr("The GNUplot executable failed to start.\nPlots won't be present in the report") );
        return "";
    }

    QFile chkGPout(pngFile);

    if ( chkGPout.size() == 0 || chkGPout.open(QIODevice::ReadOnly) == false )
    {
        QString bakCmd = cmdFileName + ".bak";
        QFile::copy( cmdFileName, bakCmd );
        QString msg = tr("GNUplot found an error in\n%1,\n\nThe actual settings could have been changed since that file has been created.\n\nSaving the command file as %2\nfor debug purposes.")
                .arg(cmdFileName)
                .arg(bakCmd);
        MYINFO << msg;
        QMessageBox::warning( as->getMainWindow(), tr("Error"), msg );
        return "";

    }


    return pngFile;
}


bool PostProc::saveData()
{
    bool ok = false;
    QFile df;


    //serializing sd[] to disk
    //wd.pushCd( as->getConfigName()  );
    wd.pushCd(destRootPath);
    QString fileName( wd.absoluteFilePath(DAILYCTR_NAME) );
    MYINFO << "Creating daily data file: " << fileName;
    df.setFileName( fileName );
    ok = df.open(QIODevice::WriteOnly | QIODevice::Truncate);
    if(ok == true)
    {
        //the header must be built with basic data types
        //in order to be always readable by different QDataStream versions
        MYINFO << "Saving magic and versions";

        DailyCounters* dc;
        QDate date;
        QDataStream ds(&df);

        ds << DAILYCTR_MAGIC;

        //convert APP_VERSION string to a number
        QString av(APP_VERSION);
        av = av.replace(".","");
        uint verNum = av.toUInt();
        ds << verNum;

        //end of header: the remaining data can be written using the
        //following QDataStream version
        ds.setVersion(QDataStream::Qt_5_0);

        //list of days covered by archive
        //the first element[0] is always invalid, but it's saved anyway
        ds << sd.keys();

        foreach(date, sd.keys())
        {
            if(date.isValid() == true)
            {
                MYINFO << "Saving data for day " << date.toString(Qt::ISODate);
                dc = sd[date];
                ok = dc->save(ds);
                if( ok == false )
                {
                    MYWARNING << "bad datastream status: " << ds.status();
                    break;
                }
            }
        }
    }
    wd.popCd();
    return ok;
}

bool PostProc::loadData()
{
    bool ok = false;
    QFile df;

    wd.pushCd( as->getConfigName()  );
    QString fileName( wd.absoluteFilePath(DAILYCTR_NAME) );
    df.setFileName( fileName );
    ok = df.open(QIODevice::ReadOnly);
    if(ok == true)
    {
        uint magic;
        QDate date;
        QDataStream ds(&df);
        DailyCounters* dc;

        //convert APP_VERSION string to a number
        QString av(APP_VERSION);
        av = av.replace(".","");
        uint wantVer = av.toUInt();
        uint appVer;

        //the header must be built with basic data types
        //in order to be always readable by different QDataStream versions
        ds >> magic;
        if(magic != DAILYCTR_MAGIC)
        {
            MYWARNING << "invalid magic: " << magic << " instead of " << DAILYCTR_MAGIC;
            wd.popCd();
            return false;
        }

        ds >> appVer;
        if(appVer != wantVer)
        {
            MYWARNING << "invalid version: " << appVer << " instead of " << wantVer;
            wd.popCd();
            return false;
        }

        //the stream format should be decided on appVer basis
        //this is the more conservative possible, since different
        //linux distros could have different QT5 releases
        ds.setVersion( QDataStream::Qt_5_0 );

        //file valid, clears completely sd[] and relative children
        //before deserializing it
        foreach(date, sd.keys())
        {
            dc = sd[date];
            if(dc != nullptr)
            {
                dc->reset();
            }
        }
        sd.clear();

        //deserializing sd[] from disk

        //days list
        QList<QDate> sdKeysRaw;
        ds >> sdKeysRaw;
        if(ds.status() != QDataStream::Ok)
        {
            wd.popCd();
            return false;
        }

        MYINFO << "Loaded " << sdKeysRaw.count() << " days in sd[] from " << fileName;

        //days directories present in archive
        QString dirNameMask = "\?\?\?\?-\?\?-\?\?";
        QStringList filter;
        filter << dirNameMask;
        QStringList allDirs = wd.entryList( filter, XQDir::Dirs, XQDir::Name );
        MYINFO << "Found " << allDirs.count() << " daily directories in archive";


        //forgets days no more present in archive
        QStringList allDays;
        QString day;
        QList<QDate> sdKeys;
        foreach (date, sdKeysRaw)
        {
            if(date.isValid() == true)
            {
                day = date.toString(Qt::ISODate);
                if (allDirs.contains(day) == false)
                {
                    MYINFO << "Forgetting " << day << " because it's no more present in archive";
                    continue;
                }
                sdKeys.append(date);
            }
        }

        //now loads days data
        foreach (date, sdKeys)
        {
            if(date.isValid() == true)
            {
                dc = new DailyCounters;
                MY_ASSERT( nullptr !=  dc);
                sd[date] = dc;
                ok = dc->load(ds);
                if( ok == false )
                {
                    MYWARNING << "bad datastream status: " << ds.status();
                    wd.popCd();
                    return false;
                }
            }
            MYINFO << "Loaded " << date.toString(Qt::ISODate) << " data to sd[]";
        }
    }
    wd.popCd();
    return ok;

}

bool PostProc::archiveDailyData(QStringList& selectedFiles, QMap< uint,QVector<uint> >& shotsToday, QString subDirName)
{

    QVector<uint> hourGroup;
    QString subs;
    uint event = 0;

    MYINFO << "archiving " << subDirName << " events";
    for(uint h = 0; h < 24; h++)
    {
        MYINFO << " Hour: " << h;
        hourGroup = shotsToday.value(h);

        foreach (event, hourGroup)
        {
            MYINFO << " Event#: " << event;

            //extension not specified
            //subs = subs.sprintf("_%05u.", event);
            subs = QString("_%1.").arg(event, 5, 10, QChar('0'));
            QStringList filters = selectedFiles.filter(subs);

            QString fileSrc;
            foreach (fileSrc, filters)
            {
                QString pathSrc = wd.absoluteFilePath(fileSrc);
                QString fileDest = QString("%1/%2")
                        .arg( subDirName )
                        .arg(fileSrc);
                QString pathDest = wd.absoluteFilePath(fileDest);
                bool ok = wd.copy(pathSrc, pathDest);
                if(ok == false)
                {
                    MYWARNING << "Error copying file "
                               << pathSrc
                               << " to "
                               << pathDest;

                    wd.popCd();
                    return false;
                }

                ok = QFile::remove(pathSrc);
                if(ok == false)
                {
                    MYWARNING << "Error removing file " << pathSrc ;

                    wd.popCd();
                    return false;
                }
            }

        }
    }

    return true;
}
