/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.
    

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 356                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-11-02 19:12:07 +01#$: Date of last commit
$Id: postproc.h 356 2018-11-02 18:12:07Z gmbertani $
*******************************************************************************/

#ifndef POSTPROC_H
#define POSTPROC_H

#include <QtCore>
#include <QtGui>
//#include <QtPrintSupport>
#include "settings.h"

namespace Ui {
    class MainWindow;
}

extern bool console;

class Control;


/**
 * @brief The PP_TASK enum
 *
 * PostProc is a worker that can execute long
 * jobs in background.
 */

enum PP_TASK
{
    PP_NONE,
    PP_MIDNIGHT,
    PP_FLATTEN
};

/**
 * Events statistic counters
 *
 *
*/
class DailyCounters
{
public:
    QDate   day;

    ///events in this day, grouped by cathegory
    QVector<uint> allEvents;
    QVector<uint> allGoods;
    QVector<uint> allOvers;
    QVector<uint> allUnders;
    QVector<uint> allFakes;

    ///associative arrays:
    /// daily events grouped by cathegory and splitten by hour
    QMap< uint,QVector<uint> > hEvents;
    QMap< uint,QVector<uint> > hGoods;
    QMap< uint,QVector<uint> > hOvers;
    QMap< uint,QVector<uint> > hUnders;
    QMap< uint,QVector<uint> > hFakes;

    ///events lasting by hour
    QMap< uint,QVector<uint> > hLastings;

    ///maximum absolute S power by hour
    QMap<uint,float> hS;

    ///average N power by hour
    QMap<uint,float> hN;

    ///shots filenames by cathegory
    QStringList   shotsOver;
    QStringList   shotsUnder;
    QStringList   shotsFake;

    ///CSV records of each event by phase
    QMap<uint,QStringList> dataBegin;
    QMap<uint,QStringList> dataPeak;
    QMap<uint,QStringList> dataEnd;


    DailyCounters()
    {
        reset();
    }

    void reset();
    bool save(QDataStream &ds);
    bool load(QDataStream &ds);
};



class PostProc : public QThread
{
    Q_OBJECT

protected:

    Settings*       as;
    Control*        ac;

    bool stopProc;

    PP_TASK         task;

    XQDir           wd;             //working directory
    QPdfWriter*     pdfReport;      //printable PDF report
    QDate           processDT;      //yesterday UTC

    QMap<QDate, DailyCounters*> sd;       //stored data by date

    /*
     * daily data storage directory tree
     *
     * .echoes/configName/yesterday  => destRootPath (Level 1)
     *
     * destRootPath/destStatsPath   (Level 2)
     *
     *              destGPpath/overDir    (Level 3)
     *                        /underDir
     *                         /fakeDir
     *
     *              destShotsPath/overDir  (Level 3)
     *                        /underDir
     *                         /fakeDir
     *
     *
     *
     */

    //absolute paths of working subdirs
    QString destRootPath;       ///level 0 (configuration)
    QString destDatePath;       ///level 1 (date)
    QString destStatsPath;      ///level 2
    QString destGPspecPath;     ///level 2
    QString destGPpowPath;      ///level 2
    QString destShotsPath;      ///level 2

    QString extResourcesPath;      ///location of css file and other auxiliary files


    //destination folder for flatten archive
    XQDir flatDir;

    ///
    /// \brief createDailyReport
    /// \return
    /// collects all the data generated with the actual configuration,
    /// filling up the <sd> list. The file is saved in archive, under
    /// the actual configuration subdirectory
    bool createDailyReport();

    ///
    /// \brief plotToPNG
    /// \param shotName
    /// \param rangeLo
    /// \param rangeHi
    /// \param msLasting
    /// \return
    ///
    QString plotToPNG(QString shotName , int rangeLo, int rangeHi, int msLasting);


    ///
    /// \brief checkExpiredData
    /// \return
    ///
    int checkExpiredData();

    ///
    /// \brief oldFiles
    /// \return
    /// returns a list of absolute filepaths older
    /// than <data lasting>
    QStringList selectSubDirs(QDate &from, QDate &to, bool absolutePaths = true);

    ///
    /// \brief parseDateFromName
    /// \param fileName
    /// \param prefix
    /// \param cfg
    /// \param start
    /// \return event nr. if defined, otherwise 1 if ok and 0 for failure
    ///
    uint parseEventDateFromName(QString fileName, const QString& prefix, const QString& ext, const QString& cfg, QDate &start);


    ///
    /// \brief midnightTasks
    /// \return
    ///
    bool midnightTasks();

    ///
    /// \brief selectFiles
    /// \param prefix
    /// \param suffix
    /// \return
    ///
    QStringList selectFiles(const QString &prefix, const QString &suffix);

    ///
    /// \brief moveTo
    /// \param mask
    /// \param prefix
    /// \param ext
    /// \param destPath
    /// \return
    ///
    bool moveTo(const QString &prefix, const QString &ext, QString destPath);

    ///
    /// \brief classify
    /// \param statCSV
    /// \param destPath
    /// \return
    ///
    bool archiveData(QDate &refDate, QString& destPath);


    ///
    /// \brief archiveShots
    /// \param selectedFiles
    /// \param shotsToday
    /// \param subDirName
    /// \return
    ///
    bool archiveDailyData(QStringList& selectedFiles, QMap< uint,QVector<uint> >& shotsToday, QString subDirName);

    /// \brief selectData
    /// \param from
    /// \param to
    /// \return
    ///
    QList<DailyCounters *> selectData(QDate from, QDate to);

    ///
    /// \brief mapBackColor
    /// \param value
    /// \return
    ///
    /// returns the RGB color corresponding <value>
    /// in a colorgramme-like dynamic palette where 0 is always black
    /// and maxCount is white.
    QString mapBackColor(uint maxCount, uint count);

    ///
    /// \brief mapForeColor
    /// \param value
    /// \return
    ///
    /// returns the RGB color corresponding <value>
    /// to the foreground color for text in colorgramme
    QString mapForeColor(uint maxCount, uint count);


    ///
    /// \brief drawCgrams
    /// \param pngFile
    /// \param days
    /// \param set
    /// \return
    ///
    QString drawCgrams(QString pngFile, QList<QDate> &days, QList<QList<uint> > &set);

    ///
    /// \brief plotCgrams
    /// \param pngFile
    /// \param days
    /// \param set
    /// \return
    ///
    QString plotCgrams(QString pngFile, QList<QDate>& days, QList< QList<uint> >& set );

    ///
    /// \brief chkBoxStateSimple
    /// \param state
    /// \return
    ///
    QString chkBoxStateSimple(int chkState );

    ///
    /// \brief flatten
    /// \return
    /// creates a flatten copy of the archive
    ///
    bool flatten();

public:
    explicit PostProc(Settings* appSettings, Control* appControl, XQDir workDir, QObject *parent=0);

    ~PostProc();

    void run();

    ///
    /// \brief setWdir
    /// \param wdir
    ///
    void setWdir(const XQDir& wdir)
    {
        wd = wdir;
    }

    ///
    /// \brief stop
    ///
    void stop()
    {
        stopProc = true;
    }

    ///
    /// \brief createFullReport
    /// \return
    /// note: must run in GUI thread
    bool createFullReport( QString reportName );

    ///
    /// \brief setFlatPath
    /// \param dest
    /// set destination folder for flattened archive
    void setFlatDir(const XQDir& dest)
    {
        flatDir = dest;
    }

    ///
    /// \brief setTask
    /// \param ppt
    ///
    void setTask( PP_TASK ppt )
    {
        task = ppt;
    }

    ///
    /// \brief archiveExists
    /// \param from
    /// \param to
    /// \return
    ///
    bool archiveExists(QDate &from, QDate &to);

    ///
    /// \brief saveData
    /// \return
    ///
    bool saveData();

    ///
    /// \brief loadData
    /// \return
    ///
    bool loadData();


signals:
    ///
    /// \brief status
    /// \param msg
    /// updates status bar in MainWindow
    void status(const QString& msg);

    ///
    /// \brief autoReport
    ///
    void autoReport();
};

#endif // POSTPROC_H
