/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.
    

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 351                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-10-22 00:13:47 +02#$: Date of last commit
$Id: ruler.cpp 351 2018-10-21 22:13:47Z gmbertani $
*******************************************************************************/


#include "ruler.h"

Ruler::Ruler(Settings* appSettings, RULER_ORIENT orient, QWidget *parent) :
    QWidget(parent)
{
    setObjectName("Ruler widget");
    MY_ASSERT( nullptr !=  appSettings);

    as = appSettings;
    ro = orient;

    pixMinTicks   = 0;
    pixMajTicks   = 0;
    hzMajTicks  = 0;
    hzMinTicks  = 0;
    hzPerPixel  = 0;

    rangeLo = 0;
    rangeHi = 0;

    lowBound      = 0;
    highBound     = 0;
    coverage = 0;
    rect = parent->contentsRect();

    //self resizes to match the entire contents area
    //of parent widget
    resize( rect.width(), rect.height() );

    notches = as->getNotches();

    setContextMenuPolicy( Qt::DefaultContextMenu );
    setMouseTracking(true);
    defaultTip = toolTip();

    pup = nullptr;


}

Ruler::~Ruler()
{
    //delete img;

}

void Ruler::mouseMoveEvent(QMouseEvent *event)
{
    int hz = static_cast<int>((event->x() * hzPerPixel) + lowBound);

    //if the mouse is hovering a notch marker
    //a tooltip with its parameters is showed

    Notch n;
    QString newTip;
    int tol = 5 * static_cast<int>(hzPerPixel);

    for(int i = 0; i < notches.count(); i++)
    {
        n = notches.at(i);
        if(hz > (n.begin - tol) &&
           hz < (n.end + tol))
        {
            newTip = QString(tr("Notch#%1\nat %2 Hz\n%3 Hz wide"))
                    .arg(i+1).arg(n.freq).arg(n.width);
            setToolTip(newTip);
            return;

        }
    }



    if(hz > rangeLo &&
       hz < (rangeLo + tol + tol))
    {
        newTip = tr("Start of event detection range");
        setToolTip(newTip);
        return;
    }

    if(hz > (rangeHi - tol - tol) &&
       hz < rangeHi)
    {
        newTip = tr("End of event detection range");
        setToolTip(newTip);
        return;
    }

    setToolTip(defaultTip);
}


void Ruler::slotUpdateNotches(int index)
{
    MY_ASSERT(pup != nullptr)

    Notch nn = pup->getNotch();
    notches.replace(index, nn);
    MYINFO << "updating notch #" << index+1;
    as->setNotches(notches);
    update();
}


void Ruler::contextMenuEvent(QContextMenuEvent *event)
{
    int hz = static_cast<int>(event->x() * hzPerPixel)  +lowBound;
    Notch n;

    //int tol = static_cast<int>( ceil(as->getResolution()) ) * 10;
    int tol = 5 * static_cast<int>(hzPerPixel);

    int index = -1;

    //Checks if the mouse has been clicked near an already
    //defined notch, otherwise considers it's a new one.


    for(int i = 0; i < notches.count(); i++)
    {
        n = notches.at(i);
        if(hz > (n.begin - tol) &&
           hz < (n.end + tol))
        {
            //opens an existing one
            index = i;
            break;
        }
    }

    if( index == -1 )
    {
        //no marker found at this frequency
        //asks to create a brand new one
        n.freq = hz;
        n.width = 0;
        n.begin = hz;
        n.end = hz;

        //keeps the list ordered by  freq
        for(int i = 0; i < notches.count(); i++)
        {
            Notch nn = notches.at(i);
            if(n.freq < nn.freq)
            {
                notches.insert(i, n);
                MYINFO << "inserted new notch as #" << index+1;
                index = i;
                break;
            }
        }

        if(index == -1)
        {
            index = notches.count();
            MYINFO << "appended new notch as #" << index+1;
            notches.append(n);
        }

    }

    pup = new NotchPopup(n, as, this);
    pup->setWindowFlags(Qt::Tool|Qt::FramelessWindowHint);
    pup->move(event->globalPos().x(), event->globalPos().y() + 30);
    pup->setIndex(index);

    connect( pup, SIGNAL( notchChanged(int) ), this, SLOT( slotUpdateNotches(int) ) );
    int result = pup->exec();

    if( result == QDialog::Accepted )
    {
        n = pup->getNotch();
        if( n.width != 0 )
        {
            slotUpdateNotches(index);
        }
        else
        {
            //removes the element from the list
            MYINFO << "deleting notch #" << pup->getIndex();
            notches.removeAt( pup->getIndex() );
        }
    }
    else
    {
        //aborting notch
        MYINFO << "aborting notch #" << pup->getIndex();
        notches.removeAt( pup->getIndex() );
    }

    as->setNotches(notches);
    update();

    delete pup;
    pup = nullptr;
}


void Ruler::setGrid(int maxTicks, int hzFrom, int hzTo)
{
    rect = parentWidget()->contentsRect();
    resize( rect.width(), rect.height() );

    hzMinTicks = HORZ_MIN_TICKS_STEP;

    notches = as->getNotches();

    int zoom = as->getZoom();
    coverage = hzTo - hzFrom;
    center = hzFrom + (coverage / 2);
    lowBound =  hzFrom;
    highBound = hzTo;

    hzMinTicks = static_cast<int>(zoom / 10.0);

    int ticks = static_cast<int>( as->pround( coverage / hzMinTicks ) );
    int mult = 5;
    int minTicks = 10 * maxTicks;   //max nr. of minor ticks
    while(ticks > minTicks)
    {
        hzMinTicks *= mult;
        ticks = static_cast<int>( as->pround( coverage / hzMinTicks) );
        mult=(mult==5)? 2 : 5;              //1000-->5000-->10000-->50000-->etc.
    }

    hzMajTicks = hzMinTicks * 10;

    //converting hz to pixels:
    hzPerPixel = static_cast<double>( coverage ) / ( contentsRect().width()  );  //Hz per pixel
    as->setHpp( hzPerPixel );


    pixMinTicks = static_cast<int>( as->pround( static_cast<double>( hzMinTicks ) / hzPerPixel ));     //pixels per minor tick
    pixMajTicks = static_cast<int>( as->pround( static_cast<double>( hzMajTicks ) / hzPerPixel ));     //pixels per major tick

    MYINFO << "pixMinTicks=" << pixMinTicks << " pixMajTicks=" << pixMajTicks
           << " hzPerPixel=" << hzPerPixel << " width=" << width();

    update();
}

void Ruler::paintEvent( QPaintEvent *pe )
{
    Q_UNUSED(pe) //avoids warning unused parameter

    QPainter p(this);

    int xm = 0, oldXm = 0;
    int hsw = 0;

    p.setFont( parentWidget()->font() );

    QPoint p1,p2,p3;
    QString hzString;
    QFontMetrics fm( p.font() );
    QImage qi;

    QPen penText, penLine;
    penLine.setColor(Qt::gray);
    penText.setColor(Qt::cyan);

    Notch n;

    if(ro == RO_DOWN || ro == RO_UP)
    {        
        //drawing markers
        for( double hz =  lowBound; hz <= highBound; hz += 1 )
        {
            xm = static_cast<int>( as->pround( (hz - lowBound) / hzPerPixel ) ) + 2;
            p1.setX(xm);
            p2.setX(xm);
            p3.setX(xm);

            if(static_cast<int>(hz) == rangeLo)
            {
                //left peak threshold limit
                qi.load(":rangeLo16");
                p.drawImage(xm - qi.width(), 2, qi);
            }

            if(static_cast<int>(hz) == rangeHi)
            {
                //right peak threshold limit
                qi.load(":rangeHi16");
                p.drawImage(xm, 2, qi);
            }

            if(static_cast<int>(hz) == center)
            {
                //center frequency marker
                qi.load(":index");
                p.drawImage(xm - (qi.width()/2), 2, qi);
            }
        }


        //adding notches
        QPixmap marker(":/notch");
        QString label;
        MY_ASSERT(marker.isNull() == false)
        penLine.setColor(Qt::red);
        penLine.setWidth(3);
        p.setPen( penLine );
        foreach (n, notches)
        {
            xm = static_cast<int>(as->pround( (n.freq - lowBound) / hzPerPixel )) - (marker.width() / 2);
            p1.setX(xm);
            p1.setY(0);

            //center marker
            p.drawPixmap(p1, marker);

            //notch width

            xm = static_cast<int>( as->pround( (n.begin - lowBound) / hzPerPixel ) );
            p2.setX( xm );
            p2.setY( 2 );
            xm = static_cast<int>( as->pround( (n.end - lowBound) / hzPerPixel ) );
            p3.setX( xm );
            p3.setY( 2 );

            p.drawLine(p2,p3);

        }

        //drawing scale
        penLine.setColor(Qt::gray);
        penText.setColor(Qt::cyan);
        penLine.setWidth(1);

        for( double hz = lowBound; hz <= highBound; hz += 1 )
        {
            xm = static_cast<int>( as->pround( (hz - lowBound) / hzPerPixel ) ) + 2;
            p1.setX(xm);
            p2.setX(xm);
            p3.setX(xm);

            if( fmod(hz, hzMajTicks) == 0.0 )
            {
                //major ticks

                p1.setY(0);
                p2.setY(rect.height());
                p.setPen(penLine);
                p.drawLine(p1,p2);

                hzString = QString("%1.%2")
                        .arg( static_cast<int>(hz)/1000 )
                        .arg( (static_cast<int>(hz) % 1000), 3,10,QChar('0'));

                //hsw = fm.horizontalAdvance( hzString ) / 2 + 3;
                hsw = (fm.boundingRect( hzString ).width() / 2) + 3;

                if((xm - oldXm) > (2 * hsw))
                {
                    //avoids numbers cluttering
                    p3.setX( xm - hsw + 3 );
                    p3.setY(rect.height() - 16);
                    p.setPen(penText);
                    p.drawText( p3, hzString );     //strings are always drawn horizontal
                    oldXm = xm;
                }

            }
            else if( fmod(hz, 5 * hzMinTicks) == 0.0 )
            {
                //half ticks
                p1.setY(rect.height() - 2);
                p2.setY(rect.height() - 15);
                p.setPen(penLine);
                p.drawLine(p1,p2);
            }
            else if( fmod(hz, hzMinTicks) == 0.0 )
            {
                //minor ticks
                p1.setY(rect.height() - 2);
                p2.setY(rect.height() - 8);
                p.setPen(penLine);
                p.drawLine(p1,p2);
            }
        }

    }

    if(ro == RO_LEFT || ro == RO_RIGHT)
    {
        //todo
    }

}

