/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.
    

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html





*******************************************************************************
$Rev:: 367                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-11-23 18:22:34 +01#$: Date of last commit
$Id: settings.cpp 367 2018-11-23 17:22:34Z gmbertani $
*******************************************************************************/

#include <float.h>
//#include "fftw3.h"
#include "settings.h"




Settings::Settings(QObject *parent) :
    QObject(parent)
{
    setObjectName("Settings object");

    //main window
    configName          = DEFAULT_CONFIG_NAME;
    geom                = DEFAULT_MAINWND_GEOM;

    //device settings:
    device              = DEFAULT_DEVICE;
    sampleRate          = DEFAULT_RATE;
    tune                = DEFAULT_TUNE;
    gain                = DEFAULT_GAIN;
    ppmError            = DEFAULT_PPMERR;
    ssBypass            = DEFAULT_SSBYPASS;

    //FFT settings:
    ftw                 = FTW_RECTANGLE;
    resolution          = DEFAULT_RESOLUTION;
    fftBins             = DEFAULT_FFT_BINS;

    //output settings:
    acqMode             = DEFAULT_ACQ_MODE;
    interval            = DEFAULT_INTVL;
    shots               = DEFAULT_MAX_SHOTS;
    recTime             = DEFAULT_RECTIME;
    after               = DEFAULT_AUTO_AFTER;
    thresholdsBehavior  = DEFAULT_THRESHOLDS;
    upThreshold         = DEFAULT_AUTO_UP_THRESHOLD;
    dnThreshold         = DEFAULT_AUTO_DN_THRESHOLD;
    tRange              = DEFAULT_AUTO_TRANGE;
    genGNUplot          = DEFAULT_GNUPLOT;
    genScreenshots      = DEFAULT_SCREENSHOTS;
    minFree             = DEFAULT_FREE_SPACE;
    plotType            = DEFAULT_PLOT_TYPE;
    avgdScans           = DEFAULT_N_SCANS;
    joinTime            = DEFAULT_JOIN_TIME;
    GNUplotPalette      = DEFAULT_GNUPLOT_PALETTE;

    //report settings
    QDate todayUTC = QDateTime::currentDateTimeUtc().date();
    coverTo             = todayUTC.toString(Qt::ISODate);  //to today
    todayUTC = todayUTC.addDays(-1);
    coverFrom           = todayUTC.toString(Qt::ISODate);  //from yesterday

    underdenseBelowMs   = DEFAULT_UNDERDENSE_MS;
    overdenseBelowSec   = DEFAULT_OVERDENSE_SEC;
    colorGraph          = DEFAULT_COLORGRAPH;
    siteInfos           = DEFAULT_SITEINFOS;
    overdenseShots      = DEFAULT_OVERDENSE_SHOTS;
    underdenseShots     = DEFAULT_UNDERDENSE_SHOTS;
    overdensePlots      = DEFAULT_OVERDENSE_PLOTS;
    underdensePlots     = DEFAULT_UNDERDENSE_PLOTS;
    shotsDetails        = DEFAULT_SHOTS_DETAILS;
    summary             = DEFAULT_SUMMARY;
    restart             = DEFAULT_RESTART;
    autoReport          = DEFAULT_AUTO_REPORT;
    carrierFilter       = DEFAULT_CARRIER_FILTER;
    lightingsFilter     = DEFAULT_LIGHTINGS_FILTER;
    saturationFilter    = DEFAULT_SATURATION_FILTER;

    //preferences
    hzTickMarks         = DEFAULT_HZ_TICKS;
    dBfsTickMarks       = DEFAULT_DBFS_TICKS;
    wfSecTickMarks      = DEFAULT_WSEC_TICKS;
    tooltips            = DEFAULT_TOOLTIPS;
    eraseLogs           = DEFAULT_ERASE_LOGS;
    eraseShots          = DEFAULT_ERASE_SHOTS;
    pingOn              = DEFAULT_PING;
    palette             = DEFAULT_WF_PALETTE;
    dataLasting         = DEFAULT_DATA_LASTING;
    GNUplotPath         = DEFAULT_GNUPLOT_PATH;
    noiseLimit          = DEFAULT_NOISE_LIMIT;
    serverPort          = DEFAULT_SERVER_PORT;
    serverAddr          = QHostAddress(DEFAULT_SERVER_ADDRESS);

    //waterfall window:
    wfOffset            = DEFAULT_OFFSET;
    wfBw                = DEFAULT_BW;
    wfZoom              = DEFAULT_ZOOM;
    brightness          = DEFAULT_BRIGHTNESS;
    contrast            = DEFAULT_CONTRAST;
    powerZoom           = DEFAULT_POWERZOOM;
    powerOffset         = DEFAULT_POWEROFFSET;
    wwGeom              = DEFAULT_WATERFALL_GEOM;
    wwPaneGeom          = DEFAULT_WATERFALL_GEOM; //must be calculated at runtime

    //these values are calculated by waterfall window
    //but to allow working in console mode, they must
    //also be stored in configuration:

    consoleMode         = false;
    dumbMode            = false;

    zoomLo              = 0;
    zoomHi              = DEFAULT_BW;

    //siteinfos
    logo                = DEFAULT_LOGO;
    latitude            = DEFAULT_LATITUDE;
    longitude           = DEFAULT_LONGITUDE;
    altitude            = DEFAULT_ALTITUDE;
    stationName         = DEFAULT_STATION_NAME;
    contact             = DEFAULT_CONTACT;
    rxSetup             = DEFAULT_RXSETUP;
    notes               = DEFAULT_NOTES;

    //local settings
    lastEventDate       = QDate::currentDate();
    lastEventNr         = 0;


    //not saved on disk
    ww = nullptr;
    mw = nullptr;
    changed = 0;
    hzPerPixel = 0;

    localSettings = nullptr;
}

Settings:: ~Settings()
{
    //save local settings before quit
    saveLocal();
    delete localSettings;
}


/**
 * WARNING!!!
 * in case a change in groups/items, this method must be updated later.
 * In this way it will be possible to convert existing rts files in
 * the new format by simply running the program and opening/saving them.
 * After conversion, this method can be updated in order
 * to load correctly the rts files with the new format.
 */
void Settings::load( QSettings& ini, bool enableNotifications )
{
    configFullPath =  ini.fileName();
    configName = QFileInfo( configFullPath ).baseName();

    //the local settings are placed in the same place
    //platform-dependent, the user has no way to change them
    //
    QString targetInstance;
    targetInstance = QString("%1-%2").arg(TARGET_NAME).arg(configName);

    localSettings = new QSettings(COMPANY_NAME, targetInstance);
    MY_ASSERT( nullptr !=  localSettings)

    loadLocal();


    //device settings:
    ini.beginGroup( "Device settings" );
    device          = ini.value("Device", DEFAULT_DEVICE).toString();
    sampleRate      = ini.value("SampleRate", DEFAULT_RATE).toInt();
    tune            = ini.value("Tune",DEFAULT_TUNE).toInt();
    gain            = ini.value("Gain", DEFAULT_GAIN).toInt();
    ppmError        = ini.value("PpmError", DEFAULT_PPMERR).toInt();
    ini.endGroup();

    //FFT settings:
    ini.beginGroup( "FFT settings" );
    ftw             = static_cast<FFT_WINDOWS>( ini.value("FFTwindow", FTW_RECTANGLE).toInt() );
    resolution      = ini.value("Resolution", DEFAULT_RESOLUTION).toLongLong();
    fftBins         = ini.value("FFTbins", DEFAULT_FFT_BINS).toInt();
    ssBypass        = ini.value("SsBypass", DEFAULT_SSBYPASS).toInt();
    wfBw            = ini.value("Bandwidth", DEFAULT_BW).toInt();
    ini.endGroup();

    //output settings:
    ini.beginGroup( "Output settings" );
    acqMode         = ini.value("AcquisitionMode", DEFAULT_ACQ_MODE).toInt();
    interval        = ini.value("SamplingInterval", DEFAULT_INTVL).toInt();
    shots           = ini.value("MaxShots", DEFAULT_MAX_SHOTS).toInt();
    recTime         = ini.value("RecordingTime", DEFAULT_RECTIME).toInt();
    after           = ini.value("ShotsAfter", DEFAULT_AUTO_AFTER).toInt();
    thresholdsBehavior   = ini.value("ThresholdsBehavior", DEFAULT_THRESHOLDS).toInt();
    upThreshold     = ini.value("ShotUpThreshold",DEFAULT_AUTO_UP_THRESHOLD).toInt();
    dnThreshold     = ini.value("ShotDnThreshold",DEFAULT_AUTO_DN_THRESHOLD).toInt();
    tRange          = ini.value("ShotFrequencyRange", DEFAULT_AUTO_TRANGE).toInt();
    genScreenshots  = ini.value("GenerateScreenshots", DEFAULT_SCREENSHOTS).toInt();
    genGNUplot      = ini.value("GenerateGNUplot", DEFAULT_GNUPLOT).toInt();
    minFree         = ini.value("MinFree", DEFAULT_FREE_SPACE).toInt();
    plotType        = ini.value("PlotType", DEFAULT_PLOT_TYPE).toInt();
    avgdScans       = ini.value("AveragedScans", DEFAULT_N_SCANS).toInt();
    joinTime        = ini.value("JoinTime", DEFAULT_JOIN_TIME).toInt();
    GNUplotPalette  = ini.value("GnuPlotPalette", DEFAULT_GNUPLOT_PALETTE).toString();
    ini.endGroup();

    //report settings:
    ini.beginGroup( "Report settings" );
    QDate todayUTC = QDateTime::currentDateTimeUtc().date();
    coverTo             = ini.value( "CoverTo", todayUTC.toString(Qt::ISODate) ).toString();
    coverFrom           = ini.value( "CoverFrom",  todayUTC.toString(Qt::ISODate)).toString();
    underdenseBelowMs   = ini.value("UnderdenseBelowMs", DEFAULT_UNDERDENSE_MS).toInt();
    overdenseBelowSec   = ini.value("OverdenseBelowSec", DEFAULT_OVERDENSE_SEC).toInt();
    colorGraph          = ini.value("ColorGraph", DEFAULT_COLORGRAPH).toInt();
    siteInfos           = ini.value("SiteInfos", DEFAULT_SITEINFOS).toInt();
    overdenseShots      = ini.value("OverdenseShots", DEFAULT_OVERDENSE_SHOTS).toInt();
    underdenseShots     = ini.value("UnderdenseShots", DEFAULT_UNDERDENSE_SHOTS).toInt();
    overdensePlots      = ini.value("OverdensePlots", DEFAULT_OVERDENSE_PLOTS).toInt();
    underdensePlots     = ini.value("UnderdensePlots", DEFAULT_UNDERDENSE_PLOTS).toInt();
    shotsDetails        = ini.value("ShotsDetails", DEFAULT_SHOTS_DETAILS).toInt();
    summary             = ini.value("SummaryTable", DEFAULT_SUMMARY).toInt();
    restart             = ini.value("AcqRestart", DEFAULT_RESTART).toInt();
    autoReport          = ini.value("AutoReport", DEFAULT_AUTO_REPORT).toInt();
    carrierFilter       = ini.value("CarrierFilter", DEFAULT_CARRIER_FILTER).toInt();
    lightingsFilter     = ini.value("LightingsFilter", DEFAULT_LIGHTINGS_FILTER).toInt();
    saturationFilter    = ini.value("SaturationFilter", DEFAULT_SATURATION_FILTER).toInt();
    ini.endGroup();

    //preferences settings:
    ini.beginGroup( "Preferences" );
    wfSecTickMarks  = ini.value("WaterfallSecTickMarks", DEFAULT_WSEC_TICKS).toInt();
    hzTickMarks     = ini.value("WaterfallHzTickMarks", DEFAULT_HZ_TICKS).toInt();
    dBfsTickMarks   = ini.value("PowergraphDbmTickMarks", DEFAULT_DBFS_TICKS).toInt();
    tooltips        = ini.value("TooltipsShown", DEFAULT_TOOLTIPS).toInt();
    pingOn          = ini.value("EnablePing", DEFAULT_PING).toInt();
    palette         = ini.value("Palette", DEFAULT_WF_PALETTE).toInt();
    eraseLogs       = ini.value("EraseLogs", DEFAULT_ERASE_LOGS).toInt();
    eraseShots      = ini.value("EraseShots", DEFAULT_ERASE_SHOTS).toInt();
    dataLasting     = ini.value("DataLasting", DEFAULT_DATA_LASTING).toInt();
    noiseLimit      = ini.value("NoiseLimit", DEFAULT_NOISE_LIMIT).toInt();
    serverPort      = ini.value("ServerPort", DEFAULT_SERVER_PORT).toInt();
    serverAddr      = QHostAddress( ini.value("ServerAddress", DEFAULT_SERVER_ADDRESS).toString() );
    geom            = ini.value("MainGeom", DEFAULT_MAINWND_GEOM).toRect();

    ini.endGroup();

    //siteinfo settings
    ini.beginGroup( "Site infos" );
    logo           =  ini.value("StationLogo",  DEFAULT_LOGO).toString();
    latitude        = ini.value("Latitude",     DEFAULT_LATITUDE).toDouble();
    longitude       = ini.value("Longitude",    DEFAULT_LONGITUDE).toDouble();
    altitude        = ini.value("Altitude",     DEFAULT_ALTITUDE).toInt();
    stationName     = ini.value("StationName",  DEFAULT_STATION_NAME).toString();
    contact         = ini.value("Contact",      DEFAULT_CONTACT).toString();
    rxSetup         = ini.value("RxSetup",      DEFAULT_RXSETUP).toString();
    notes           = ini.value("Notes",        DEFAULT_NOTES).toString();
    ini.endGroup();


    //waterfall window
    ini.beginGroup( "Waterfall settings" );
    wfOffset        = ini.value("WaterfallOffset", DEFAULT_OFFSET).toInt();
    wfZoom          = ini.value("WaterfallZoom", DEFAULT_ZOOM).toInt();
    brightness      = ini.value("WaterfallBrightness", DEFAULT_BRIGHTNESS).toInt();
    contrast        = ini.value("WaterfallContrast", DEFAULT_CONTRAST).toInt();
    powerOffset     = ini.value("WaterfallPowerOffset", DEFAULT_POWEROFFSET).toInt();
    powerZoom       = ini.value("WaterfallPowerZoom", DEFAULT_POWERZOOM).toInt();    
    wwGeom          = ini.value("WaterfallGeom", DEFAULT_WATERFALL_GEOM).toRect();
    wwPaneGeom      = ini.value("WaterfallPaneGeom", DEFAULT_WATERFALL_GEOM).toRect();
    zoomLo          = ini.value("WaterfallZoomLowerBound", 0).toInt();
    zoomHi          = ini.value("WaterfallZoomHigherBound", DEFAULT_BW).toInt();

    ini.endGroup();

    //fix zoom values < 1.0 (==10)
    wfZoom = (wfZoom > DEFAULT_ZOOM)? wfZoom : DEFAULT_ZOOM;

    //notches
    ini.beginGroup( "Notch filters" );
    int sz = ini.beginReadArray("Notches");
    Notch nt;
    for (int i = 0; i < sz; ++i)
    {
         ini.setArrayIndex(i);
         nt.freq = ini.value("Frequency").toInt();
         nt.width = ini.value("Bandwidth").toInt();
         nt.begin = ini.value("BeginHz").toInt();
         nt.end = ini.value("EndHz").toInt();
         notches.append(nt);
    }

    ini.endArray();
    ini.endGroup();

    MYINFO << "load(" << ini.fileName() << ")" << endl;

    if(enableNotifications == true)
    {
        emit notifyOpened();
    }

    changed = 0;
}

void Settings::save( QSettings& ini, bool enableNotifications  )
{
    if(localSettings == nullptr)
    {
        configFullPath =  ini.fileName();
        configName = QFileInfo( configFullPath ).baseName();

        //the local settings are placed in the same place
        //platform-dependent, the user has no way to change them
        //
        QString targetInstance;
        targetInstance = QString("%1-%2").arg(TARGET_NAME).arg(configName);

        localSettings = new QSettings(COMPANY_NAME, targetInstance);
        MY_ASSERT( nullptr !=  localSettings)
    }


    //device settings:
    ini.beginGroup( "Device settings" );
    ini.setValue( "Device", device );
    ini.setValue( "SampleRate", sampleRate );
    ini.setValue( "Tune", tune );
    ini.setValue( "Gain", gain );
    ini.setValue( "PpmError", ppmError );
    ini.endGroup();

    //FFT settings:
    ini.beginGroup( "FFT settings" );
    ini.setValue( "FFTwindow", static_cast<int>( ftw ) );
    ini.setValue( "FFTbins", fftBins );
    ini.setValue( "Resolution", resolution );
    ini.setValue( "SsBypass", ssBypass );
    ini.setValue( "Bandwidth", wfBw );
    ini.endGroup();

    //output settings:
    ini.beginGroup( "Output settings" );
    ini.setValue( "AcquisitionMode", acqMode );
    ini.setValue( "SamplingInterval", interval );
    ini.setValue( "MaxShots", shots );
    ini.setValue( "RecordingTime", recTime );
    ini.setValue( "ShotsAfter", after );
    ini.setValue( "ThresholdsBehavior", thresholdsBehavior );
    ini.setValue( "ShotUpThreshold", upThreshold );
    ini.setValue( "ShotDnThreshold", dnThreshold );
    ini.setValue( "ShotFrequencyRange", tRange );
    ini.setValue( "GenerateScreenshots", genScreenshots );
    ini.setValue( "GenerateGNUplot", genGNUplot );
    ini.setValue( "MinFree", minFree );
    ini.setValue( "PlotType", plotType );
    ini.setValue( "AveragedScans", avgdScans );
    ini.setValue( "JoinTime", joinTime );
    ini.setValue( "GnuPlotPalette", GNUplotPalette );
    ini.endGroup();

    //report settings:
    ini.beginGroup( "Report settings" );
    ini.setValue( "CoverFrom", coverFrom );
    ini.setValue( "CoverTo", coverTo );
    ini.setValue( "UnderdenseBelowMs", underdenseBelowMs );
    ini.setValue( "OverdenseBelowSec", overdenseBelowSec );
    ini.setValue( "ColorGraph", colorGraph );
    ini.setValue( "SiteInfos", siteInfos );
    ini.setValue( "OverdenseShots", overdenseShots );
    ini.setValue( "UnderdenseShots", underdenseShots );
    ini.setValue( "OverdensePlots", overdensePlots );
    ini.setValue( "UnderdensePlots", underdensePlots );
    ini.setValue( "ShotsDetails", shotsDetails );
    ini.setValue( "SummaryTable", summary );
    ini.setValue( "AcqRestart", restart );
    ini.setValue( "AutoReport", autoReport );
    ini.setValue( "CarrierFilter", carrierFilter );
    ini.setValue( "LightingsFilter", lightingsFilter );
    ini.setValue( "SaturationFilter", saturationFilter );
    ini.endGroup();

    //preferences
    ini.beginGroup( "Preferences" );
    ini.setValue( "WaterfallSecTickMarks", wfSecTickMarks );
    ini.setValue( "WaterfallHzTickMarks", hzTickMarks );
    ini.setValue( "PowergraphDbmTickMarks", dBfsTickMarks );
    ini.setValue( "TooltipsShown", tooltips );
    ini.setValue( "EnablePing", pingOn );
    ini.setValue( "Palette", palette );
    ini.setValue( "EraseLogs", eraseLogs );
    ini.setValue( "EraseShots", eraseShots );
    ini.setValue( "DataLasting", dataLasting );
    ini.setValue( "NoiseLimit", noiseLimit );
    ini.setValue( "ServerPort", serverPort );
    ini.setValue( "ServerAddress", serverAddr.toString() );
    ini.setValue(  "MainGeom", geom );

    ini.endGroup();

    //siteinfos
    ini.beginGroup( "Site infos" );
    ini.setValue( "StationLogo", logo );
    ini.setValue( "Latitude", latitude );
    ini.setValue( "Longitude", longitude );
    ini.setValue( "Altitude", altitude );
    ini.setValue( "StationName", stationName );
    ini.setValue( "Contact", contact );
    ini.setValue( "RxSetup", rxSetup );
    ini.setValue( "Notes", notes );
    ini.endGroup();

    //waterfall window
    ini.beginGroup( "Waterfall settings" );
    ini.setValue( "WaterfallOffset", wfOffset );
    ini.setValue( "WaterfallZoom", wfZoom );
    ini.setValue( "WaterfallBrightness", brightness );
    ini.setValue( "WaterfallContrast", contrast );
    ini.setValue( "WaterfallPowerZoom", powerZoom );
    ini.setValue( "WaterfallPowerOffset", powerOffset );    
    ini.setValue(  "WaterfallGeom", wwGeom );
    ini.setValue(  "WaterfallPaneGeom", wwPaneGeom );
    ini.setValue(  "WaterfallZoomLowerBound", zoomLo );
    ini.setValue(  "WaterfallZoomHigherBound", zoomHi );
    ini.endGroup();

    //notches
    ini.beginGroup( "Notch filters" );
    int sz = notches.count();
    ini.beginWriteArray("Notches");
    Notch nt;
    for (int i = 0; i < sz; ++i)
    {
         ini.setArrayIndex(i);
         nt = notches.at(i);
         ini.setValue("Frequency", nt.freq);
         ini.setValue("Bandwidth", nt.width);
         ini.setValue("BeginHz", nt.begin);
         ini.setValue("EndHz", nt.end);
    }

    ini.endArray();
    ini.endGroup();

    ini.sync();

    MYINFO << "save(" << ini.fileName() << ")" << endl;

    if(ini.fileName() != DEFAULT_CONFIG_NAME)
    {
        if(enableNotifications == true)
        {
            emit notifySavedAs();
        }
    }

    changed = 0;
}


void Settings::loadLocal()
{

    MYINFO << "loadLocal(" << localSettings->fileName() << ")" << endl;
    localSettings->beginGroup( "LocalSettings" );
    lastEventDate   = localSettings->value("LastEventDate", QDate::currentDate()).toDate();
    lastEventNr     = localSettings->value("LastEventNr", 0).toInt();
    GNUplotPath     = localSettings->value("GNUplotPath", DEFAULT_GNUPLOT_PATH).toString();
    localSettings->endGroup();

}



void Settings::saveLocal()
{
    MYINFO << "saveLocal(" << localSettings->fileName() << ")" << endl;
    localSettings->beginGroup( "LocalSettings" );
    localSettings->setValue( "LastEventDate", lastEventDate );
    localSettings->setValue( "LastEventNr", lastEventNr );
    localSettings->setValue( "GNUplotPath", GNUplotPath );
    localSettings->endGroup();
}



//main window settings


void Settings::loadConfig( QString& name, bool enableNotifications )
{
    //opens a new configuration

    if (true /*name != configName*/)
    {
        configFullPath = name;
        MYINFO << "getConfig(" << configFullPath << ")" << endl;
        configName = QFileInfo( configFullPath ).baseName();
        QSettings ini( configFullPath, QSettings::IniFormat );
        load(ini, enableNotifications);
    }
}

void Settings::setConfigName(XQDir wd, QString& name  )
{
    //perform a SaveAs

    configName = name;
    MYINFO << "setConfigName(" << configName << ")" << endl;
    QString rts = configName+CONFIG_EXT;
    QSettings ini( wd.absoluteFilePath(rts), QSettings::IniFormat );
    save(ini, true);
}



//device settings

void Settings::setDevice( const QString& dev, bool enableNotifications )
{
    if (dev.size() > 0)
    {
        device = dev;
        MYDEBUG << "setDevice(" << dev << ")" << endl;

        if(enableNotifications == true)
        {
            emit notifySetDevice();
        }
        changed++;
    }
}

QString Settings::getDeviceNoIdx()
{
    QString chopped = device.section(':', 1);
    return chopped;
}


void Settings::setSampleRate(int value  , bool enableNotifications)
{
    if (value != sampleRate)
    {       
        //actual FFT resolution in points:
        float totalBins = static_cast<float>(wfBw / getResolution());

        sampleRate = value;
        MYDEBUG << "setSampleRate(" << value << ") bandwidth is set equal to SR" << endl;
        wfBw = sampleRate;


        //the resolution changes to mantain FFT bins unchanged
        double dRes = static_cast<double>(wfBw / totalBins);
        MYDEBUG << "resolution is " << dRes << " Hz" << endl;
        resolution = static_cast<qint64>(dRes * 10e8);
        if(enableNotifications == true)
        {
            emit notifySetSampleRate();
        }

        changed++;
    }
}


void Settings::setTune( int value, bool enableNotifications )
{
    if (value != tune)
    {
        MY_ASSERT(tune >= 0)
        tune = value;
        MYDEBUG << "setTune(" << value << ")" << endl;

        if(enableNotifications == true)
        {
            emit notifySetTune();
        }
        changed++;
    }
}



void Settings::setGain( int value  )
{
    if (value != gain)
    {
        gain = value;
        MYDEBUG << "setGain(" << value << ")" << endl;
        emit notifySetGain();
        changed++;
    }
}

void Settings::setError(double value)
{
    //max excursion: -327.68  to +327.67
    int iValue = static_cast<int>(value * 100.0);
    if ( iValue != ppmError )
    {
        ppmError = iValue;
        MYDEBUG << "setError(" << iValue << ")" << endl;
        emit notifySetError();
        changed++;
    }
}


//FFT settings
void Settings::setWindow( int value  )
{
    if ( value != static_cast<int>(ftw) )
    {
        ftw = static_cast<FFT_WINDOWS>( value );
        MYDEBUG << "setWindow(" << value << ")" << endl;
        emit notifySetWindow();
        changed++;
    }
}



void Settings::decResolution( bool enableNotifications )
{
    //actual FFT resolution in points:
    int checkBins = 0;

    int exp = static_cast<int>( ceilf( log2f( fftBins ) ) );
    exp++;
    checkBins = (1 << exp);
    if( checkBins <= MAX_FFT_BINS )
    {
        double dRes = static_cast<double>(wfBw) / checkBins;
        fftBins = checkBins;
        MYDEBUG << "decResolution+(" << dRes << ") new fftBins=" << fftBins << endl;
        resolution = static_cast<qint64>(dRes * 10e8);
        if(enableNotifications == true)
        {
            emit notifySetResolution();
        }
        changed++;
    }
}

void Settings::setResBins( int FFTbins, bool enableNotifications )
{
    //actual FFT resolution in points:
    int checkBins = 0;

    int exp = static_cast<int>( ceilf( log2f( FFTbins ) ) );
    checkBins = (1 << exp);
    double dRes = static_cast<double>(wfBw) / checkBins;
    fftBins = checkBins;
    MYDEBUG << "setResBins+(" << fftBins << ") new resolution=" << dRes << endl;
    resolution = static_cast<qint64>(dRes * 10e8);
    if(enableNotifications == true)
    {
        emit notifySetResolution();
    }
    changed++;
}

void Settings::incResolution( bool enableNotifications )
{
    //actual FFT resolution in points:
    int checkBins = 0;

    int exp = static_cast<int>( ceilf( log2f( fftBins ) ) );
    if(exp > 2)
    {
        exp--;
    }
    checkBins = (1 << exp);
    double dRes = static_cast<double>(wfBw) / checkBins;
    fftBins = checkBins;
    MYDEBUG << "incResolution+(" << dRes << ") new fftBins=" << fftBins << endl;
    resolution = static_cast<qint64>(dRes * 10e8);
    if(enableNotifications == true)
    {
        emit notifySetResolution();
    }
    changed++;
}

void Settings::updateResolution( bool enableNotifications )
{
    //actual FFT resolution in points:
    int checkBins = 0;

    int exp = static_cast<int>( ceilf( log2f( fftBins ) ) );
    checkBins = (1 << exp);
    double dRes = static_cast<double>(wfBw) / checkBins;
    fftBins = checkBins;
    MYDEBUG << "updateResolution() fftBins=" << fftBins << " resolution=" << dRes << endl;
    resolution = static_cast<qint64>(dRes * 10e8);
    if(enableNotifications == true)
    {
        emit notifySetResolution();
    }
    changed++;
}



void Settings::setSsBypass( int value  )
{
    if (value != ssBypass)
    {
        ssBypass = value;
        MYDEBUG << "ssBypass(" << value << ")" << endl;
        emit notifySsBypass();
        changed++;
    }
}


void Settings::setBandwidth( int value )
{
    if (value != wfBw)
    {

        double totalBins = wfBw / getResolution();
        wfBw = value;
        MYDEBUG << "setBandwidth(" << value << ")" << endl;

        //the FFT resolution must be adjusted to maintain the same totalBins:
        double dRes = wfBw / totalBins;

        MYDEBUG << "adjusted resolution after bandwidth change =" << dRes << " - totalBins=" << totalBins << endl;
        resolution = static_cast<qint64>(dRes * 10e8);
        emit notifySetBandwidth();
        emit notifySetResolution();
        changed++;
    }
}


//output settings

void Settings::setAcqMode( int value  )
{
    if (value != acqMode)
    {
        acqMode = value;
        MYDEBUG << "notifySetAcqMode(" << value << ")" << endl;
        emit notifySetAcqMode();
        changed++;
    }
}

void Settings::setInterval( int value  )
{
    if (value != interval)
    {
        interval = value;
        MYDEBUG << "setInterval(" << value << ")" << endl;
        emit notifySetInterval();
        changed++;
    }
}

void Settings::setShots( int value )
{
    if (value != shots)
    {
        shots = value;
        MYDEBUG << "setShots(" << value << ")" << endl;
        emit notifySetShots();
        changed++;
    }
}


void Settings::setRecTime( int value )
{
    if (value != recTime)
    {
        recTime = value;
        MYDEBUG << "setRecTime(" << value << ")" << endl;
        emit notifySetRecTime();
        changed++;
    }
}


void Settings::setAfter( int value )
{
    if (value != after)
    {
        after = value;
        MYDEBUG << "setAfter(" << value << ")" << endl;
        emit notifySetAfter();
        changed++;
    }
}


void Settings::setThresholdsBehavior(int value)
{
    if (value != thresholdsBehavior)
    {
        thresholdsBehavior = value;
        MYDEBUG << "setThresholdsBehavior(" << value << ")" << endl;
        changed++;
        emit notifySetThresholdsBehavior();
    }
}

void Settings::setUpThreshold(double value, bool enableNotifications )
{
    int iValue = static_cast<int>(value * 100.0);
    if ( iValue != upThreshold )
    {
        upThreshold = iValue;
        MYDEBUG << "setUpThreshold(" << iValue << ")" << endl;

        if(enableNotifications == true)
        {
            emit notifySetUpThreshold();
        }
        changed++;
    }
}

void Settings::setDnThreshold(double value, bool enableNotifications  )
{
    int iValue = static_cast<int>(value * 100.0);
    if ( iValue != dnThreshold )
    {
        dnThreshold = iValue;
        MYDEBUG << "setDnThreshold(" << iValue << ")" << endl;

        if(enableNotifications == true)
        {
            emit notifySetDnThreshold();
        }
        changed++;
    }
}


void Settings::setTrange( int value )
{
    if (value != tRange)
    {
        tRange = value;
        MYDEBUG << "setTrange(" << value << ")" << endl;
        emit notifySetTrange();
        changed++;
    }
}

void Settings::setGNUplot( int value  )
{
    if (value != genGNUplot)
    {
        genGNUplot = value;
        MYDEBUG << "setGNUplot(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setScreenshots( int value  )
{
    if (value != genScreenshots)
    {
        genScreenshots = value;
        MYDEBUG << "setScreenshots(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setMinFree(int value )
{
    if (value != minFree)
    {
        minFree = value;
        MYDEBUG << "setMinFree(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setPlotType( int value )
{
    if (value != plotType)
    {
        plotType = value;
        MYDEBUG << "setPlotType(" << value << ")" << endl;
        changed++;
    }
}



void Settings::setAvgdScans( int value )
{
    if (value != avgdScans)
    {
        avgdScans = value;
        MYDEBUG << "setAvgdScans(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setJoinTime( int value )
{
    if (value != joinTime)
    {
        joinTime = value;
        MYDEBUG << "setJoinTime( " << value << " )" << endl;
        changed++;
    }
}


void Settings::setGNUplotPalette( const QString& gpp )
{
    if (gpp.size() > 0 &&
            gpp != GNUplotPalette)
    {
        GNUplotPalette = gpp;
        MYDEBUG << "setGnuplotPalette(" << gpp << ")" << endl;
        changed++;
    }
}

void Settings::setGNUplotPath( const QString& path )
{
    if (path.size() > 0)
    {
        GNUplotPath = path;
        MYDEBUG << "seGNUplotPath(" << path << ")" << endl;
        changed++;
    }
}



//report settings:

void Settings::setCoverFrom(QDate cf)
{
    if (cf != getCoverFrom() &&
        cf < getCoverTo() )
    {
        coverFrom = cf.toString(Qt::ISODate);
        MYDEBUG << "setCoverFrom(" << coverFrom << ")" << endl;
        changed++;
    }
}


void Settings::setCoverTo( QDate ct )
{
    if (ct != getCoverTo() &&
        ct > getCoverFrom())
    {
        coverTo = ct.toString(Qt::ISODate);
        MYDEBUG << "setCoverTo(" << coverTo << ")" << endl;
        changed++;
    }
}

void Settings::setUnderdenseMs( int value  )
{
    if (value != underdenseBelowMs)
    {
        underdenseBelowMs = value;
        MYDEBUG << "setUnderdenseMs(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setOverdenseSec( int value  )
{
    if (value != overdenseBelowSec)
    {
        overdenseBelowSec = value;
        MYDEBUG << "setOverdenseSec(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setColorGraph( int value  )
{
    if (value != colorGraph)
    {
        colorGraph = value;
        MYDEBUG << "setColorGraph(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setSiteInfos( int value  )
{
    if (value != siteInfos)
    {
        siteInfos = value;
        MYDEBUG << "setSiteInfos(" << value << ")" << endl;
        changed++;
    }
}



void Settings::setOverdenseShots( int value  )
{
    if (value != overdenseShots)
    {
        overdenseShots = value;
        MYDEBUG << "setOverdenseShots(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setUnderdenseShots( int value  )
{
    if (value != underdenseShots)
    {
        underdenseShots = value;
        MYDEBUG << "setUnderdenseShots(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setOverdensePlots( int value  )
{
    if (value != overdensePlots)
    {
        overdensePlots = value;
        MYDEBUG << "setOverdensePlots(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setUnderdensePlots( int value  )
{
    if (value != underdensePlots)
    {
        underdensePlots = value;
        MYDEBUG << "setUnderdensePlots(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setShotsDetails( int value  )
{
    if (value != shotsDetails)
    {
        shotsDetails = value;
        MYDEBUG << "setShotsDetails(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setSummary( int value  )
{
    if (value != summary)
    {
        summary = value;
        MYDEBUG << "setSummary(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setRestart( int value )
{
    if (value != restart)
    {
        restart = value;
        MYDEBUG << "setRestart(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setAutoReport( int value )
{
    if (value != autoReport)
    {
        autoReport = value;
        MYDEBUG << "setAutoReport(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setCarrierFilter( int value )
{
    if (value != carrierFilter)
    {
        carrierFilter = value;
        MYDEBUG << "setCarrierFilter(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setLightingsFilter( int value )
{
    if (value != lightingsFilter)
    {
        lightingsFilter = value;
        MYDEBUG << "setLightingsFilter(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setSaturationFilter( int value )
{
    if (value != saturationFilter)
    {
        saturationFilter = value;
        MYDEBUG << "setSaturationFilter(" << value << ")" << endl;
        changed++;
    }
}



//waterfall settings:

void Settings::setOffset( int value )
{
    if (value != wfOffset)
    {
        wfOffset = value;
        MYDEBUG << "setOffset(" << value << ")" << endl;
        emit notifySetOffset();
        changed++;
    }
}


void Settings::setZoom( int value )
{
    if (value != wfZoom)
    {
        wfZoom = value;
        MYDEBUG << "setZoom(" << value << ")" << endl;
        emit notifySetZoom();
        changed++;
    }
}


void Settings::setZoomedRange( int lo, int hi )
{
    //these values are calculated by waterfall window
    //but to allow working in console mode, they must
    //also be stored in configuration
    if(zoomLo <= zoomHi)
    {
        zoomLo = lo;
        zoomHi = hi;
    }
}


void Settings::setBrightness( int value )
{
    if (value != brightness)
    {
        brightness = value;
        MYDEBUG << "setBrightness(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setContrast( int value )
{
    if (value != contrast)
    {
        contrast = value;
        MYDEBUG << "setContrast(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setPowerOffset( int value )
{
    if (value != powerOffset)
    {
        powerOffset = value;
        MYDEBUG << "setPowerOffset(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setPowerZoom( int value )
{
    if (value != powerZoom)
    {
        powerZoom = value;
        MYDEBUG << "setPowerZoom(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setNotches( QList<Notch>& ntList )
{
    notches = ntList;
    MYDEBUG << "setNotches(" << ntList.count() << " notches )";
    changed++;
}


//preferences

void Settings::setSec(int value , bool enableNotifications)
{
    if (value != wfSecTickMarks)
    {
        wfSecTickMarks = value;
        MYDEBUG << "setSec(" << value << ")" << endl;
        if(enableNotifications == true)
        {
            emit notifySetSec();
        }
        changed++;
    }
}

void Settings::setHz(int value , bool enableNotifications)
{
    if (value != hzTickMarks)
    {
        hzTickMarks = value;
        MYDEBUG << "setHz(" << value << ")" << endl;
        if(enableNotifications == true)
        {
            emit notifySetHz();
        }
        changed++;
    }
}

void Settings::setDbfs( int value, bool enableNotifications )
{
    if (value != dBfsTickMarks)
    {
        dBfsTickMarks = value;
        MYDEBUG << "setDbm(" << value << ")" << endl;
        if(enableNotifications == true)
        {
            emit notifySetDbfs();
        }
        changed++;
    }
}


void Settings::setTooltips( int value )
{
    if (value != tooltips)
    {
        tooltips = value;
        MYDEBUG << "setTooltips(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setPing( int value )
{
    if (value != pingOn)
    {
        pingOn = value;
        MYDEBUG << "setPing(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setPalette( int value )
{
    if (value != palette)
    {
        palette = value;
        MYDEBUG << "setPalette(" << value << ")" << endl;
        changed++;
    }
}


void Settings::setDataLasting( int value )
{
    if (value != dataLasting)
    {
        dataLasting = value;
        MYDEBUG << "setDataLasting(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setNoiseLimit(int value)
{
    if (value != noiseLimit)
    {
        noiseLimit = value;
        MYDEBUG << "setNoiseLimit(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setEraseLogs( int value )
{
    if (value != eraseLogs)
    {
        eraseLogs = value;
        MYDEBUG << "setEraseLogs(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setEraseShots( int value )
{
    if (value != eraseShots)
    {
        eraseShots = value;
        MYDEBUG << "setEraseShots(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setServerPort( int value  )
{
    if (value != serverPort)
    {
        serverPort = value;
        MYDEBUG << "setServerPort(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setServerAddress(const QString& ipAddr)
{
    QHostAddress value(ipAddr);
    if (value != serverAddr)
    {
        serverAddr = value;
        MYDEBUG << "setServerAddress(" << ipAddr << ")" << endl;
        changed++;
    }
}

//site infos
void Settings::setLogo( QString logoPix )
{
    logo = logoPix;
    MYDEBUG << "setLogo()" << endl;
    changed++;
}

void Settings::setStationName( QString value )
{
    stationName = value;
    MYDEBUG << "setStationName(" << value << ")" << endl;
    changed++;
}


void Settings::setContact( QString value )
{
    contact = value;
    MYDEBUG << "setContact(" << value << ")" << endl;
    changed++;
}


void Settings::setRxSetup( QString value )
{
    rxSetup = value;
    MYDEBUG << "setRxSetup(" << value << ")" << endl;
    changed++;
}


void Settings::setNotes( QString value )
{
    notes = value;
    MYDEBUG << "setNotes(" << value << ")" << endl;
    changed++;
}



void Settings::setLat( double value )
{
    if ( fpCmp(value, latitude) != 0 )
    {
        latitude = value;
        MYDEBUG << "setLat(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setLong(double value )
{
    if ( fpCmp(value, longitude) != 0 )
    {
        longitude = value;
        MYDEBUG << "setLong(" << value << ")" << endl;
        changed++;
    }
}

void Settings::setAlt( int value )
{
    if ( value != altitude )
    {
        altitude = value;
        MYDEBUG << "setAlt(" << value << ")" << endl;
        changed++;
    }
}




//-------------------------------------------------------------------
// local settings
//-------------------------------------------------------------------


int Settings::getLastEventNr()
{
    if(lastEventDate != QDate::currentDate())
    {
        return 0;
    }
    return lastEventNr;
}

void Settings::setLastEventNr( int ev )
{
    lastEventDate = QDate::currentDate();
    lastEventNr = ev;
    saveLocal();
}


void Settings::setMainGeometry( QRect newGeom  )
{
    if(newGeom != geom)
    {
        geom = newGeom;
        MYDEBUG << "setGeometry(" << geom << ")" << endl;
        //saveLocal();
    }
}

void Settings::setWfGeometry(QRect newWwGeom, QRect newWwPaneGeom  )
{
    if(newWwGeom != wwGeom ||
       newWwPaneGeom != wwPaneGeom)
    {
        wwGeom = newWwGeom;
        wwPaneGeom = newWwPaneGeom;
        MYDEBUG << "setWfGeometry(" << wwGeom << ")";
        MYDEBUG << " waterfall pane widget geometry: " << wwPaneGeom << ")" << endl;
        emit notifySetWfGeometry();
        //saveLocal();
    }
}




//-------------------------------------------------------------------
// utility methods
//-------------------------------------------------------------------


float Settings::lowestElem( QList<float>* elems )
{
    float lowVal = FLT_MAX;
    //int lowElem = 0;

    for (int i = 0; i < elems->count(); ++i)
    {
        if(elems->at(i) < lowVal)
        {
            lowVal = elems->at(i);
            //lowElem = i;
        }
    }
    return lowVal;
}

float Settings::highestElem( QList<float>* elems )
{
    float highVal = -FLT_MAX;
    //int highElem = 0;

    for (int i = 0; i < elems->count(); ++i)
    {
        if(elems->at(i) > highVal)
        {
            highVal = elems->at(i);
            //highElem = i;
        }
    }
    return highVal;
}


float Settings::funcGen( QMap<float,float>* func, float xInput)
{
    float  ratio = 0;
    float  sideX = 0;
    float  sideY = 0;
    float  yOutput = 0;

    // Ckeck the minimum numbers of items
    if(func->count() < 2)
    {
        // Impossible to interpolate
        yOutput = 0;
        return yOutput;
    }

    QList<float> yValues = func->values();
    QList<float> xValues = func->keys();

    // Check Lo Limit
    float lowest = lowestElem(&xValues);
    if(xInput < lowest)
    {
        //clipping underflows
        yOutput = func->value(lowest);
        return yOutput;
    }

    // Check Hi Limit
    float highest = highestElem(&xValues);
    if(xInput > highest)
    {
        //clipping overflows
        yOutput = func->value(highest);
        return yOutput;
    }

    for(int i = 0; i < func->count(); i++)
    {
        if(xInput >= xValues.at(i) && xInput <= xValues.at(i+1) )
        {
            ratio = (
                    (yValues.at(i+1) - yValues.at(i)) /
                    (xValues.at(i+1) - xValues.at(i))
                );

            sideX = xInput - xValues.at(i);
            sideY = (sideX * ratio);

            yOutput = sideY + yValues.at(i);
            return yOutput;
        }
    }

    /* Error */
    yOutput = 0;
    return yOutput;
}

float Settings::pround(float x, int precision)
{
    float power = powf(10, precision);
    float xpowered = x * power;
    float xlated;
    float frac = modff(xpowered, &xlated);
    bool minus = (xlated < 0) ? true : false;

    if(minus == false && frac > 0.555555F)
    {
        xlated += 1;
    }

    if(minus == true && frac < -0.555555F)
    {
        xlated -= 1;
    }
    xlated = xlated / power;
    return xlated;
}


QString Settings::thresholdBehaviorString()
{
    switch( thresholdsBehavior )
    {
        case TB_ABSOLUTE:
            return tr("absolute");

        case TB_DIFFERENTIAL:
            return tr("differential");

        case TB_AUTOMATIC:
            return tr("automatic");

    }

    return QString("FAILURE");
}


int Settings::fpCmp( double val1, double val2, int decimals )
{
    double multiplier = pow(10, decimals);
    int i1 = static_cast<int>( val1 * multiplier );
    int i2 = static_cast<int>( val2 * multiplier );
    return i2-i1;
}

int Settings::fpCmpf( float val1, float val2, int decimals )
{
    float multiplier = powf(10, decimals);
    int i1 = static_cast<int>( val1 * multiplier );
    int i2 = static_cast<int>( val2 * multiplier );
    return i2-i1;
}

int Settings::fuzzyCompare(float a, float b, float tolerance)
{
    float diff = fabs(a-b);
    float percDiff = fabs((diff * 100) / a);

    if(percDiff > tolerance)
    {
        return (a > b)? 1 : -1;
    }

    return 0;
}

