/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.
    

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html





*******************************************************************************
$Rev:: 367                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-11-23 18:22:34 +01#$: Date of last commit
$Id: settings.h 367 2018-11-23 17:22:34Z gmbertani $
*******************************************************************************/

#ifndef SETTINGS_H
#define SETTINGS_H

#include <QtCore>
#include <QtNetwork>
#include <QtGui>
#include "xqdir.h"
#include "setup.h"

using namespace std;

#define MYDEBUG             qDebug()
#define MYINFO              qInfo()
#define MYWARNING           qWarning()
#define MYCRITICAL          qCritical()
#define MYFATAL(str,arg)    qFatal(str,arg)



///
/// \brief The StatInfos class
/// container for program's statistics
class StatInfos
{
public:

    struct StatControl
    {
        uint loop;                  // progressive loops count
        uint scans;                 // progressive scans count
        uint noDataBufs;            // progressive count of missing FFT data to display
        uint nans;                  // progressive count of NANs
        uint infinite;              // progressive count of infinites
        int totalTime;              // ms spent in last scan
        int deadTime;               // ms spent since quitting wfdBfs til next tick
        int scanDelta;              // ms spent in wfdBfs() waiting a scan to analyze
        int dBfsDelta;              // ms spent in wfdBfs() producing a scan line for waterfall
        int csvDelta;               // ms spent in wfdBfs() when saving statistic data to csv
        int outDelta;               // ms spent in wfdBfs() when dumping gnuplot data to file

    } sc;

    struct StatRadio
    {
        uint nans;                                    /// progressive count of NANs
        uint overflows;                               /// progressive count of overflows
        uint elapsedLoops;                            /// progressive counter of radio thread loops

        int inputSamples;                               /// resampler input samples
        int outputSamples;                              /// resampler output samples
        int FFTbins;                                    /// total FFT  bins (N)
        float binSize;                                  /// size of each bin

        int acqTime;                                  /// minimum ms needed to get a sample from dongle
        int waitScanMs;                               /// ms lost while waiting for a free scan bufs
        int waitIQms;                                 /// ms lost while waiting for IQ samples
        int resampMs;                                 /// ms taken by resampling IQ data
        int FFTms;                                    /// ms spent in FFT

    } sr;

    struct StatIQdevice
    {
        int droppedSamples;                           // number of I/Q pairs missing
        int timeFreeBuf;                              // time spent waiting for a free buffer
    } ss;

    StatInfos()
    {
        memset(&sc, 0, sizeof(sc));
        memset(&sr, 0, sizeof(sr));
        memset(&ss, 0, sizeof(ss));
    }

};


///
/// \brief The Notch class
///
class Notch
{
public:
    Notch()
    {
        freq = 0;
        width = 0;
        begin = 0;
        end = 0;
    }

    bool operator ==(const Notch& n2)
    {
        return (freq == n2.freq &&
                width == n2.width &&
                begin == n2.begin &&
                end == n2.end);
    }

    bool operator !=(const Notch& n2)
    {
        return !(freq == n2.freq &&
                width == n2.width &&
                begin == n2.begin &&
                end == n2.end);
    }

    int freq;           //notch frequency center
    int width;          //notch band width (= notchEnd-notchBeg)
    int begin;          //notch band begin [Hz]
    int end;            //notch band end [Hz]
};




class Waterfall;
class MainWindow;

///
//Initialization states
enum APP_STATE
{
    AST_NONE,           //nothing changed
    AST_INIT,           //app just started
    AST_ERROR,          //temporary error
    AST_CANNOT_OPEN,    //app and device found but could not open it

    AST_OPEN,           //device opening ok, waiting for acquisition start
    AST_STOPPED,        //acquisition stopped
    AST_STARTED,        //acquisition started

    AST_TOTAL_STATES    //final stub
};

///
/// \brief The FFT_WINDOWS enum
///
enum FFT_WINDOWS
{
    FTW_RECTANGLE,
    FTW_HAMMING,
    FTW_BLACKMAN,
    FTW_BLACKMAN_HARRIS,
    FTW_HANN_POISSON,
    FTW_YOUSSEF,
    FTW_BARTLETT,

    FTW_HANN,
    FTW_BLACKMAN_HARRIS_7,
    FTW_FLATTOP,
};

///
/// \brief The PLOT_TYPES enum
///
enum PLOT_TYPES
{
    PT_2D_MAPPED,       //default: 3D data (time/freq/power) mapped on 2D graph (x=freq, y=time, color=power) like screenshots
    PT_3D_REAL,         //3D data (time/freq/power) in perspective (x=freq, y=time, z=power, color=power)
    PT_2D_POWER         //continuous float 2D graph : S&N power vs. time
};

///
/// \brief The E_THRESHOLDS_BEHAVIOR enum
///
enum E_THRESHOLDS_BEHAVIOR
{
    TB_DIFFERENTIAL,
    TB_ABSOLUTE,
    TB_AUTOMATIC
};

///
/// \brief The E_ACQ_MODE enum
///
enum E_ACQ_MODE
{
    AM_INVALID,
    AM_CONTINUOUS,      //continuous S/N acquisition
    AM_PERIODIC,        //periodic acquisition, with dumps taken every X seconds
    AM_AUTOMATIC        //automatic acquisition, with dumps taken when S-N exceeds a threshold
};

///
/// \brief The Settings class
///
class Settings : public QObject
{
    Q_OBJECT

    //hw settings for RTL-SDR devices


    //bool directSampling;
    //bool offsetTuning;

    int changed;               //< if >0 settings changed, save() must be called

    //settings:

    //main window settings:
    QRect       geom;               //< mainwindow geometry
    QString     configName;         //< configuration file name without extension
    QString     configFullPath;     //<  configuration file's absolute path

    //device settings:
    QString     device;         //< RTL-SDR dongle active
    int         sampleRate;     //< dongle sample rate [Hz]
    int         wfBw;           //< portion of SR subsampled for display
    int         tune;           //< RTL-SDR tuning
    int         gain;           //< [dBfs/10]
    int         ppmError;       //< tuner error compensation [ppm/100]
    int         ssBypass;       //< subsampler bypass

    //FFT settings:
    FFT_WINDOWS ftw;            //< FFT window function
    int         interval;       //< waterfall display interval [mS]
    qint64      resolution;     //< FFT resolution [Hz * 10e8]
    int         fftBins;        //< FFT resolution in points

    //output settings:
    int         acqMode;        //< acquisition mode
    int         shots;          //< max number of shots to take before self-stop
    int         recTime;        //< size of each shot [sec] = before+after
    int         after;          //< number of seconds to record after trigger
    int         upThreshold;    //< peaks detection S-N threshold [dBfs / 100] or [%]
    int         dnThreshold;    //< peaks termination S-N threshold [dBfs / 100] or [%]
    int         tRange;         //< peaks detection S-N frequency range
    int         genGNUplot;     //< generate GNUplot output files
    int         genScreenshots; //< generate screenshots
    int         eraseLogs;      //< erase all log files before acquisition restarts
    int         eraseShots;     //< erase all shots before acquisition restarts
    int         plotType;       //< type of gnuplot graph produced
    int         dataLasting;    //< data file lasting in days
    int         avgdScans;      //< Noise filter slope
    int         joinTime;       //< minimum time between consecutive events
    int         minFree;        //< minimum space to keep free on working drive [GB]
    QString     GNUplotPalette; //< the gnuplot palette generated by dbmPalette

    //report settings
    QString     coverFrom;      //<  report coverage from UTC date
    QString     coverTo;        //<  report coverage to UTC date

    int         underdenseBelowMs;  //<  lasting in mS of the longest underdense event
    int         overdenseBelowSec;  //<  lasting in secs of the longest overdense event

    int         colorGraph;         //<  generate ColorGramme-style hourly rates graph
    int         siteInfos;          //<  includes site informations and station setup in report
    int         overdenseShots;     //<  includes overdense events shots in report
    int         underdenseShots;    //<  includes underdense events shots in report
    int         overdensePlots;     //<  includes overdense events plots in report
    int         underdensePlots;    //<  includes underdense events plots in report
    int         shotsDetails;       //<  includes the statistic CSV data with each shot
    int         summary;            //<  includes a summary index to all the events
    int         restart;            //<  restarts acquisition once generated the report
    int         autoReport;         //<  produces automatically the report
    int         carrierFilter;      //<  enables fakes filter for carriers
    int         lightingsFilter;    //<  enables fakes filter for lightings
    int         saturationFilter;   //<  enables fakes filter for saturations

    //preferences:
    int     hzTickMarks;         //<  show tick marks on X axis (Frequency Hz - one tick every 1000 Hz)
    int     dBfsTickMarks;       //<  show tick marks on Y axis (total power dBfs)
    int     wfSecTickMarks;      //<  show tick marks on Y axis (absolute time GMT - one tick every second)
    int     tooltips;            //<  tooltips visible if nonzero
    int     pingOn;              //< ping sound enabling
    int     thresholdsBehavior;  //< thresholds behavior: absolute, differential, automatic
    int     palette;             //< 0=B/W, 1=color
    int     noiseLimit;          //< maximum noise alloweed for automatic capturing
    int     serverPort;          //< dongle server port

    QHostAddress    serverAddr; //< dongle server IP address

    //waterfall window  settings:
    int     wfOffset;           //< center frequency in detailed waterfall/power windows (relative to tune frequency)
    int     wfZoom;             //< portion of bandwidth displayed in waterfall, in tenths of X
    int     brightness;         //< palette settings - brightness percentage 0..100%
    int     contrast;           //< palette settings - contrast percentage 0..100%
    int     powerZoom;          //< power graphs: gain of power axle
    int     powerOffset;        //< power graphs: offset of power plotting
    QList<Notch> notches;      //< notch filters list

    //site infos settings:
    QString  logo;              //< station logo
    double   latitude;          //< station position: latitude [decimal degrees]
    double   longitude;         //< station position: longitude [decimal degrees]
    int      altitude;          //< station position: altitude [m]
    QString  stationName;       //< station name
    QString  contact;           //< station contact / post address
    QString  rxSetup;           //< station hw setup
    QString  notes;             //< further notes about this report

    //local settings
    int     lastEventNr;        //< number of last event catched in automatic mode
    QDate   lastEventDate;      //< date of the last event catched in automatic mode
    QRect   wwGeom;             //< waterfall window geometry
    QRect   wwPaneGeom;         //< waterfall widget pane geometry
    QString     GNUplotPath;    //< path to the gnuplot executable
    QSettings*  localSettings;  //< nonportable machine-dependent configuration

    //temporary storage
    bool    consoleMode;        //< application started in console mode (-c)
    bool    dumbMode;           //< application started in dumb mode (-d)

    int     zoomLo;             //< lowest freq. showed in waterfall - NOT SAVED ON DISK
    int     zoomHi;             //< highest freq. showed in waterfall - NOT SAVED ON DISK
    float  hzPerPixel;         //< Hz per pixel (waterfall resolution) - NOT SAVED ON DISK
    Waterfall*  ww;             //< waterfall window pointer - NOT SAVED ON DISK
    MainWindow* mw;             //< main window pointer - NOT SAVED ON DISK

public:
    explicit Settings(QObject *parent = nullptr);
    ~Settings();

    //save/load configuration settings AND local settings files
    void load(QSettings& ini, bool enableNotifications = true );
    void save( QSettings& ini, bool enableNotifications = true );

    void loadConfig(QString& name , bool enableNotifications = true);

    int changes() const
    {
        return changed;
    }

public slots:
    //getters

    //mainwindow settings
    QString getConfigName()     const { return configName    ; }
    QString getConfigFullPath() const { return configFullPath; }


    //device settings:

    QString getDeviceNoIdx();     //< returns the device name without index

    QString getDevice()     const { return device        ; }
    int getSampleRate()     const { return sampleRate    ; }
    int getTune()           const { return tune          ; }
    int getGain()           const { return gain          ; }
    float getError()        const { return static_cast<float>(ppmError) / 100.0F ; }

    //FFT settings:
    FFT_WINDOWS getWindow() const { return ftw           ; }
    double getResolution()   const
    {
        double res =  static_cast<double>(resolution);
        res /= 10e8 ;
        return res;
    }
    int getFFTbins()        const { return fftBins       ; }
    int getSsBypass()       const { return ssBypass      ; }
    int getBandwidth()      const { return wfBw          ; }


    //output settings
    int getAcqMode()        const { return acqMode       ; }
    int getInterval()       const { return interval      ; }
    int getShots()          const { return shots         ; }
    int getRecTime()        const { return recTime       ; }
    int getAfter()          const { return after         ; }
    float getUpThreshold()  const { return static_cast<float>(upThreshold) / 100.0F   ; }
    float getDnThreshold()  const { return static_cast<float>(dnThreshold) / 100.0F   ; }
    int getTrange()         const { return tRange        ; }
    int getGNUplot()        const { return genGNUplot    ; }
    int getScreenshots()    const { return genScreenshots; }
    float getMinFree()      const { return minFree       ; }
    int getPlotType()       const { return plotType      ; }
    int getAvgdScans()      const { return avgdScans   ; }
    int getJoinTime()       const { return joinTime      ; }
    QString getGNUplotPalette()  const { return GNUplotPalette; }
    QString getGNUplotPath()     const { return GNUplotPath; }

    //report settings
    QDate getCoverFrom() const { return QDate::fromString(coverFrom, Qt::ISODate); }
    QDate getCoverTo()   const { return QDate::fromString(coverTo, Qt::ISODate); }
    int getUnderdenseMs()    const { return underdenseBelowMs; }
    int getOverdenseSec()    const { return overdenseBelowSec; }
    int getColorGraph()      const { return colorGraph; }
    int getSiteInfos()       const { return siteInfos; }
    int getOverdenseShots()  const { return overdenseShots; }
    int getUnderdenseShots() const { return underdenseShots; }
    int getOverdensePlots()  const { return overdensePlots; }
    int getUnderdensePlots() const { return underdensePlots; }
    int getShotsDetails()    const { return shotsDetails; }
    int getSummary()         const { return summary; }
    int getRestart()         const { return restart; }
    int getAutoReport()      const { return autoReport; }
    int getCarrierFilter()   const { return carrierFilter; }
    int getLightingsFilter() const { return lightingsFilter; }
    int getSaturationFilter() const { return saturationFilter; }

    //preferences
    int getHz()             const { return hzTickMarks   ; }
    int getDbfs()           const { return dBfsTickMarks  ; }
    int getSec()            const { return wfSecTickMarks ; }
    int getTooltips()       const { return tooltips     ; }
    int getEraseLogs()      const { return eraseLogs     ; }
    int getEraseShots()     const { return eraseShots    ; }
    int getPing()           const { return pingOn   ; }
    int getPalette()        const { return palette   ; }
    int getDataLasting()    const { return dataLasting   ; }
    int getThresholdsBehavior()  const { return thresholdsBehavior; }
    int getNoiseLimit()          const { return noiseLimit   ; }
    int getServerPort()          const { return serverPort   ; }
    QHostAddress getServerAddress()  const { return serverAddr; }

    //waterfall window
    int getOffset()         const { return wfOffset      ; }
    int getZoom()           const { return wfZoom        ; }
    int getBrightness()     const { return brightness    ; }
    int getContrast()       const { return contrast      ; }
    int getPowerZoom()      const { return powerZoom     ; }
    int getPowerOffset()    const { return powerOffset   ; }
    QList<Notch> getNotches() const { return notches; }

    //site infos settings:
    QString getLogo()       const { return logo ; }
    double getLat()         const { return latitude ; }
    double getLong()        const { return longitude ; }
    int getAlt()         const { return altitude ; }

    QString getStationName()  const { return stationName ; }
    QString getContact()      const { return contact   ; }
    QString getRxSetup()      const { return rxSetup   ; }
    QString getNotes()        const { return notes   ; }


    bool getConsoleMode() const
    {
        return consoleMode;
    }

    bool getDumbMode() const
    {
        return dumbMode;
    }

    //save/load local settings file only
    void loadLocal();
    void saveLocal();

    //local settings
    QDate getLastEventDate() const { return lastEventDate;  }

    void setLastEventNr(int ev );
    int getLastEventNr();

    void setWfGeometry( QRect newWwGeom, QRect newWwPaneGeom);
    QRect getWwGeometry()   const { return wwGeom        ; }
    QRect getWwPaneGeometry()   const { return wwPaneGeom     ; }

    void setMainGeometry( QRect newGeom);
    QRect getMainGeometry() const { return geom          ; }


    //utilities

    ///
    /// \brief fpCmpf compares two floats up to the
    /// required decimals
    /// \param val1
    /// \param val2
    /// \param decimals
    /// \return 0: val1==val2, <0: val1<val2, >0: vali1>val2
    ///
    int fpCmpf( float val1, float val2, int decimals=3 );

    ///
    /// \brief fpCmp compares two doubles up to the
    /// required decimals
    /// \param val1
    /// \param val2
    /// \param decimals
    /// \return
    ///
    int fpCmp( double val1, double val2, int decimals=5 );



    ///
    /// \brief funcGen
    /// \param func
    /// \param xInput
    /// \return
    /// table lookup function generator
    ///
    float funcGen( QMap<float,float>* func, float xInput);


    ///
    /// \brief lowestElem
    /// \param elems
    /// \return
    /// returns the lowest element in a list
    ///
    float lowestElem( QList<float>* elems );


    ///
    /// \brief highestElem
    /// \param elems
    /// \return
    /// returns the highest element in a list
    ///
    float highestElem( QList<float>* elems );


    ///
    /// \brief thresholdBehaviorString
    /// \return
    ///
    QString thresholdBehaviorString();    ///absolute, differential, automatic

    ///
    /// \brief clearChanged
    /// resets the settings changed flag
    void clearChanged()
    {
        changed = false;
    }



    ///
    /// \brief getWaterfallWindow
    /// \return
    /// The class Settings holds the pointer of windows (no disk storage)
    Waterfall* getWaterfallWindow() const
    {
        return ww;
    }

    ///
    /// \brief setWaterfallWindow
    /// \param w
    ///
    void setWaterfallWindow(Waterfall* w)
    {
        MY_ASSERT( nullptr !=  w);
        ww = w;
    }

    //The class Settings holds the pointer of windows (no disk storage)
    MainWindow* getMainWindow() const
    {
        return mw;
    }

    ///
    /// \brief setMainWindow
    /// \param w
    ///
    void setMainWindow(MainWindow* w)
    {
        MY_ASSERT( nullptr !=  w);
        mw = w;
    }


    ///
    /// \brief pround
    /// \param x
    /// \param precision
    /// \return
    /// generic rounding function
    ///
    float pround(float x, int precision = 1);

    ///
    /// \brief fuzzyCompare
    /// \param a
    /// \param b
    /// \param tolerance (percentage)
    /// \return 1 if a>b,
    ///         -1 if a<b,
    ///         0 if a-b is within +/-tolerance % of a,
    ///
    int fuzzyCompare(float a, float b, float tolerance);


    //
    ///
    /// \brief iround
    /// \param x
    /// \return
    /// generic rounding function
    ///
    inline int iround(float x)
    {
        return static_cast<int>( pround(x,0) );
    }


    void setConsoleMode(bool c)
    {
        consoleMode = c;
    }

    void setDumbMode(bool d)
    {
        dumbMode = d;
    }

    ///
    /// \brief setHpp
    /// \param hpp
    /// Hz per pixel (waterfall resolution)
    void setHpp(float hpp)
    {
        hzPerPixel = hpp;
    }


    ///
    /// \brief getHpp
    /// \return
    ///
    float getHpp() const
    {
        return hzPerPixel;
    }

    ///
    /// \brief getFracPpmErr
    /// \return
    ///
    void getZoomedRange( int& lo, int& hi ) const
    {
        lo = zoomLo; hi = zoomHi;
    }

    //setters for:

    //mainwindow settings
    void setConfigName(XQDir wd, QString& name);
    void setZoomedRange( int lo, int hi );

    //device settings
    void setDevice          ( const QString& dev, bool enableNotifications = true );
    void setSampleRate      ( int value, bool enableNotifications = true );
    void setTune            ( int value, bool enableNotifications = true );
    void setGain            ( int value );
    void setError           (double value );
    void setSsBypass        ( int value );

    //FFT settings:
    void setWindow          ( int value);
    void setResBins         ( int FFTbins, bool enableNotifications = true );
    void updateResolution  ( bool enableNotifications = true  );
    //FFT resolution pushbuttons:
    void incResolution(bool enableNotifications = true);
    void decResolution(bool enableNotifications = true);

    //output settings
    void setAcqMode         ( int value);
    void setInterval        ( int value);
    void setShots           ( int value);
    void setRecTime         ( int value);
    void setAfter           ( int value);
    void setThresholdsBehavior( int value );
    void setUpThreshold     (double value, bool enableNotifications = true);
    void setDnThreshold     (double value, bool enableNotifications = true);
    void setTrange          ( int value);
    void setGNUplot         ( int value);
    void setScreenshots     ( int value);
    void setMinFree         (int value);
    void setEraseLogs       ( int value);
    void setEraseShots      ( int value);
    void setPlotType        ( int value);
    void setDataLasting     ( int value);
    void setAvgdScans       ( int value);
    void setJoinTime        ( int value);
    void setGNUplotPalette  ( const QString& gpp);
    void setGNUplotPath     ( const QString& path);

    //report settings
    void setCoverFrom       ( QDate cf);
    void setCoverTo         ( QDate ct);
    void setUnderdenseMs    ( int value);
    void setOverdenseSec    ( int value);
    void setColorGraph      ( int value);
    void setSiteInfos       ( int value);
    void setOverdenseShots  ( int value);
    void setUnderdenseShots ( int value);
    void setOverdensePlots  ( int value);
    void setUnderdensePlots ( int value);
    void setShotsDetails    ( int value);
    void setSummary         ( int value);
    void setRestart         ( int value);
    void setAutoReport      ( int value );
    void setCarrierFilter   ( int value );
    void setLightingsFilter ( int value );
    void setSaturationFilter ( int value );

    //waterfall window
    void setOffset          ( int value);
    void setZoom            ( int value);
    void setBandwidth       ( int value);
    void setBrightness      ( int value);
    void setContrast        ( int value);
    void setPowerZoom       ( int value);
    void setPowerOffset     ( int value);
    void setNotches         ( QList<Notch>& ntList);


    //preferences
    void setHz              ( int value, bool enableNotifications = true);
    void setDbfs            ( int value, bool enableNotifications = true);
    void setSec             ( int value, bool enableNotifications = true);
    void setTooltips        ( int value);
    void setPing            ( int value );
    void setPalette         ( int value );
    void setNoiseLimit      ( int value  );
    void setServerPort      ( int value  );
    void setServerAddress   ( const QString& ipAddr  );

    //siteinfo settings:
    void setLogo            (QString logoPix);
    void setLat             (double value);
    void setLong            (double value);
    void setAlt             (int value);
    void setStationName     ( QString value);
    void setContact         ( QString value);
    void setRxSetup         ( QString value);
    void setNotes           ( QString value);


signals:

    /*
     * notification signals are required only for controls having an immediate
     * feedback even when acquisition is running
     */

    void notifyOpened           ();       ///configuration loaded from disk, GUI needs refresh.
    void notifySavedAs          ();
    void notifySetWfGeometry    ();
    void notifySetOffset        ();
    void notifySetZoom          ();
    void notifySetBandwidth     ();

    //device settings:
    void notifySetSampleRate    ();
    void notifySetTune          ();
    void notifySetError         ();
    void notifySetDevice        ();
    void notifySetGain          ();
    void notifySsBypass         ();

    //FFT settings
    void notifySetResolution    ();
    void notifySetWindow        ();

    //output settings
    void notifySetAcqMode       ();
    void notifySetInterval      ();
    void notifySetUpThreshold   ();
    void notifySetDnThreshold   ();
    void notifySetTrange        ();
    void notifySetAfter         ();
    void notifySetRecTime       ();
    void notifySetShots         ();
    void notifySetThresholdsBehavior();

    //preferences settings
    void notifySetHz            ();
    void notifySetDbfs          ();
    void notifySetSec           ();

};

#endif // SETTINGS_H
