/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.


    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 268                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-07-19 07:45:48 +02#$: Date of last commit
$Id: syncrx.cpp 268 2018-07-19 05:45:48Z gmbertani $
*******************************************************************************/

#include "setup.h"
#include "rtl-sdr.h"
#include "syncrx.h"

bool SyncRx::lookForDongleStatic( QList<DeviceMap>* devices )
{
    MYINFO << __func__;

    MY_ASSERT( nullptr !=  devices)

    QString buf;
    uint i;
    uint deviceCount = 0;
    char vendor[256]={"vendor"}, product[256]={"product"}, serial[256]={"serial"};
    int retVal = 0;
    DeviceMap dm;

    foreach(dm, *devices)
    {
        if(dm.name.contains("RTL", Qt::CaseInsensitive))
        {
            deviceCount++;
        }
    }

    if(deviceCount == 0)
    {
        //if the dongles are already known
        //avoids to rescan the usb bus

        deviceCount = rtlsdr_get_device_count();
        if (deviceCount > 0)
        {
            //RTL-SDR dongle found
            MYINFO << "Found " << deviceCount << "device(s):";


            for (i = 0; i < deviceCount; i++)
            {

                retVal = rtlsdr_get_device_usb_strings(i, vendor, product, serial);
                if(retVal != 0)
                {
                    MYINFO << "rtlsdr_get_device_usb_strings(" << i << ") returned " << retVal;
                    continue;
                }

                buf = QString(product);
                if(buf.contains("RTL", Qt::CaseInsensitive))
                {
                    buf=QString("%1: %2 %3  SN: %4")
                            .arg(i)
                            .arg(vendor)
                            .arg(product)
                            .arg(serial);

                    dm.name = buf;
                    dm.index = static_cast<int>(i);
                    devices->append(dm);
                    MYINFO << buf;
                }
                else
                {
                    buf=QString("%1: %2 %3  SN: %4")
                            .arg(i)
                            .arg(vendor)
                            .arg(product)
                            .arg(serial);

                    MYINFO << "Device (" << buf << ") is not a RTL-SDR, ignoring it";
                }

            }
        }
    }
    return (devices->count() > 0);
}



SyncRx::SyncRx(Settings* appSettings, Pool<IQbuf*>* bPool, QThread *parentThread, int devIndex) :
    IQdevice(appSettings, bPool, parentThread)
{
    setObjectName("SyncRx thread");
    owner = parentThread;

    MYINFO << "parentThread=" << parentThread << " parent()=" << parent();

    ready = false;
    dev = nullptr;
    sentBuffersCount = 0;
    sysFaultCount = 0;
    droppedSamples = 0;
    timeFreeBuf = 0;
    iqBufRoundSize = 0;

    cfgChangesCount = as->changes();
    serverSocket = nullptr;
    clients.clear();

    //experimental
    iqRecordOn = false;
    int retVal = rtlsdr_open( &dev, static_cast<uint32_t>( devIndex ) );
    if (retVal < 0)
    {
       MYCRITICAL << "ERROR: Failed opening rtlsdr device " << as->getDevice() << " rtlsdr_open() returned " << retVal;
       emit status( tr("Failed opening rtlsdr device."), AST_CANNOT_OPEN );
       return;
    }
    MYINFO << "RTL-SDR Dongle opened";
    retVal = rtlsdr_reset_buffer(dev);
    if (retVal < 0)
    {
        MYCRITICAL << "Failed reset of rtlsdr buffers, rtlsdr_reset_buffer() returned: " << retVal;
        emit status( tr("Failed reset of rtlsdr buffers."), AST_ERROR );
        rtlsdr_close(dev);
        return;
    }

    retVal = rtlsdr_set_agc_mode(dev, 0);
    if (retVal != 0)
    {
        MYCRITICAL << "ERROR: Failed turning off AGC, rtlsdr_set_agc_mode() returned: " << retVal;
        emit status( tr("Failed turning AGC off."), AST_ERROR );
        rtlsdr_close(dev);
        return;
    }


    //refill gains list
    int count = rtlsdr_get_tuner_gains(dev, nullptr);
    if (count > 0)
    {
        int* garr = new int[count];
        MY_ASSERT( nullptr !=  garr)

        gains.append("AUTO");

        rtlsdr_get_tuner_gains(dev, garr);
        for (int i = 0; i < count; i++)
        {
            //gains are in tenth of dB in order to store them as ints
            //but we need to display dB
            QString gs;
            gs = gs.asprintf("%.2f", static_cast<double>( garr[i] ) / 10.0 );
            gains.append( gs );
            MYDEBUG << "adding gain " << gs << " to ui->cbGain";
        }

        delete[] garr;
    }
    else
    {
        MYCRITICAL << "Failed getting gains from device.";
        emit status( tr("Failed getting gains from device.") );
        rtlsdr_close(dev);
        return;
    }


    //sample rates list
    rates.append("226000");
    rates.append("250000");
    rates.append("300000");
    rates.append("1000000");
    rates.append("1200000");
    rates.append("1500000");
    rates.append("1800000");
    rates.append("2000000");
    rates.append("2200000");
    rates.append("2500000");
    rates.append("2800000");
    rates.append("3000000");
    rates.append("3200000");

    //The device is now partially initialized.
    //The device parameters must be set first
    //in order to allow acquisition thread starting


}

SyncRx::~SyncRx()
{
    //freeing clients references
    clients.clear();

    //destroying server
    if(serverSocket != nullptr)
    {
        serverSocket->deleteLater();
        //serverSocket = nullptr;
    }

    if(dev != nullptr)
    {
        rtlsdr_close(dev);
        MYINFO << "RTL-SDR Dongle closed";
        dev = nullptr;
    }
}


void SyncRx::run()
{
    //dev must be opened before starting this thread
    MY_ASSERT(dev != nullptr)

    MYINFO << "Dongle acquisition thread started";

    int nRead = 0;
    int bufIdx = -1;
    int iqBufSize = 0;
    int retVal = 0;
    IQbuf* buf = nullptr;


    timeAcq.start();

    //recording stuff
    //the purpose is to record iq buffers in a file with real meteor echies inside
    //in order to be replayed later by funcgen without dongle for easier parameters tuning
    XQDir recDir( XQDir::temp() );
    QString recName = recDir.absoluteFilePath("echoesIQsamples.dat");
    QFile rec( recName );
    QDataStream ds(&rec);
    uint count  = 0;
    const uint magic = 0xAA55F00F;

    if(iqRecordOn)
    {
        if (rec.open(QIODevice::WriteOnly) == false)
        {
            MYWARNING << "IQ recording required but unable to open file " << recName;
            iqRecordOn = false;
        }
        else
        {
            MYINFO << "IQ recording active on file " << recName;
        }
    }

    //this instance is a dongle server, so it can read the samples directly from
    //a RTL-SDR dongle and forward them - besides the instance threads,
    //to other instances via UDP
    if(serverSocket == nullptr)
    {
        startServer();
    }

    if(as->getDumbMode())
    {
        MYINFO << "Dumb mode selected, this instance runs as pure server, no events catching";
    }

    for(;;)
    {
        if( isFinished() == true || isInterruptionRequested() == true )
        {
            MYINFO << "SyncRx thread stopped";
            if(iqRecordOn)
            {
                rec.close();
            }

            //stopAllClients();

            if(iqRecordOn)
            {
                rec.close();
            }
            return;
        }

        if(serverSocket->isValid())
        {
            readPendingClientDatagrams();
        }

        //TODO: only changes in device+fft tabs should be notified
        if(as->changes() != cfgChangesCount ||
           iqBufSize == 0 )
        {
            cfgChangesCount = as->changes();

            //resolution = BW / FFTbins
            //iqBufSize = SR x (FFTbins / BW) = SR / resolution

            iqBufSize = static_cast<int>(as->getSampleRate() / as->getResolution());

            //the buffer size passed to rtlsdr_read_sync() must be multiple of 512
            //to avoid overflow errors.
            //The samples exceeding iqBufSize will be stored anyway because the physical
            //size of each buffer is 3,2 Msamples but the exceeding part will be ignored by
            //following stages.

            iqBufRoundSize = iqBufSize;
            if((iqBufSize % 512) > 0)
            {
                iqBufRoundSize = iqBufSize + (512 - (iqBufSize % 512));
            }
            shareDeviceConfig();
        }

        bufIdx = pb->take();
        if(bufIdx != -1)
        {
            timeFreeBuf = static_cast<int>( timeAcq.restart() );

            buf = pb->getElem(bufIdx);
            if(buf != nullptr)
            {

                if( buf->doublets() != iqBufSize || iqBufRoundSize == 0 )
                {
                    buf->setBufSize( iqBufSize );

                    //the buffer size passed to rtlsdr_read_sync() must be multiple of 512
                    //to avoid overflow errors.
                    //The samples exceeding iqBufSize will be stored because the physical
                    //size of each buffer is 3,2 Msamples but will be ignored by
                    //following stages.
                    iqBufRoundSize = iqBufSize;
                    if((iqBufSize % 512) > 0)
                    {
                        iqBufRoundSize = iqBufSize + (512 - (iqBufSize % 512));
                        buf->setBufSize( iqBufRoundSize );
                    }

                }

                tsGen = QDateTime::currentDateTimeUtc();
                QString s = tsGen.toString( "dd/MM/yyyy;hh:mm:ss.zzz");
                buf->setTimeStamp( s );
                if(dev != nullptr)
                {
                    //dongle server
                    if((iqBufSize == 0) || (iqBufRoundSize == 0))
                    {
                        MYCRITICAL << "iqBufSize = " << iqBufSize << " to = " << iqBufRoundSize;
                    }
                    retVal = rtlsdr_read_sync(dev, buf->data(), iqBufRoundSize, &nRead);

                    if(retVal != 0)
                    {
                        sysFaultCount++;
                        MYWARNING << "SyncRx:rtlsdr_read_sync() returned " << retVal << " progressive: " << sysFaultCount;
                        pb->release(bufIdx);
                    }
                    else
                    {
                        MYDEBUG << "SyncRx:rtlsdr_read_sync() returned a " << buf->bSize() << " bytes long data buffer";
                        sysFaultCount = 0;
                        if (nRead > 0)
                        {
                            droppedSamples = buf->bSize() - nRead;
                        }

                        if(serverSocket != nullptr && clients.count() > 0)
                        {
                            shareSamples(buf);
                        }


                        if(iqRecordOn)
                        {
                            //for debug only: recording IQ buffers to a file
                            //each record starts with magic and progressive counter
                            //then the complete buffer follows
                            ds << magic;
                            ds << count;
                            ds.writeRawData( reinterpret_cast<char*>(buf->data()), buf->bSize() );
                            ds << buf->data();
                            count++;
                        }

                        if(as->getDumbMode())
                        {
                            //in dumb mode, the samples are sent to external clients only
                            //no captures are performed locally
                            pb->release(bufIdx);
                        }
                        else
                        {
                            //forwards the samples to the Radio class for events processing
                             pb->forward(bufIdx);
                        }

                    }
                }

            }
            timeAcq.restart();
        }

        if(sysFaultCount >= MAX_SYS_ERRORS)
        {
            emit fault(retVal);
            break;
        }

        QThread::yieldCurrentThread();
    }

    if(serverSocket == nullptr)
    {
        MYINFO << "Server socket closed";
        delete serverSocket;
        serverSocket = nullptr;
    }
}


bool SyncRx::isDeviceOpen()
{
    if(dev != nullptr)
    {
        return true;
    }
    return false;
}


bool SyncRx::getStats( StatInfos* stats )
{
    stats->ss.droppedSamples = droppedSamples;
    stats->ss.timeFreeBuf = timeFreeBuf;
    return true;
}


void SyncRx::slotParamsChange()
{
    //it can happen exiting from test patterns:
    if(as->getTune() < MIN_TUNE_HZ)
    {
        tune = DEFAULT_TUNE;
        as->setTune(tune);
    }

    bw = as->getBandwidth();
    lowBound   =  as->getTune() - (bw / 2);
    highBound  =  as->getTune() + (bw / 2);

    MYINFO << "Subsampled range (band)" << lowBound << " to " << highBound << " Hz";

    emit status( tr("Ready.") );

    //ending of parameter change sequence
    ready = true;
}

bool SyncRx::slotSetTune()
{
    int fails = 0;
    int retVal = 0;

    ready = false;

    int newTune = as->getTune();

    if(isRunning() == true)
    {
        if(pb != nullptr)
        {
            //flushes the unprocessed iq buffers
            pb->flush();
        }
    }

    int freq = static_cast<int>( rtlsdr_get_center_freq( dev ) );
    tune = newTune;

    if(tune < MIN_TUNE_HZ)
    {
        MYWARNING << "Stored tuning frequency " << tune << " is too low, taking the default value 144MHz";
        tune = DEFAULT_TUNE;
    }

    if (tune != freq)
    {
        MYINFO << "Actual tuning: " << freq << " Hz, changing to " << tune << " Hz";

        uchar dump[BUFFER_DUMP];
        int nRead;

        rtlsdr_set_offset_tuning(dev, 0);
        int ot = rtlsdr_get_offset_tuning(dev);
        MYINFO << "Offset tuning status: " << ot;

        retVal = rtlsdr_set_center_freq(dev, static_cast<uint32_t>( tune ) );
        if(retVal >= 0)
        {
            usleep(5000);
            retVal = rtlsdr_read_sync(dev, &dump, BUFFER_DUMP, &nRead);
            if (nRead != BUFFER_DUMP || retVal < 0)
            {
                MYCRITICAL << "Failed retuning to " << tune << "Hz"
                    << "flushing buffer unsuccessful, rtlsdr_read_sync() returned: " << retVal;
                emit status( tr("Failed retuning - flushing buffer unsuccessful."), AST_ERROR );
                fails++;
            }

            //Tuning OK
            as->setTune(tune);
            slotParamsChange();
        }
        else
        {
            MYCRITICAL << "Failed retuning to " << tune << "Hz, rtlsdr_set_center_freq() returned: " << retVal;
            emit status( tr("Failed retuning."), AST_ERROR );
            fails++;
        }

    }
    else if (freq <= 0)
    {
        MYCRITICAL << "Failed getting tuner frequency, freq: " << freq;
        emit status( tr("Failed getting tuner frequency."), AST_ERROR );
        fails++;
    }

    return (fails == 0);
}

bool SyncRx::slotSetSampleRate()
{   
    int fails = 0;
    int retVal = 0;
    ready = false;

    if(dev == nullptr)
    {
        MYINFO << "device not opened, doing nothing" ;
        return false;
    }

    QString newSr;
    newSr.setNum( as->getSampleRate() );

    if( !rates.contains( newSr ))
    {
        MYWARNING << "Sample rate not supported:" << newSr << " taking the default value";
        newSr.setNum( DEFAULT_RATE );
    }

    sr = static_cast<int>( rtlsdr_get_sample_rate(dev) );
    if( sr != newSr.toInt() )
    {
        sr = newSr.toInt();
        retVal = rtlsdr_set_sample_rate(dev, static_cast<uint32_t>( sr ));
        if(retVal >= 0)
        {
            as->setSampleRate(sr);
            MYINFO << "Sample rate set to " << sr << " Hz";
            bw = sr;

            slotParamsChange();
        }
        else
        {
            MYCRITICAL << "Failed setting sample rate to " << newSr << " Hz, rtlsdr_set_sample_rate() returned: " << retVal;
            emit status( tr("Failed setting sample rate."), AST_ERROR );
            fails++;
        }
    }

    return (retVal >= 0);
}

bool SyncRx::slotSetGain()
{
    int fails = 0;
    int retVal = 0;
    ready = false;

    if(dev == nullptr)
    {
        MYINFO << "device not opened, doing nothing" ;
        return false;
    }

    QString sGain;
    float fGain = static_cast<float>(as->getGain()) / 10;
    sGain.setNum(fGain, 'f', 2);
    if( as->getGain() != AUTO_GAIN && !gains.contains( sGain ) )
    {
        bool ok;
        MYWARNING << "Gain level not supported:" << sGain;
        sGain = gains.last();
        MYINFO << " taking as default value the highest supported by this device: " << sGain;
        fGain = sGain.toFloat(&ok);
        MY_ASSERT(ok);
        gain = static_cast<int>(fGain * 10);
        as->setGain(gain);
    }

    //set gain mode
    if( fGain < 0.0F )
    {
        //negative values set auto gain mode
        retVal = rtlsdr_set_tuner_gain_mode(dev, 0);
        if(retVal == 0)
        {
            gain = -10;
            as->setGain(gain);
            MYINFO << "Automatic gain set" ;
        }
        else
        {
            MYCRITICAL << "Failed setting automatic gain, error: " << retVal ;
            emit status( tr("Failed setting automatic gain."), AST_ERROR );
            fails++;
        }
    }
    else
    {
        //set manual gain
        retVal = rtlsdr_set_tuner_gain_mode(dev, 1);
        if(retVal == 0)
        {
            MYINFO << "Manual gain setting" ;

            retVal = rtlsdr_set_tuner_gain(dev, as->getGain());   //set gain in tenths of dB
            if (retVal == 0)
            {
                MYINFO << "Setting tuner gain to " << sGain << " dB (" << as->getGain() << " [dB/10] )" ;
            }

            int gainCheck = rtlsdr_get_tuner_gain(dev);   //set gain in tenths of dB

            if( retVal != 0 || gainCheck != as->getGain()  )
            {
                MYCRITICAL << "Failed setting tuner gain " << sGain << " dB, rtlsdr_set_tuner_gain() returned: " << retVal ;
                emit status( tr("Failed setting tuner gain."), AST_ERROR );
                fails++;
            }
            else
            {
                MYINFO << "Tuner gain set to " << sGain << " dB" ;
            }

        }
        else
        {
            MYCRITICAL << "Failed setting manual gain, rtlsdr_set_tuner_gain_mode() returned: " << retVal ;
            emit status( tr("Failed setting manual gain."), AST_ERROR );
            fails++;
        }
    }


    if(fails > 0)
    {
        //sets the gain to default in case of errors
        gain = DEFAULT_GAIN;
        as->setGain(gain);
    }
    slotParamsChange();

    return (fails == 0);
}

bool SyncRx::slotSetError()
{
    int fails = 0;
    int retVal = 0;
    ready = false;

    float newErr = as->getError();

    //ppm error
    retVal = rtlsdr_get_freq_correction(dev);
    float err = newErr;
    float intErr = 0.0;

    //the integer part of the ppm error is used to set the frequency correction into the dongle
    std::modf(err, &intErr);
    int dtoi = static_cast<int>(intErr);

    if ( retVal != dtoi )
    {
        retVal = rtlsdr_set_freq_correction(dev, dtoi);
        if (retVal >= 0)
        {
            MYINFO << "Tuning error set to " << intErr << " ppm";
            as->setError(newErr);
            slotParamsChange();
        }
        else
        {
            MYCRITICAL << "Failed setting tuning error to " << intErr << " ppm, rtlsdr_set_freq_correction() returned:" << retVal;
            emit status( tr("Failed setting tuning error ppm."), AST_ERROR );
            fails++;
        }
    }
    return (fails == 0);
}


bool SyncRx::slotSetBandwidth()
{
    ready = false;
    int newBw = as->getBandwidth();

    if( bw != newBw && bw <= sr )
    {
        bw = newBw;
        as->setBandwidth(newBw);

        slotParamsChange();
    }
    return true;
}



bool SyncRx::startServer()
{

    serverSocket = new QUdpSocket();
    bool result = serverSocket->bind(static_cast<quint16>(as->getServerPort()));
    MY_ASSERT( nullptr !=  serverSocket);
    if(result)
    {
        MYINFO << "### Ready to accept UDP connections on port " << as->getServerPort() << " state:" << serverSocket->state();
    }
    else
    {
        MYCRITICAL << "### Failed listening for UDP connections on port " << as->getServerPort() << ":" << serverSocket->errorString();
    }
    return result;
}

void SyncRx::readPendingClientDatagrams()
{
    qint64 result = 0;
    //MYDEBUG << "### Socket state: " << serverSocket->state();
    if(serverSocket->state() == QAbstractSocket::BoundState)
    {
        //MYDEBUG << "### Checking for clients requests...";
        while (serverSocket->hasPendingDatagrams())
        {
            qint64 size = serverSocket->pendingDatagramSize();
            if(size <= 0)
            {
                MYWARNING << "pendingDatagramSize <= 0 (QTBUG-49301) socket badly closed on client - "
                    << serverSocket->errorString();
                requestInterruption();
                break;
            }

            ClientDescriptor* candidateClient = new ClientDescriptor;
            QByteArray datagram(static_cast<int>(size),0);
            try
            {
                result = serverSocket->readDatagram(datagram.data(),
                                     static_cast<qint64>(datagram.size()),
                                     &candidateClient->address, &candidateClient->port);

                if(result < 0)
                {
                    MYCRITICAL << "### Failed reading " << datagram.size() << " bytes : " << serverSocket->errorString();
                    return;
                }

                ClientDescriptor* c = nullptr;

                //Expected data from clients are always in form
                //<client name>.<payload> for instance "perseids.START"
                //Unknown clients must annunce themselves first by sending
                //<client name>.<client name> for instance "perseids.perseids"
                QStringList data = QString( datagram.simplified() ).split(".");
                QString clientName = data.at(0);
                QString clientData = data.at(1);

                //checks if the client is already in distribution list
                bool alreadyKnown = false;
                foreach(c, clients)
                {
                    if(c->name == clientName)
                    {
                        MYINFO << "### Client " << c->name << " sent something";
                        c->address = candidateClient->address;
                        c->port = candidateClient->port;
                        alreadyKnown = true;
                    }
                }

                if(alreadyKnown == true)
                {
                    if(clientData.contains("QUIT"))
                    {
                        //the client requires disconnection
                        MYINFO << "### Releasing client " << c->name;
                        clients.removeAll(c);
                    }
                    else if(clientData.contains("START"))
                    {
                        MYINFO << "### Start sending samples to client " << c->name;
                        c->active = true;
                    }
                    else if(clientData.contains("STOP"))
                    {
                        MYINFO << "### Stop sending samples to client " << c->name;
                        c->active = false;
                    }

                    delete candidateClient;
                    candidateClient = nullptr;
                }

                if(clientData == clientName)//datagram contains a new client name
                {
                    //new client to be served
                    //the scan flux is closed by default
                    if(alreadyKnown == false)
                    {
                        candidateClient->name = clientName;
                        candidateClient->active = false;
                        clients.append(candidateClient);
                        MYINFO << "### Accepting UDP connection from new client " << candidateClient->name
                               << " at " << candidateClient->address.toString() << " : " << candidateClient->port;

                        //sends always the client the device configuration back to client
                        sendDeviceConfig(candidateClient);
                    }
                    else
                    {
                        //sends always the client the device configuration back to client
                        sendDeviceConfig(c);
                    }

                }


            }
            catch ( QException& e )
            {
                MYCRITICAL << "Exception " << e.what();
            }
        }
    }
}

void SyncRx::shareDeviceConfig()
{
    MYINFO << __func__;
    ClientDescriptor* client = nullptr;
    try
    {
        foreach(client, clients)
        {
            sendDeviceConfig(client);
        }
    }
    catch ( QException& e )
    {
        MYCRITICAL << "Exception " << e.what();
    }
}

void SyncRx::shareSamples(IQbuf* buf)
{
    MYDEBUG << __func__;
    sentBuffersCount++;
    ClientDescriptor* client = nullptr;
    QByteArray baIQbuf; //serialized IQbuf to be sent via UDP socket, to be broken down in datagrams, 512 bytes max each.
    QByteArray baDgram; //single datagram
    QByteArray baHeader; //serialized IQbuf header

    //preparing header
    QDataStream hdr(&baHeader, QIODevice::WriteOnly);
    qint16 syncWord = SYNC_IQBUF;

    hdr << syncWord;
    hdr << buf->myUID();
    hdr << buf->timeStamp();

    //preparing serialized IQbuf
    QDataStream ds(&baIQbuf, QIODevice::WriteOnly);
    ds.writeRawData( reinterpret_cast<char*>( buf->data() ), static_cast<uint32_t>( buf->bSize() ));

    try
    {

        foreach(client, clients)
        {
            if(client->active == true)
            {
                //Sending header
                qint64 sent = serverSocket->writeDatagram(baHeader, client->address, client->port);
                if(sent != baHeader.size())
                {
                    MYCRITICAL << "Sending header to " << client->name << " (" << client->address.toString() << "/" << client->port << ")" << endl
                               << " not complete, sent " << sent << " bytes instead of " << baHeader.size() << " bytes "
                               << endl  << " error: " << serverSocket->errorString();
                    break;
                }

                //sending data samples
                int byteIndex = 0, sliceIndex = 0;
                int sliceSize = (baIQbuf.size() > MAX_DGRAM_SIZE)? MAX_DGRAM_SIZE : baIQbuf.size();
                MYDEBUG << "baIQbuf.size=" << baIQbuf.size() << " sliceSize=" << sliceSize;
                while(byteIndex < baIQbuf.size())
                {
                    baDgram = baIQbuf.mid(byteIndex, sliceSize);


                    sent = serverSocket->writeDatagram(baDgram, client->address, client->port);
                    if(sent < 0)
                    {
                        MYWARNING << "Sending to " << client->name << " (" << client->address.toString() << "/" << client->port << ")" << endl
                                   << " not complete, sent " << sent << " bytes instead of " << baDgram.size()
                                   << endl  << " error: " << serverSocket->errorString();

                        MYINFO << "Resending slice#" << sliceIndex << ", byteIndex = " << byteIndex;
                    }
                    else
                    {
                        //prepare to send the next slice
                        MYDEBUG << "Sent slice#" << sliceIndex;
                        byteIndex += baDgram.size();
                        sliceIndex++;
                    }

                }
            }
        }

        MYDEBUG << "Sent buffer #" << sentBuffersCount;
    }
    catch ( QException& e )
    {
        MYCRITICAL << "Exception " << e.what();
    }
}


void SyncRx::stopAllClients()
{
    MYINFO << __func__;
    QByteArray baStop;
    baStop.append("STOP");

    ClientDescriptor* client = nullptr;
    try
    {
        //notifies all the connected clients that the server is terminating
        foreach(client, clients)
        {
            serverSocket->writeDatagram(baStop,  client->address, client->port);
            MYINFO << "Sending STOP to client " << client->address.toString() << ":" << client->port;
            client->active = false;
        }
    }
    catch ( QException& e )
    {
        MYCRITICAL << "Exception " << e.what();
    }
}


void SyncRx::sendDeviceConfig(ClientDescriptor* client)
{
    qint64 sent = 0;

    QByteArray baMyConfig;
    QDataStream cfg(&baMyConfig, QIODevice::WriteOnly);
    qint16 syncWord = SYNC_SRVCONF;
    cfg << syncWord;
    cfg << as->getDevice();
    cfg << as->getSampleRate();
    cfg << as->getGain();
    cfg << as->getError();
    cfg << as->getTune();
    cfg << iqBufRoundSize;
    cfg << getPackedGains();
    cfg << getPackedRates();

    MYINFO << "gains: " << gains;
    MYINFO << "rates: " << rates;
    MYINFO << "sizeof CFG: " << baMyConfig.size();


    MYINFO << "### Sending device config to " << client->address.toString() << ":" << client->port;

    sent = serverSocket->writeDatagram(baMyConfig,  client->address, client->port);

    if(sent != baMyConfig.size())
    {
        MYCRITICAL << "### Sending to " << client->address.toString() << " not complete, sent (=" << sent
                   << ") bytes < sizeof(dc)(=" << baMyConfig.size() << ") bytes";
    }
}
