/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.
    

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 149                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-03-14 00:08:31 +01#$: Date of last commit
$Id: syncrx.h 149 2018-03-13 23:08:31Z gmbertani $
*******************************************************************************/

#ifndef SYNCRX_H
#define SYNCRX_H

#include <QtCore>
#include <QtNetwork>
#include <cmath>
#include "rtl-sdr.h"
#include "iqdevice.h"

///
/// \brief The ClientDescriptor struct
///
struct ClientDescriptor
{
    QString         name;       /// descriptive only, needs not to be unique
    QHostAddress    address;    /// address+port must be unique in clients list
    quint16         port;
    bool            active;     /// true = the client is ready to receive the samples flux
};



///
/// \brief The SyncRx class
///
class SyncRx : public IQdevice
{
    Q_OBJECT

protected:
    bool bufReady;
    bool iqRecordOn;                             /// iq samples recording active

    int droppedSamples;                           /// number of I/Q pairs missing
    int timeFreeBuf;                              /// time spent waiting for a free buffer
    int cfgChangesCount;                         /// configuration changes counter for clients notification
    int sentBuffersCount;                       /// progressive of buffers sent to clients
    int sysFaultCount;                          /// progressive nr. of errors returned by librtlsdr primitives
    int iqBufRoundSize;                         /// rx buffers size
    rtlsdr_dev_t *dev;                             /// RTL-SDR opened device
    QElapsedTimer timeAcq;

    QUdpSocket* serverSocket;
    QList<ClientDescriptor*> clients;


    ///
    /// \brief startServer
    ///
    bool startServer();

    ///
    /// \brief sendDeviceConfig
    /// \param client
    ///
    void sendDeviceConfig(ClientDescriptor* client);

    ///
    /// \brief stopAllClients
    ///
    void stopAllClients();

    ///
    /// \brief broadcastSamples
    /// \param buf
    ///
    void shareSamples(IQbuf* buf);

    ///
    /// \brief readPendingClientDatagrams
    ///
    void readPendingClientDatagrams();

    ///
    /// \brief broadcastDeviceConfig
    ///
    void shareDeviceConfig();


public:

    ///
    /// \brief SyncRx
    /// \param appSettings
    /// \param bPool
    /// \param parent
    ///
    explicit SyncRx(Settings* appSettings, Pool<IQbuf*>* bPool, QThread *parentThread = nullptr, int devIndex = 0);

    ~SyncRx();

    ///
    /// \brief lookForDongleStatic
    /// \param devices
    /// \return
    ///
    static bool lookForDongleStatic(QList<DeviceMap> *devices );

    ///
    /// \brief run
    ///
    virtual void run();

    ///
    /// \brief className
    /// \return
    ///
    virtual const char* className() const
    {
        return static_cast<const char*>( "SyncRx" );
    }

    ///
    /// \brief deviceOpen
    /// \return
    ///
    virtual bool isDeviceOpen();

    ///
    /// \brief getStats
    /// \param stats
    /// \return
    ///
    virtual bool getStats( StatInfos* stats );


public slots:

    ///
    /// \brief paramsChange
    ///
    virtual void slotParamsChange();

    ///
    /// \brief slotSetTune
    /// \return
    ///
    virtual bool slotSetTune();
    ///
    /// \brief slotSetSampleRate
    /// \return
    ///
    virtual bool slotSetSampleRate();

    ///
    /// \brief slotSetGain
    /// \return
    ///
    virtual bool slotSetGain();

    ///
    /// \brief slotSetError
    /// \return
    ///
    virtual bool slotSetError();

    ///
    /// \brief slotSetBandwidth
    /// \return
    ///
    virtual bool slotSetBandwidth();



};

#endif // SYNCRX_H
