/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.


    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html




*******************************************************************************
$Rev:: 369                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-11-24 10:01:36 +01#$: Date of last commit
$Id: waterfall.cpp 369 2018-11-24 09:01:36Z gmbertani $
*******************************************************************************/
#include "ui_waterfall.h"
#include "control.h"
#include "ruler.h"
#include "dbmpalette.h"
#include "mainwindow.h"
#include "waterfall.h"

Waterfall::Waterfall(Settings* appSettings, Control* appControl, MainWindow* mainWnd, QWidget *parent) :
    QWidget(parent),
    ui(new Ui::Waterfall)
{
    setObjectName("Waterfall widget");
    ui->setupUi(this);

    MY_ASSERT( nullptr !=  appSettings);
    as = appSettings;

    MY_ASSERT( nullptr !=  appControl);
    ac = appControl;

    MY_ASSERT( nullptr !=  mainWnd);
    m = mainWnd;

    scanList.clear();
    storedList.clear();

    //defaults
    fTicks = WF_FREQ_TICKS;
    tTicks = WF_TIME_TICKS;

    init = true;

    yScans = 0;

    debouncePbShot = 0;

    sigBlock = false;
    samplerate = as->getSampleRate();   
    bandwidth = samplerate;
    maxBW = samplerate;
    zoomCoverage = bandwidth / 2;
    tuneZoom = DEFAULT_ZOOM;
    tuneOffset = 0;
    maxOff = 0;
    minOff = 0;

    //tilt = {0, 20,10,10,5,5,5,3,3,2,2};
    tilt = {0, 30, 30, 10, 10, 5, 5, 5, 3, 3, 3};

    fromHz = as->getTune() - (samplerate / 2);
    toHz = as->getTune() + (samplerate / 2);

    lowest = as->getTune() - (as->getBandwidth() / 2);
    highest = as->getTune() + (as->getBandwidth() / 2);
    zoomLo = lowest;
    zoomHi = highest;

    powBeg = MIN_DBFS;
    powEnd = MAX_DBFS;

    storedS = 0;
    storedN = 0;
    storedDiff = 0;
    storedAvgDiff = 0;

    avgNow = 0;
    QString frameStyle = "border-radius: 4px;  border: 1px solid white;";

    //create a 200 colors matching a 120 dBfs range palette
    ui->subFramePalette->setFrameStyle(QFrame::Box);
    ui->subFramePalette->setStyleSheet(frameStyle);
    dp = new DbmPalette(as, DEFAULT_DBFS_STEP, ui->subFramePalette);
    MY_ASSERT( nullptr !=  dp);

    //create waterfall diagram
    //the parent is a subframe because I needed to leave some
    //space (top margin) above the diagram to align it to the power graph at right
    ui->subFrameWf->setFrameStyle(QFrame::Box);
    ui->subFrameWf->setStyleSheet(frameStyle);
    ww = new WaterfallWidget(as, dp, ui->subFrameWf);
    MY_ASSERT( nullptr !=  ww);

    //frequency ruler is horizontal not scrolling
    ui->frameFreq->setFrameStyle(QFrame::Box);
    ui->frameFreq->setStyleSheet(frameStyle);
    hr = new Ruler(as, RO_DOWN, ui->frameFreq);
    MY_ASSERT( nullptr !=  hr);

    //ui->wPanoramic->setFrameStyle(QFrame::Box);
    //ui->wPanoramic->setStyleSheet(frameStyle);

    //ui->wPower->setFrameStyle(QFrame::Box);
    //ui->wPower->setStyleSheet(frameStyle);

    //instantaneous spectra graph (formerly InstantSpectra class)

    hcv = new QChartView();
    MY_ASSERT( nullptr !=  hcv);

    //vertical power axis
    haY = new QValueAxis;
    MY_ASSERT( nullptr !=  haY);

    //horizontal frequency axis
    haX = new QValueAxis;
    MY_ASSERT( nullptr !=  haX);

    hSerie = new QLineSeries();
    MY_ASSERT( nullptr !=  hSerie);

    hcv->setRenderHint(QPainter::Antialiasing);
    hcv->setSceneRect(ui->wPanoramic->rect());

    hChart = hcv->chart();
    hChart->legend()->hide();

    QMargins marH(20,0,0,0);
    hChart->setMargins(marH);
    hChart->setTheme( QChart::ChartThemeDark );
    hChart->addSeries(hSerie);

    //power axis (Y)
    haY->setRange(powBeg, powEnd);    
    haY->setLabelFormat("%- 3d    ");

#if (QT_VERSION >= QT_VERSION_CHECK(5, 12, 0))
    haY->setTickType(QValueAxis::TicksDynamic);
    haY->setTickAnchor(0);
    haY->setTickInterval(tilt.at(1));
#else
    haY->setTickCount(5);
    haY->applyNiceNumbers();
#endif


    hChart->addAxis(haY, Qt::AlignLeft);
    hSerie->attachAxis(haY);
    //hChart->setAxisY(haY, hSerie); deprecated

    //frequency axis (X)
    int points = static_cast<int>(hChart->plotArea().width());
    haX->setRange( 0, points - 1);
    //haX->applyNiceNumbers();
    haX->hide();
    hChart->addAxis(haX, Qt::AlignBottom);
    hSerie->attachAxis(haX);
    //hChart->setAxisX(haX, hSerie); deprecated

    //average power and peak power vs. time graph (formerly PowerGraph class)
    vcv = new QChartView();
    MY_ASSERT( nullptr !=  vcv);

    vSerieS = new QLineSeries();
    MY_ASSERT( nullptr !=  vSerieS);

    vSerieN = new QLineSeries();
    MY_ASSERT( nullptr !=  vSerieN);

    //vertical time axis
    vaY = new QValueAxis;
    MY_ASSERT( nullptr !=  vaY);

    //horizontal power axis
    vaX = new QValueAxis;
    MY_ASSERT( nullptr !=  vaX);

    vcv->setRenderHint(QPainter::Antialiasing);
    vcv->setSceneRect(ui->wPower->rect());

    vChart = vcv->chart();
    vChart->legend()->hide();

    vChart->addSeries(vSerieS);
    vChart->addSeries(vSerieN);
    QMargins marV(10,0,0,10);
    vChart->setMargins(marV);
    vChart->setTheme( QChart::ChartThemeDark );

    //time axis (Y) in seconds. It must scroll
    //side-by-side with the waterfall
    points = static_cast<int>( vChart->plotArea().height() );
    vaY->setRange(-points, 0);
    vaY->setLabelFormat("%-03d ");
    vaY->hide();
    vChart->addAxis(vaY, Qt::AlignLeft);
    vSerieN->attachAxis(vaY);
    //vChart->setAxisY(vaY, vSerieN); deprecated
    vSerieS->attachAxis(vaY);

    //power axis (X)
    vaX->setRange(powBeg, powEnd);
    vaX->setLabelFormat("  %-03d ");

#if (QT_VERSION >= QT_VERSION_CHECK(5, 12, 0))
    vaX->setTickType(QValueAxis::TicksDynamic);
    vaX->setTickAnchor(0);
    vaX->setTickInterval(tilt.at(1));
#else
    vaX->setTickCount(5);
    vaX->applyNiceNumbers();
#endif

    vaX->setLabelsAngle(90);
    vChart->addAxis(vaX, Qt::AlignBottom);
    vSerieN->attachAxis(vaX);
    //vChart->setAxisX(vaX, vSerieN); deprecated
    vSerieS->attachAxis(vaX);

    //not needed in constructor, since it's executed only once
    //static const QMetaMethod notifyOpenedSignal = QMetaMethod::fromSignal( &Settings::notifyOpened );
    if(true /* !isSignalConnected( notifyOpenedSignal ) */)
    {
        //plugging signals, only once

        //notifications
        connect( as,    SIGNAL( notifyOpened() ),         this,   SLOT( slotConfigChanged() ) );
        connect( as,    SIGNAL( notifySetDbfs() ),        this,   SLOT( slotRuntimeChange() ) );
        connect( as,    SIGNAL( notifySetHz() ),          this,   SLOT( slotRuntimeChange() ) );
        connect( as,    SIGNAL( notifySetSec() ),         this,   SLOT( slotRuntimeChange() ) );
        connect( as,    SIGNAL( notifySetTune() ),        this,   SLOT( slotRuntimeChange() ) );
        connect( as,    SIGNAL( notifySetSampleRate() ),  this,   SLOT( slotRuntimeChange() ) );
        connect( as,    SIGNAL( notifySetBandwidth() ),   this,   SLOT( slotRuntimeChange() ) );
        connect( as,    SIGNAL( notifySetResolution() ),  this,   SLOT( slotRuntimeChange() ) );
        connect( as,    SIGNAL( notifySsBypass() ),       this,   SLOT( slotRuntimeChange() ) );

        //pushbuttons
        connect( ui->pbShot,    SIGNAL( pressed() ), this,  SLOT( slotManualShot() ) );

        //sliders
        connect( ui->hsTuneOffset,  SIGNAL( valueChanged(int) ), this,  SLOT( slotSetOffset(int) ) );
        connect( ui->hsTuneZoom,    SIGNAL( valueChanged(int) ), this,  SLOT( slotSetZoom(int) ) );
        connect( ui->hsBrightness,  SIGNAL( valueChanged(int) ), this,  SLOT( slotSetBrightness(int) ) );
        connect( ui->hsContrast,    SIGNAL( valueChanged(int) ), this,  SLOT( slotSetContrast(int) ) );

        connect( ui->hsPowerOffset,  SIGNAL( valueChanged(int) ), this,  SLOT( slotSetPowerOffset(int) ) );
        connect( ui->hsPowerZoom,    SIGNAL( valueChanged(int) ), this,  SLOT( slotSetPowerZoom(int) ) );

        connect( ac,    SIGNAL( run(int,int) ),  this,   SLOT( slotStartAcq(int,int) ) );
        connect( ac,    SIGNAL( stop() ), this,   SLOT( slotStopAcq() ) );

        connect( vaY,   SIGNAL( rangeChanged(qreal,qreal) ), this,   SLOT( slotRangeChanged(qreal, qreal) ) );
    }


}

Waterfall::~Waterfall()
{
    delete vcv;
    delete hcv;
    delete dp;
    delete ww;
    delete hr;
    delete ui;
}

QString Waterfall::GNUplotPalette()
{
    return dp->GNUplotPalette();
}

void Waterfall::GUIinit(void)
{
    reinit = true;

    //parametrize instant spectra graph ----------------------

    //hcv->setSizePolicy( QSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding ));
    ui->saPanoramic->setWidget(hcv);
    hSerie->clear();

    //parametrize S/N vs. time graph -----------------------

    ui->saPower->setWidget(vcv);
    vSerieN->clear();
    vSerieS->clear();

    //line colors must be the same used in legend
    hSerie->setColor( Qt::white );
    vSerieS->setColor( Qt::red );       //peak power    (S)
    vSerieN->setColor( Qt::cyan );      //average power (N)

    //inizializes the status of every control
    //starting from sliders
    int b = as->getBrightness();
    ui->hsBrightness->setValue(b);
    ui->lbBrightness->setNum(b);

    int c = as->getContrast();
    ui->hsContrast->setValue(c);
    ui->lbContrast->setNum(c);

    ui->hsPowerZoom->setMinimum(1);
    ui->hsPowerZoom->setMaximum(10);
    int pg = as->getPowerZoom();
    ui->hsPowerZoom->setValue(pg);

#if (QT_VERSION >= QT_VERSION_CHECK(5, 12, 0))
    haY->setTickInterval(tilt.at(pg));
    vaX->setTickInterval(tilt.at(pg));
#endif

    ui->hsPowerOffset->setMinimum(MIN_DBFS);
    ui->hsPowerOffset->setMaximum(MAX_DBFS);
    int po = as->getPowerOffset();
    ui->hsPowerOffset->setValue(po);


    //these controls are independent from radio settings
    slotSetBrightness(b);
    slotSetContrast(c);
    slotSetPowerZoom(pg);
    slotSetPowerOffset(po);

    setGeometry( as->getWwGeometry() );
    setDetectionRange(lowest, highest);

    //trick to avoid the offset and zoom self-reset when setting bw
    int bw = as->getBandwidth();
    int offset = as->getOffset();
    int zoom = as->getZoom();
    int tune = as->getTune();

    ui->lbBW->setNum(bw);

    fromHz = tune - (samplerate / 2);
    toHz = tune + (samplerate / 2);

    lowest = tune - (bw / 2);
    highest = tune + (bw / 2);

    //the zoom value is expressed as _tenths_ of X
    //zoomCoverage = (bw/zoom) / 2;
    zoomCoverage = static_cast<int>((bw/( static_cast<float>(zoom)/10.0F )) / 2.0F);
    minOff = lowest+zoomCoverage;
    maxOff = highest-zoomCoverage;
    offsetCoverage = ((maxOff - minOff)/2) > 1 ? ((maxOff - minOff)/2) : 2;
    tuneOffset = offset;

    zoomLo = tune + tuneOffset - zoomCoverage;
    zoomHi = tune + tuneOffset + zoomCoverage;
    as->setZoomedRange(zoomLo, zoomHi);
    slotUpdateGrid();

    if(minOff == maxOff)
    {
        //when zoom = DEFAULT_ZOOM
        ui->hsTuneOffset->setEnabled(false);
        tuneOffset = 0;
    }
    else
    {
        ui->hsTuneOffset->setEnabled(true);
    }

    ui->hsTuneOffset->setValue(tuneOffset);

    int minHz = hr->getMinHz();
    int step = calcSliderStep( minHz );

    ui->hsTuneOffset->setSingleStep( step );
    ui->hsTuneOffset->setPageStep( step * 10 );
    ui->hsTuneOffset->setMaximum( offsetCoverage );
    ui->hsTuneOffset->setMinimum( -offsetCoverage );
    ui->lbFoff->setNum(tuneOffset);

    ui->hsTuneZoom->setValue(zoom);

    slotSetOffset( tuneOffset );
    slotSetZoom(zoom);


    takeShot(0,0,0,0,0,"");

    ui->lbCapturing->hide();

    QString s;
    s.setNum( as->getLat() );
    ui->lbLatVal->setText(s);
    s.setNum( as->getLong() );
    ui->lbLongVal->setText(s);

    s = QString("%1.%2").arg(zoom/10).arg(zoom%10);
    ui->lbFzoom->setText(s);
    s.setNum(zoomCoverage * 2);
    ui->lbFzoomHz->setText(s);

    slotUpdateLogo();

    MYDEBUG << "*** END of GUI init***" << endl;

    slotRuntimeChange();


}


void Waterfall::slotUpdateLogo()
{
    MYDEBUG << "slotUpdateLogo()";
    QString logoFile = as->getLogo();
    QPixmap logo(logoFile);
    ui->lbLogo->setText("");
    ui->lbLogo->setPixmap( logo );
}

void Waterfall::slotConfigChanged()
{
    //init = true;
}

void Waterfall::slotRangeChanged(qreal min, qreal max)
{

    if(yScans % PG_ALIGNMENT_EVERY_SCANS == 0)
    {
        MYDEBUG << "ReLayouting - slotRangeChanged(min=" << min << " max=" << max << " )  yScans=" << yScans;
        min = ceil(min);
        max = ceil(max);
        MYDEBUG << "setRange(min=" << min << " max=" << max << " )";
        vaY->setRange(min, max);

    }
    else
    {
        MYDEBUG << "slotRangeChanged(min=" << min << " max=" << max << " )  yScans=" << yScans;
    }
}

void Waterfall::setCapturing(bool isInhibited, bool isCapturing)
{
    if(!isInhibited)
    {
        ui->lbNoCap->setVisible(false);
        ui->lbCapturing->setVisible(isCapturing);
    }
    else
    {
        ui->lbCapturing->setVisible(false);
        ui->lbNoCap->setVisible(true);
    }
}

void Waterfall::setScreenshotCoverage(int secs)
{
    if(secs == 0)
    {
        ui->lbSsCov->setText(tr("measuring..."));
    }
    else
    {
        ui->lbSsCov->setNum(secs);
    }
}



void Waterfall::slotStartAcq(int shotNr, int totalShots)
{

    yScans = 0;
    vSerieN->clear();
    vSerieS->clear();
    int points = static_cast<int>(vChart->plotArea().height());
    vaY->setRange(-points, 0);

    QString s;

    if(shotNr == 0)
    {
        ui->lbShot->setText("00/00");
        return;
    }

    if(totalShots == 0)
    {
        //manual shots
        //s.sprintf("%02i", shotNr);
        s = QString("%1").arg(shotNr, 2, 10, QChar('0'));
        ui->lbShot->setText(s);
        ui->lbTaken->setText( tr("Manual shots taken:") );
    }
    else
    {
        //s.sprintf("%02i/%02i", shotNr,totalShots);
        s = QString("%1/%2").arg(shotNr, 2, 10, QChar('0')).arg(totalShots, 2, 10, QChar('0'));
        ui->lbShot->setText(s);
        ui->lbTaken->setText( tr("Automatic shots taken:") );
    }

    ui->lbTaken->show();
    ui->lbShot->show();

    slotRuntimeChange();
}

void Waterfall::slotStopAcq()
{
    MYDEBUG << "Total points in vertical sidegraph vSerieN: " << vSerieN->count();
    MYDEBUG << "Total points in vertical sidegraph vSerieS: " << vSerieS->count();
    MYDEBUG << "Total Yscans: " << yScans;
}



void Waterfall::slotSetOffset( int value )
{
    //rounds value by single step
    int sStep = ui->hsTuneOffset->singleStep();

    int mod = value % sStep;
    if (mod != 0)
    {
        value = ( mod > (sStep / 2) )?
                (value - mod + sStep) :
                (value - mod);
    }

    as->setOffset(value);
    slotRuntimeChange();
}


void Waterfall::slotSetZoom( int value )
{
    //zoom is a value 1..100x
    //rounds value by single step
    int sStep = ui->hsTuneZoom->singleStep();
    int mod = value % sStep;
    if (mod != 0)
    {
        value = ( mod > (sStep / 2) )?
                (value - mod + sStep) :
                (value - mod);
    }

    //resets the offset each time the zoom changes

    as->setZoom(value);
    slotRuntimeChange();
}



void Waterfall::slotRuntimeChange()
{

    QString s;
    int sr = -1, bw = -1, off = 0, zoom = -1, tune = 0;

    MYDEBUG << __func__ ;

    blockAllSignals(true);
    setGeometry( as->getWwGeometry() );

    //dp->updatePalette();

    sr = as->getSampleRate();
    bw = as->getBandwidth();
    zoom = as->getZoom();
    off = as->getOffset();
    tune = as->getTune();

    if(sr != samplerate)
    {
        //SR changed: if lower than BW is reset to maximum value, equal to SR
        maxBW = sr;
        if(bandwidth > maxBW)
        {
            bw = maxBW;
        }

        if(init == false && reinit == false)
        {
            //consequently, the zoom is set back to 1 and the offset to zero
            zoom = DEFAULT_ZOOM;
            off = 0;
        }

        //adjusts the BW,offset and zoom controls accordingly
        fromHz = tune - (sr / 2);
        toHz = tune + (sr / 2);
        lowest = tune - (bw / 2);
        highest = tune + (bw / 2);

        //the zoom value is expressed as _tenths_ of X
        //zoomCoverage = (bw/zoom) / 2;
        zoomCoverage = static_cast<int>((bw/( static_cast<float>(zoom)/10.0 )) / 2.0);
        minOff = lowest+zoomCoverage;
        maxOff = highest-zoomCoverage;
        offsetCoverage = ((maxOff - minOff)/2) > 1 ? ((maxOff - minOff)/2) : 2;



         ui->lbBW->setNum(bw);

        as->setBandwidth(bw);
        bandwidth = bw;

        ui->hsTuneOffset->setMaximum( offsetCoverage );
        ui->hsTuneOffset->setMinimum( -offsetCoverage );
        ui->hsTuneOffset->setTickInterval( bw / 10 );
        ui->hsTuneOffset->setValue( off );
        s.setNum(off /*-tune*/);
        ui->lbFoff->setText(s);
        as->setOffset(off);
        tuneOffset = off;

        //the zoom slider min/max values are fixed
        //only the value can change
        ui->hsTuneZoom->setValue( zoom );
        s = QString("%1.%2").arg(zoom/10).arg(zoom%10);
        ui->lbFzoom->setText(s);
        s.setNum(zoomCoverage * 2);
        ui->lbFzoomHz->setText(s);
        as->setZoom(zoom);
        tuneZoom = zoom;

        samplerate = sr;

    }

    if(bw != bandwidth)
    {
        if(init == false && reinit == false)
        {
            //BW changed: the zoom is set back to 1 and the offset to zero
            zoom = DEFAULT_ZOOM;
            off = 0;
        }

        lowest = tune - (bw / 2);
        highest = tune + (bw / 2);
        //the zoom value is expressed as _tenths_ of X
        //zoomCoverage = (bw/zoom) / 2;
        zoomCoverage = static_cast<int>((bw/( static_cast<float>(zoom)/10.0F )) / 2.0F);
        minOff = lowest+zoomCoverage;
        maxOff = highest-zoomCoverage;
        offsetCoverage = ((maxOff - minOff)/2) > 1 ? ((maxOff - minOff)/2) :21;

        ui->hsTuneOffset->setMaximum( offsetCoverage );
        ui->hsTuneOffset->setMinimum( -offsetCoverage );
        ui->hsTuneOffset->setTickInterval( bw / 10 );
        ui->hsTuneOffset->setValue( off );
        s.setNum(off);
        ui->lbFoff->setText(s);
        as->setOffset(off);
        tuneOffset = off;

        //the zoom slider min/max values are fixed
        //only the value can change
        ui->hsTuneZoom->setValue( zoom );
        s = QString("%1.%2").arg(zoom/10).arg(zoom%10);
        ui->lbFzoom->setText(s);
        s.setNum(zoomCoverage * 2);
        ui->lbFzoomHz->setText(s);
        as->setZoom(zoom);
        tuneZoom = zoom;

         ui->lbBW->setNum(bw);

        as->setBandwidth( bw );
        bandwidth = bw;

    }

    if(zoom != tuneZoom)
    {
        //the zoom value is expressed as _tenths_ of X
        //zoomCoverage = (bw/zoom) / 2;
        zoomCoverage = static_cast<int>((bw/( static_cast<float>(zoom)/10.0 )) / 2.0);
        minOff = lowest+zoomCoverage;
        maxOff = highest-zoomCoverage;
        offsetCoverage = ((maxOff - minOff)/2) > 1 ? ((maxOff - minOff)/2) : 2;


        ui->hsTuneOffset->setMaximum( offsetCoverage );
        ui->hsTuneOffset->setMinimum( -offsetCoverage );
        ui->hsTuneOffset->setTickInterval( zoomCoverage / 5 );
        ui->hsTuneOffset->setValue( off );
        s.setNum(off);
        ui->lbFoff->setText(s);
        as->setOffset(off);
        tuneOffset = off;

        //the zoom slider min/max values are fixed
        //only the value can change
        ui->hsTuneZoom->setValue( zoom );
        s = QString("%1.%2").arg(zoom/10).arg(zoom%10);
        ui->lbFzoom->setText(s);
        s.setNum(zoomCoverage * 2);
        ui->lbFzoomHz->setText(s);
        as->setZoom(zoom);
        if(zoom == DEFAULT_ZOOM)
        {
            off = 0;
        }
        tuneZoom = zoom;
    }

    if(off != tuneOffset)
    {
        //offset changed
        ui->hsTuneOffset->setValue( off );
        s.setNum(off);
        ui->lbFoff->setText(s);
        as->setOffset(off);
        tuneOffset = off;
    }

    if(minOff == maxOff)
    {
        //when zoom = DEFAULT_ZOOM
        ui->hsTuneOffset->setEnabled(false);
    }
    else
    {
        ui->hsTuneOffset->setEnabled(true);
    }

    int centerFreq = tune+off;
    MYINFO << " Central freq = tune(" << tune << ") + offset(" << off << ") = " << centerFreq << " bandwidth=" << bandwidth;

    ui->lbFreq->setNum(centerFreq);

    ui->lbRes->setNum( static_cast<double>( as->pround( as->getResolution(), 2 ) ) );
    ui->lbPoints->setNum( as->getFFTbins() );

    s.setNum( as->getLat() );
    ui->lbLatVal->setText(s);
    s.setNum( as->getLong() );
    ui->lbLongVal->setText(s);


    if(as->getDbfs() != 0)
    {
        //shows the power grid in both graphs
        haY->show();
        vaX->show();

    }
    else
    {
        haY->hide();
        vaX->hide();
    }

    zoomLo = tune + tuneOffset - zoomCoverage;
    zoomHi = tune + tuneOffset + zoomCoverage;
    as->setZoomedRange(zoomLo, zoomHi);
    haX->setRange(zoomLo, zoomHi);

    int step = calcSliderStep( hr->getMinHz() );

    ui->hsTuneOffset->setSingleStep( step );
    ui->hsTuneOffset->setPageStep( step * 10 );
    ui->hsTuneOffset->setMaximum( offsetCoverage );
    ui->hsTuneOffset->setMinimum( -offsetCoverage );

    slotUpdateGrid();
    blockAllSignals(false);

    init = false;
    reinit = false;
}

void Waterfall::slotSetBrightness( int value )
{
    MYDEBUG << "linear input value=" << value;
    as->setBrightness(value);
    dp->setDbmPalette(value, as->getContrast());
    //dp->updatePalette();
    //slotRuntimeChange();
}


void Waterfall::slotSetContrast( int value )
{
    MYDEBUG << "linear input value=" << value;
    as->setContrast(value);
    dp->setDbmPalette(as->getBrightness(), value);
    //dp->updatePalette();
    //slotRuntimeChange();
}

void Waterfall::slotSetPowerZoom( int zoom )
{
    MYDEBUG << "linear input value=" << zoom;
    as->setPowerZoom(zoom);
    int off = as->getPowerOffset();
    int delta = static_cast<int>(fabs(MAX_DBFS - MIN_DBFS) / zoom);
    powBeg = off - (delta/2);
    powEnd = off + (delta/2);
    haY->setRange(powBeg, powEnd);
    vaX->setRange(powBeg, powEnd);

#if (QT_VERSION >= QT_VERSION_CHECK(5, 12, 0))
   vaX->setTickInterval(tilt.at(zoom));
   haY->setTickInterval(tilt.at(zoom));
#endif

    /*
    //power/time series reset
    yScans = 0;
    vSerieN->clear();
    vSerieS->clear();
    int vp = vChart->plotArea().height();
    vaY->setRange(-vp, 0);
    */

    update();
    MYDEBUG << "powBeg=" << powBeg << " powEnd=" << powEnd;
}

void Waterfall::slotSetPowerOffset(int dbm )
{
    MYDEBUG << "linear input value=" << dbm;
    as->setPowerOffset(dbm);
    int zoom = as->getPowerZoom();
    int delta = static_cast<int>( fabs(MAX_DBFS - MIN_DBFS) / zoom );
    powBeg = dbm - (delta/2);
    powEnd = dbm + (delta/2);
    haY->setRange(powBeg, powEnd);
    vaX->setRange(powBeg, powEnd);

    /*
    //power/time series reset
    yScans = 0;
    vSerieN->clear();
    vSerieS->clear();
    int vp = vChart->plotArea().height();
    vaY->setRange(-vp, 0);
    */
    update();
    MYDEBUG << "powBeg=" << powBeg << " powEnd=" << powEnd;
}





void Waterfall::slotShowStatus( int appState )
{
    if(appState <= AST_CANNOT_OPEN)
    {
        ui->hsTuneOffset->setEnabled(false);
        ui->hsTuneZoom->setEnabled(false);
    }
    else
    {
        //ui->hsBW->setEnabled(true);
        ui->hsTuneOffset->setEnabled(true);
        ui->hsTuneZoom->setEnabled(true);
    }
}

void Waterfall::cleanAll()
{
    ww->clean();
    dp->updatePalette();
}

void Waterfall::slotPlotPower(float maxDbm, float avgDbm, float avgDiff, bool raisingEdge)
{
    //max number of visible points
    uint vp = static_cast<uint>( vChart->plotArea().height() );

    int S = static_cast<int>( maxDbm );
    int N = static_cast<int>( avgDbm );
    int iDiff = S-N;
    int iAvgDiff = static_cast<int>( avgDiff );

    if(raisingEdge)
    {
        storedS = S;
        storedN = N;
        storedDiff = iDiff;
        storedAvgDiff = iAvgDiff;
    }

    ui->lbSval->setText( QString("%1 dBfs").arg(S) );
    ui->lbNval->setText( QString("%1 dBfs").arg(N) );
    ui->lbDiffVal->setText( QString("%1 dBfs").arg(iDiff) );
    ui->lbAvgDiffVal->setText( QString("%1 dBfs").arg(iAvgDiff) );


    vSerieS->append( S, yScans );
    vSerieN->append( N, yScans );

    yScans++;

    /*
     * QtCharts scrolling issue (feature or bug?)
     * Each scroll() causes a rangeChanged() signal().
     * The range scroll amount is not exactly the same
     * given to scroll(), I don't know why, maybe a cumulative
     * rounding error.
     * The result is that the range loses the alignment
     * to yScans and the resulting graph, after half hour
     * of work becomes clearly no more aligned to waterfall.
     *
     * For workaround, I implemented a slot to receive the
     * rangeChanged() signals and issue setRange() to fix
     * the misalignments. Once each 500 scans, it does a
     * rounding.
     */

    vChart->scroll(0.0, 1.0);


    //the oldest point, when no more visible, is deleted
    //from both series
    if( yScans >  vp )
    {
        vSerieN->removePoints( 0, 1 );
        vSerieS->removePoints( 0, 1 );
    }

    /* manages X counter overflow */
    if(yScans == UINT_MAX)
    {
        MYWARNING << "Power graph yScans overflow, resetting graph";
        yScans = 0;
        vSerieN->clear();
        vSerieS->clear();
    }

    avgNow = static_cast<int>(N);
    debouncePbShot -= (debouncePbShot > 0)? 1 : 0;
}

void Waterfall::slotAppendPixel(float dBfs)
{
    //update waterfall
    if(ww->appendPixel(dBfs) == true)
    {
        scanList.append( QPointF( scanList.size(), dBfs ) );
    }
}

void Waterfall::slotRefresh( QString& timeStamp, bool raisingEdge )
{

    hSerie->replace( scanList );
    haX->setRange(0, scanList.size()-1);
    if(raisingEdge == true)
    {
        storedList = scanList;
    }
    scanList.clear();

    ww->refresh(timeStamp);

    ui->lbUTC->setText( ac->getDateTime() );
}

void Waterfall::slotReplay()
{
    if(storedList.count() > 0)
    {
        scanList = storedList;
    }

    ui->lbSval->setText( QString("%1 dBfs").arg(storedS) );
    ui->lbNval->setText( QString("%1 dBfs").arg(storedN) );
    ui->lbDiffVal->setText( QString("%1 dBfs").arg(storedDiff) );
    ui->lbAvgDiffVal->setText( QString("%1 dBfs").arg(storedAvgDiff) );

    hSerie->replace( scanList );
    haX->setRange(0, scanList.size()-1);
    storedList.clear();
    scanList.clear();
}

void Waterfall::slotUpdateGrid()
{
    MYDEBUG << "slotUpdateGrid()";

    dp->refresh();
    hr->setGrid( fTicks, zoomLo, zoomHi );
    ww->setGrid(hr->getMinPix(), tTicks, hr->getMajHz(), hr->getMinHz());

    palBeg = 0;
    palEnd = 0;
    palExc = 0;
    dp->paletteExtension(palBeg, palEnd, palExc);
    as->setWfGeometry( geometry(), ww->geometry() );
}

void Waterfall::resizeEvent ( QResizeEvent* event )
{
    if(init == true)
    {
        //skips the first resizeEvent() issued by the system
        MYDEBUG << "event->size=" << event->size() << " geometry()=" << geometry() << " IGNORED";
    }
    else
    {
        MYDEBUG << "event->size=" << event->size() << " geometry()=" << geometry();
        MYDEBUG << "widget waterfall : " << ww->size();
        MYDEBUG << "widget hChart: " << hChart->size();
        MYDEBUG << "plot area hChart: " << hChart->plotArea();


        ww->setGrid(hr->getMinPix(), tTicks, hr->getMajHz(), hr->getMinHz(), true);
        as->setWfGeometry( geometry(), ww->geometry() );

        int points = static_cast<int>(hChart->plotArea().width());
        haX->setRange(0, points-1);

        points = static_cast<int>(vChart->plotArea().height());
        vaY->setRange(-points, 0);

        yScans = 0;
        vSerieN->clear();
        vSerieS->clear();
        points = static_cast<int>( vChart->plotArea().height() );
        vaY->setRange(-points, 0);
        emit waterfallResized();
        slotUpdateGrid();
    }
}

void Waterfall::moveEvent ( QMoveEvent * event )
{
    if(init == true)
    {
        //skips the first moveEvent() issued by the system
        MYDEBUG << "event->pos=" << event->pos() << " geometry()=" << geometry() << " IGNORED";
    }
    else
    {
        MYDEBUG << "event->pos=" << event->pos() << " geometry()=" << geometry();
        as->setWfGeometry( geometry(), ww->geometry() );
    }
}


void Waterfall::blockAllSignals(bool block)
{
    MYDEBUG << "blockAllSignals(" << block << ")";
    sigBlock = block;

    ui->hsBrightness->blockSignals(block);
    ui->hsContrast->blockSignals(block);
    ui->hsPowerZoom->blockSignals(block);
    ui->hsPowerOffset->blockSignals(block);
    ui->hsTuneOffset->blockSignals(block);
    ui->hsTuneZoom->blockSignals(block);
    ui->pbShot->blockSignals(block);
}

void Waterfall::slotManualShot()
{
    //forwards the request for manual shot to Control
    //because the filename must be built by it
    //emit requestManualShot();
    if( debouncePbShot == 0 )
    {
        ac->slotTakeWfManShot();
        debouncePbShot = MANSHOT_DEBOUNCE_TICKS;
    }
}


void Waterfall::takeShot(int shotNr, int totalShots, float maxDbm, float avgDbm, float avgDiff, QString shotFileName )
{
    QString s;

    if( ac->getAcquisitionMode() == AM_AUTOMATIC )
    {
        //redisplay the instantSpectra at moment of the event
        //for the sake of screenshot completeness
        slotReplay();
    }

    if(shotNr == 0)
    {
        ui->lbShot->setText("00/00");
        return;
    }

    if(totalShots == 0)
    {
        //manual shots
        //s.sprintf("%02i", shotNr);
        s = QString("%1").arg(shotNr, 2, 10, QChar('0'));
        ui->lbShot->setText(s);
        ui->lbTaken->setText( tr("Manual shots taken:") );
    }
    else
    {        
        //automatic shots
        //s.sprintf("%02i/%02i", shotNr,totalShots);
        s = QString("%1/%2").arg(shotNr, 2, 10, QChar('0')).arg(totalShots, 2, 10, QChar('0'));
        ui->lbShot->setText(s);
        ui->lbTaken->setText( tr("Automatic shots taken:") );


        int S = static_cast<int>( maxDbm );
        int N = static_cast<int>( avgDbm );
        int iDiff = S-N;
        int iAvgDiff = static_cast<int>( avgDiff );

        ui->lbSval->setText( QString("%1 dBfs").arg(S) );
        ui->lbNval->setText( QString("%1 dBfs").arg(N) );
        ui->lbDiffVal->setText( QString("%1 dBfs").arg(iDiff) );
        ui->lbAvgDiffVal->setText( QString("%1 dBfs").arg(iAvgDiff) );

    }
    ui->lbTaken->show();
    ui->lbShot->show();

    show();
    setFocus();

    QPixmap shot = grab();
    shot.save( shotFileName, "PNG"); //takes a shot of the entire window

}



QSize Waterfall::getWidgetSize()
{
    return ww->size();
}

int Waterfall::calcSliderStep( int base )
{
    int pow10 = 0, step = 1;
    int test = static_cast<int>( pow(10.0, pow10) );

    for( ; pow10 < 5; pow10++ )
    {
        if( base % test == 0 )
        {

            step = test;
            test = 5 * step;

            if( base % test == 0 )
            {
                step = test;
                test = 2 * step;
            }
            else
            {
                break;
            }
        }
        else
        {
            break;
        }
    }

    return step;
}
