/******************************************************************************

    Echoes is a RF spectrograph for RTL-SDR devices designed for meteor scatter
    Copyright (C) 2018  Giuseppe Massimo Bertani gmbertani(a)users.sourceforge.net


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.
    

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, http://www.gnu.org/copyleft/gpl.html






*******************************************************************************
$Rev:: 369                      $: Revision of last commit
$Author:: gmbertani             $: Author of last commit
$Date:: 2018-11-24 10:01:36 +01#$: Date of last commit
$Id: waterfall.h 369 2018-11-24 09:01:36Z gmbertani $
*******************************************************************************/

#ifndef WATERFALL_H
#define WATERFALL_H

#include <QtCore>
#include <QtGui>
#include <QtWidgets>
#include <QtCharts>

#include "settings.h"
#include "ruler.h"
#include "waterfallwidget.h"

QT_CHARTS_USE_NAMESPACE

class DbmPalette;
class Preferences;
class Control;

namespace Ui {
    class Waterfall;
}

class MainWindow;

class Waterfall : public QWidget
{
    Q_OBJECT

private:
    Ui::Waterfall *ui;          //<Waterfall window form

protected:
    Settings* as;               //<current application settings
    MainWindow* m;              //<parent window

    Control*    ac;             //<application control

    WaterfallWidget* ww;        //<Waterfall widget
    DbmPalette* dp;             //<color/value palette
    Ruler* hr;                  //<horizontal (fixed) frequency ruler

    bool init;                  //set in constructor: some code must be executed only when called by constructor
    bool reinit;                //set in GUIinit() to skip tune zoom/offset reset at resolution change

    bool sigBlock;              //<signals blocked if true


    int samplerate;             //<last known sampling rate
    int bandwidth;              //<last known bandwidth
    int tuneOffset;             //<last known tune offset
    int tuneZoom;               //<last known tune zoom

    int maxBW;                  //<maximum bw alloweed = SR

    int zoomCoverage;           //<relative frequency coverage of waterfall at each side of central frequency

    int offsetCoverage;         //<relative excursion of offset slider at each side of tune frequency

    int minOff;                 //<offset slider absolute frequency excursion start [hz]
    int maxOff;                 //<offset slider absolute frequency excursion end [hz]

    int zoomLo;                 //<minimum absolute frequency showed in waterfall [Hz] If zoom = 1 it's equal to lowest
    int zoomHi;                 //<maximum absolute frequency showed in waterfall [Hz] If zoom = 1 it's equal to highest

    int lowest;                 //<minimum absolute frequency showable in waterfall with zoom = 1 [hz]  If BW == SR equal to fromHz
    int highest;                //<maximum absolute frequency showable in waterfall with zoom = 1 [hz]  If BW == SR equal to toHz

    int fromHz;                 //<absolute minimum frequency showable in waterfall when BW == SR
    int toHz;                   //<absolute maximum frequency showable in waterfall when BW == SR

    int fTicks;                 //<frequency grid ticks
    int tTicks;                 //<time grid ticks
    uint yScans;                 //<number of rows plotted in the waterfall since acquisition start

    int avgNow;                 //<last average read

    int     palBeg;             //<color palette begin value [dBfs]
    int     palEnd;             //<color palette end value [dBfs]
    int     palExc;             //<color palette extension [dBfs]

    int     powBeg;             //<lowest power value shown on graphs [dBfs]
    int     powEnd;             //<highest power value shown on graphs [dBfs]
    int     debouncePbShot;         //<debounce delayer for manual shot pushbutton

    int storedS;                //< S value to show during replay
    int storedN;                //< N value to show during replay
    int storedDiff;             //< S-N value to show during replay
    int storedAvgDiff;          //< average S-N value to show during replay


    QList<QPointF> scanList;    //<frequency points list to draw he instananeous spectra graph
    QList<QPointF> storedList;  //<saved scanList to replay the peaks during shots

    QVector<int> tilt;          //<zoom vs. graphs tick interval lookup table

    //<instantaneous spectra graph (formerly InstantSpectra class)
    QChartView* hcv;
    QChart *hChart;
    QLineSeries* hSerie;
    QValueAxis *haX, *haY;

    //<average power (N), peak power (S) and difference vs. time graph (formerly PowerGraph class)
    QChartView* vcv;
    QChart *vChart;
    QLineSeries* vSerieS;
    QLineSeries* vSerieN;
    QLineSeries* vSerieD;
    QValueAxis *vaX, *vaY;


    virtual void resizeEvent ( QResizeEvent* event );
    virtual void moveEvent ( QMoveEvent * event );


    ///
    /// \brief calcSliderStep
    /// \param base
    /// \return
    ///
    int calcSliderStep(int base);

public:
    explicit Waterfall(Settings* appSettings, Control* appControl, MainWindow* mainWnd, QWidget *parent = nullptr);
    ~Waterfall();



    ///
    /// \brief GUIinit
    ///
    void GUIinit();

    ///
    /// \brief blockAllSignals
    /// \param block
    ///
    void blockAllSignals(bool block);

    ///
    /// \brief takeShot
    /// \param shotNr
    /// \param totalShots
    /// \param maxDbm
    /// \param avgDbm
    /// \param avgDiff
    /// \param shotFileName
    ///
    void takeShot(int shotNr, int totalShots, float maxDbm, float avgDbm, float avgDiff, QString shotFileName = "unnamed.png" );

    ///
    /// \brief cleanAll
    ///
    void cleanAll();

    ///
    /// \brief setCapturing
    /// \param isInhibited
    /// \param isCapturing
    ///
    void setCapturing(bool isInhibited, bool isCapturing);

    ///
    /// \brief setDetectionRange
    /// \param start
    /// \param end
    ///
    void setDetectionRange(int start, int end)
    {
        hr->setDetectionRange(start, end);
    }

    ///
    /// \brief scansNumber
    /// \return
    ///
    int scansNumber() const
    {
        return ww->scans();
    }

    ///
    /// \briefsetScreenshotCoverage
    /// \param secs
    ///
    void setScreenshotCoverage(int secs);


    ///
    /// \brief GNUplotPalette
    /// \return
    ///
    QString GNUplotPalette();

    ///
    /// \brief getWidgetSize
    /// \return
    ///
    QSize getWidgetSize();


public slots:
    //TODO: many of these slots do not need to be slot anymore

    ///
    /// \brief slotSetBrightness
    /// \param value
    ///
    void slotSetBrightness( int value );

    ///
    /// \brief slotSetContrast
    /// \param value
    ///
    void slotSetContrast( int value );

    ///
    /// \brief slotRuntimeChange
    ///
    void slotRuntimeChange();

    ///
    /// \brief slotShowStatus
    /// \param appState
    ///
    void slotShowStatus( int appState );

    ///
    /// \brief slotRefresh
    /// \param timeStamp
    /// \param raisingEdge
    ///
    void slotRefresh(QString& timeStamp , bool raisingEdge);

    ///
    /// \brief slotReplay
    ///
    void slotReplay();

    ///
    /// \brief slotAppendPixel
    /// \param dBfs
    ///
    void slotAppendPixel( float dBfs );

    ///
    /// \brief slotUpdateGrid
    ///
    void slotUpdateGrid();

    ///
    /// \brief slotSetOffset
    /// \param value
    ///
    void slotSetOffset( int value );

    ///
    /// \brief slotSetZoom
    /// \param value
    ///
    void slotSetZoom( int value );

    ///
    /// \brief slotManualShot
    ///
    void slotManualShot();

    ///
    /// \brief slotPlotPower
    /// \param maxDbm
    /// \param avgDbm
    /// \param avgDiff
    ///
    void slotPlotPower(float maxDbm, float avgDbm, float avgDiff, bool raisingEdge);

    ///
    /// \brief slotStartAcq
    /// \param shotNr
    /// \param totalShots
    ///
    void slotStartAcq(int shotNr, int totalShots);

    ///
    /// \brief slotStopAcq
    ///
    void slotStopAcq();

    ///
    /// \brief slotSetPowerOffset
    /// \param dbm
    ///
    void slotSetPowerOffset(int dbm );

    ///
    /// \brief slotSetPowerZoom
    /// \param zoom
    ///
    void slotSetPowerZoom(int zoom );

    ///
    /// \brief slotConfigChanged
    ///
    void slotConfigChanged();

    ///
    /// \brief slotRangeChanged
    /// \param min
    /// \param max
    ///
    void slotRangeChanged(qreal min, qreal max);

    ///
    /// \brief slotUpdateLogo
    ///
    void slotUpdateLogo();

signals:
    ///
    /// \brief requestManualShot
    ///
    void requestManualShot();

    ///
    /// \brief waterfallResized
    ///
    void waterfallResized();
};

#endif // WATERFALL_H
